Package: poorman
Type: Package
Title: A Poor Man's Dependency Free Recreation of 'dplyr'
Version: 0.2.8
Authors@R: c(
    person("Nathan", "Eastwood", "", "nathan.eastwood@icloud.com",
           role = c("aut", "cre")),
    person("Etienne", "Bacher", , "etienne.bacher@protonmail.com", role = "ctb",
           comment = c(ORCID = "0000-0002-9271-5075")))
Maintainer: Nathan Eastwood <nathan.eastwood@icloud.com>
Description: A replication of key functionality from 'dplyr' and the wider 'tidyverse' using only 'base'.
URL: https://nathaneastwood.github.io/poorman/,
        https://github.com/nathaneastwood/poorman
BugReports: https://github.com/nathaneastwood/poorman/issues
Depends: R (>= 3.3)
Suggests: knitr, rmarkdown, roxygen2, tinytest
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.2.3
VignetteBuilder: knitr
Language: en-GB
NeedsCompilation: no
Packaged: 2026-07-30 16:04:53 UTC; nathan
Author: Nathan Eastwood [aut, cre],
  Etienne Bacher [ctb] (ORCID: <https://orcid.org/0000-0002-9271-5075>)
Repository: CRAN
Date/Publication: 2026-07-31 00:00:36 UTC
Built: R 4.5.2; ; 2026-07-31 01:41:03 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(arrange,data.frame)
S3method(default_missing,data.frame)
S3method(default_missing,default)
S3method(distinct,data.frame)
S3method(distinct,grouped_df)
S3method(fill,data.frame)
S3method(fill,grouped_df)
S3method(filter,data.frame)
S3method(filter,grouped_df)
S3method(glimpse,data.frame)
S3method(glimpse,default)
S3method(group_by,data.frame)
S3method(group_by_drop_default,default)
S3method(group_by_drop_default,grouped_df)
S3method(group_data,data.frame)
S3method(group_data,grouped_df)
S3method(mutate,data.frame)
S3method(mutate,grouped_df)
S3method(nest_by,data.frame)
S3method(nest_by,grouped_df)
S3method(recode,character)
S3method(recode,factor)
S3method(recode,numeric)
S3method(relocate,data.frame)
S3method(rename,data.frame)
S3method(rename_with,data.frame)
S3method(replace_na,data.frame)
S3method(replace_na,default)
S3method(slice,data.frame)
S3method(slice,grouped_df)
S3method(slice_head,data.frame)
S3method(slice_head,grouped_df)
S3method(slice_max,data.frame)
S3method(slice_max,grouped_df)
S3method(slice_min,data.frame)
S3method(slice_min,grouped_df)
S3method(slice_sample,data.frame)
S3method(slice_sample,grouped_df)
S3method(slice_tail,data.frame)
S3method(slice_tail,grouped_df)
S3method(summarise,data.frame)
S3method(summarise,grouped_df)
S3method(summarize,data.frame)
S3method(summarize,grouped_df)
S3method(transmute,data.frame)
S3method(transmute,grouped_df)
S3method(ungroup,data.frame)
S3method(ungroup,grouped_df)
S3method(union_all,data.frame)
S3method(union_all,default)
export("%>%")
export(across)
export(add_count)
export(add_tally)
export(all_of)
export(anti_join)
export(any_of)
export(arrange)
export(between)
export(bind_cols)
export(bind_rows)
export(case_when)
export(coalesce)
export(column_to_rownames)
export(contains)
export(count)
export(cumall)
export(cumany)
export(cume_dist)
export(cummean)
export(cur_column)
export(cur_data)
export(cur_data_all)
export(cur_group)
export(cur_group_id)
export(cur_group_rows)
export(dense_rank)
export(desc)
export(distinct)
export(ends_with)
export(everything)
export(fill)
export(filter)
export(first)
export(full_join)
export(glimpse)
export(group_by)
export(group_by_drop_default)
export(group_cols)
export(group_data)
export(group_indices)
export(group_keys)
export(group_rows)
export(group_size)
export(group_split)
export(group_vars)
export(groups)
export(has_rownames)
export(if_all)
export(if_any)
export(if_else)
export(inner_join)
export(lag)
export(last)
export(last_col)
export(lead)
export(left_join)
export(lst)
export(matches)
export(min_rank)
export(mutate)
export(n)
export(n_distinct)
export(n_groups)
export(na_if)
export(near)
export(nest_by)
export(nth)
export(ntile)
export(num_range)
export(peek_vars)
export(percent_rank)
export(pivot_longer)
export(pivot_wider)
export(pull)
export(recode)
export(recode_factor)
export(relocate)
export(remove_rownames)
export(rename)
export(rename_with)
export(replace_na)
export(right_join)
export(row_number)
export(rowid_to_column)
export(rownames_to_column)
export(select)
export(semi_join)
export(slice)
export(slice_head)
export(slice_max)
export(slice_min)
export(slice_sample)
export(slice_tail)
export(starts_with)
export(summarise)
export(summarize)
export(tally)
export(transmute)
export(ungroup)
export(union_all)
export(unite)
export(where)
export(with_groups)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(utils,head)

<!-- README.md is generated from README.Rmd. Please edit that file -->

<a name="readme-top"></a>

# {poorman} <a href='https://nathaneastwood.github.io/tags/poorman/'><img src='man/figures/logo.png' align="right" height="139" /></a>

[![CRAN
status](https://www.r-pkg.org/badges/version/poorman)](https://cran.r-project.org/package=poorman)
[![Dependencies](https://tinyverse.netlify.app/badge/poorman)](https://cran.r-project.org/package=poorman)
![CRAN downloads](https://cranlogs.r-pkg.org/badges/poorman)
[![R-CMD-check](https://github.com/nathaneastwood/poorman/actions/workflows/check-standard.yaml/badge.svg)](https://github.com/nathaneastwood/poorman/actions/workflows/check-standard.yaml)
[![codecov](https://codecov.io/gh/nathaneastwood/poorman/branch/master/graph/badge.svg?token=YPQSSEEHZJ)](https://app.codecov.io/gh/nathaneastwood/poorman)

<blockquote align="center">
I’d seen my father. He was a poor man, and I watched him do astonishing
things. - Sidney Poitier
</blockquote>

## Overview

{poorman} is a grammar of data manipulation, providing dependency free
versions of [{dplyr}](https://github.com/tidyverse/dplyr) verbs that
help you solve the most common data manipulation challenges:

-   `select()` picks variables based on their names.
-   `mutate()` adds new variables that are functions of existing
    variables.
-   `filter()` picks cases based on their values.
-   `summarise()` reduces multiple values down to a single summary.
-   `arrange()` changes the ordering of the rows.

{poorman} attempts to replicate the {dplyr} API exactly such that your
{dplyr} code will still run even if you use {poorman} in its place. In
addition to replicating {dplyr} functionality, {poorman} implements
other functionality from the wider {tidyverse} such as select helpers
and the pipe, `%>%`.

For more details on the functionality available within {poorman}, check
out the {poorman} series of blog posts
[here](https://nathaneastwood.github.io/tags/poorman/).

<p align="right">
(<a href="#readme-top">back to top</a>)
</p>

## Installation

You can install:

-   the development version from
    [GitHub](https://github.com/nathaneastwood/poorman) with

``` r
# install.packages("remotes")
remotes::install_github("nathaneastwood/poorman")
```

-   the latest release from CRAN with

``` r
install.packages("poorman")
```

<p align="right">
(<a href="#readme-top">back to top</a>)
</p>

## Docker

If you’d like to try out the latest version of the package on CRAN using
Docker, you can run the latest image with:

``` bash
docker run --rm -it nathaneastwood/poorman
```

<p align="right">
(<a href="#readme-top">back to top</a>)
</p>

## Usage

``` r
library(poorman, warn.conflicts = FALSE)
# 
#   I'd seen my father. He was a poor man, and I watched him do astonishing things.
#     - Sidney Poitier

mtcars %>%
  select(mpg, wt, starts_with("c")) %>%
  mutate(kpl = (1.609 * mpg) / 3.785, wt_kg = wt * 453.5924) %>%
  filter(mpg > 28)
#                 mpg    wt cyl carb      kpl    wt_kg
# Fiat 128       32.4 2.200   4    1 13.77321 997.9033
# Honda Civic    30.4 1.615   4    2 12.92301 732.5517
# Toyota Corolla 33.9 1.835   4    1 14.41086 832.3421
# Lotus Europa   30.4 1.513   4    2 12.92301 686.2853

mtcars %>%
  group_by(am, cyl) %>%
  summarise(mean_mpg = mean(mpg), sd_mpg = sd(mpg)) %>%
  ungroup()
#   am cyl mean_mpg    sd_mpg
# 1  0   4 22.90000 1.4525839
# 2  0   6 19.12500 1.6317169
# 3  0   8 15.05000 2.7743959
# 4  1   4 28.07500 4.4838599
# 5  1   6 20.56667 0.7505553
# 6  1   8 15.40000 0.5656854
```

<p align="right">
(<a href="#readme-top">back to top</a>)
</p>

## Related Work

-   [{dplyr}](https://github.com/tidyverse/dplyr)
-   [{bplyr}](https://github.com/yonicd/bplyr) - imports {magrittr} and
    {rlang}; it prepends functions with `b_*()`, e.g. `b_select()`.
-   [{tbltools}](https://github.com/mkearney/tbltools) - imports
    {magrittr} and appends `*_data()` to each of its functions,
    e.g. `select_data()`.

<p align="right">
(<a href="#readme-top">back to top</a>)
</p>

#' Group by one or more variables
#'
#' Determine the groups within a `data.frame` to perform operations on. [ungroup()] removes the grouping levels.
#'
#' @param .data `data.frame`. The data to group.
#' @param ... One or more unquoted column names to group/ungroup the data by.
#' @param .add `logical(1)`. When `FALSE` (the default) `group_by()` will override existing groups. To add to existing
#' groups, use `.add = TRUE`.
#' @param .drop `logical(1)`. Drop groups formed by factor levels that don't appear in the data? The default is `TRUE`
#' except when `.data` has been previously grouped with `.drop = FALSE`. See [group_by_drop_default()] for details.
#'
#' @examples
#' group_by(mtcars, am, cyl)
#' ungroup(mutate(group_by(mtcars, am, cyl), sumMpg = sum(mpg)))
#' mtcars %>%
#'   group_by(am, cyl) %>%
#'   mutate(sumMpg = sum(mpg)) %>%
#'   ungroup()
#' mtcars %>%
#'   group_by(carb) %>%
#'   filter(any(gear == 5))
#'
#' # You can group by expressions: this is just short-hand for
#' # a mutate() followed by a group_by()
#' mtcars %>% group_by(vsam = vs + am)
#'
#' @return
#' When using [group_by()], a `data.frame`, grouped by the grouping variables.
#'
#' @name group_by
#' @export
group_by <- function(.data, ..., .add = FALSE, .drop = group_by_drop_default(.data)) {
  UseMethod("group_by")
}

#' @export
group_by.data.frame <- function(.data, ..., .add = FALSE, .drop = group_by_drop_default(.data)) {
  vars <- dotdotdot(..., .impute_names = TRUE)
  if (all(vapply(vars, is.null, FALSE))) {
    res <- groups_set(.data, NULL)
    class(res) <- class(res)[!(class(res) %in% "grouped_df")]
    return(res)
  }
  new_cols <- add_group_columns(.data, vars)
  res <- new_cols$data
  groups <- new_cols$groups
  if (isTRUE(.add)) groups <- union(group_vars(.data), groups)
  unknown <- !(groups %in% colnames(res))
  if (any(unknown)) stop("Invalid groups: ", groups[unknown])
  if (length(groups) > 0L) {
    res <- groups_set(res, groups, .drop)
    class(res) <- union("grouped_df", class(res))
  }
  res
}

#' Default value for .drop argument of group_by
#'
#' @param .tbl A `data.frame`.
#'
#' @examples
#' group_by_drop_default(iris)
#'
#' iris %>%
#'   group_by(Species) %>%
#'   group_by_drop_default()
#'
#' iris %>%
#'   group_by(Species, .drop = FALSE) %>%
#'   group_by_drop_default()
#'
#' @return `TRUE` unless `.tbl` is a grouped `data.frame` that was previously obtained by `group_by(.drop = FALSE)`
#'
#' @export
group_by_drop_default <- function(.tbl) {
  UseMethod("group_by_drop_default")
}

#' @export
group_by_drop_default.default <- function(.tbl) {
  TRUE
}

#' @export
group_by_drop_default.grouped_df <- function(.tbl) {
  tryCatch({
    !identical(attr(group_data(.tbl), ".drop"), FALSE)
  }, error = function(e) {
    TRUE
  })
}

# -- Helpers -------------------------------------------------------------------

add_group_columns <- function(.data, vars) {
  vars <- vars[!vapply(vars, is.null, FALSE)]
  types <- do.call(c, lapply(vars, typeof))
  test <- any(types == "language")
  needs_mutate <- if (test) unname(which(types == "language")) else NULL
  if (!is.null(needs_mutate)) {
    .data <- do.call(mutate, c(list(.data = ungroup(.data)), vars[needs_mutate]))
  }
  list(data = .data, groups = names(vars))
}

expect_equal(
  unite(data.frame(x = "a", y = "b", stringsAsFactors = FALSE), z, x:y),
  data.frame(z = "a_b", stringsAsFactors = FALSE),
  info = "unite() pastes columns together and removes old columns"
)

expect_equal(
  unite(data.frame(x = "a", y = "b", stringsAsFactors = FALSE), x, x:y),
  data.frame(x = "a_b", stringsAsFactors = FALSE),
  info = "unite() does not remove new col in case of name clash"
)

df <- data.frame(g = 1, x = "a", stringsAsFactors = FALSE) %>% group_by(g)
rs <- df %>% unite(x, x)
expect_equal(df, rs, info = "unite() preserves grouping")
expect_equal(class(df), class(rs), info = "unite() preserves grouping")
expect_equal(group_vars(df), group_vars(rs), info = "unite() preserves grouping")

df <- data.frame(g = 1, x = "a", stringsAsFactors = FALSE) %>% group_by(g)
rs <- df %>% unite(gx, g, x)
expect_equal(rs$gx, "1_a", info = "unite() drops grouping when needed")
expect_equal(group_vars(rs), character(), info = "unite() drops grouping when needed")

expect_equal(
  unite(data.frame(x = "a", y = "b", stringsAsFactors = FALSE), "z"),
  data.frame(z = "a_b", stringsAsFactors = FALSE),
  info = "Passing no columns uses all columns"
)

df <- data.frame(x = c("a", "a", NA, NA), y = c("b", NA, "b", NA), stringsAsFactors = FALSE)
out <- unite(df, "z", x:y, na.rm = TRUE)
expect_equal(out$z, c("a_b", "a", "b", ""), info = "unite() can remove missing vars on request")

df <- data.frame(
  x = c("x", "y", "z"),
  lgl = NA,
  dbl = NA_real_,
  chr = NA_character_,
  stringsAsFactors = FALSE
)
expect_equal(
  unite(df, "out", any_of(c("x", "lgl")), na.rm = TRUE)$out,
  c("x", "y", "z"),
  info = "unite() can remove missing vars on request even for logical NAs"
)
expect_equal(
  unite(df, "out", any_of(c("x", "dbl")), na.rm = TRUE)$out,
  c("x", "y", "z"),
  info = "unite() can remove missing vars on request even for numeric NAs"
)
expect_equal(
  unite(df, "out", any_of(c("x", "chr")), na.rm = TRUE)$out,
  c("x", "y", "z"),
  info = "unite() can remove missing vars on request even for character NAs"
)

df <- data.frame(x = c(1, 1, 1, 1), y = c(1, 1, 2, 2), z = c(1, 2, 1, 2))
expect_equal(
  distinct(df),
  df,
  info = "distinct() is equivalent to unique.data.frame() when .keep_all is FALSE"
)

df <- data.frame(x = c(1, 1, 1, 1), y = c(1, 1, 2, 2), z = c(1, 2, 1, 2))
expect_equal(
  distinct(df, x, .keep_all = FALSE),
  unique(df["x"]),
  info = "distinct() for single column works as expected"
)
expect_equal(
  distinct(df, y, .keep_all = FALSE),
  data.frame(y = 1:2),
  info = "distinct() for single column works as expected"
)

expect_equal(
  ncol(distinct(data.frame())),
  0L,
  info = "distinct() works for 0-column data.frames"
)

expect_equal(
  distinct(data.frame(x = c(1, 1), y = c(2, 2))),
  data.frame(x = 1, y = 2),
  info = "If no variables are specified, test distinct() uses all columns"
)

expect_equal(
  data.frame(x = c(1, 1, 1), y = 3:1) %>% distinct(x),
  data.frame(x = 1),
  info = "distinct() keeps only specified columns"
)

expect_equal(
  data.frame(x = c(1, 1, 1), y = 3:1) %>% distinct(x, .keep_all = TRUE),
  data.frame(x = 1, y = 3L),
  info = "distinct() keeps all columns when .keep_all = TRUE"
)

expect_equal(
  data.frame(a = 1:3, b = 4:6) %>% distinct(a, a) %>% colnames(),
  "a",
  "distinct() doesn't duplicate columns"
)

expect_equal(
  row.names(distinct(df, y)),
  c("1", "2"),
  info = "distinct() reset row names when they are numeric"
)

expect_equal(
  row.names(distinct(head(mtcars, 3), mpg)),
  c("Mazda RX4", "Datsun 710"),
  info = "distinct() keeps row names when they are character"
)

expect_equal(
  data.frame(a = 1:3, b = 4:6) %>% group_by(a) %>% distinct(a) %>% colnames(),
  "a",
  info = "Grouped distinct() doesn't duplicate columns"
)

expect_equal(
  data.frame(g = c(1, 2), x = c(1, 2)) %>% group_by(g) %>% distinct(x) %>% colnames(),
  c("g", "x"),
  info = "Grouped distinct() always includes group columns"
)

df <- data.frame(g = c(1, 2), x = c(1, 2))
expect_equal(
  df %>% distinct() %>% group_by(g),
  df %>% group_by(g) %>% distinct(),
  info = "Empty grouped distinct() equivalent to empty ungrouped"
)

df <- data.frame(g = c(1, 2), x = c(1, 2))
expect_equal(
  df %>% distinct(aa = g * 2),
  df %>% mutate(aa = g * 2) %>% distinct(aa),
  info = "distinct() on a new, mutated variable is equivalent to mutate() followed by distinct()"
)

df <- data.frame(g = c(1, 2), x = c(1, 2))
expect_equal(
  df %>% distinct(aa = g),
  df %>% mutate(aa = g) %>% distinct(aa),
  info = "distinct() on a new, copied variable is equivalent to mutate() followed by distinct()"
)

expect_equal(
  data.frame(a = c(1, 2), b = c(1, 2)) %>% distinct(a, b * 2),
  data.frame(a = c(1, 2), `b * 2` = c(2, 4), check.names = FALSE),
  info = "Non-named mutations arguments are returned"
)

expect_equal(
  data.frame(a = c(1, 2), b = c(1, 2)) %>% distinct(a, b2 = b * 2),
  data.frame(a = c(1, 2), b2 = c(2, 4), check.names = FALSE),
  info = "Named mutations arguments are returned"
)

expect_equal(
  structure(list(), .Names = character(0), row.names = c(NA, -2L), class = "data.frame") %>%
    distinct() %>%
    nrow(),
  1L,
  info = "distinct() handles 0 columns edge case"
)

expect_equal(
  names(distinct(data.frame(x = 1:2, y = 3:4), y, x)),
  c("x", "y"),
  info = "distinct() preserves order of the input variables"
)

gf <- data.frame(x = c(1, 1, 2, 2), y = c(1, 1, 2, 2)) %>% group_by(x)
expect_equal(
  gf %>% distinct() %>% group_vars(),
  "x",
  info = "distinct() preserves grouping"
)

expect_equal(
  group_vars(distinct(gf, x = x + 2)),
  "x",
  info = "Grouped distinct() preserves grouping"
)

rm(df, gf)

df <- data.frame(x = 1:2)
gf <- group_by(df, x)
expect_equal(class(with_groups(df, x, mutate)), "data.frame", info = "restores original class #1")
expect_equal(class(with_groups(gf, x, mutate)), c("grouped_df", "data.frame"), info = "restores original class #2")

gf <- group_by(data.frame(x = 1:2), x)
out <- gf %>% with_groups(NULL, mutate, y = mean(x))
expect_equal(out$y, c(1.5, 1.5), info = ".groups = NULL ungroups")

expect_identical(
  with_groups(mtcars, identity(2), mutate, disp = disp / sd(disp)),
  with_groups(mtcars, 2, mutate, disp = disp / sd(disp)),
  info = ".groups is defused with context"
)

expect_equal(
  mtcars %>% select(MilesPerGallon = mpg, NumberOfGears = gear),
  {
    res <- mtcars[, c("mpg", "gear")]
    colnames(res) <- c("MilesPerGallon", "NumberOfGears")
    res
  },
  info = "Test renaming columns when selecting"
)

expect_equal(
  mtcars %>% select(MilesPerGallon = 1, NumberOfGears = 10),
  {
    res <- mtcars[, c("mpg", "gear")]
    colnames(res) <- c("MilesPerGallon", "NumberOfGears")
    res
  },
  info = "Test renaming columns when selecting using integers"
)

expect_equal(
  mtcars %>% select(MilesPerGallon = "mpg", NumberOfGears = "gear"),
  {
    res <- mtcars[, c("mpg", "gear")]
    colnames(res) <- c("MilesPerGallon", "NumberOfGears")
    res
  },
  info = "Test renaming columns when selecting using characters"
)

expect_equal(
  mtcars %>% select(MilesPerGallon = 1, Horsepower = "hp", NumberOfGears = gear),
  {
    res <- mtcars[, c("mpg", "hp", "gear")]
    colnames(res) <- c("MilesPerGallon", "Horsepower", "NumberOfGears")
    res
  },
  info = "Test renaming columns when selecting using a mixture of integer, character and nse names"
)

expect_equal(
  mtcars %>% select(MilesPerGallon = -mpg) %>% colnames(),
  colnames(mtcars)[-1L],
  info = "Negative column selections do not affect renaming"
)

expect_equal(
  mtcars %>% select(Columns = -mpg:-gear) %>% colnames(),
  "carb",
  info = "Negative column selections do not affect renaming in multiple column selections"
)

expect_equal(
  mtcars %>% select(Col = mpg:disp) %>% colnames(),
  c("Col1", "Col2", "Col3"),
  info = "Renaming columns during multiple column selection"
)

expect_equal(
  mtcars %>% select(11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1),
  mtcars[, rev(colnames(mtcars))],
  info = "Test order of selection is preserved"
)

expect_equal(
  mtcars %>% select(1, 1, 1, 1),
  mtcars[, "mpg", drop = FALSE],
  info = "Test selecting the same column multiple times only returns it once"
)

expect_equal(
  mtcars %>% select(mpg:disp, Gears = gear),
  {
    res <- mtcars[, c("mpg", "cyl", "disp", "gear")]
    colnames(res)[4] <- "Gears"
    res
  },
  info = "Ensure columns are named correctly when not all columns are to be renamed"
)

res <- suppressMessages(mtcars %>% group_by(cyl) %>% select(MilesPerGallon = mpg))
expect_equal(
  colnames(res),
  c("cyl", "MilesPerGallon"),
  info = "Ensure columns are named correctly when renaming grouped data"
)

expect_equal(
  transmute(.data = mtcars, -mpg, neg_am = -am),
  {
    mtcars[["-mpg"]] <- mtcars$mpg * -1
    mtcars[, "neg_am"] <- mtcars$am * -1
    mtcars[, c("-mpg", "neg_am")]
  },
  info = "Transmute works with non-named parameters"
)

expect_equal(
  mtcars %>% transmute(mpg2 = mpg * 2),
  {
    mtcars[, "mpg2"] <- mtcars$mpg * 2
    mtcars[, "mpg2", drop = FALSE]
  },
  info = "Transmute single column"
)

expect_equal(
  mtcars %>% transmute(mpg2 = mpg * 2, mpg3 = mpg * 3),
  {
    mtcars[, "mpg2"] <- mtcars$mpg * 2
    mtcars[, "mpg3"] <- mtcars$mpg * 3
    mtcars[, c("mpg2", "mpg3")]
  },
  info = "Transmute multiple columns"
)

expect_equal(
  mtcars %>% group_by(am) %>% transmute(sumMpg = sum(mpg)) %>% colnames(),
  c("am", "sumMpg"),
  info = "transmute() keeps groups"
)

expect_equal(
  colnames(rename_with(mtcars, toupper, 1)),
  c("MPG", colnames(mtcars)[2:ncol(mtcars)]),
  info = "rename_with() can select columns"
)

expect_equal(
  colnames(rename_with(mtcars, toupper, mpg)),
  c("MPG", colnames(mtcars)[2:ncol(mtcars)]),
  info = "rename_with() can select columns"
)

expect_equal(
  colnames(rename_with(mtcars, gsub, 1, pattern = "p", replacement = "P")),
  c("mPg", colnames(mtcars)[2:ncol(mtcars)]),
  info = "rename_with() passes ... along"
)

expect_error(
  mtcars %>% rename_with(.fn = function(x) rep("X", ncol(mtcars))),
  info = "Must not return duplicate columns"
)

expect_equal(
  mtcars %>% group_by(am, cyl) %>% rename_with(toupper) %>% group_vars(),
  c("CYL", "AM"),
  info = "Group names are updated"
)

expect_equal(
  mtcars %>% filter(am == 1),
  mtcars[mtcars$am == 1, ],
  info = "Test single logical filter"
)

expect_equal(
  mtcars %>% filter(mpg > 28 & am == 1),
  mtcars[mtcars$mpg > 28 & mtcars$am == 1, ],
  info = "Test multiple logical filters; sep = &"
)

expect_equal(
  mtcars %>% filter(mpg > 28, am == 1),
  mtcars[mtcars$mpg > 28 & mtcars$am == 1, ],
  info = "Test multiple logical filters; sep = ,"
)

expect_equal(
  mtcars %>% filter(cyl == 4 | cyl == 8),
  mtcars[mtcars$cyl == 4 | mtcars$cyl == 8, ],
  info = "Test multiple logical filters; sep = |"
)

expect_equal(
  mtcars %>% filter(!(cyl %in% c(4, 6)), am != 0),
  mtcars[!(mtcars$cyl %in% c(4, 6)) & mtcars$am != 0, ],
  info = "Test multiple logical filters; negation"
)

expect_equal(
  {
    sp1 <- "virginica"
    filter(iris, Species == sp1)
  },
  iris[iris$Species == "virginica", ],
  info = "Test filters work with variable inputs"
)

expect_equal(
  {
    sp2 <- levels(iris$Species)
    lapply(sp2, function(x) {
      filter(iris, Species == x)
    })
  },
  {
    res <- split(iris, iris$Species)
    names(res) <- NULL
    res
  },
  info = "Test filter works in an anonymous function (#9)"
)

expect_equal(
  {
    sp3 <- unique(as.character(iris$Species))
    lapply(sp2, function(x) {
      filter(iris, Species == x)
    })
  },
  {
    res <- split(iris, iris$Species)
    names(res) <- NULL
    res
  },
  info = "Test filter works in an anonymous function; factors (#9)"
)

expect_equal(
  filter(.data = mtcars),
  mtcars,
  info = "filter() returns the input data if no parameters are given"
)

# Grouped Operations
expect_equal(
  mtcars %>% group_by(carb) %>% filter(any(mpg > 28)) %>% ungroup(),
  {
    rows <- rownames(mtcars)
    res <- do.call(rbind, unname(lapply(split(mtcars, mtcars$carb), function(x) x[any(x$mpg > 28), ])))
    res[rows[rows %in% rownames(res)], ]
  },
  info = "Test grouped filters"
)

expect_equal(
  mtcars %>% group_by(cyl, am) %>% filter(cyl %in% c(4, 8), .preserve = TRUE) %>% group_rows(),
  list(
    c(4L, 5L, 15L), c(1L, 12L, 13L, 14L, 20L, 21L, 22L, 25L), integer(0), integer(0),
    c(2L, 3L, 6L, 7L, 8L, 9L, 10L, 11L, 16L, 17L, 18L, 19L), 23:24
  ),
  info = ".preserve = TRUE returns the correct row indicies"
)

expect_equal(
  mtcars %>% group_by(cyl, am) %>% filter(cyl %in% c(4, 8), .preserve = FALSE) %>% group_rows(),
  list(
    c(4L, 5L, 15L), c(1L, 12L, 13L, 14L, 20L, 21L, 22L, 25L),
    c(2L, 3L, 6L, 7L, 8L, 9L, 10L, 11L, 16L, 17L, 18L, 19L), 23:24
  ),
  info = ".preserve = FALSE returns the correct row indicies"
)

# Errors

expect_error(
  filter(.data = mtcars, mpg > 20, am = 1),
  info = "Check for named arguments works"
)

mt_gears <- mtcars %>% group_by(gear)

expect_equal(
  mtcars %>% mutate(n = n()),
  {
    res <- mtcars
    res$n <- 32L
    res
  },
  info = "n() works within mutate()"
)

expect_equal(
  mt_gears %>% mutate(n = n()) %>% ungroup(),
  {
    rows <- rownames(mtcars)
    res <- do.call(rbind, unname(lapply(split(mtcars, mtcars$gear), function(x) {
      x$n <- nrow(x)
      x
    })))
    res[rows[rows %in% rownames(res)], ]
  },
  info = "n() works within a grouped mutate()"
)

expect_equal(
  mtcars %>% filter(mpg < n() / 2),
  mtcars[mtcars$mpg < 16, ],
  info = "n() works within filter()"
)

expect_equal(
  mt_gears %>% filter(n() < 10) %>% ungroup(),
  {
    do.call(rbind, unname(lapply(split(mtcars, mtcars$gear), function(x) {
      x[nrow(x) < 10, ]
    })))
  },
  info = "n() works within a grouped filter()"
)

expect_equal(
  mtcars %>% summarise(n = n()),
  data.frame(n = 32),
  info = "n() works within summarise()"
)

expect_equal(
  mt_gears %>% summarise(n = n()) %>% ungroup(),
  data.frame(gear = c(3, 4, 5), n = c(15, 12, 5)),
  info = "n() works within a grouped summarise()"
)

expect_error(
  mtcars %>% select(n())
)

rm(mt_gears)

df <- data.frame(x = 1:10, y = 1:10)
gf <- group_by(df, x)
expect_identical(
  arrange(df),
  df,
  info = "empty arrange() returns input for data.frames"
)
expect_identical(
  arrange(gf),
  gf,
  info = "empty arrange() returns input for grouped_df"
)

df <- data.frame(a = numeric(0))
expect_equal(
  arrange(df, a),
  df,
  info = "arrange() can sort empty data frame"
)

df <- data.frame(x = c(2, 1, NA))
expect_equal(
  df %>% arrange(x) %>% pull(),
  c(1, 2, NA),
  info = "local arrange() sorts missing values to the end"
)
expect_equal(
  df %>% arrange(desc(x)) %>% pull(),
  c(2, 1, NA),
  info = "local arrange() sorts missing values to the end"
)

# -- column types --------------------------------------------------------------

df <- data.frame(x = 1:3, y = 3:1 + 2i)
expect_equal(
  arrange(df, y),
  data.frame(x = 3:1, y = 1:3 + 2i),
  info = "arrange() handles complex columns"
)

TestS4 <- suppressWarnings(setClass("TestS4", contains = "integer"))
setMethod(
  "[",
  'TestS4',
  function(x, i, ...) TestS4(unclass(x)[i, ...])
)
df <- data.frame(x = 1:3, y = TestS4(3:1))
expect_equal(
  arrange(df, y),
  data.frame(x = 3:1, y = TestS4(1:3)),
  info = "arrange handles S4 classes"
)
removeClass("TestS4")

# -- data ----------------------------------------------------------------------

df1 <- data.frame(x = 1:3, y = 3:1)
df2 <- df1 %>% group_by(x)
expect_true(
  inherits(arrange(df1, x), "data.frame"),
  info = "arrange() preserves input class for data.frames"
)
expect_true(
  inherits(arrange(df2, x), "grouped_df"),
  info = "arrange() preserves input class for grouped_dfs"
)

df <- data.frame(g = c(2, 2, 1, 1), x = c(1, 3, 2, 4))
res <- df %>% group_by(g) %>% arrange(x)
expect_true(
  inherits(res, "grouped_df"),
  info = "arrange() updates the grouping structure: 1"
)
expect_equal(
  group_rows(res),
  list(c(2L, 4L), c(1L, 3L)),
  info = "arrange() updates the grouping structure: 2"
)

expect_equal(
  mtcars %>% arrange(mpg),
  mtcars[order(mtcars$mpg), ],
  info = "Test ascending column arrangement"
)

expect_equal(
  mtcars %>% arrange(-mpg),
  mtcars[order(-mtcars$mpg), ],
  info = "Test descending column arrangement"
)

expect_equal(
  mtcars %>% arrange(cyl, mpg),
  mtcars[order(mtcars$cyl, mtcars$mpg), ],
  info = "Test multiple column arrangement"
)

expect_equal(
  mtcars %>% arrange(cyl, -mpg),
  mtcars[order(mtcars$cyl - mtcars$mpg), ],
  info = "Test multiple ascending and descending column arrangement"
)

# data masking
df <- data.frame(x = 1:4, y = 5:8)
expect_equal(
  arrange(df, -x * y),
  structure(list(x = 4:1, y = 8:5), row.names = 1:4, class = "data.frame"),
  info = "arrange() can add and arrange by new columns (#89)"
)

# Grouped Operations
df <- data.frame(g = c(2, 2, 1, 1), x = c(1, 3, 2, 4))
res <- df %>% group_by(g) %>% arrange(x)
expect_true(
  inherits(res, "grouped_df"),
  info = "arrange() keeps the grouping class"
)

expect_equal(
  group_rows(res),
  list(c(2L, 4L), c(1L, 3L)),
  info = "arrange() keeps the grouping structure"
)

# .by_group
df <- data.frame(g = c(2, 1, 2, 1), x = 4:1)
gf <- group_by(df, g)
expect_equal(
  arrange(gf, x),
  structure(
    list(g = c(1, 2, 1, 2), x = 1:4),
    groups = structure(
      list(g = c(1, 2), .rows = list(c(1L, 3L), c(2L, 4L))), row.names = 1:2, class = "data.frame", .drop = TRUE
    ),
    row.names = 1:4,
    class = c("grouped_df", "data.frame")
  ),
  info = "grouped arrange() ignores group_by groups"
)
expect_equal(
  arrange(gf, x, .by_group = TRUE),
  structure(
    list(g = c(1, 1, 2, 2), x = c(1L, 3L, 2L, 4L)),
    groups = structure(list(g = c(1, 2), .rows = list(1:2, 3:4)), row.names = 1:2, class = "data.frame", .drop = TRUE),
    row.names = 1:4, class = c("grouped_df", "data.frame")
  ),
  info = "grouped arrange() ignores group, unless requested with .by_group"
)

# with character columns
df <- data.frame(x = c("a", "b", "a", "b"),
                 y = c("c", "c", "c", "d"),
                 z = c(4, 2, 1, 3))
expect_equal(
  df %>% arrange(-x),
  data.frame(x = c("b", "b", "a", "a"),
             y = c("c", "d", "c", "c"),
             z = c(2, 3, 4, 1))
)
expect_equal(
  df %>% arrange(-x, y),
  data.frame(x = c("b", "b", "a", "a"),
             y = c("c", "d", "c", "c"),
             z = c(2, 3, 4, 1)) %>%
    structure(row.names = 1:4)
)
expect_equal(
  df %>% arrange(-x, y, z),
  data.frame(x = c("b", "b", "a", "a"),
             y = c("c", "d", "c", "c"),
             z = c(2, 3, 1, 4)) %>%
    structure(row.names = 1:4)
)

expect_equal(
  df %>% arrange(desc(x)),
  data.frame(x = c("b", "b", "a", "a"),
             y = c("c", "d", "c", "c"),
             z = c(2, 3, 4, 1))
)
expect_equal(
  df %>% arrange(desc(x), y),
  data.frame(x = c("b", "b", "a", "a"),
             y = c("c", "d", "c", "c"),
             z = c(2, 3, 4, 1)) %>%
    structure(row.names = 1:4)
)
expect_equal(
  df %>% arrange(desc(x), y, z),
  data.frame(x = c("b", "b", "a", "a"),
             y = c("c", "d", "c", "c"),
             z = c(2, 3, 1, 4)) %>%
    structure(row.names = 1:4)
)

# all missings left unchanged 
df <- data.frame(
  lgl = c(NA, NA),
  int = c(NA_integer_, NA),
  dbl = c(NA_real_, NA),
  chr = c(NA_character_, NA)
)

down <- fill(df, lgl, int, dbl, chr)
up <- fill(df, lgl, int, dbl, chr, .direction = "up")

expect_identical(down, df)
expect_identical(up, df)


# filled down from last non-missing
df <- data.frame(x = c(NA, 1, NA, 2, NA, NA))

out <- fill(df, x)
expect_equal(out$x, c(NA, 1, 1, 2, 2, 2))

out <- fill(df, x, .direction = "up")
expect_equal(out$x, c(1, 1, 2, 2, NA, NA))

out <- fill(df, x, .direction = "downup")
expect_equal(out$x, c(1, 1, 1, 2, 2, 2))

out <- fill(df, x, .direction = "updown")
expect_equal(out$x, c(1, 1, 2, 2, 2, 2))


# missings filled up for each vector
df <- data.frame(
  lgl = c(NA, T),
  int = c(NA, 1L),
  dbl = c(NA, 1),
  chr = c(NA, "a")
)
out <- fill(df, everything(), .direction = "up")
expect_equal(out$lgl, c(TRUE, TRUE))
expect_equal(out$int, c(1L, 1L))
expect_equal(out$dbl, c(1, 1))
expect_equal(out$chr, c("a", "a"))


# NaN treated as missing
df <- data.frame(x = c(1, NaN, 2))

out <- fill(df, x)
expect_identical(out$x, c(1, 1, 2))

out <- fill(df, x, .direction = "up")
expect_identical(out$x, c(1, 2, 2))

# respect grouping
df <- data.frame(x = c(1, 1, 2), y = c(1, NA, NA))
out <- df %>%
  group_by(x) %>%
  fill(y)
expect_equal(out$y, c(1, 1, NA))

expect_equal(
  mtcars %>% nest_by(am),
  {
    res <- data.frame(am = c(0L, 1L))
    res[["data"]] <- list(mtcars[mtcars$am == 0, -9], mtcars[mtcars$am == 1, -9])
    res %>% group_by(am)
  },
  info = "nest_by() takes group names"
)

expect_equal(
  mtcars %>% group_by(am) %>% nest_by(),
  {
    res <- data.frame(am = c(0L, 1L))
    res[["data"]] <- list(mtcars[mtcars$am == 0, -9], mtcars[mtcars$am == 1, -9])
    res %>% group_by(am)
  },
  info = "nest_by() works on grouped data sets"
)

expect_error(
  mtcars %>% nest_by(am, .keep = "true"),
  info = "`.keep` should be `logical(1)`"
)

expect_error(
  mtcars %>% nest_by(am, .key = TRUE),
  info = "`.key` should be `character(1)`"
)

expect_error(
  mtcars %>% group_by(am) %>% nest_by(cyl),
  info = "Can't re-group while nesting. Either `ungroup()` first or don't supply arguments to `nest_by()`"
)

expect_error(
  mtcars %>% group_by(fake_column_name)
)

res <- suppressMessages(mtcars %>% group_by(am, cyl) %>% select(mpg))
expect_equal(
  colnames(res),
  c("am", "cyl", "mpg"),
  info = "Test that groups persist in select() when no groups are selected"
)

res <- suppressMessages(mtcars %>% group_by(am, cyl) %>% select(mpg, cyl))
expect_equal(
  colnames(res),
  c("am", "mpg", "cyl"),
  info = "Test that groups persist in select() when only some groups are selected"
)

res <- mtcars %>% group_by(am, cyl) %>% relocate(gear, .before = mpg)
expect_equal(
  colnames(res),
  c("gear", "mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "carb"),
  info = "Test that groups persist in relocate()"
)

res <- mtcars %>% group_by(am, cyl) %>% rename(Gears = gear)
expect_equal(
  colnames(res),
  c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "Gears", "carb"),
  info = "Test that groups persist in rename()"
)

res <- mtcars %>% group_by(tmp = am * cyl)
expect_equal(
  colnames(res),
  c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb", "tmp"),
  info = "group_by() can create new columns #1"
)
expect_equal(
  res$tmp,
  mtcars$am * mtcars$cyl,
  info = "group_by() can create new columns #2"
)

# NULL group
expect_equal(
  group_by(mtcars, NULL),
  mtcars,
  info = "NULL group returns the original data.frame"
)
res <- group_by(mtcars, am, cyl)
expect_equal(
  class(group_by(res, NULL)),
  "data.frame",
  info = "group_by(., NULL) ungroups data #1"
)
expect_equal(
  attr(group_by(res, NULL), "groups", exact = TRUE),
  NULL,
  info = "group_by(., NULL) ungroups data #2"
)

rm(res)

# .drop = TRUE ---------------------------------------------------

res <- iris %>%
  filter(Species == "setosa") %>%
  group_by(Species, .drop = TRUE)
expect_identical(
  group_data(res),
  structure(
    list(
      Species = structure(1L, .Label = c("setosa", "versicolor",  "virginica"), class = "factor"),
      .rows = list(1:50)
    ),
    row.names = 1L, class = "data.frame", .drop = TRUE
  ),
  info = "group_by(.drop = TRUE) drops empty groups"
)
expect_true(group_by_drop_default(res))


res <- iris %>%
  filter(Species == "setosa") %>%
  group_by(Species, .drop = TRUE)
res2 <- filter(res, Sepal.Length > 5)
expect_true(group_by_drop_default(res2), info = "grouped data.frames remember their .drop")

res3 <- filter(res, Sepal.Length > 5, .preserve = FALSE)
expect_true(group_by_drop_default(res3), info = "grouped data.frames remember their .drop")

res4 <- group_by(res3, Species)
expect_true(group_by_drop_default(res4), info = "grouped data.frames remember their .drop")
expect_equal(nrow(group_data(res4)), 1L, info = "grouped data.frames remember their .drop")

res <- iris %>%
  filter(Species == "setosa") %>%
  group_by(Species, .drop = FALSE)
expect_false(group_by_drop_default(res), info = "grouped data frames remember their .drop = FALSE")

res2 <- res %>%
  group_by(Species)
expect_false(group_by_drop_default(res2), info = "grouped data frames remember their .drop = FALSE")

df <- data.frame(x = ordered("x"))
drop <- df %>% group_by(x) %>% group_data()
nodrop <- df %>% group_by(x, .drop = FALSE) %>% group_data()
expect_equal(is.ordered(drop$x), is.ordered(nodrop$x), info = "group_by(.drop = FALSE) preserve ordered factors")
expect_true(is.ordered(nodrop$x), info = "group_by(.drop = FALSE) preserve ordered factors")

df <- data.frame(
  f1 = factor("a", levels = c("a", "b", "c")),
  f2 = factor("d", levels = c("d", "e", "f", "g")),
  x = 42
)

res <- df %>%
  group_by(f1, f2, .drop = TRUE)
expect_equal(n_groups(res), 1L, info = "summarise maintains the .drop attribute")

res2 <- summarise(res, x = sum(x))
expect_equal(n_groups(res2), 1L, info = "summarise maintains the .drop attribute")
expect_true(group_by_drop_default(res2), info = "summarise maintains the .drop attribute")

df1 <- group_by(data.frame(
  f1 = factor(c("a", "b"), levels = c("a", "b", "c")),
  x  = 42:43
), f1, .drop = TRUE)
df2 <- group_by(data.frame(
  f1 = factor(c("a"), levels = c("a", "b", "c")),
  y = 1
), f1, .drop = TRUE)
res <- left_join(df1, df2, by = "f1")
expect_equal(n_groups(res), 2L, info = "joins maintain the .drop attribute")

df2 <- group_by(data.frame(
  f1 = factor(c("a", "c"), levels = c("a", "b", "c")),
  y = 1:2
), f1, .drop = TRUE)
res <- full_join(df1, df2, by = "f1")
expect_equal(n_groups(res), 3L, info = "joins maintain the .drop attribute")

d <- data.frame(
  f1 = factor("b", levels = c("a", "b", "c")),
  f2 = factor("g", levels = c("e", "f", "g")),
  x  = 48
)
res <- group_by(group_by(d, f1, .drop = TRUE), f2, .add = TRUE)
expect_equal(n_groups(res), 1L, info = "group_by(add = TRUE) sets .drop if the origonal data was .drop = TRUE")
expect_true(group_by_drop_default(res), info = "group_by(add = TRUE) sets .drop if the origonal data was .drop = TRUE")

df <- data.frame(x = 1:2, y = 1:2) %>%
  structure(class = c("grouped_df", "data.frame"))
expect_true(group_by_drop_default(df), info = "group_by_drop_default() is forgiving about corrupt grouped df")

res <- data.frame(x = c("apple", NA, "banana"), y = 1:3, stringsAsFactors = FALSE) %>%
  group_by(x) %>%
  group_data()
expect_identical(res$x, c("apple", "banana", NA_character_), info = "group_by() puts NA groups last in STRSXP")
expect_identical(res$.rows, list(1L, 3L, 2L), info = "group_by() puts NA groups last in STRSXP")

expect_equal(
  mtcars %>% nest_by(am),
  {
    res <- data.frame(am = c(0L, 1L))
    res[["data"]] <- list(mtcars[mtcars$am == 0, -9], mtcars[mtcars$am == 1, -9])
    res %>% group_by(am)
  },
  info = "nest_by() takes group names"
)

expect_equal(
  mtcars |> select(mpg, cyl, am) |> nest_by(am, cyl),
  {
    res <- data.frame(
      am = c(0, 0, 0, 1, 1, 1),
      cyl = c(4, 6, 8, 4, 6, 8)
    )
    res[["data"]] <- list(
      mtcars[mtcars$am == 0 & mtcars$cyl == 4, "mpg", drop = FALSE],
      mtcars[mtcars$am == 0 & mtcars$cyl == 6, "mpg", drop = FALSE],
      mtcars[mtcars$am == 0 & mtcars$cyl == 8, "mpg", drop = FALSE],
      mtcars[mtcars$am == 1 & mtcars$cyl == 4, "mpg", drop = FALSE],
      mtcars[mtcars$am == 1 & mtcars$cyl == 6, "mpg", drop = FALSE],
      mtcars[mtcars$am == 1 & mtcars$cyl == 8, "mpg", drop = FALSE]
    )
    group_by(res, am)
  },
  info = "nest_by() works with only one non-group column (#124)"
)

expect_equal(
  mtcars %>% group_by(am) %>% nest_by(),
  {
    res <- data.frame(am = c(0L, 1L))
    res[["data"]] <- list(mtcars[mtcars$am == 0, -9], mtcars[mtcars$am == 1, -9])
    res %>% group_by(am)
  },
  info = "nest_by() works on grouped data sets"
)

expect_error(
  mtcars %>% nest_by(am, .keep = "true"),
  info = "`.keep` should be `logical(1)`"
)

expect_error(
  mtcars %>% nest_by(am, .key = TRUE),
  info = "`.key` should be `character(1)`"
)

expect_error(
  mtcars %>% group_by(am) %>% nest_by(cyl),
  info = "Can't re-group while nesting. Either `ungroup()` first or don't supply arguments to `nest_by()`"
)

#' Compare two numeric vectors
#'
#' This is a safe way of comparing if two vectors of floating point numbers are (pairwise) equal.  This is safer than
#' using `==`, because it has a built in tolerance.
#'
#' @param x,y Numeric vectors to compare
#' @param tol Tolerance of comparison.
#'
#' @examples
#' sqrt(2) ^ 2 == 2
#' near(sqrt(2) ^ 2, 2)
#'
#' @export
near <- function(x, y, tol = .Machine$double.eps^0.5) {
  abs(x - y) < tol
}

#' Determine the grouping structure of the data
#' @param x A `data.frame`
#' @param groups `character(n)`. The names of the grouping columns.
#' @return If `groups` is `NULL` or of length 0, then `x`; otherwise `x` with the attribute `"groups"`.
#' @noRd
groups_set <- function(x, groups, drop = group_by_drop_default(x)) {
  attr(x, "groups") <- if (is.null(groups) || length(groups) == 0L) {
    NULL
  } else {
    calculate_groups(x, groups, drop)
  }
  x
}

#' Get group details
#' @param x A `data.frame`.
#' @return A `data.frame` containing the grouping columns where each row represents a single combination of groups and
#' a column called `.rows` which contains the row numbers each group combination appears on.
#' @noRd
get_group_details <- function(x) {
  groups <- attr(x, "groups", exact = TRUE)
  if (is.null(groups)) character(0) else groups
}

#' Check if an R object has groups
#' @return `logical(1)`
#' @noRd
has_groups <- function(x) {
  groups <- group_vars(x)
  if (length(groups) == 0L) FALSE else TRUE
}

#' @param fn `character(1)`. The function to apply to each group.
#' @param .data A `data.frame`.
#' @param ... Arguments to be passed to `fn`.
#' @noRd
apply_grouped_function <- function(fn, .data, drop = FALSE, ...) {
  groups <- group_vars(.data)
  grouped <- split_into_groups(.data, groups, drop)
  res <- do.call(rbind, unname(lapply(grouped, fn, ...)))
  if (any(groups %in% colnames(res))) {
    class(res) <- c("grouped_df", class(res))
    res <- groups_set(res, groups[groups %in% colnames(res)])
  }
  res
}

#' Calculate the groups
#' @param data A `data.frame`.
#' @param groups `character(n)`. The names of the groups.
#' @return
#' The columns give the values of the grouping variables. The last column, always called `.rows`, is a list of integer
#' vectors that gives the location of the rows in each group.
#' @noRd
calculate_groups <- function(data, groups, drop = group_by_drop_default(data)) {
  data <- ungroup(data)

  unknown <- setdiff(groups, colnames(data))
  if (length(unknown) > 0L) {
    stop(sprintf("`groups` missing from `data`: %s.", paste0(groups, collapse = ", ")))
  }

  unique_groups <- unique(data[, groups, drop = FALSE])
  is_factor <- do.call(c, lapply(unique_groups, function(x) is.factor(x)))
  n_comb <- nrow(unique_groups)
  rows <- rep(list(NA), n_comb)
  data_groups <- interaction(data[, groups, drop = TRUE])
  for (i in seq_len(n_comb)) {
    rows[[i]] <- which(data_groups %in% interaction(unique_groups[i, groups]))
  }

  if (!isTRUE(drop) && any(is_factor)) {
    na_lvls <- do.call(
      expand.grid,
      lapply(unique_groups, function(x) if (is.factor(x)) levels(x)[!(levels(x) %in% x)] else NA)
    )
    unique_groups <- rbind(unique_groups, na_lvls)
    for (i in seq_len(nrow(na_lvls))) {
      rows[[length(rows) + 1]] <- integer(0)
    }
  }

  unique_groups[[".rows"]] <- rows
  unique_groups <- unique_groups[do.call(order, lapply(groups, function(x) unique_groups[, x])), , drop = FALSE]
  rownames(unique_groups) <- NULL
  structure(unique_groups, .drop = drop)
}

is.grouped_df <- function(x) {
  inherits(x, "grouped_df")
}

#' Convert values to NA
#'
#' This is a translation of the SQL command `NULLIF`. It is useful if you want to convert an annoying value to `NA`.
#'
#' @param x The vector to modify.
#' @param y The value to replace with `NA`.
#'
#' @seealso
#' [coalesce()] to replace missing values within subsequent `vector`(s) of value(s). [replace_na()] to replace `NA` with
#' a value.
#'
#' [replace_na()] to replace `NA` with a value.
#'
#' [recode()] to more generally replace values.
#'
#' @return
#' A modified version of `x` that replaces any values that are equal to `y` with `NA`.
#'
#' @examples
#' na_if(1:5, 5:1)
#'
#' x <- c(1, -1, 0, 10)
#' 100 / x
#' 100 / na_if(x, 0)
#'
#' y <- c("abc", "def", "", "ghi")
#' na_if(y, "")
#'
#' # na_if() is particularly useful inside mutate(),
#' # and is meant for use with vectors rather than entire data.frames
#' mtcars %>%
#'   mutate(cyl = na_if(cyl, 6))
#'
#' @export
na_if <- function(x, y) {
  y_len <- length(y)
  x_len <- length(x)
  if (!(y_len %in% c(1L, x_len))) stop("`y` must be length ", x_len, " (same as `x`) or 1, not ", y_len)
  x[x == y] <- NA
  x
}

x <- c(0, 1, 0)
expect_equal(
  na_if(x, 0),
  c(NA, 1, NA),
  info = "scalar y replaces all matching x"
)
expect_equal(
  na_if(x, 1),
  c(0, NA, 0),
  info = "scalar y replaces all matching x"
)

expect_error(
  na_if(x, 1:10),
  info = "x and y must be the same length"
)
expect_error(
  na_if(1, 1:2),
  info = "x and y must be the same length"
)

#' Mutating Joins
#'
#' @description
#' The mutating joins add columns from `y` to `x`, matching rows based on the keys:
#'
#' * `inner_join()`: includes all rows in `x` and `y`.
#' * `left_join()`: includes all rows in `x`.
#' * `right_join()`: includes all rows in `y`.
#' * `full_join()`: includes all rows in `x` or `y`.
#'
#' If a row in `x` matches multiple rows in `y`, all the rows in `y` will be returned once for each matching row in `x`.
#'
#' @param x,y The `data.frame`s to join.
#' @param by A character vector of variables to join by. If `NULL`, the default, `*_join()` will do a natural join,
#' using all variables with common names across the two tables. A message lists the variables so that you can check
#' they're right (to suppress the message, simply explicitly list the variables that you want to join).
#'
#' To join by different variables on x and y use a named vector. For example, `by = c("a" = "b")` will match `x.a` to
#' `y.b`.
#'
#' To join by multiple variables, use a vector with length > 1. For example, `by = c("a", "b")` will match `x$a` to
#' `y$a` and `x$b` to `y$b`. Use a named vector to match different variables in `x` and `y`. For example,
#' `by = c("a" = "b", "c" = "d")` will match `x$a` to `y$b` and `x$c` to `y$d`.
#'
#' To perform a cross-join, generating all combinations of `x` and `y`, use `by = character()`.
#' @param suffix `character(2)`. If there are non-joined duplicate variables in `x` and `y`, these suffixes will be
#' added to the output to disambiguate them.
#' @param ... Additional arguments to pass to [merge()]
#' @param keep `logical(1)`. Should the join keys from both `x` and `y` be preserved in the output? Only applies to
#' `left_join()`, `right_join()`, and `full_join()`.
#' @param na_matches Should `NA` and `NaN` values match one another?
#'
#' The default, `"na"`, treats two `NA` or `NaN` values as equal, like `%in%`, [match()], [merge()].
#'
#' Use `"never"` to always treat two `NA` or `NaN` values as different, like joins for database sources, similarly to
#' `merge(incomparables = FALSE)`.
#'
#' @return
#' A `data.frame`. The order of the rows and columns of `x` is preserved as much as possible. The output has the
#' following properties:
#'
#' * For `inner_join()`, a subset of `x` rows.
#'   For `left_join()`, all `x` rows.
#'   For `right_join()`, a subset of `x` rows, followed by unmatched `y` rows.
#'   For `full_join()`, all `x` rows, followed by unmatched `y` rows.
#' * For all joins, rows will be duplicated if one or more rows in `x` matches multiple rows in `y`.
#' * Output columns include all `x` columns and all `y` columns. If columns in `x` and `y` have the same name (and
#'   aren't included in `by`), `suffix`es are added to disambiguate.
#' * Output columns included in `by` are coerced to common type across `x` and `y`.
#' * Groups are taken from `x`.
#'
#' @examples
#' # If a row in `x` matches multiple rows in `y`, all the rows in `y` will be
#' # returned once for each matching row in `x`
#' df1 <- data.frame(x = 1:3)
#' df2 <- data.frame(x = c(1, 1, 2), y = c("first", "second", "third"))
#' df1 %>% left_join(df2)
#'
#' # By default, NAs match other NAs so that there are two
#' # rows in the output of this join:
#' df1 <- data.frame(x = c(1, NA), y = 2)
#' df2 <- data.frame(x = c(1, NA), z = 3)
#' left_join(df1, df2)
#'
#' # You can optionally request that NAs don't match, giving a
#' # a result that more closely resembles SQL joins
#' left_join(df1, df2, na_matches = "never")
#'
#' @name mutate_joins
NULL

#' @rdname mutate_joins
#' @export
inner_join <- function(x, y, by = NULL, suffix = c(".x", ".y"), ..., na_matches = c("na", "never")) {
  join_worker(x = x, y = y, by = by, suffix = suffix, sort = FALSE, ..., keep = FALSE, na_matches = na_matches)
}

#' @rdname mutate_joins
#' @export
left_join <- function(x, y, by = NULL, suffix = c(".x", ".y"), ..., keep = FALSE, na_matches = c("na", "never")) {
  join_worker(x = x, y = y, by = by, suffix = suffix, all.x = TRUE, ..., keep = keep, na_matches = na_matches)
}

#' @rdname mutate_joins
#' @export
right_join <- function(x, y, by = NULL, suffix = c(".x", ".y"), ..., keep = FALSE, na_matches = c("na", "never")) {
  join_worker(x = x, y = y, by = by, suffix = suffix, all.y = TRUE, ..., keep = keep, na_matches = na_matches)
}

#' @rdname mutate_joins
#' @export
full_join <- function(x, y, by = NULL, suffix = c(".x", ".y"), ..., keep = FALSE, na_matches = c("na", "never")) {
  join_worker(x = x, y = y, by = by, suffix = suffix, all = TRUE, ..., keep = keep, na_matches = na_matches)
}

join_worker <- function(x, y, by = NULL, suffix = c(".x", ".y"), keep = FALSE, na_matches = c("na", "never"), ...) {
  na_matches <- match.arg(arg = na_matches, choices = c("na", "never"), several.ok = FALSE)
  incomparables <- if (na_matches == "never") NA else NULL
  x[, ".join_id"] <- seq_len(nrow(x))
  merged <- if (is.null(by)) {
    by <- intersect(names(x), names(y))
    join_message(by)
    merge(
      x = x, y = y, by = by, suffixes = suffix, incomparables = incomparables, ...
    )[, union(names(x), names(y)), drop = FALSE]
  } else if (is.null(names(by))) {
    merge(x = x, y = y, by = by, suffixes = suffix, incomparables = incomparables, ...)
  } else {
    merge(x = x, y = y, by.x = names(by), by.y = by, suffixes = suffix, incomparables = incomparables, ...)
  }
  merged <- merged[order(merged[, ".join_id"]), colnames(merged) != ".join_id", drop = FALSE]
  if (isTRUE(keep)) {
    keep_pos <- match(by, names(merged))
    x_by <- paste0(by, suffix[1L])
    colnames(merged)[keep_pos] <- x_by
    merged[, paste0(by, suffix[2L])] <- merged[, x_by]
  }
  rownames(merged) <- NULL
  reconstruct_attrs(merged, x)
}

join_message <- function(by) {
  if (length(by) > 1L) {
    message("Joining, by = c(\"", paste0(by, collapse = "\", \""), "\")\n", sep = "")
  } else {
    message("Joining, by = \"", by, "\"\n", sep = "")
  }
}

df <- data.frame(
  e = 1,
  f = factor(c(1, 1, 2, 2), levels = 1:3),
  g = c(1, 1, 2, 2),
  x = c(1, 2, 1, 4)
) %>%
  group_by(e, f, g, .drop = FALSE)

expect_equal(group_size(filter(df, f == 1)), c(2, 0, 0), info = "filter keeps zero length groups")
expect_equal(group_size(slice(df, 1)), c(1, 1, 0), info = "slice keeps zero length groups")

# expect_equal(
#   ungroup(count(filter(df, f == 1))),
#   data.frame(e = 1, f = factor(1:3), g = c(1, 2, NA), n = c(2L, 0L, 0L)),
#   info = "filtering retains labels for zero length groups"
# )

# expect_equal(
#   ungroup(count(slice(df, 1))),
#   data.frame(e = 1, f = factor(1:3), g = c(1, 2, NA), n = c(1L, 1L, 0L)),
#   info = "slicing retains labels for zero length groups"
# )

expect_equal(group_size(mutate(df, z = 2)), c(2, 2, 0), info = "mutate keeps zero length groups")

# expect_equal(nrow(summarise(df, z = n())), 3L, info = "summarise returns a row for zero length groups")

# expect_equal(group_size(arrange(df)), c(2, 2, 0), info = "arrange keeps zero length groups")
# expect_equal(group_size(arrange(df, x)), c(2, 2, 0), info = "arrange keeps zero length groups")

# gg <- bind_rows(df, df)
# expect_equal(group_size(gg), c(4L, 4L, 0L), info = "bind_rows respect the drop attribute of grouped df")

# test_that("joins respect zero length groups", {
#   df1 <- data.frame(f = factor( c(1,1,2,2), levels = 1:3), x = c(1,2,1,4)) %>%
#     group_by(f)

#   df2 <- data.frame(f = factor( c(2,2,3,3), levels = 1:3), y = c(1,2,3,4)) %>%
#     group_by(f)

#   expect_equal(group_size(left_join( df1, df2, by = "f")),  c(2,4))
#   expect_equal(group_size(right_join( df1, df2, by = "f")),  c(4,2))
#   expect_equal(group_size(full_join( df1, df2, by = "f")),  c(2,4,2))
#   expect_equal(group_size(anti_join( df1, df2, by = "f")),  c(2))
#   expect_equal(group_size(inner_join( df1, df2, by = "f")),  c(4))


#   df1 <- data.frame(f = factor( c(1,1,2,2), levels = 1:3), x = c(1,2,1,4)) %>%
#     group_by(f, .drop = FALSE)
#   df2 <- data.frame(f = factor( c(2,2,3,3), levels = 1:3), y = c(1,2,3,4)) %>%
#     group_by(f, .drop = FALSE)

#   expect_equal(group_size(left_join( df1, df2, by = "f")),  c(2,4,0))
#   expect_equal(group_size(right_join( df1, df2, by = "f")),  c(0,4,2))
#   expect_equal(group_size(full_join( df1, df2, by = "f")),  c(2,4,2))
#   expect_equal(group_size(anti_join( df1, df2, by = "f")),  c(2,0,0))
#   expect_equal(group_size(inner_join( df1, df2, by = "f")),  c(0,4,0))
# })

df <- data.frame(x = factor(1:3, levels = 1:4)) %>% group_by(x, .drop = FALSE)
expect_equal(n_groups(df), 4, info = "n_groups respects zero-length groups")

# df <- data.frame(x = factor(rep(1:3, each = 10), levels = 1:4))
# out <- df %>%
#   group_by(x, .drop = FALSE) %>%
#   summarise(n = n())
# expect_equal(out$n, c(10L, 10L, 10L, 0L), info = "summarise respects zero-length groups")

df1 <- data.frame(
  id = c(1, 3, 5, 7, 9),
  name = LETTERS[1:5],
  height = c(1, 2, 2, 2, 2),
  stringsAsFactors = FALSE
)

df2 <- data.frame(
  id = c(2, 4, 5, 7),
  name = LETTERS[1:4],
  weight = c(2, 3, 4, 5),
  stringsAsFactors = FALSE
)

expect_equal(
  {
    invisible(suppressMessages(res <- df1 %>% inner_join(df2)))
    res
  },
  data.frame(
    id = c(5, 7),
    name = LETTERS[3:4],
    height = c(2, 2),
    weight = c(4, 5),
    stringsAsFactors = FALSE
  ),
  info = "Inner join"
)

expect_equal(
  {
    invisible(suppressMessages(res <- df1 %>% left_join(df2)))
    res
  },
  data.frame(
    id = c(1, 3, 5, 7, 9),
    name = LETTERS[1:5],
    height = c(1, 2, 2, 2, 2),
    weight = c(NA, NA, 4, 5, NA),
    stringsAsFactors = FALSE
  ),
  info = "Left join"
)

expect_equal(
  {
    invisible(suppressMessages(res <- df1 %>% right_join(df2)))
    res
  },
  data.frame(
    id = c(5, 7, 2, 4),
    name = LETTERS[c(3, 4, 1, 2)],
    height = c(2, 2, NA, NA),
    weight = c(4, 5, 2, 3),
    stringsAsFactors = FALSE
  ),
  info = "Right join"
)

expect_equal(
  {
    invisible(suppressMessages(res <- df1 %>% full_join(df2)))
    res
  },
  data.frame(
    id = c(1, 3, 5, 7, 9, 2, 4),
    name = c("A", "B", "C", "D", "E", "A", "B"),
    height = c(1, rep(2, 4), NA, NA),
    weight = c(NA, NA, 4, 5, NA, 2, 3),
    stringsAsFactors = FALSE
  ),
  info = "Full join"
)

df1 <- data.frame(
  a = c(1, 2, 3, 4, 5, 6),
  b = c("a", "b", "f", "e", "r", "h"),
  d = c("q", "l", "o", "n", "q", "z"),
  stringsAsFactors = FALSE
)

df2 <- data.frame(
  a = c(1, 2, 3, 4, 5, 6),
  d = c("q", "l", "o", "n", "q", "z"),
  stringsAsFactors = FALSE
)

expect_equal(
  df1 %>% left_join(df2, by = c("a", "d")),
  data.frame(
    a = c(1, 2, 3, 4, 5, 6),
    d = c("q", "l", "o", "n", "q", "z"),
    b = c("a", "b", "f", "e", "r", "h"),
    stringsAsFactors = FALSE
  ),
  info = "Explicit `by` parameter"
)

expect_equal(
  df1 %>% left_join(df2, by = "a", suffix = c("_x", "_y")) %>% colnames(),
  c("a", "b", "d_x", "d_y"),
  info = "Suffix changes"
)

df1 <- data.frame(
  a1 = c(1, 2, 3, 4, 5, 6),
  b = c("a", "b", "f", "e", "r", "h"),
  d = c("q", "l", "o", "n", "q", "z"),
  stringsAsFactors = FALSE
)

df2 <- data.frame(
  a2 = c(1, 2, 3, 4, 5, 6),
  d = c("q", "l", "o", "n", "q", "z"),
  stringsAsFactors = FALSE
)

expect_equal(
  df1 %>% left_join(df2, by = c("a1" = "a2", "d" = "d")),
  data.frame(
    a1 = c(1, 2, 3, 4, 5, 6),
    d = c("q", "l", "o", "n", "q", "z"),
    b = c("a", "b", "f", "e", "r", "h"),
    stringsAsFactors = FALSE
  ),
  info = "Non-matching column names"
)

df1 <- data.frame(x = c(1, NA), y = 2)
df2 <- data.frame(x = c(1, NA), z = 3)
expect_equal(
  invisible(suppressMessages(res <- full_join(df1, df2, suffix = c(".xx", ".yy"), keep = TRUE, na_matches = "never"))),
  data.frame(
    x.xx = c(1, NA, NA),
    y = c(2, 2, NA),
    z = c(3, NA, 3),
    x.yy = c(1, NA, NA)
  ),
  info = "keep = TRUE keeps the join columns"
)

#' Select variables with a function
#'
#' This selection helper selects the variables for which a function returns `TRUE`.
#'
#' @param fn A function that returns `TRUE` or `FALSE`.
#'
#' @examples
#' iris %>% select(where(is.numeric))
#' iris %>% select(where(function(x) is.numeric(x)))
#' iris %>% select(where(function(x) is.numeric(x) && mean(x) > 3.5))
#'
#' @return A vector of `integer` column positions which are the result of the `fn` evaluation.
#'
#' @seealso [select_helpers]
#'
#' @export
where <- function(fn) {
  if (!is.function(fn)) {
    stop(deparse_var(fn), " is not a valid predicate function.")
  }
  preds <- unlist(lapply(
    select_env$.data,
    function(x, fn) {
      do.call("fn", list(x))
    },
    fn
  ))
  if (!is.logical(preds)) stop("`where()` must be used with functions that return `TRUE` or `FALSE`.")
  data_cols <- select_env$get_colnames()
  cols <- data_cols[preds]
  which(data_cols %in% cols)
}

expect_error(
  iris %>% select(is.numeric),
  info = "Predicate functions must be wrapped in `where()`"
)

df <- data.frame(
  is.col1 = 1:5, is.col2 = letters[1:5],
  check.names = FALSE,
  stringsAsFactors = FALSE
)
expect_equal(
  df %>% select(is.col1),
  df[, "is.col1", drop = FALSE],
  info = "Columns starting with is. are still selected"
)

expect_error(
  df %>% select(where(is.blah)),
  info = "Non-existent predicates are found out"
)

expect_equal(
  df %>% select(where(is.character)),
  df[, "is.col2", drop = FALSE],
  info = "The user can select columns using predicates"
)

df <- data.frame(a = 1:10, b = 101:110)
expect_equal(
  df %>% select(where(function(x) mean(x) > 10)),
  df[, "b", drop = FALSE],
  info = "Custom functions can be used in where()"
)

expect_error(
  df %>% select(where(function(x) if (mean(x) > 10) "yes")),
  info = "where() must be used with functions that return `TRUE` or `FALSE`"
)

rm(df)

#' Do values in a numeric vector fall in specified range?
#'
#' This is a shortcut for `x >= left & x <= right`.
#'
#' @param x A `numeric` vector of values.
#' @param left,right Boundary values.
#'
#' @examples
#' between(1:12, 7, 9)
#'
#' x <- rnorm(1e2)
#' x[between(x, -1, 1)]
#'
#' @return A `logical` vector the same length as `x`.
#'
#' @export
between <- function(x, left, right) {
  if (!is.null(attr(x, "class")) && !inherits(x, c("Date", "POSIXct"))) {
    warning("`between()` called on numeric vector with S3 class")
  }
  if (!is.double(x)) x <- as.numeric(x)
  x >= as.numeric(left) & x <= as.numeric(right)
}

expect_equal(
  between(1, NA, 1),
  NA,
  info = "Return NA if any argument is NA"
)

x <- runif(1e3)
expect_equal(
  between(x, 0.25, 0.5),
  x >= 0.25 & x <= 0.5,
  info = "Return the expected results"
)
rm(x)

expect_warning(
  between(structure(c(1, 5), class = "foo"), 1, 3),
  "numeric vector with S3 class",
  info = "Expect warning for numeric vectors with a class"
)

expect_warning(
  between(factor("x"), 1, 2),
  info = "Expect a warning for S3 class"
)

expect_silent(between(Sys.Date(), 1, 3), info = "Expect no warning for Date")

expect_silent(between(Sys.time(), 1, 3), info = "Expect no warning for POSIXct")

#' Count the number of unique values in a set of vectors
#'
#' This is the equivalent of `length(unique(x))` for multiple vectors.
#'
#' @param ... Vectors of values.
#' @param na.rm `logical(1)`. If `TRUE` missing values don't count.
#'
#' @examples
#' x <- sample(1:10, 1e5, rep = TRUE)
#' length(unique(x))
#' n_distinct(x)
#'
#' @export
n_distinct <- function(..., na.rm = FALSE) {
  res <- do.call(cbind, list(...))
  if (isTRUE(na.rm)) res <- res[!is.na(res), , drop = FALSE]
  nrow(unique(res))
}

expect_equal(
  sapply(iris, n_distinct),
  sapply(iris, function(.) length(unique(.))),
  info = "n_distinct() gives the correct results on a data.frame"
)

df_var <- data.frame(
  l = c(TRUE, FALSE, FALSE),
  i = c(1, 1, 2),
  d = Sys.Date() + c(1, 1, 2),
  f = factor(letters[c(1, 1, 2)]),
  n = c(1, 1, 2) + 0.5,
  t = Sys.time() + c(1, 1, 2),
  c = letters[c(1, 1, 2)],
  stringsAsFactors = FALSE
)
expect_equal(
  sapply(df_var, n_distinct),
  sapply(df_var, function(.) length(unique(.))),
  info = "n_distinct() gives correct results for key types"
)

expect_equal(
  n_distinct(c(1.0, NA, NA)),
  2,
  info = "n_distinct() treats NA correctly"
)

expect_equal(
  n_distinct(1, 1:4),
  4,
  info = "n_distinct() recycles length 1 vectors"
)
expect_equal(
  n_distinct(1:4, 1),
  4,
  info = "n_distinct() recycles length 1 vectors"
)

x <- iris$Sepal.Length
y <- iris$Sepal.Width
expect_equal(
  n_distinct(iris$Sepal.Length, iris$Sepal.Width),
  n_distinct(x, y),
  info = "n_distinct() handles unnamed"
)

expect_identical(
  data.frame(x = 42) %>% summarise(n = n_distinct(.data$x)),
  data.frame(n = 1L),
  info = "n_distinct() respects .data"
)

rm(df_var, x, y)

expect_equal(
  n_distinct(c(0, 1, 1, 2, 3, 5, 8, NA), na.rm = TRUE),
  6L,
  info = "na.rm removes NAs"
)

expect_equal(
  n_distinct(list(c(1, 1, 2), c(1, 2, 3))),
  2L,
  info = "n_distinct() can handle nested lists"
)

#' Cumulative versions of any, all, and mean
#'
#' {poorman} provides `cumall()`, `cumany()`, and `cummean()` to complete R's set of cumulative functions.
#'
#' @section Cumulative logical functions:
#'
#' These are particularly useful in conjunction with `filter()`:
#'
#' * `cumall(x)`: all cases until the first `FALSE`.
#' * `cumall(!x)`: all cases until the first `TRUE`.
#' * `cumany(x)`: all cases after the first `TRUE`.
#' * `cumany(!x)`: all cases after the first `FALSE`.
#'
#' @param x For `cumall()` and `cumany()`, a logical vector; for `cummean()` an integer or numeric vector.
#'
#' @return A vector the same length as `x`.
#'
#' @examples
#' # `cummean()` returns a numeric/integer vector of the same length
#' # as the input vector.
#' x <- c(1, 3, 5, 2, 2)
#' cummean(x)
#' cumsum(x) / seq_along(x)
#'
#' # `cumall()` and `cumany()` return logicals
#' cumall(x < 5)
#' cumany(x == 3)
#'
#' # `cumall()` vs. `cumany()`
#' df <- data.frame(
#'   date = as.Date("2020-01-01") + 0:6,
#'   balance = c(100, 50, 25, -25, -50, 30, 120)
#' )
#' # all rows after first overdraft
#' df %>% filter(cumany(balance < 0))
#' # all rows until first overdraft
#' df %>% filter(cumall(!(balance < 0)))
#'
#' @export
cummean <- function(x) {
  res <- vector(mode = "numeric", length = length(x))
  for (i in seq_along(x)) {
    res[[i]] <- mean(x[seq(from = 1, to = i, by = 1)])
  }
  res
}

#' @rdname cummean
#' @export
cumany <- function(x) {
  x <- as.logical(x)
  res <- vector(mode = "logical", length = length(x))
  for (i in seq_along(x)) {
    res[[i]] <- any(x[seq(from = 1, to = i, by = 1)] == TRUE)
  }
  res
}

#' @rdname cummean
#' @export
cumall <- function(x) {
  if (any(!is.logical(x))) return(rep(NA, length(x)))
  res <- vector(mode = "logical", length = length(x))
  for (i in seq_along(x)) {
    res[[i]] <- all(x[seq(from = 1, to = i, by = 1)] == TRUE)
  }
  res
}

batman <- c(NA, NA, NA, NA, NA)
expect_true(all(is.na(cumany(batman))), info = "cumany() handles all NA vectors")
expect_true(all(is.na(cumall(batman))), info = "cumall() handles all NA vectors")

expect_identical(
  cumall(c(TRUE, NA, FALSE, NA)),
  c(TRUE, NA, FALSE, FALSE),
  info = "cumall: handles NAs #1"
)

expect_identical(
  cumall(c(FALSE, NA, TRUE)),
  c(FALSE, FALSE, FALSE),
  info = "cumall: handles NAs #2"
)

expect_identical(
  cumall(c(NA, TRUE)),
  c(NA, NA),
  info = "cumall: handles NAs #3"
)

expect_identical(
  cumall(c(NA, FALSE)),
  c(NA, FALSE),
  info = "cumall: handles NAs #4"
)

expect_identical(
  cumany(c(TRUE, NA, FALSE)),
  c(TRUE, TRUE, TRUE),
  info = "cumany: handles NAs #1"
)

expect_identical(
  cumany(c(FALSE, NA, TRUE)),
  c(FALSE, NA, TRUE),
  info = "cumall: handles NAs #2"
)

# scalars
expect_true(is.na(cumall(NA)), info = "cumall: NA")
expect_true(is.na(cumany(NA)), info = "cumany: NA")
expect_true(cumall(TRUE), info = "cumall: TRUE")
expect_false(cumall(FALSE), info = "cumall: FALSE")
expect_true(cumany(TRUE), info = "cumany: TRUE")
expect_false(cumany(FALSE), info = "cumany: FALSE")

# degenerate cases
expect_identical(cumall(logical()), logical(), info = "cumall: empty input returns logical(0)")
expect_identical(cumany(logical()), logical(), info = "cumany: empty input returns logical(0)")

x <- as.raw(c(2L, 9L, 0L))
class(x) <- "logical"
expect_identical(
  cumall(x),
  x == TRUE,
  "cumall: behaviour of degenerate logical vectors mimics that of base R functions"
)
expect_identical(
  cumany(x),
  c(TRUE, TRUE, TRUE),
  "cumany: behaviour of degenerate logical vectors mimics that of base R functions"
)

x <- 1:5
expect_equal(cummean(x), c(1, 1.5, 2, 2.5, 3), info = "cummean returns expected output")
expect_equal(cummean(x), cumsum(x) / seq_along(x), info = "cummean is consistent with cumsum() and seq_along()")
x <- c(1, 3, NA, 5, 2, 2)
expect_equal(cummean(x), c(1, 2, NA, NA, NA, NA), info = "cummean handles NAs correctly")
expect_equal(cummean(numeric()), numeric(), info = "empty input returns numeric(0)")

#' Windowed Rank Functions
#'
#' Six variations on ranking functions, mimicking the ranking functions described in SQL2003. They are currently
#' implemented using the built in [rank()] function. All ranking functions map smallest inputs to smallest outputs. Use
#' `desc()` to reverse the direction.
#'
#' @details
#' * `cume_dist()`: a cumulative distribution function. Proportion of all values less than or equal to the current rank.
#' * `dense_rank()`: like `min_rank()`, but with no gaps between ranks
#' * `min_rank()`: equivalent to `rank(ties.method = "min")`
#' * `ntile()`: a rough rank, which breaks the input vector into `n` buckets. The size of the buckets may differ by up
#'    to one, larger buckets have lower rank.
#' * `percent_rank()`: a number between `0` and `1` computed by rescaling `min_rank` to `[0, 1]`
#' * `row_number()`: equivalent to `rank(ties.method = "first")`
#'
#' @param x A vector of values to rank. Missing values are left as is. If you want to treat them as the smallest or
#' largest values, replace with `Inf` or `-Inf` before ranking.
#' @param n `integer(1)`. The number of groups to split up into.
#'
#' @examples
#' x <- c(5, 1, 3, 2, 2, NA)
#' row_number(x)
#' min_rank(x)
#' dense_rank(x)
#' percent_rank(x)
#' cume_dist(x)
#'
#' ntile(x, 2)
#' ntile(1:8, 3)
#'
#' # row_number can be used with single table verbs without specifying x
#' # (for data frames and databases that support windowing)
#' mutate(mtcars, row_number() == 1L)
#' mtcars %>% filter(between(row_number(), 1, 10))
#'
#' @name window_rank
NULL

#' @rdname window_rank
#' @export
cume_dist <- function(x) {
  rank(x, ties.method = "max", na.last = "keep") / sum(!is.na(x))
}

#' @rdname window_rank
#' @export
dense_rank <- function(x) {
  match(x, sort(unique(x)))
}

#' @rdname window_rank
#' @export
min_rank <- function(x) {
  rank(x, ties.method = "min", na.last = "keep")
}

#' @rdname window_rank
#' @export
ntile <- function(x = row_number(), n) {
  if (!missing(x)) x <- row_number(x)
  len <- length(x) - sum(is.na(x))
  n <- as.integer(floor(n))
  if (len == 0L) {
    rep(NA_integer_, length(x))
  } else {
    n_larger <- as.integer(len %% n)
    size <- len / n
    larger_size <- as.integer(ceiling(size))
    smaller_size <- as.integer(floor(size))
    larger_threshold <- larger_size * n_larger
    bins <- if_else(
      x <= larger_threshold,
      (x + (larger_size - 1L)) / larger_size,
      (x + (-larger_threshold + smaller_size - 1L)) / smaller_size + n_larger
    )
    as.integer(floor(bins))
  }
}

#' @rdname window_rank
#' @export
percent_rank <- function(x) {
  (min_rank(x) - 1) / (sum(!is.na(x)) - 1)
}

#' @rdname window_rank
#' @export
row_number <- function(x) {
  if (missing(x)) seq_len(n()) else rank(x, ties.method = "first", na.last = "keep")
}

expect_equal(
  cume_dist(c(1:3, NA)),
  c(1 / 3, 2 / 3, 1, NA),
  info = "cume_dist() ignores NAs"
)

expect_equal(
  ntile(c(1:3, NA, NA, NA), 3),
  c(1:3, NA, NA, NA),
  info = "ntile() ignores number of NAs"
)
expect_equal(
  ntile(c(1:3, NA, NA, NA), 1),
  c(1L, 1L, 1L, NA, NA, NA),
  info = "ntile() ignores number of NAs"
)

expect_equal(
  ntile(numeric(), 3),
  integer(),
  info = "ntile() always returns an integer"
)
expect_equal(
  ntile(NA, 3),
  NA_integer_,
  info = "ntile() always returns an integer"
)

expect_equal(ntile(1, 5), 1, info = "ntile() puts large groups first")
expect_equal(ntile(1:2, 5), 1:2, info = "ntile() puts large groups first")
expect_equal(ntile(1:3, 5), 1:3, info = "ntile() puts large groups first")
expect_equal(ntile(1:4, 5), 1:4, info = "ntile() puts large groups first")
expect_equal(ntile(1:5, 5), 1:5, info = "ntile() puts large groups first")
expect_equal(ntile(1:6, 5), c(1, 1:5), info = "ntile() puts large groups first")
expect_equal(ntile(1, 7), 1, info = "ntile() puts large groups first")
expect_equal(ntile(1:2, 7), 1:2, info = "ntile() puts large groups first")
expect_equal(ntile(1:3, 7), 1:3, info = "ntile() puts large groups first")
expect_equal(ntile(1:4, 7), 1:4, info = "ntile() puts large groups first")
expect_equal(ntile(1:5, 7), 1:5, info = "ntile() puts large groups first")
expect_equal(ntile(1:6, 7), 1:6, info = "ntile() puts large groups first")
expect_equal(ntile(1:7, 7), 1:7, info = "ntile() puts large groups first")
expect_equal(ntile(1:8, 7), c(1, 1:7), info = "ntile() puts large groups first")

expect_equal(
  percent_rank(c(1:3, NA)),
  c(0, 0.5, 1, NA),
  info = "percent_rank() ignores NAs"
)

df <- data.frame(
  x = 1:10, y = seq(1, 10, by = 1),
  g = rep(c(1, 2), each = 5), s = c(letters[1:3], LETTERS[1:5], letters[4:5])
)
expect_equal(
  mutate(df, rank = min_rank(desc(x)))$rank,
  10:1,
  info = "desc() is correctly handled by window functions"
)
expect_equal(
  mutate(group_by(df, g), rank = min_rank(desc(x)))$rank,
  rep(5:1, 2),
  info = "desc() is correctly handled by window functions"
)
expect_equal(
  mutate(df, rank = row_number(desc(x)))$rank,
  10:1,
  info = "desc() is correctly handled by window functions"
)
expect_equal(
  mutate(group_by(df, g), rank = row_number(desc(x)))$rank,
  rep(5:1, 2),
  info = "desc() is correctly handled by window functions"
)

expect_equal(
  {
    data.frame(
      id = rep(c(1, 2), each = 5), value = c(1, 1, 2, 5, 0, 6, 4, 0, 0, 2),
      s = c(letters[1:2], LETTERS[1:4], letters[2:5])
    ) %>%
      group_by(id) %>%
      mutate(var = row_number(value)) %>%
      pull(var)
  },
  c(2, 3, 4, 5, 1, 5, 4, 1, 2, 3),
  info = "row_number() gives correct results"
)

g <- group_by(mtcars, cyl)
expect_equal(
  mutate(g, rn = row_number()),
  mutate(g, rn = 1:n()),
  info = "row_number() works with 0 arguments"
)

expect_true(
  {
    test <- data.frame(
      Name = c("a", "b", "c", "d", "e"),
      ID = c(1, 1, 1, 1, 1),
      expression = c(NaN, NaN, NaN, NaN, NaN)
    ) %>%
      group_by(ID) %>%
      mutate(rank = min_rank(expression))
    all(is.na(test$rank))
  },
  info = "min_rank() handles columns full of NaN"
)

df <- data.frame(x = 1:42)
expect_identical(
  mutate(df, nt = ntile(n = 9)),
  mutate(df, nt = ntile(row_number(), n = 9)),
  info = "ntile() works with one argument"
)
df <- group_by(data.frame(x = 1:42, g = rep(1:7, each = 6)), g)
expect_identical(
  mutate(df, nt = ntile(n = 4)),
  mutate(df, nt = ntile(row_number(), n = 4)),
  info = "ntile() works with one argument"
)

data <- data.frame(x = c(1, 2, NA, 1, 0, NA))
res <- data %>%
  mutate(
    min_rank = min_rank(x),
    percent_rank = percent_rank(x),
    dense_rank = dense_rank(x),
    cume_dist = cume_dist(x),
    ntile = ntile(x, 2),
    row_number = row_number(x)
  )
expect_true(all(is.na(res$min_rank[c(3, 6)])), info = "min_rank() deals correctly with NA")
expect_equal(
  res$min_rank[c(1, 2, 4, 5)],
  c(2L, 4L, 2L, 1L),
  info = "min_rank() deals correctly with NA"
)
expect_true(all(is.na(res$dense_rank[c(3, 6)])), info = "dense_rank() deals correctly with NA")
expect_equal(
  res$dense_rank[c(1, 2, 4, 5)],
  c(2L, 3L, 2L, 1L),
  info = "dense_rank() deals correctly with NA"
)
expect_true(all(is.na(res$percent_rank[c(3, 6)])), info = "percent_rank() deals correctly with NA")
expect_equal(
  res$percent_rank[c(1, 2, 4, 5)],
  c(1 / 3, 1, 1 / 3, 0),
  info = "percent_rank() deals correctly with NA"
)
expect_true(all(is.na(res$cume_dist[c(3, 6)])), info = "cume_dist() deals correctly with NA")
expect_equal(
  res$cume_dist[c(1, 2, 4, 5)],
  c(.75, 1, .75, .25),
  info = "cume_dist() deals correctly with NA"
)
expect_true(all(is.na(res$ntile[c(3, 6)])), info = "ntile() deals correctly with NA")
expect_equal(
  res$ntile[c(1, 2, 4, 5)],
  c(1L, 2L, 2L, 1L),
  info = "ntile() deals correctly with NA"
)
expect_true(all(is.na(res$row_number[c(3, 6)])), info = "row_number() deals correctly with NA")
expect_equal(
  res$row_number[c(1, 2, 4, 5)],
  c(2L, 4L, 3L, 1L),
  info = "row_number() deals correctly with NA"
)
data <- data.frame(
  x = rep(c(1, 2, NA, 1, 0, NA), 2),
  g = rep(c(1, 2), each = 6)
)
res <- data %>%
  group_by(g) %>%
  mutate(
    min_rank = min_rank(x),
    percent_rank = percent_rank(x),
    dense_rank = dense_rank(x),
    cume_dist = cume_dist(x),
    ntile = ntile(x, 2),
    row_number = row_number(x)
  )
expect_true(all(is.na(res$min_rank[c(3, 6, 9, 12)])), info = "min_rank() deals correctly with NA")
expect_equal(
  res$min_rank[c(1, 2, 4, 5, 7, 8, 10, 11)],
  rep(c(2L, 4L, 2L, 1L), 2),
  info = "min_rank() deals correctly with NA"
)
expect_true(all(is.na(res$dense_rank[c(3, 6, 9, 12)])), info = "dense_rank() deals correctly with NA")
expect_equal(
  res$dense_rank[c(1, 2, 4, 5, 7, 8, 10, 11)],
  rep(c(2L, 3L, 2L, 1L), 2),
  info = "dense_rank() deals correctly with NA"
)
expect_true(all(is.na(res$percent_rank[c(3, 6, 9, 12)])), info = "percent_rank() deals correctly with NA")
expect_equal(
  res$percent_rank[c(1, 2, 4, 5, 7, 8, 10, 11)],
  rep(c(1 / 3, 1, 1 / 3, 0), 2),
  info = "percent_rank() deals correctly with NA"
)
expect_true(all(is.na(res$cume_dist[c(3, 6, 9, 12)])), info = "cume_dist() deals correctly with NA")
expect_equal(
  res$cume_dist[c(1, 2, 4, 5, 7, 8, 10, 11)],
  rep(c(.75, 1, .75, .25), 2),
  info = "cume_dist() deals correctly with NA"
)
expect_true(all(is.na(res$ntile[c(3, 6, 9, 12)])), info = "ntile() deals correctly with NA")
expect_equal(
  res$ntile[c(1, 2, 4, 5, 7, 8, 10, 11)],
  rep(c(1L, 2L, 2L, 1L), 2),
  info = "ntile() deals correctly with NA"
)
expect_true(all(is.na(res$row_number[c(3, 6, 9, 12)])), info = "row_number() deals correctly with NA")
expect_equal(
  res$row_number[c(1, 2, 4, 5, 7, 8, 10, 11)],
  rep(c(2L, 4L, 3L, 1L), 2),
  info = "row_number() deals correctly with NA"
)

rm(data, df, g, res, test)

#' Select Grouping Variables
#'
#' This selection helper matches grouping variables. It can be used within [select()] and [relocate()] selections.
#'
#' @seealso
#' [groups()] and [group_vars()] for retrieving the grouping variables outside selection contexts.
#'
#' @examples
#' mtcars %>% group_by(am, cyl) %>% select(group_cols())
#'
#' @export
group_cols <- function() {
  match(group_vars(select_env$.data), select_env$get_colnames())
}

#' A General Vetorised `if()`
#'
#' This function allows you to vectorise multiple [if_else()] statements. It is an R equivalent of the SQL `CASE WHEN`
#' statement. If no cases match, `NA` is returned.
#'
#' @param ... A sequence of two-sided formulas. The left hand side (LHS) determines which values match this case. The
#' right hand side (RHS) provides the replacement value.
#'
#' The LHS must evaluate to a logical vector. The RHS does not need to be logical, but all RHSs must evaluate to the
#' same type of vector.
#'
#' Both LHS and RHS may have the same length of either 1 or n. The value of n must be consistent across all cases. The
#' case of `n == 0` is treated as a variant of `n != 1`.
#'
#' `NULL` inputs are ignored.
#'
#' @return
#' A vector of length 1 or n, matching the length of the logical input or output vectors, with the type (and attributes)
#' of the first RHS. Inconsistent lengths or types will generate an error.
#'
#' @examples
#' x <- 1:50
#' case_when(
#'   x %% 35 == 0 ~ "fizz buzz",
#'   x %% 5 == 0 ~ "fizz",
#'   x %% 7 == 0 ~ "buzz",
#'   TRUE ~ as.character(x)
#' )
#'
#' # Like an if statement, the arguments are evaluated in order, so you must
#' # proceed from the most specific to the most general. This won't work:
#' case_when(
#'   TRUE ~ as.character(x),
#'   x %%  5 == 0 ~ "fizz",
#'   x %%  7 == 0 ~ "buzz",
#'   x %% 35 == 0 ~ "fizz buzz"
#' )
#'
#' # If none of the cases match, NA is used:
#' case_when(
#'   x %%  5 == 0 ~ "fizz",
#'   x %%  7 == 0 ~ "buzz",
#'   x %% 35 == 0 ~ "fizz buzz"
#' )
#'
#' # Note that NA values in the vector x do not get special treatment. If you want
#' # to explicitly handle NA values you can use the `is.na` function:
#' x[2:4] <- NA_real_
#' case_when(
#'   x %% 35 == 0 ~ "fizz buzz",
#'   x %% 5 == 0 ~ "fizz",
#'   x %% 7 == 0 ~ "buzz",
#'   is.na(x) ~ "nope",
#'   TRUE ~ as.character(x)
#' )
#'
#' # All RHS values need to be of the same type. Inconsistent types will throw an error.
#' # This applies also to NA values used in RHS: NA is logical, use
#' # typed values like NA_real_, NA_complex, NA_character_, NA_integer_ as appropriate.
#' case_when(
#'   x %% 35 == 0 ~ NA_character_,
#'   x %% 5 == 0 ~ "fizz",
#'   x %% 7 == 0 ~ "buzz",
#'   TRUE ~ as.character(x)
#' )
#' case_when(
#'   x %% 35 == 0 ~ 35,
#'   x %% 5 == 0 ~ 5,
#'   x %% 7 == 0 ~ 7,
#'   TRUE ~ NA_real_
#' )
#'
#' # case_when() evaluates all RHS expressions, and then constructs its
#' # result by extracting the selected (via the LHS expressions) parts.
#' # In particular NaN are produced in this case:
#' y <- seq(-2, 2, by = .5)
#' case_when(
#'   y >= 0 ~ sqrt(y),
#'   TRUE   ~ y
#' )
#'
#' \dontrun{
#' case_when(
#'   x %% 35 == 0 ~ 35,
#'   x %% 5 == 0 ~ 5,
#'   x %% 7 == 0 ~ 7,
#'   TRUE ~ NA
#' )
#' }
#'
#' # case_when is particularly useful inside mutate when you want to
#' # create a new variable that relies on a complex combination of existing
#' # variables
#' mtcars %>%
#'   mutate(
#'     efficient = case_when(
#'       mpg > 25 ~ TRUE,
#'       TRUE ~ FALSE
#'     )
#'   )
#'
#' @export
case_when <- function(...) {
  fs <- list(...)
  lapply(fs, function(x) if (!inherits(x, "formula")) stop("`case_when()` requires formula inputs."))
  n <- length(fs)
  if (n == 0L) stop("No cases provided.")
  query <- vector("list", n)
  value <- vector("list", n)
  default_env <- parent.frame()
  for (i in seq_len(n)) {
    query[[i]] <- eval(fs[[i]][[2]], envir = default_env)
    value[[i]] <- eval(fs[[i]][[3]], envir = default_env)
    if (!is.logical(query[[i]])) stop(fs[[i]][[2]], " does not return a `logical` vector.")
  }
  m <- validate_case_when_length(query, value, fs)
  out <- value[[1]][rep(NA_integer_, m)]
  replaced <- rep(FALSE, m)
  for (i in seq_len(n)) {
    out <- replace_with(out, query[[i]] & !replaced, value[[i]], NULL)
    replaced <- replaced | (query[[i]] & !is.na(query[[i]]))
  }
  out
}

validate_case_when_length <- function(query, value, fs) {
  lhs_lengths <- lengths(query)
  rhs_lengths <- lengths(value)
  all_lengths <- unique(c(lhs_lengths, rhs_lengths))
  if (length(all_lengths) <= 1L) return(all_lengths[[1L]])
  non_atomic_lengths <- all_lengths[all_lengths != 1L]
  len <- non_atomic_lengths[[1L]]
  if (length(non_atomic_lengths) == 1L) return(len)
  inconsistent_lengths <- non_atomic_lengths[-1L]
  lhs_problems <- lhs_lengths %in% inconsistent_lengths
  rhs_problems <- rhs_lengths %in% inconsistent_lengths
  problems <- lhs_problems | rhs_problems
  if (any(problems)) {
    stop(
      "The following formulas must be length ", len, " or 1, not ",
      paste(inconsistent_lengths, collapse = ", "), ".\n    ",
      paste(fs[problems], collapse = "\n    ")
    )
  }
}

x <- 1:3
expect_equal(
  case_when(
    x <= 1 ~ 1,
    x <= 2 ~ 2,
    x <= 3 ~ 3
  ),
  c(1, 2, 3),
  info = "case_when() matches values in order"
)

x <- 1:3
expect_equal(
  case_when(
    x <= 1 ~ 1,
    x <= 2 ~ 2
  ),
  c(1, 2, NA),
  info = "case_when() unmatched gets missing value"
)

x <- c(1:3, NA)
expect_equal(
  case_when(
    x <= 1 ~ 1,
    x <= 2 ~ 2,
    is.na(x) ~ 0
  ),
  c(1, 2, NA, 0),
  info = "case_when() missing values can be replaced"
)

expect_equal(
  case_when(
    c(TRUE, FALSE, NA) ~ 1:3,
    TRUE ~ 4L
  ),
  c(1L, 4L, 4L),
  info = "case_when() NA conditions"
)

expect_equal(
  case_when(
    TRUE ~ 1:3,
    FALSE ~ 4:6
  ),
  1:3,
  info = "case_when() atomic conditions 1"
)
expect_equal(
  case_when(
    NA ~ 1:3,
    TRUE ~ 4:6
  ),
  4:6,
  info = "case_when() atomic conditions 2"
)

expect_equal(
  case_when(
    TRUE ~ integer(),
    FALSE ~ integer()
  ),
  integer(),
  info = "case_when() zero-length conditions and values 1"
)
expect_equal(
  case_when(
    logical() ~ 1,
    logical() ~ 2
  ),
  numeric(),
  info = "case_when() zero-length conditions and values 2"
)

res <- data.frame(a = 1:3) %>%
  mutate(b = (function(x) case_when(x < 2 ~ TRUE, TRUE ~ FALSE))(a)) %>%
  pull()
expect_equal(res, c(TRUE, FALSE, FALSE), info = "case_when() can be used in anonymous functions")

res <- mtcars %>% mutate(efficient = case_when(mpg > 25 ~ TRUE, TRUE ~ FALSE)) %>% pull()
expect_equal(
  res,
  c(
    FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
    FALSE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE
  ),
  info = "case_when() can be used inside mutate()"
)

expect_error(
  case_when(c(TRUE, FALSE) ~ 1:3, c(FALSE, TRUE) ~ 1:2),
  info = "case_when() must return the correct amount of RHS values."
)

expect_error(
  case_when(50 ~ 1:3),
  info = "case_when() must return a logical vector."
)

expect_error(
  case_when(paste(50)),
  info = "case_when() requires formula inputs."
)

expect_error(
  case_when(),
  info = "No cases provided."
)

#' Fill in missing values with previous or next value
#'
#' Fills missing values in selected columns using the next or previous entry. This is useful in the common output format
#' where values are not repeated, and are only recorded when they change.
#'
#' Missing values are replaced in atomic vectors; `NULL`s are replaced in lists.
#'
#' @param data A `data.frame`.
#' @param ... Columns to fill.
#' @param .direction Direction in which to fill missing values. Currently either `"down"` (the default), `"up"`,
#' `"downup"` (i.e. first down and then up) or `"updown"` (first up and then down).
#'
#' @examples
#' # Value (year) is recorded only when it changes
#' sales <- data.frame(
#'   quarter = c(
#'     "Q1", "Q2", "Q3", "Q4", "Q1", "Q2", "Q3", "Q4", "Q1", "Q2",
#'     "Q3", "Q4", "Q1", "Q2", "Q3", "Q4"
#'   ),
#'   year = c(2000, NA, NA, NA, 2001, NA, NA, NA, 2002, NA, NA, NA, 2004, NA, NA, NA),
#'   sales = c(
#'     66013, 69182, 53175, 21001, 46036, 58842, 44568, 50197, 39113, 41668, 30144,
#'     52897, 32129, 67686, 31768, 49094
#'   )
#' )
#'
#' # `fill()` defaults to replacing missing data from top to bottom
#' sales %>% fill(year)
#'
#' # Value (pet_type) is missing above
#' tidy_pets <- data.frame(
#'   rank = c(1L, 2L, 3L, 4L, 5L, 6L, 1L, 2L, 3L, 4L, 5L, 6L),
#'   pet_type = c(NA, NA, NA, NA, NA, "Dog", NA, NA, NA, NA, NA, "Cat"),
#'   breed = c(
#'     "Boston Terrier", "Retrievers (Labrador)", "Retrievers (Golden)",
#'     "French Bulldogs", "Bulldogs", "Beagles", "Persian", "Maine Coon",
#'     "Ragdoll", "Exotic", "Siamese", "American Short"
#'   )
#' )
#'
#' # For values that are missing above you can use `.direction = "up"`
#' tidy_pets %>%
#'   fill(pet_type, .direction = "up")
#'
#' # Value (n_squirrels) is missing above and below within a group
#' squirrels <- data.frame(
#'   group = c(1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3),
#'   name = c(
#'     "Sam", "Mara", "Jesse", "Tom", "Mike", "Rachael", "Sydekea",
#'     "Gabriela", "Derrick", "Kara", "Emily", "Danielle"
#'   ),
#'   role = c(
#'     "Observer", "Scorekeeper", "Observer", "Observer", "Observer",
#'     "Observer", "Scorekeeper", "Observer", "Observer", "Scorekeeper",
#'     "Observer", "Observer"
#'   ),
#'   n_squirrels = c(NA, 8, NA, NA, NA, NA, 14, NA, NA, 9, NA, NA)
#' )
#'
#' # The values are inconsistently missing by position within the group
#' # Use .direction = "downup" to fill missing values in both directions
#' squirrels %>%
#'   group_by(group) %>%
#'   fill(n_squirrels, .direction = "downup") %>%
#'   ungroup()
#'
#' # Using `.direction = "updown"` accomplishes the same goal in this example
#'
#' @name fill
#' @export
fill <- function(data, ..., .direction = c("down", "up", "downup", "updown")) {
  UseMethod("fill")
}

#' @export
fill.data.frame <- function(data, ..., .direction = c("down", "up", "downup", "updown")) {

  col_pos <- select_positions(data, ..., .group_pos = TRUE)
  if (length(col_pos) == 0) return(data)
  .direction <- match.arg(.direction)

  # If "updown", then we apply direction = "up" for all non-NA, and then
  # we apply direction = "down" for all non-NA (and similarly for "downup").
  # So it's not the same as apply "up" and "down" consecutively for each non-NA

  steps <- switch(
    .direction,
    "up" = "up",
    "down" = "down",
    "updown" = c("up", "down"),
    "downup" = c("down", "up")
  )

  names(col_pos) <- NULL

  for (i in col_pos) {

    for (k in steps) {

      # code adapted from https://stackoverflow.com/a/13810615/11598948

      if (all(is.na(data[[i]]))) next

      if (k == "up") data[[i]] <- rev(data[[i]])
      ind <- which(!is.na(data[[i]]))

      if (is.na(data[[i]][1])) {
        ind <- c(1, ind)
      }
      rep_times <- diff(c(ind, length(data[[i]]) + 1))
      data[[i]] <- rep(data[[i]][ind], times = rep_times)
      if (k == "up") data[[i]] <- rev(data[[i]])
    }
  }

  return(data)
}

#' @export
fill.grouped_df <- function(data, ...) {
  context$group_env <- parent.frame(n = 1)
  on.exit(rm(list = c("group_env"), envir = context), add = TRUE)
  rows <- rownames(data)
  res <- apply_grouped_function("fill", data, drop = TRUE, ...)
  res[rows, , drop = FALSE]
}

context <- new.env()

# Data
context$setup <- function(.data) context$.data <- .data
context$get_data <- function() context$.data
context$get_columns <- function(cols) context$.data[, cols, drop = FALSE]
context$cur_column <- NULL
context$get_nrow <- function() nrow(context$.data)
context$get_colnames <- function() colnames(context$.data)
context$is_grouped <- function() has_groups(context$.data)
context$as_env <- function() {
  if (any(is_nested(context$.data))) {
    lapply(as.list(context$.data), function(x) if (is.data.frame(x[[1]])) x[[1]] else x)
  } else {
    context$.data
  }
}
context$group_env <- NULL
context$clean <- function() {
  rm(list = c(".data"), envir = context)
  if (!is.null(context$cur_column)) rm(list = c("cur_column"), envir = context)
}

#' Context dependent expressions
#'
#' @description
#' These functions return information about the "current" group or "current" variable, so only work inside specific
#' contexts like [summarise()] and [mutate()].
#'
#' @section `data.table`:
#' If you're familiar with `data.table`:
#'
#' * `cur_data()` <-> `.SD`
#' * `cur_group_id()` <-> `.GRP`
#' * `cur_group()` <-> `.BY`
#' * `cur_group_rows()` <-> `.I`
#'
#' @examples
#' df <- data.frame(
#'   g = sample(rep(letters[1:3], 1:3)),
#'   x = runif(6),
#'   y = runif(6),
#'   stringsAsFactors = FALSE
#' )
#' gf <- df %>% group_by(g)
#'
#' gf %>% summarise(n = n())
#'
#' gf %>% mutate(id = cur_group_id())
#' gf %>% summarise(row = cur_group_rows())
#' gf %>% summarise(data = list(cur_group()))
#' gf %>% summarise(data = list(cur_data()))
#' gf %>% summarise(data = list(cur_data_all()))
#'
#' gf %>% mutate(across(everything(), ~ paste(cur_column(), round(.x, 2))))
#'
#' @seealso
#' See [group_data()] for equivalent functions that return values for all groups.
#'
#' @name context
NULL

#' @description
#' * `n()` gives the number of observations in the current group.
#' @rdname context
#' @export
n <- function() {
  check_context("`n()`", context$.data)
  context$get_nrow()
}

#' @description
#' * `cur_data()` gives the current data for the current group (excluding grouping variables).
#' @rdname context
#' @export
cur_data <- function() {
  check_context("`cur_data()`", context$.data)
  data <- context$get_data()
  data[, !(colnames(data) %in% group_vars(data)), drop = FALSE]
}

#' @description
#' * `cur_data_all()` gives the current data for the current group (including grouping variables).
#' @rdname context
#' @export
cur_data_all <- function() {
  check_context("`cur_data_all()`", context$.data)
  ungroup(context$get_data())
}

#' @description
#' * `cur_group()` gives the group keys, a single row `data.frame` containing a column for each grouping variable and
#'   its value.
#' @rdname context
#' @export
cur_group <- function() {
  check_context("`cur_group()`", context$.data)
  data <- context$get_data()
  res <- data[1L, group_vars(data), drop = FALSE]
  rownames(res) <- NULL
  res
}

#' @description
#' * `cur_group_id()` gives a unique numeric identifier for the current group.
#' @rdname context
#' @export
cur_group_id <- function() {
  check_context("`cur_group_id()`", context$.data)
  data <- context$get_data()
  res <- data[1L, group_vars(data), drop = FALSE]
  details <- get_group_details(data)
  details[, ".group_id"] <- seq_len(nrow(details))
  res <- suppressMessages(semi_join(details, res))
  res[, ".group_id"]
}

#' @description
#' * `cur_group_rows()` gives the rows the groups appear in the data.
#' @rdname context
#' @export
cur_group_rows <- function() {
  check_context("`cur_group_rows()`", context$.data)
  data <- context$get_data()
  res <- data[1L, group_vars(data), drop = FALSE]
  res <- suppressMessages(semi_join(get_group_details(data), res))
  unlist(res[, ".rows"])
}

#' @description
#' * `cur_column()` gives the name of the current column (in [across()] only).
#' @rdname context
#' @export
cur_column <- function() {
  check_context("`cur_column()`", context$cur_column, "`across`")
  context$cur_column
}

## -- Helpers ----------------------------------------------------------------------------------------------------------

check_context <- function(fn, context, name = NULL) {
  if (is.null(context)) {
    stop(fn, " must only be used inside ", if (is.null(name)) "poorman verbs" else name)
  }
}

df <- data.frame(
  g = c("a", "b", "b", "c", "c", "c"),
  x = c(1, 4, 5, 7, 3, 6),
  y = c(23, 46, 7, 4, 56, 12),
  stringsAsFactors = FALSE
)
gf <- df %>% group_by(g)

expect_equal(
  df %>% summarise(n()),
  structure(list(`n()` = 6L), class = "data.frame", row.names = c(NA, -1L)),
  info = "n() correctly counts the number of rows"
)

expect_equal(
  gf %>% summarise(n = n()) %>% ungroup(),
  structure(list(g = c("a", "b", "c"), n = 1:3), row.names = c(NA, -3L), class = "data.frame"),
  info = "n() works within a grouped context"
)

expect_equal(
  {
    res <- gf %>% mutate(id = cur_group_id()) %>% ungroup()
    res[, "id"]
  },
  c(1L, 2L, 2L, 3L, 3L, 3L),
  info = "cur_group_id() works as expected"
)

expect_equal(
  {
    res <- gf %>% summarise(row = cur_group_rows())
    res[, "row"]
  },
  1L:6L,
  info = "cur_group_rows() works as expected"
)

expect_equal(
  {
    res <- gf %>% summarise(data = list(cur_group()))
    res[, "data"]
  },
  list(
    structure(list(g = "a"), row.names = c(NA, -1L), class = "data.frame"),
    structure(list(g = "b"), row.names = c(NA, -1L), class = "data.frame"),
    structure(list(g = "c"), row.names = c(NA, -1L), class = "data.frame")
  ),
  info = "cur_group() works as expected"
)

expect_equal(
  {
    res <- gf %>% summarise(data = list(cur_data()))
    res[, "data"]
  },
  list(
    structure(list(x = 1, y = 23), class = "data.frame", row.names = 1L),
    structure(list(x = c(4, 5), y = c(46, 7)), class = "data.frame", row.names = 2:3),
    structure(list(x = c(7, 3, 6), y = c(4, 56, 12)), class = "data.frame", row.names = 4:6)
  ),
  info = "cur_data() works as expected"
)

# errors -----------------------------------------------------------------------
expect_error(
  mtcars %>% select(n()),
  info = "`n()` cannot be used outside of the context"
)

expect_error(
  mtcars %>% select(cur_data()),
  info = "`cur_data()` cannot be used outside of the context"
)

expect_error(
  mtcars %>% select(cur_group()),
  info = "`cur_group()` cannot be used outside of the context"
)

expect_error(
  mtcars %>% select(cur_group_id()),
  info = "`cur_group_id()` cannot be used outside of the context"
)

expect_error(
  mtcars %>% select(cur_group_rows()),
  info = "`cur_group_rows()` cannot be used outside of the context"
)

#' Perform an operation with temporary groups
#'
#' This function allows you to modify the grouping variables for a single operation.
#'
#' @param .data A `data.frame`.
#' @param .groups <[`poor-select`][select_helpers]> One or more variables to group by. Unlike [group_by()], you can
#' only group by existing variables, and you can use `poor-select` syntax like `c(x, y, z)` to select multiple
#' variables.
#'
#' Use `NULL` to temporarily **un**group.
#' @param .f A `function` to apply to regrouped data. Supports lambda-style `~` syntax.
#' @param ... Additional arguments passed on to `.f`.
#'
#' @examples
#' df <- data.frame(g = c(1, 1, 2, 2, 3), x = runif(5))
#' df %>% with_groups(g, mutate, x_mean = mean(x))
#' df %>% with_groups(g, ~ mutate(.x, x_mean = mean(x)))
#'
#' df %>%
#'   group_by(g) %>%
#'   with_groups(NULL, mutate, x_mean = mean(x))
#'
#' # NB: grouping can't be restored if you remove the grouping variables
#' df %>%
#'   group_by(g) %>%
#'   with_groups(NULL, mutate, g = NULL)
#'
#' @export
with_groups <- function(.data, .groups, .f, ...) {
  .groups <- eval_select_pos(.data = .data, .cols = substitute(.groups))
  val <- as_symbols(names(.data)[.groups])
  out <- do.call(group_by, c(list(.data = .data), val))
  .f <- as_function(.f)
  out <- .f(out, ...)
  reconstruct_attrs(out, .data)
}

#' Recode values
#'
#' @description
#' This is a vectorised version of [switch()]: you can replace `numeric` values based on their position or their name,
#' and `character` or `factor` values only by their name. This is an S3 generic: `{poorman}` provides methods for
#' `numeric`,  `character`, and `factor`s. For `logical` vectors, use [if_else()]. For more complicated criteria, use
#' [case_when()].
#'
#' You can use `recode()` directly with `factor`s; it will preserve the existing order of levels while changing the
#' values. Alternatively, you can use `recode_factor()`, which will change the order of levels to match the order of
#' replacements.
#'
#' This is a direct port of the `dplyr::recode()` function.
#'
#' @param .x A vector to modify
#' @param ... Replacements. For `character` and `factor` `.x`, these should be named and replacement is based only on
#' their name. For `numeric` `.x`, these can be named or not. If not named, the replacement is done based on position
#' i.e. `.x` represents positions to look for in replacements. See examples.
#'
#' When named, the argument names should be the current values to be replaced, and the argument values should be the new
#' (replacement) values.
#'
#' All replacements must be the same type, and must have either length one or the same length as `.x`.
#' @param .default If supplied, all values not otherwise matched will be given this value. If not supplied and if the
#' replacements are the same type as the original values in `.x`, unmatched values are not changed. If not supplied and
#' if the replacements are not compatible, unmatched values are replaced with `NA`.
#'
#' `.default` must be either length 1 or the same length as `.x`.
#' @param .missing If supplied, any missing values in `.x` will be replaced by this value. Must be either length 1 or
#' the same length as `.x`.
#' @param .ordered `logical(1)`. If `TRUE`, `recode_factor()` creates an ordered `factor`.
#'
#' @return A vector the same length as `.x`, and the same type as
#'   the first of `...`, `.default`, or `.missing`.
#'   `recode_factor()` returns a factor whose levels are in the same order as
#'   in `...`. The levels in `.default` and `.missing` come last.
#'
#' @seealso
#' [na_if()] to replace specified values with a `NA`.
#'
#' [coalesce()] to replace missing values with a specified value.
#'
#' [replace_na()] to replace `NA` with a value.
#'
#' @examples
#' # For character values, recode values with named arguments only. Unmatched
#' # values are unchanged.
#' char_vec <- sample(c("a", "b", "c"), 10, replace = TRUE)
#' recode(char_vec, a = "Apple")
#' recode(char_vec, a = "Apple", b = "Banana")
#'
#' # Use .default as replacement for unmatched values. Note that NA and
#' # replacement values need to be of the same type.
#' recode(char_vec, a = "Apple", b = "Banana", .default = NA_character_)
#'
#' # Throws an error as NA is logical, not character.
#' \dontrun{
#' recode(char_vec, a = "Apple", b = "Banana", .default = NA)
#' }
#'
#' # For numeric values, named arguments can also be used
#' num_vec <- c(1:4, NA)
#' recode(num_vec, `2` = 20L, `4` = 40L)
#'
#' # Or if you don't name the arguments, recode() matches by position.
#' # (Only works for numeric vector)
#' recode(num_vec, "a", "b", "c", "d")
#' # .x (position given) looks in (...), then grabs (... value at position)
#' # so if nothing at position (here 5), it uses .default or NA.
#' recode(c(1, 5, 3), "a", "b", "c", "d", .default = "nothing")
#'
#' # Note that if the replacements are not compatible with .x,
#' # unmatched values are replaced by NA and a warning is issued.
#' recode(num_vec, `2` = "b", `4` = "d")
#' # use .default to change the replacement value
#' recode(num_vec, "a", "b", "c", .default = "other")
#' # use .missing to replace missing values in .x
#' recode(num_vec, "a", "b", "c", .default = "other", .missing = "missing")
#'
#' # For factor values, use only named replacements
#' # and supply default with levels()
#' factor_vec <- factor(c("a", "b", "c"))
#' recode(factor_vec, a = "Apple", .default = levels(factor_vec))
#'
#' # Use recode_factor() to create factors with levels ordered as they
#' # appear in the recode call. The levels in .default and .missing
#' # come last.
#' recode_factor(num_vec, `1` = "z", `2` = "y", `3` = "x")
#' recode_factor(num_vec, `1` = "z", `2` = "y", `3` = "x", .default = "D")
#' recode_factor(num_vec, `1` = "z", `2` = "y", `3` = "x", .default = "D", .missing = "M")
#'
#' # When the input vector is a compatible vector (character vector or
#' # factor), it is reused as default.
#' recode_factor(letters[1:3], b = "z", c = "y")
#' recode_factor(factor(letters[1:3]), b = "z", c = "y")
#'
#' @export
recode <- function(.x, ..., .default = NULL, .missing = NULL) {
  UseMethod("recode")
}

#' @export
recode.numeric <- function(.x, ..., .default = NULL, .missing = NULL) {
  values <- dotdotdot(...)

  nms <- have_name(values)
  if (all(nms)) {
    vals <- as.double(names(values))
  } else if (all(!nms)) {
    vals <- seq_along(values)
  } else {
    stop("Either all values must be named, or none must be named.")
  }

  n <- length(.x)
  template <- find_template(values, .default, .missing)
  res <- template[rep(NA_integer_, n)]
  replaced <- rep(FALSE, n)

  for (i in seq_along(values)) {
    res <- replace_with(res, .x == vals[i], values[[i]], paste0("Vector ", i))
    replaced[.x == vals[i]] <- TRUE
  }

  .default <- validate_recode_default(.default, .x, res, replaced)
  res <- replace_with(res, !replaced & !is.na(.x), .default, "`.default`")
  res <- replace_with(res, is.na(.x), .missing, "`.missing`")
  res
}

#' @export
recode.character <- function(.x, ..., .default = NULL, .missing = NULL) {
  .x <- as.character(.x)
  values <- dotdotdot(...)
  val_names <- names(values)
  have_names <- have_name(values)
  if (!all(have_names)) {
    bad <- which(!have_names) + 1L
    stop("Argument", if (length(bad) > 1L) "s", " ", paste(bad, sep = ", "), " must be named, not unnamed.")
  }

  n <- length(.x)
  template <- find_template(values, .default, .missing)
  res <- template[rep(NA_integer_, n)]
  replaced <- rep(FALSE, n)

  for (nm in val_names) {
    res <- replace_with(res, .x == nm, values[[nm]], paste0("`", nm, "`"))
    replaced[.x == nm] <- TRUE
  }

  .default <- validate_recode_default(.default, .x, res, replaced)
  res <- replace_with(res, !replaced & !is.na(.x), .default, "`.default`")
  res <- replace_with(res, is.na(.x), .missing, "`.missing`")
  res
}

#' @export
recode.factor <- function(.x, ..., .default = NULL, .missing = NULL) {
  values <- dotdotdot(...)
  if (length(values) == 0) stop("No replacements provided.")

  have_names <- have_name(values)
  if (!all(have_names)) {
    bad <- which(!have_names) + 1
    stop(bad, " must be named, not unnamed.")
  }
  if (!is.null(.missing)) {
    stop("`.missing` is not supported for factors.")
  }

  n <- length(levels(.x))
  template <- find_template(values, .default, .missing)
  res <- template[rep(NA_integer_, n)]
  replaced <- rep(FALSE, n)

  for (nm in names(values)) {
    res <- replace_with(res, levels(.x) == nm, values[[nm]], paste0("`", nm, "`"))
    replaced[levels(.x) == nm] <- TRUE
  }
  .default <- validate_recode_default(.default, .x, res, replaced)
  res <- replace_with(res, !replaced, .default, "`.default`")

  if (is.character(res)) {
    levels(.x) <- res
    .x
  } else {
    res[as.integer(.x)]
  }
}

have_name <- function(x) {
  nms <- names(x)
  if (is.null(nms)) rep(FALSE, length(x)) else !(nms == "" | is.na(nms))
}

compact <- function(.x) Filter(length, .x)

find_template <- function(values, .default = NULL, .missing = NULL) {
  x <- compact(c(values, .default, .missing))
  if (length(x) == 0L) {
    stop("No replacements provided.")
  }
  x[[1]]
}

validate_recode_default <- function(default, x, res, replaced) {
  default <- recode_default(x, default, res)
  if (is.null(default) && sum(replaced & !is.na(x)) < length(res[!is.na(x)])) {
    warning(
      "Unreplaced values treated as NA as .x is not compatible. ",
      "Please specify replacements exhaustively or supply .default",
      call. = FALSE
    )
  }
  default
}

recode_default <- function(x, default, res) {
  UseMethod("recode_default")
}

recode_default.default <- function(x, default, res) {
  same_type <- identical(typeof(x), typeof(res))
  if (is.null(default) && same_type) x else default
}

recode_default.factor <- function(x, default, res) {
  if (is.null(default)) {
    if ((is.character(res) || is.factor(res)) && is.factor(x)) {
      levels(x)
    } else {
      res[NA_integer_]
    }
  } else {
    default
  }
}

#' @rdname recode
#' @export
recode_factor <- function(.x, ..., .default = NULL, .missing = NULL, .ordered = FALSE) {
  recoded <- recode(.x, ..., .default = .default, .missing = .missing)

  values <- dotdotdot(...)
  all_levels <- unique(c(values, recode_default(.x, .default, recoded), .missing))
  recoded_levels <- if (is.factor(recoded)) levels(recoded) else unique(recoded)
  levels <- intersect(all_levels, recoded_levels)

  factor(recoded, levels, ordered = .ordered)
}

expect_equal(
  recode(1:2, "a", "b"),
  c("a", "b"),
  info = "positional substitution works"
)

expect_equal(
  recode(1:2, `2` = "b", `1` = "a"),
  c("a", "b"),
  "names override positions"
)


x1 <- letters[1:3]
expect_equal(
  recode(x1, a = "apple", .default = NA_character_),
  c("apple", NA, NA),
  info = "named substitution works for character"
)
x2 <- factor(x1)
expect_equal(
  recode(x2, a = "apple", .default = NA_character_),
  factor(c("apple", NA, NA)),
  info = "named substitution works for factor"
)

expect_equal(
  recode(c(1, NA), "a"),
  c("a", NA),
  info = "missing values replaced by missing argument"
)
expect_equal(
  recode(c(1, NA), "a", .missing = "b"),
  c("a", "b"),
  info = "missing values replaced by missing argument"
)
expect_equal(
  recode(c(letters[1:3], NA), .missing = "A"),
  c("a", "b", "c", "A"),
  info = "missing values replaced by missing argument"
)

expect_warning(
  expect_equal(recode(c(1, 2), "a"), c("a", NA))
)
expect_equal(
  recode(c(1, 2), "a", .default = "b"),
  c("a", "b"),
  info = "unmatched value replaced by default argument"
)
expect_equal(
  recode(letters[1:3], .default = "A"),
  c("A", "A", "A"),
  info = "unmatched value replaced by default argument"
)

expect_equal(
  recode(c(1, 2, NA), "a", .default = "b", .missing = "c"),
  c("a", "b", "c"),
  info = "missing and default place nicely together"
)

expect_equal(recode("x", x = "a"), "a", info = "can give name x")

x <- rep(1:3, 3)
expect_equal(
  recode(x, `3` = 10L, .default = x),
  rep(c(1L, 2L, 10L), 3),
  info = ".default works when not all values are named"
)

x <- letters[1:3]
expect_equal(recode(x, a = "A"), c("A", "b", "c"), info = ".default is aliased to .x when missing and compatible")
n <- 1:3
expect_equal(recode(n, `1` = 10L), c(10L, 2L, 3L), info = ".default is aliased to .x when missing and compatible")

x <- letters[1:3]
expect_warning(
  expect_equal(recode(x, a = 1), c(1L, NA, NA)),
  info = ".default is not aliased to .x when missing and not compatible"
)
n <- 1:3
expect_warning(
  expect_equal(recode(n, `1` = "a"), c("a", NA, NA)),
  info = ".default is not aliased to .x when missing and not compatible"
)

expect_warning(
  recode(1:3, `1` = "a"),
  info = "treated as NA"
)
expect_warning(
  recode_factor(letters[1:3], b = 1, c = 2),
  info = "conversion of unreplaced values to NA gives warning"
)

# factor ------------------------------------------------------------------

expect_equal(
  recode(factor(letters[1:3]), a = "A"),
  factor(c("A", "b", "c")),
  info = "default .default works with factors"
)

f <- factor(letters[1:3])
expect_equal(recode(f, a = 1, b = 2, c = 3), c(1, 2, 3), info = "can recode factor to double")
expect_equal(recode(f, a = 1, b = 2), c(1, 2, NA), info = "can recode factor to double")
expect_equal(recode(f, a = 1, b = 2, .default = 99), c(1, 2, 99), info = "can recode factor to double")

x <- c(1:3, NA)
expect_warning(
  expect_equal(
    recode_factor(x, `1` = "z", `2` = "y"),
    factor(c("z", "y", NA, NA), levels = c("z", "y"))
  ),
  info = "recode_factor() handles .missing and .default levels"
)
expect_equal(
  recode_factor(x, `1` = "z", `2` = "y", .default = "D"),
  factor(c("z", "y", "D", NA), levels = c("z", "y", "D")),
  info = "recode_factor() handles .missing and .default levels"
)
expect_equal(
  recode_factor(x, `1` = "z", `2` = "y", .default = "D", .missing = "M"),
  factor(c("z", "y", "D", "M"), c("z", "y", "D", "M")),
  info = "recode_factor() handles .missing and .default levels"
)

expected <- factor(c("a", "z", "y"), levels = c("z", "y", "a"))
x1 <- letters[1:3]
x2 <- factor(x1)
expect_equal(recode_factor(x1, b = "z", c = "y"), expected, info = "recode_factor() handles vector .default")
expect_equal(recode_factor(x2, b = "z", c = "y"), expected, info = "recode_factor() handles vector .default")
expect_equal(
  recode_factor(x1, b = "z", c = "y", .default = x1),
  expected,
  info = "recode_factor() handles vector .default"
)
expect_equal(
  recode_factor(x2, b = "z", c = "y", .default = x1),
  expected,
  info = "recode_factor() handles vector .default"
)

expect_equal(
  recode(factor(letters[1:4]), d = "c", b = "a"),
  factor(c("a", "a", "c", "c"), levels = c("a", "c")),
  info = "can recode factor with redundant levels"
)
expect_equal(
  recode_factor(letters[1:4], d = "c", b = "a"),
  factor(c("a", "a", "c", "c"), levels = c("c", "a")),
  info = "can recode factor with redundant levels"
)

# Errors --------------------------------------------
expect_error(recode(factor("a"), a = 5, .missing = 10), info = ".missing is not supported for factors")
expect_error(recode("a", b = 5, "c"), info = "All arguments must be named")
expect_error(recode(factor("a"), b = 5, "c"), info = "All arguments must be named")
expect_error(recode(1:5), info = "No replacements provided")
expect_error(recode("a"), info = "No replacements provided")
expect_error(recode(factor("a")), info = "No replacements provided")

#' Apply a function (or functions) across multiple columns
#'
#' @description
#' `across()` makes it easy to apply the same transformation to multiple columns, allowing you to use [select()]
#' semantics inside in "data-masking" functions like [summarise()] and [mutate()].
#'
#' `if_any()` and `if_all()` are used to apply the same predicate function to a selection of columns and combine the
#' results into a single logical vector.
#'
#' `across()` supersedes the family of `dplyr` "scoped variants" like `summarise_at()`, `summarise_if()`, and
#' `summarise_all()` and therefore these functions will not be implemented in `poorman`.
#'
#' @param cols,.cols <[`poor-select`][select_helpers]> Columns to transform. Because `across()` is used within functions
#' like `summarise()` and `mutate()`, you can't select or compute upon grouping variables.
#' @param .fns Functions to apply to each of the selected columns.
#' Possible values are:
#'
#' - `NULL`, to returns the columns untransformed.
#' - A function, e.g. `mean`.
#' - A lambda, e.g. `~ mean(.x, na.rm = TRUE)`
#' - A list of functions/lambdas, e.g. `list(mean = mean, n_miss = ~ sum(is.na(.x))`
#'
#' Within these functions you can use [cur_column()] and [cur_group()] to access the current column and grouping keys
#' respectively.
#' @param ... Additional arguments for the function calls in `.fns`.
#' @param .names A glue specification that describes how to name the output
#'  columns. This can use `{.col}` to stand for the selected column name, and `{.fn}`
#'  to stand for the name of the function being applied. The default (`NULL`) is
#'  equivalent to "`{.col}`" for the single function case and "`{.col}_{.fn}`" for
#'  the case where a list is used for `.fns`.
#'
#' @return
#' `across()` returns a `data.frame` with one column for each column in `.cols` and each function in `.fns`.
#'
#' `if_any()` and `if_all()` return a logical vector.
#'
#' @examples
#' # across() -----------------------------------------------------------------
#' iris %>%
#'   group_by(Species) %>%
#'   summarise(across(starts_with("Sepal"), mean))
#' iris %>%
#'   mutate(across(where(is.factor), as.character))
#'
#' # Additional parameters can be passed to functions
#' iris %>%
#'   group_by(Species) %>%
#'   summarise(across(starts_with("Sepal"), mean, na.rm = TRUE))
#'
#' # A named list of functions
#' iris %>%
#'   group_by(Species) %>%
#'   summarise(across(starts_with("Sepal"), list(mean = mean, sd = sd)))
#'
#' # Use the .names argument to control the output names
#' iris %>%
#'   group_by(Species) %>%
#'   summarise(
#'     across(starts_with("Sepal"),
#'     mean,
#'     .names = c("mean_sepal_length", "mean_sepal_width"))
#'   )
#'
#' # if_any() and if_all() ----------------------------------------------------
#' iris %>%
#'   filter(if_any(ends_with("Width"), ~ . > 4))
#' iris %>%
#'   filter(if_all(ends_with("Width"), ~ . > 2))
#'
#' @export
across <- function(.cols = everything(), .fns = NULL, ..., .names = NULL) {
  setup <- setup_across(substitute(.cols), .fns, .names)
  # deal with {.fn} and {.col} in .names
  if (length(setup$names) == 1 && grepl("\\{\\.col\\}|\\{\\.fn\\}", setup$names)) {
    ref <- setup$names
    id <- 1
    fn_names <- unique(names(setup$funs))
    for (i in seq_along(setup$cols)) {
      .col <- setup$cols[i]
      for (j in seq_along(fn_names)) {
        .fn <- fn_names[j]
        setup$names[id] <- gluestick(ref)
        id <- id + 1
      }
    }
  }
  cols <- setup$cols
  n_cols <- length(cols)
  if (n_cols == 0L) return(data.frame())
  funs <- setup$funs
  data <- context$get_columns(cols)
  names <- setup$names
  if (is.null(funs)) {
    data <- data.frame(data)
    if (is.null(names)) {
      return(data)
    } else {
      return(setNames(data, names))
    }
  }
  n_fns <- length(funs)
  res <- vector(mode = "list", length = n_fns * n_cols)
  k <- 1L
  for (i in seq_len(n_cols)) {
    context$cur_column <- cols[[i]]
    col <- data[[i]]
    for (j in seq_len(n_fns)) {
      res[[k]] <- funs[[j]](col, ...)
      k <- k + 1L
    }
  }
  if (is.null(names(res))) names(res) <- names
  as.data.frame(res)
}

#' @rdname across
#' @export
if_any <- function(.cols, .fns = NULL, ..., .names = NULL) {
  df <- do.call(across, list(.cols = substitute(.cols), .fns = .fns, ..., .names = .names))
  if (nrow(df) == 0L) return(FALSE)
  check_if_types(df)
  Reduce(`|`, df)
}

#' @rdname across
#' @export
if_all <- function(.cols, .fns = NULL, ..., .names = NULL) {
  df <- do.call(across, list(.cols = substitute(.cols), .fns = .fns, ..., .names = .names))
  if (nrow(df) == 0L) return(FALSE)
  check_if_types(df)
  Reduce(`&`, df)
}

check_if_types <- function(df) {
  types <- vapply(df, class, NA_character_)
  not_logical <- types != "logical"
  if (any(not_logical)) {
    stop(
      "Cannot convert the following columns to <logical>:\n    ",
      paste0(colnames(df)[not_logical], " <", types, "> ", collapse = "\n    ")
    )
  }
}

# -- helpers -----------------------------------------------------------------------------------------------------------

setup_across <- function(.cols, .fns, .names) {
  cols <- eval_select_pos(.data = context$.data, .cols, .group_pos = FALSE)
  cols <- context$get_colnames()[cols]
  if (context$is_grouped()) cols <- setdiff(cols, group_vars(context$.data))

  funs <- if (is.null(.fns)) NULL else if (!is.list(.fns)) list(.fns) else .fns
  if (is.null(funs)) return(list(cols = cols, funs = funs, names = .names))
  f_nms <- names(funs)
  if (is.null(f_nms) && !is.null(.fns)) names(funs) <- seq_along(funs)
  if (any(nchar(f_nms) == 0L)) {
    miss <- which(nchar(f_nms) == 0L)
    names(funs)[miss] <- miss
    f_nms <- names(funs)
  }

  funs <- lapply(funs, as_function)

  names <- if (!is.null(.names)) {
    .names
  } else {
    if (length(funs) == 1L && is.null(f_nms)) {
      cols
    } else {
      nms <- do.call(paste, c(rev(expand.grid(names(funs), cols)), sep = "_"))
      if (length(nms) == 0L) nms <- NULL
      nms
    }
  }

  list(cols = cols, funs = funs, names = names)
}

options(stringsAsFactors = FALSE)

expect_equal(
  data.frame(x = 1) %>% mutate(across()),
  data.frame(x = 1),
  info = "across() works on one column data.frame"
)

expect_equivalent(
  mtcars %>% group_by(cyl) %>% summarise(across(starts_with("c"), mean)) %>% ungroup(),
  data.frame(cyl = c(4, 6, 8), carb_1 = c(1.54545455, 3.42857, 3.5)),
  tolerance = 0.00001,
  info = "across() does not select grouping variables"
)

gf <- data.frame(x = 1, y = 2, z = 3, s = "") %>% group_by(x)
expect_named <- function(x, y, info = NA_character_) expect_equal(colnames(x), y, info = info)
expect_named(
  summarise(gf, across()),
  c("x", "y", "z", "s"),
  "across() correctly names output columns"
)
expect_named(
  summarise(gf, across(where(is.numeric), mean)),
  c("x", "y", "z"),
  "across() correctly names output for non-named functions"
)
expect_named(
  summarise(gf, across(where(is.numeric), list(mean = mean, sum = sum))),
  c("x", "y_mean", "y_sum", "z_mean", "z_sum"),
  info = "across() correctly names output columns for named lists of functions"
)
expect_named(
  summarise(gf, across(where(is.numeric), list(mean = mean, sum = sum),
                       .names = "my_col_{.col}_{.fn}")),
  c("x", "my_col_y_mean", "my_col_y_sum", "my_col_z_mean", "my_col_z_sum"),
  info = "across() correctly names output columns for named lists of functions and custom names"
)
expect_named(
  summarise(gf, across(where(is.numeric), list(mean = mean, sum))),
  c("x", "y_mean", "y_2", "z_mean", "z_2"),
  info = "across() correctly names output columns for partially named lists of functions"
)
expect_named(
  summarise(gf, across(where(is.numeric), list(mean, sum = sum))),
  c("x", "y_1", "y_sum", "z_1", "z_sum"),
  info = "across() correctly names output columns for partially named lists of functions"
)
expect_named(
  summarise(gf, across(where(is.numeric), list(mean, sum = sum),
                       .names = "my_col_{.col}_{.fn}")),
  c("x", "my_col_y_1", "my_col_y_sum", "my_col_z_1", "my_col_z_sum"),
  info = "across() correctly names output columns for partially named lists of functions and custom names"
)
expect_named(
  summarise(gf, across(where(is.numeric), list(mean, sum))),
  c("x", "y_1", "y_2", "z_1", "z_2"),
  info = "across() correctly names output columns for non-named lists of functions"
)
expect_named(
  summarise(gf, across(where(is.numeric), list(mean, sum), .names = "my_col_{.col}_{.fn}")),
  c("x", "my_col_y_1", "my_col_y_2", "my_col_z_1", "my_col_z_2"),
  info = "across() correctly names output columns for non-named lists of functions and custom names"
)

expect_identical(
  data.frame(x = 1:2, y = c("a", "b")) %>% summarise(across(everything(), list(cls = class, type = is.numeric))),
  data.frame(x_cls = "integer", x_type = TRUE, y_cls = "character", y_type = FALSE),
  info = "across() result locations are aligned with column names"
)

expect_equal(
  summarise(data.frame(x = c(1, NA)), across(everything(), mean, na.rm = TRUE)),
  data.frame(x = 1),
  info = "across() passes ... to functions"
)

expect_equal(
  summarise(data.frame(x = c(1, NA)), across(everything(), list(mean = mean, median = median), na.rm = TRUE)),
  data.frame(x_mean = 1, x_median = 1),
  info = "across() passes ... to functions"
)

df <- data.frame(x = 1)
expect_equal(
  mutate(df, across(x, `+`, 1)),
  data.frame(x = 2),
  info = "across() passes unnamed arguments following .fns as ..."
)

expect_equal(
  summarize(data.frame(x = c(1, 2)), across(x, tail, n = 1)),
  data.frame(x = 2),
  info = "across() avoids simple argument name collisions with ..."
)

df <- data.frame(a = 1)
expect_equal(
  mutate(df, x = ncol(across(where(is.numeric))), y = ncol(across(where(is.numeric)))),
  data.frame(a = 1, x = 1L, y = 2L),
  info = "across() works sequentially"
)
expect_equal(
  mutate(df, a = "x", y = ncol(across(where(is.numeric)))),
  data.frame(a = "x", y = 0L),
  info = "across() works sequentially"
)
expect_equal(
  mutate(df, x = 1, y = ncol(across(where(is.numeric)))),
  data.frame(a = 1, x = 1, y = 2L),
  info = "across() works sequentially"
)

expect_named(
  mutate(data.frame(a = 1, b = 2), a = 2, x = across())$x,
  c("a", "b"),
  info = "across() retains original ordering"
)

expect_equal(
  mutate(data.frame(a = 1, b = 2), x = ncol(across(where(is.numeric))) + ncol(across(a))),
  data.frame(a = 1, b = 2, x = 3),
  info = "across() can be used twice in the same expression"
)

expect_equal(
  mutate(data.frame(a = 1, b = 2), x = ncol(across(where(is.numeric))), y = ncol(across(a))),
  data.frame(a = 1, b = 2, x = 2, y = 1),
  info = "across() can be used in separate expressions"
)

df <- data.frame(g = 1:2, a = 1:2, b = 3:4) %>% group_by(g)
expect_equal(
  mutate(df, x = if_else(cur_group_id() == 1L, across(a)$a, across(b)$b)),
  structure(
    list(g = 1:2, a = 1:2, b = 3:4, x = c(1L, 4L)),
    groups = structure(list(g = 1:2, .rows = list(1L, 2L)),
      row.names = 1:2,
      class = "data.frame",
      .drop = TRUE
    ), row.names = 1:2, class = c("grouped_df", "data.frame")
  ),
  info = "across() usage can depend on the group id"
)

df <- data.frame(g = rep(1:2, each = 2), a = 1:4) %>% group_by(g)
expect_identical(
  mutate(df, data.frame(x = across(where(is.numeric), mean)$a, y = across(where(is.numeric), max)$a)),
  mutate(df, x = mean(a), y = max(a)),
  info = "across() internal cache key depends on all inputs"
)

expect_error(
  data.frame(x = 1) %>% summarise(across(everything(), "foo")),
  info = "across() fails on non-function"
)

expect_equal(
  mutate(data.frame(), across()),
  data.frame(),
  info = "across() works with empty data.frames"
)

# Formula approach

expect_equal(
  data.frame(x = 1, y = 2) %>% summarise(across(everything(), ~rep(42, .))),
  data.frame(x = rep(42, 2), y = rep(42, 2)),
  info = "across() uses tidy recycling rules"
)

expect_error(
  data.frame(x = 2, y = 3) %>% summarise(across(everything(), ~rep(42, .))),
  info = "across() uses tidy recycling rules #2"
)

expect_equal(
  iris %>% group_by(Species) %>% summarise(across(starts_with("Sepal"), ~mean(., na.rm = TRUE))) %>% ungroup(),
  data.frame(
    Species = structure(1:3, .Label = c("setosa", "versicolor", "virginica"), class = "factor"),
    Sepal.Length = c(5.006, 5.936, 6.588),
    Sepal.Width = c(3.428, 2.77, 2.974)
  ),
  info = "purrr style formulas work"
)

expect_equal(
  mtcars %>% summarise(across(starts_with("m"), list(~mean(.x), ~sd(.x)))),
  data.frame(mpg_1 = mean(mtcars$mpg), mpg_2 = sd(mtcars$mpg)),
  info = "list of purrr style formulas works"
)

expect_equal(
  mtcars %>% summarise(across(starts_with("m"), list("mean", "sd"))),
  data.frame(mpg_1 = mean(mtcars$mpg), mpg_2 = sd(mtcars$mpg)),
  info = "character vector style functions work"
)

# if_all() and if_any()

d <- data.frame(x = 10, y = 10)
expect_error(filter(d, if_all(x:y, identity)), info = "if_all() enforces logical in filter")
expect_error(filter(d, if_any(x:y, identity)), info = "if_any() enforces logical in filter")
expect_error(mutate(d, ok = if_all(x:y, identity)), info = "if_all() enforces logical in mutate")
expect_error(mutate(d, ok = if_any(x:y, identity)), info = "if_any() enforces logical in mutate")

d <- data.frame(x = c(1, 5, 10, 10), y = c(0, 0, 0, 10), z = c(10, 5, 1, 10))
res <- mutate(.data = d, any = if_any(x:z, ~ . > 8), all = if_all(x:z, ~ . > 8))
expect_equal(res$any, c(TRUE, FALSE, TRUE, TRUE), info = "if_any() can be used in mutate")
expect_equal(res$all, c(FALSE, FALSE, FALSE, TRUE), info = "if_all() can be used in mutate")

df <- expand.grid(
  x = c(TRUE, FALSE, NA), y = c(TRUE, FALSE, NA)
)
expect_identical(
  filter(df, x & y),
  filter(df, if_all(c(x, y), identity)),
  info = "if_all() respects filter()-like NA handling"
)
expect_identical(
  filter(df, x | y),
  filter(df, if_any(c(x, y), identity)),
  info = "if_any() respects filter()-like NA handling"
)

expect_equal(
  mtcars %>% filter(if_any(contains("Width"), ~ . > 4)),
  mtcars[mtcars$mpg > 100, ],
  info = "if_any() conditions that return empty data.frames do not fail."
)
expect_equal(
  mtcars %>% filter(if_all(contains("Width"), ~ . > 4)),
  mtcars[mtcars$mpg > 100, ],
  info = "if_all() conditions that return empty data.frames do not fail."
)

#' Print a grouped `data.frame`
#'
#' A print method for grouped `data.frame`s. Uses the standard `print.data.frame()` method but also reports the groups.
#'
#' @param x An object of class `grouped_df`.
#' @param ... Additional arguments to [print()].
#' @inheritParams base::print.data.frame
#'
#' @examples
#' mtcars %>% group_by(cyl, am) %>% print()
#'
#' @noRd
print.grouped_df <- function(x, ..., digits = NULL, quote = FALSE, right = TRUE, row.names = TRUE, max = NULL) {
  class(x) <- "data.frame"
  print(x, ..., digits = digits, quote = quote, right = right, row.names = row.names, max = max)
  cat("\nGroups: ", paste(group_vars(x), collapse = ", "), "\n\n")
}

#' Build a list
#'
#' @description
#' `lst()` constructs a list, similar to [base::list()], but where components
#' are built sequentially. When defining a component, you can refer to components
#' created earlier in the call. `lst()` also generates missing names
#' automatically.
#'
#' @param ... Named or unnamed elements of a list. If the element is unnamed, its
#' expression will be used as its name.
#'
#' @return A named list.
#' @export
#' @examples
#' # the value of n can be used immediately in the definition of x
#' lst(n = 5, x = runif(n))
#'
#' # missing names are constructed from user's input
#' lst(1:3, z = letters[4:6], runif(3))
#'
#' a <- 1:3
#' b <- letters[4:6]
#' lst(a, b)

lst <- function(...) {
  fn_call <- match.call()
  list_to_eval <- as.list(fn_call)[-1]

  out <- vector(mode = "list", length = length(list_to_eval))
  names(out) <- names(list_to_eval)
  exprs <- lapply(substitute(list(...)), deparse)[-1]
  for (element in seq_along(list_to_eval)) {
    value <- list_to_eval[[element]]
    if (is.language(value)) {
      # need to update the environment in which the values are obtained
      # ex: lst(a = 1, a = a + 1, b = a), 'b' needs the updated value of 'a',
      # not its initial value.
      value <- eval(
        value,
        envir = if (length(out) == 0) {
          list_to_eval
        } else {
          # restrict the environment to the previous elements of the list (and
          # to the last value for each name if there are duplicated names)
          drop_dup_list(out[1:(element - 1)])
        }
      )
    }
    if (is.null(value)) {
      out[element] <- list(NULL)
    } else {
      out[[element]] <- value
    }

    # this naming part needs to happen at the end of the loop to avoid error
    # with lst(NULL)
    invalid_name <- is.null(names(out)[element]) ||
      is.na(names(out)[element]) ||
      names(out)[element] == ""

    if (invalid_name) {
      if (exprs[[element]] != "NULL" || (exprs[[element]] == "NULL" && is.null(out[[element]]))) {
        names(out)[element] <- exprs[[element]]
      }
    }
  }
  out
}

#' Drop List Duplicated
#' If several elements of a list have the same name, only keep the last one with this name.
#' @examples
#' list(a = 1, a = 2, b = 1)
#' # list(a = 2, b = 1)
#' @noRd
drop_dup_list <- function(x) {

  list_names <- names(x)
  if (identical(list_names, unique(list_names))) return(x)

  count <- table(list_names)
  dupes <- names(count[count > 1])
  uniques <- names(count[count == 1])

  to_drop <- do.call(c, lapply(
    dupes,
    function(x) {
      matches <- which(list_names == x)
      matches[-length(matches)]
    }
  ))
  x[uniques] <- Filter(Negate(is.null), x[uniques])

  return(x[-to_drop])

}

expect_equal(lst(NULL), list("NULL" = NULL))
expect_identical(lst(a = NULL), list(a = NULL))
expect_identical(
  lst(NULL, b = NULL, 1:3),
  list("NULL" = NULL, b = NULL, "1:3" = 1:3)
)

expect_identical(lst(a = 1, b = a), list(a = 1, b = 1))
expect_identical(lst(a = NULL, b = a), list(a = NULL, b = NULL))

expect_identical(lst(a = 1, a = a + 1, b = a), list(a = 1, a = 2, b = 2))
expect_identical(lst(b = 1, a = b, a = b + 1, b = a), list(b = 1, a = 1, a = 2, b = 2))

#' Unite Multiple Columns Into One
#'
#' Convenience function to paste together multiple columns.
#'
#' @param data A `data.frame`.
#' @param col `character(1)` or `symbol(1)`. The name of the new column.
#' @param ... The columns to unite.
#' @param sep `character(1)`. Separator to use between the values.
#' @param remove `logical(1)`. If `TRUE`, remove the input columns from the output `data.frame`.
#' @param na.rm `logical(1)`. If `TRUE`, missing values will be remove prior to uniting each value.
#'
#' @return
#' A `data.frame` with the columns passed via `...` pasted together in a new column.
#'
#' @examples
#' df <- data.frame(x = c("a", "a", NA, NA), y = c("b", NA, "b", NA))
#' df
#'
#' df %>% unite("z", x:y, remove = FALSE)
#' # To remove missing values:
#' df %>% unite("z", x:y, na.rm = TRUE, remove = FALSE)
#'
#' @importFrom stats na.omit
#'
#' @export
unite <- function(data, col, ..., sep = "_", remove = TRUE, na.rm = FALSE) {
  col <- deparse_var(col)
  cols_pos <- if (missing(...)) seq_along(data) else select_positions(data, ...)
  to_unite <- data[, cols_pos, drop = FALSE]
  data[[col]] <- if (isTRUE(na.rm)) {
    apply(to_unite, 1L, function(x) paste(stats::na.omit(x), collapse = sep))
  } else {
    apply(to_unite, 1L, paste, collapse = sep)
  }
  if (isTRUE(remove)) {
    to_rm <- setdiff(colnames(to_unite), col)
    if (is.grouped_df(data) && length(to_rm) > 0L) {
      rm_groups <- as_symbols(intersect(group_vars(data), to_rm))
      data <- do.call(ungroup, squash(list(x = data, rm_groups)))
    }
    data[, to_rm] <- NULL
  } else {
    data <- eval(bquote(relocate(data, col, .before = .(cols_pos[1]))))
  }
  data
}

### Taken from https://github.com/coolbutuseless/gluestick/blob/main/R/gluestick.R
### MIT License

#' String interpolation
#'
#' Expressions enclosed by specified delimiters will be evaluated as R code within the context of the `src`
#' data/environment. The results will then be inserted into the original string via `sprintf()` i.e. string
#' interpolation.
#'
#' @param fmt single character string containing the format specification.
#' @param src data source. An `environment`, `list`, `data.frame` or anything supported by [as.environment()]. Default:
#' [parent.frame()] i.e. the calling environment.
#' @param open,close the opening and closing character strings which delimit an expression. Default: `{}`. Note: the
#' delimiters can be more complex than just a single character.
#' @param eval `logical(1)`. Should the expressions be treated as R code to be evaluated? Default: `TRUE` means to treat
#' the expressions as R code and evaluate. If `FALSE`, then no code evaluation will ever be done and expressions will be
#' treated as only variable names in the given `src` data. This may be safer in some contexts e.g. for user supplied
#' `fmt` strings.
#'
#' @examples
#' gluestick("Hello {name}", list(name = '#RStats'))
#' gluestick("Hello ~!name!~", list(name = '#RStats'), open = "~!", close = "!~")
#' name <- '#RStats'; gluestick("Hello {name}")
#'
#' @noRd
gluestick <- function(fmt, src = parent.frame(), open = "{", close = "}", eval = TRUE) {

  nchar_open <- nchar(open)
  nchar_close <- nchar(close)

  stopifnot(exprs = {
    is.character(fmt)
    length(fmt) == 1L
    is.character(open)
    length(open) == 1L
    nchar_open > 0L
    is.character(close)
    length(close) == 1
    nchar_close > 0
  })

  open <- gsub("(.)", "\\\\\\1", open)
  close <- gsub("(.)", "\\\\\\1", close)
  re <- paste0(open, ".*?", close)

  matches <- gregexpr(re, fmt)
  exprs <- regmatches(fmt, matches)[[1]]

  exprs <- substr(exprs, nchar_open + 1L, nchar(exprs) - nchar_close)

  fmt_sprintf <- gsub(re, "%s", fmt)
  fmt_sprintf <- gsub("%(?!s)", "%%", fmt_sprintf, perl = TRUE)

  args <- if (eval) {
    lapply(exprs, function(expr) eval(parse(text = expr), envir = src))
  } else {
    unname(mget(exprs, envir = as.environment(src)))
  }

  do.call(sprintf, c(list(fmt_sprintf), args))
}

#' Forward-pipe operator
#'
#' Pipe an object forward into a function or call expression.
#'
#' @param lhs The result you are piping.
#' @param rhs Where you are piping the result to.
#'
#' @examples
#' # Basic use:
#' iris %>% head
#'
#' # Use with lhs as first argument
#' iris %>% head(10)
#'
#' # Using the dot place-holder
#' "Ceci n'est pas une pipe" %>% gsub("une", "un", .)
#'
#' # When dot is nested, lhs is still placed first:
#' sample(1:10) %>% paste0(LETTERS[.])
#'
#' # This can be avoided:
#' rnorm(100) %>% {c(min(.), mean(.), max(.))} %>% floor
#'
#' # Lambda expressions:
#' iris %>%
#'   {
#'     size <- sample(1:10, size = 1)
#'     rbind(head(., size), tail(., size))
#'   }
#'
#' # renaming in lambdas:
#' iris %>%
#'   {
#'     my_data <- .
#'     size <- sample(1:10, size = 1)
#'     rbind(head(my_data, size), tail(my_data, size))
#'   }
#'
#' @name pipe
#'
#' @author
#' Nathan Eastwood and Antoine Fabri \email{antoine.fabri@@gmail.com}.
#'
#' @export
`%>%` <- function(lhs, rhs) {
  rhs_call <- insert_dot(substitute(rhs))
  eval(rhs_call, envir = list(`.` = lhs), enclos = parent.frame())
}

#' @author Antoine Fabri
#' @noRd
insert_dot <- function(expr) {
  if (is.symbol(expr) || expr[[1]] == quote(`(`)) {
    # if a symbol or an expression inside parentheses, make it a call with dot
    # arg
    expr <- as.call(c(expr, quote(`.`)))
  } else if (length(expr) == 1) {
    # if a call without an arg, give it a dot arg
    expr <- as.call(c(expr[[1]], quote(`.`)))
  } else if (
    expr[[1]] != quote(`{`) &&
    !any(vapply(expr[-1], identical, quote(`.`), FUN.VALUE = logical(1))) &&
    !any(vapply(expr[-1], identical, quote(`!!!.`), FUN.VALUE = logical(1)))
  ) {
    # if a call with args but no dot in arg, insert one first
    expr <- as.call(c(expr[[1]], quote(`.`), as.list(expr[-1])))
  }
  expr
}

#' Reduce multiple values down to a single value
#'
#' Create one or more scalar variables summarising the variables of an existing `data.frame`. Grouped `data.frame`s will
#' result in one row in the output for each group.
#'
#' `summarise()` and `summarize()` are synonyms.
#'
#' @param .data A `data.frame`.
#' @param ... Name-value pairs of summary functions. The name will be the name of the variable in the result.
#' @param .groups `character(1)`. Grouping structure of the result.
#'
#' * `"drop_last"`: drops the last level of grouping.
#' * `"drop"`: all levels of grouping are dropped.
#' * `"keep"`: keeps the same grouping structure as `.data`.
#'
#' When `.groups` is not specified, it is chosen based on the number of rows of the results:
#' * If all the results have 1 row, you get `"drop_last"`.
#' * If the number of rows varies, you get `"keep"`.
#'
#' In addition, a message informs you of that choice, unless the result is ungrouped, the option
#' `"poorman.summarise.inform"` is set to `FALSE`.
#'
#' The value can be:
#' * A vector of length `1`, e.g. `min(x)`, `n()`, or `sum(is.na(y))`.
#' * A vector of length `n`, e.g. `quantile()`.
#'
#' @examples
#' # A summary applied to ungrouped tbl returns a single row
#' mtcars %>%
#'   summarise(mean = mean(disp), n = n())
#'
#' # Usually, you'll want to group first
#' mtcars %>%
#'   group_by(cyl) %>%
#'   summarise(mean = mean(disp), n = n())
#'
#' # You can summarise to more than one value:
#' mtcars %>%
#'    group_by(cyl) %>%
#'    summarise(qs = quantile(disp, c(0.25, 0.75)), prob = c(0.25, 0.75))
#'
#' # You use a data frame to create multiple columns so you can wrap
#' # this up into a function:
#' my_quantile <- function(x, probs) {
#'   data.frame(x = quantile(x, probs), probs = probs)
#' }
#' mtcars %>%
#'   group_by(cyl) %>%
#'   summarise(my_quantile(disp, c(0.25, 0.75)))
#'
#' # Each summary call removes one grouping level (since that group
#' # is now just a single row)
#' mtcars %>%
#'   group_by(cyl, vs) %>%
#'   summarise(cyl_n = n()) %>%
#'   group_vars()
#'
#' @name summarise
#' @export
summarise <- function(.data, ..., .groups = NULL) {
  UseMethod("summarise")
}

#' @export
summarise.data.frame <- function(.data, ..., .groups = NULL) {
  fns <- dotdotdot(...)
  context$setup(.data)
  on.exit(context$clean(), add = TRUE)
  groups_exist <- context$is_grouped()
  if (groups_exist) {
    group <- unique(context$get_columns(group_vars(context$.data)))
  }
  if (is_empty_list(fns)) {
    if (groups_exist) return(group) else return(data.frame())
  }
  res <- vector(mode = "list", length = length(fns))
  eval_env <- c(as.list(context$.data), vector(mode = "list", length = length(fns)))
  new_pos <- seq(length(context$.data) + 1L, length(eval_env), 1L)
  for (i in seq_along(fns)) {
    eval_env[[new_pos[i]]] <- do.call(with, list(eval_env, fns[[i]]))
    nms <- if (!is_named(eval_env[[new_pos[i]]])) {
      if (!is.null(names(fns)[[i]])) names(fns)[[i]] else deparse(fns[[i]])
    } else {
      NULL
    }
    if (!is.null(nms)) names(eval_env)[[new_pos[i]]] <- nms
    res[[i]] <- build_data_frame(eval_env[[new_pos[i]]], nms = nms)
  }
  res <- do.call(cbind, res)
  if (groups_exist) res <- cbind(group, res, row.names = NULL)
  res
}

#' @export
summarise.grouped_df <- function(.data, ..., .groups = NULL) {
  if (!is.null(.groups)) {
    .groups <- match.arg(arg = .groups, choices = c("drop", "drop_last", "keep"), several.ok = FALSE)
  }

  groups <- group_vars(.data)
  res <- apply_grouped_function("summarise", .data, drop = TRUE, ...)
  res <- res[arrange_rows(res, as_symbols(groups)), , drop = FALSE]

  verbose <- summarise_verbose(.groups)

  if (is.null(.groups)) {
    all_one <- as.data.frame(table(res[, groups]))
    all_one <- all_one[all_one$Freq != 0, ]
    .groups <- if (all(all_one$Freq == 1)) "drop_last" else "keep"
  }

  if (.groups == "drop_last") {
    n <- length(groups)
    if (n > 1) {
      if (verbose) summarise_inform(groups[-n])
      res <- groups_set(res, groups[-n], group_by_drop_default(.data))
    }
  } else if (.groups == "keep") {
    if (verbose) summarise_inform(groups)
    res <- groups_set(res, groups, group_by_drop_default(.data))
  } else if (.groups == "drop") {
    attr(res, "groups") <- NULL
  }

  rownames(res) <- NULL
  res
}

#' @rdname summarise
#' @export
summarize <- summarise
#' @export
summarize.data.frame <- summarise.data.frame
#' @export
summarize.grouped_df <- summarise.grouped_df

# -- Helpers -------------------------------------------------------------------

summarise_inform <- function(new_groups) {
  message(sprintf(
    "`summarise()` has grouped output by %s. You can override using the `.groups` argument.",
    paste0("'", new_groups, "'", collapse = ", ")
  ))
}

summarise_verbose <- function(.groups) {
  is.null(.groups) &&
    !identical(getOption("poorman.summarise.inform"), FALSE)
}

#' @param x A `data.frame`.
#' @return
#' When using [ungroup()], a `data.frame`.
#' @rdname group_by
#' @export
ungroup <- function(x, ...) {
  UseMethod("ungroup")
}

#' @export
ungroup.data.frame <- function(x, ...) {
  rm_groups <- deparse_dots(...)
  groups <- group_vars(x)
  if (length(rm_groups) == 0L) rm_groups <- groups
  x <- groups_set(x, groups[!(groups %in% rm_groups)])
  if (length(attr(x, "groups")) == 0L) {
    attr(x, "groups") <- NULL
    class(x) <- class(x)[!(class(x) %in% "grouped_df")]
  }
  x
}

#' @export
ungroup.grouped_df <- function(x, ...) {
  NextMethod("ungroup")
}

#' Capture unevaluated dots
#'
#' Gather the unevaluated dots into a list, storing them as is.
#'
#' @param ... Arguments to be stored in the list.
#' @param .impute_names `logical(1)`. Whether to fill any missing names of the list.
#'
#' @noRd
dotdotdot <- function(..., .impute_names = FALSE) {
  dots <- eval(substitute(alist(...)))
  if (isTRUE(.impute_names)) {
    deparse_dots <- lapply(dots, deparse)
    names_dots <- names(dots)
    unnamed <- if (is.null(names_dots)) rep(TRUE, length(dots)) else nchar(names_dots) == 0L
    names(dots)[unnamed] <- deparse_dots[unnamed]
  }
  dots
}

#' Capture unevaluated dots
#'
#' Gather the unevaluated dots into a list, storing them as characters.
#'
#' @param ... Arguments to be stored in the list.
#'
#' @noRd
deparse_dots <- function(...) {
  vapply(substitute(...()), deparse, NA_character_)
}

#' Capture variable
#'
#' Deparse a variable's value
#'
#' @param var An R object.
#' @param frame An `environment` in which `var` should be evaluated.
#'
#' @noRd
deparse_var <- function(var, frame = if (is.null(eval_env$env)) parent.frame() else eval_env$env) {
  sub_var <- eval(substitute(substitute(var)), frame)
  if (is.symbol(sub_var)) var <- as.character(sub_var)
  var
}

expect_error(
  1L %>% select("mpg"),
  info = "Test non-data.frame is not a data.frame"
)

expect_equal(
  poorman:::deparse_var("val"),
  "val",
  info = "deparse_var() works for character strings"
)

expect_equal(
  poorman:::deparse_var(val),
  "val",
  info = "deparse_var() works for symbols"
)

expect_equal(
  poorman:::deparse_var(NULL),
  NULL,
  info = "deparse_var() works for NULL values"
)

check_is_dataframe <- function(.data) {
  parent_fn <- all.names(sys.call(-1L), max.names = 1L)
  if (!is.data.frame(.data)) stop(parent_fn, " must be given a data.frame")
  invisible()
}

seq2 <- function(from, to) {
  if (length(from) != 1) stop("`from` must be length one")
  if (length(to) != 1) stop("`to` must be length one")
  if (from > to) integer() else seq.int(from, to)
}

collapse_to_sentence <- function(x) {
  len_x <- length(x)
  if (len_x == 0L) {
    stop("Length of `x` is 0")
  } else if (len_x == 1L) {
    as.character(x)
  } else if (len_x == 2L) {
    paste(x, collapse = " and ")
  } else {
    paste(paste(x[1:(len_x - 1)], collapse = ", "), x[len_x], sep = " and ")
  }
}

#' Build a `data.frame` from a variety of inputs including atomic vectors, lists and other `data.frame`s
#' @noRd
build_data_frame <- function(x, nms = NULL) {
  res <- if (is.atomic(x)) {
    data.frame(x)
  } else if (is.list(x) && !is.data.frame(x)) {
    structure(list(x = x), class = "data.frame", row.names = c(NA, -1L))
  } else if (is.data.frame(x)) {
    x
  }
  if (!is.null(nms)) colnames(res) <- nms
  res
}

#' Check whether any elements of a list are nested
#' @param lst A `list()`
#' @examples
#' is_nested(list(a = 1, b = 2, c = 3))
#' is_nested(list(a = 1, b = list(c = 2, d = 3)))
#' @noRd
is_nested <- function(lst) vapply(lst, function(x) inherits(x[1L], "list"), FALSE)

#' Remove all levels of a list
#' @noRd
squash <- function(lst) {
  do.call(c, lapply(lst, function(x) if (is.list(x) && !is.data.frame(x)) squash(x) else list(x)))
}

#' Move entries within a list up one level
#' @noRd
flatten <- function(lst) {
  nested <- is_nested(lst)
  res <- c(lst[!nested], unlist(lst[nested], recursive = FALSE))
  if (sum(nested)) Recall(res) else return(res)
}

#' Find first non-missing element
#'
#' Given a set of vectors, `coalesce()` finds the first non-missing value at each position. This is inspired by the SQL
#' `COALESCE` function which does the same thing for `NULL`s.
#'
#' @details
#' Currently, `coalesce()` type checking does not take place.
#'
#' @param ... Vectors. Inputs should be recyclable (either be length `1L` or `n`) and coercible to a common type.
#'
#' @seealso
#' [na_if()] to replace specified values to a `NA`.
#'
#' [replace_na()] to replace a `NA` with a value.
#'
#' @examples
#' # Use a single value to replace all missing vectors
#' x <- sample(c(1:5, NA, NA, NA))
#' coalesce(x, 0L)
#'
#' # Or match together a complete vector from missing pieces
#' y <- c(1, 2, NA, NA, 5)
#' z <- c(NA, NA, 3, 4, 5)
#' coalesce(y, z)
#'
#' @export
coalesce <- function(...) {
  if (missing(..1)) stop("At least one argument must be supplied.")

  vectors <- list(...)
  vectors_lens <- unique(lengths(vectors))
  if (length(vectors_lens) > 2L || (length(vectors_lens) == 2L && !1 %in% vectors_lens)) {
    stop("Vectors must all be of length 1 and/or n")
  }
  max_len <- max(vectors_lens)

  len_one <- lengths(vectors) == 1L
  vectors[len_one] <- lapply(vectors[len_one], function(x) rep(x, max_len))

  x <- vectors[[1]]
  vectors <- vectors[-1]

  for (i in seq_along(vectors)) {
    x_miss <- is.na(x)
    x[x_miss] <- vectors[[i]][x_miss]
  }
  x
}

expect_equal(
  coalesce(c(1, NA, 3, NA, 5), 0L),
  c(1, 0, 3, 0, 5),
  info = "Use a single value to replace all missing vectors"
)

expect_equal(
  {
    y <- c(1, 2, NA, NA, 5)
    z <- c(NA, NA, 3, 4, 5)
    coalesce(y, z)
  },
  1:5,
  info = "Match together a complete vector from missing pieces"
)

expect_equal(
  coalesce(NA, 2, c(1, 2, 3)),
  rep(2, 3),
  info = "Multiple vectors work"
)

expect_error(
  coalesce(c(1, NA, NA), c(1, 2)),
  info = "Differing lengths are not allowed when at least one vector is length 1: length(..1) > length(..2)"
)
expect_error(
  coalesce(c(1, NA), c(1, 2, 3)),
  info = "Differing lengths are not allowed when at least one vector is length 1: length(..1) < length(..2)"
)
expect_error(
  coalesce(NA, c(2, 3), c(1, 2, 3)),
  info = "Multiple vector lengths fails"
)

#' Subset distinct/unique rows
#'
#' Select only distinct/unique rows from a `data.frame`.
#'
#' @param .data A `data.frame`.
#' @param ... Optional variables to use when determining uniqueness. If there are multiple rows for a given combination
#' of inputs, only the first row will be preserved. If omitted, will use all variables.
#' @param .keep_all `logical(1)`. If `TRUE`, keep all variables in `.data`. If a combination of `...` is not distinct,
#' this keeps the first row of values.
#'
#' @examples
#' df <- data.frame(
#'   x = sample(10, 100, rep = TRUE),
#'   y = sample(10, 100, rep = TRUE)
#' )
#' nrow(df)
#' nrow(distinct(df))
#' nrow(distinct(df, x, y))
#'
#' distinct(df, x)
#' distinct(df, y)
#'
#' # You can choose to keep all other variables as well
#' distinct(df, x, .keep_all = TRUE)
#' distinct(df, y, .keep_all = TRUE)
#'
#' # You can also use distinct on computed variables
#' distinct(df, diff = abs(x - y))
#'
#' # The same behaviour applies for grouped data frames,
#' # except that the grouping variables are always included
#' df <- data.frame(
#'   g = c(1, 1, 2, 2),
#'   x = c(1, 1, 2, 1)
#' ) %>% group_by(g)
#' df %>% distinct(x)
#'
#' @return
#' A `data.frame` with the following properties:
#'
#' * Rows are a subset of the input but appear in the same order.
#' * Columns are not modified if `...` is empty or `.keep_all` is `TRUE`. Otherwise, `distinct()` first calls `mutate()`
#'   to create new columns.
#' * Groups are not modified.
#' * `data.frame` attributes are preserved.
#'
#' @export
distinct <- function(.data, ..., .keep_all = FALSE) {
  UseMethod("distinct")
}

#' @export
distinct.data.frame <- function(.data, ..., .keep_all = FALSE) {
  if (ncol(.data) == 0L) return(.data[1, ])
  cols <- dotdotdot(...)
  col_names <- names(cols)
  col_len <- length(cols)
  if (is.null(col_names) && col_len > 0L) names(cols) <- cols
  if (col_len == 0L) {
    res <- .data
  } else {
    mut <- mutate_df(.data, ...)
    res <- mut$data
    col_names <- names(cols)
    res <- if (!is.null(col_names)) {
      zero_names <- nchar(col_names) == 0L
      if (any(zero_names)) {
        names(cols)[zero_names] <- cols[zero_names]
        col_names <- names(cols)
      }
      suppressMessages(select(res, col_names))
    } else {
      suppressMessages(select(res, cols))
    }
  }
  res <- unique(res)
  if (isTRUE(.keep_all)) {
    res <- cbind(res, .data[rownames(res), setdiff(colnames(.data), colnames(res)), drop = FALSE])
  }
  common_cols <- c(intersect(colnames(.data), colnames(res)), setdiff(col_names, colnames(.data)))
  if (is.numeric(attr(res, "row.names"))) {
    row.names(res) <- seq_len(nrow(res))
  }
  if (length(common_cols) > 0L) res[, common_cols, drop = FALSE] else res
}

#' @export
distinct.grouped_df <- function(.data, ..., .keep_all = FALSE) {
  apply_grouped_function("distinct", .data, drop = TRUE, ..., .keep_all = .keep_all)
}

#' Filtering joins filter rows from `x` based on the presence or absence of
#' matches in `y`:
#'
#' * `semi_join()` return all rows from `x` with a match in `y`.
#' * `anti_join()` return all rows from `x` with*out* a match in `y`.
#'
#' @param x,y The `data.frame`s to join.
#' @param by A character vector of variables to join by. If `NULL`, the default,
#' `*_join()` will do a natural join, using all variables with common names
#' across the two tables. A message lists the variables so that you can check
#' they're right (to suppress the message, simply explicitly list the variables
#' that you want to join).
#'
#' @examples
#' table1 <- data.frame(
#'   pupil = rep(1:3, each = 2),
#'   test = rep(c("A", "B"), 3),
#'   score = c(60, 70, 65, 80, 85, 70),
#'   stringsAsFactors = FALSE
#' )
#' table2 <- table1[c(1, 3, 4), ]
#'
#' table1 %>% anti_join(table2, by = c("pupil", "test"))
#' table1 %>% semi_join(table2, by = c("pupil", "test"))
#'
#' @name filter_joins
NULL

#' @rdname filter_joins
#' @export
anti_join <- function(x, y, by = NULL) {
  filter_join_worker(x, y, by, type = "anti")
}

#' @rdname filter_joins
#' @export
semi_join <- function(x, y, by = NULL) {
  filter_join_worker(x, y, by, type = "semi")
}

filter_join_worker <- function(x, y, by = NULL, type = c("anti", "semi")) {
  type <- match.arg(type, choices = c("anti", "semi"), several.ok = FALSE)
  if (is.null(by)) {
    by <- intersect(names(x), names(y))
    join_message(by)
  }
  rows <- interaction(x[, by]) %in% interaction(y[, by])
  if (type == "anti") rows <- !rows
  res <- x[rows, , drop = FALSE]
  rownames(res) <- NULL
  reconstruct_attrs(res, x)
}

table1 <- data.frame(
  id = 1:5,
  animal = c("cat", "dog", "hamster", "parrot", "horse"),
  stringsAsFactors = FALSE
)
table2 <- table1[c(1, 3, 5), ]

expect_equal(
  {
    invisible(suppressMessages(res <- table1 %>% anti_join(table2)))
    res
  },
  data.frame(id = c(2, 4), animal = c("dog", "parrot"), stringsAsFactors = FALSE),
  info = "Single column anti-join"
)

expect_equal(
  {
    invisible(suppressMessages(res <- table1 %>% semi_join(table2)))
    res
  },
  data.frame(id = c(1L, 3L, 5L), animal = c("cat", "hamster", "horse"), stringsAsFactors = FALSE),
  info = "Single column semi-join"
)

table1 <- data.frame(
  pupil = rep(1:3, each = 2),
  test = rep(c("A", "B"), 3),
  score = c(60, 70, 65, 80, 85, 70),
  stringsAsFactors = FALSE
)
table2 <- table1[c(1, 3, 4), ]

expect_equal(
  table1 %>% anti_join(table2, by = c("pupil", "test")),
  data.frame(pupil = c(1, 3, 3), test = c("B", "A", "B"), score = c(70, 85, 70), stringsAsFactors = FALSE),
  info = "Multi-column anti-join"
)

expect_equal(
  table1 %>% semi_join(table2, by = c("pupil", "test")),
  data.frame(pupil = c(1L, 2L, 2L), test = c("A", "A", "B"), score = c(60, 65, 80), stringsAsFactors = FALSE),
  info = "Multi-column semi-join"
)

tbl1 <- data.frame(
  a = rep(LETTERS[1:3], 10),
  stringsAsFactors = FALSE
)
tbl2 <- data.frame(
  a = LETTERS[1:3],
  b = colours()[1:3],
  stringsAsFactors = FALSE
)
expect_equal(
  invisible(suppressMessages(res <- semi_join(tbl1, tbl2))),
  data.frame(
    a = rep(LETTERS[1:3], 10),
    stringsAsFactors = FALSE
  ),
  info = "Ensure `drop = FALSE` for single column data.frames."
)

#' Tools for working with row names
#'
#' In some quarters, it is considered best to avoid row names, because they are effectively a character column with
#' different semantics than every other column.
#' These functions allow to you detect if a `data.frame` has row names (`has_rownames()`), remove them
#' (`remove_rownames()`), or convert them back-and-forth between an explicit column (`rownames_to_column()` and
#' `column_to_rownames()`). Also included is `rowid_to_column()`, which adds a column at the start of the dataframe of
#' ascending sequential row ids starting at 1. Note that this will remove any existing row names.
#'
#' @param .data A `data.frame`.
#' @param var `character(1)`. The name of the column to use for row names.
#'
#' @returns
#' * `column_to_rownames()` always returns a `data.frame`.
#' * `has_rownames()` returns a `logical(1)`.
#' * All other functions return an object of the same class as the input.
#'
#' @examples
#' # Detect row names
#' has_rownames(mtcars)
#' has_rownames(iris)
#'
#' # Remove row names
#' remove_rownames(mtcars) %>% has_rownames()
#'
#' # Convert between row names and column
#' mtcars <- rownames_to_column(mtcars, var = "car")
#' column_to_rownames(mtcars, var = "car") %>% head()
#'
#' # Adding rowid as a column
#' rowid_to_column(iris) %>% head()
#'
#' @name rownames
#' @export
rownames_to_column <- function(.data, var = "rowname") {
  check_is_dataframe(.data)
  col_names <- colnames(.data)
  if (var %in% col_names) stop("Column `", var, "` already exists in `.data`.")
  .data[, var] <- rownames(.data)
  rownames(.data) <- NULL
  .data[, c(var, setdiff(col_names, var))]
}

#' @rdname rownames
#' @export
rowid_to_column <- function(.data, var = "rowid") {
  check_is_dataframe(.data)
  if (var %in% colnames(.data)) stop("Column `", var, "` already exists in `.data`.")
  args <- list(.data = .data, quote(row_number()), .before = 1L)
  names(args)[2L] <- var
  remove_rownames(do.call(mutate, args))
}

#' @rdname rownames
#' @export
column_to_rownames <- function(.data, var = "rowname") {
  check_is_dataframe(.data)
  if (has_rownames(.data)) stop("`.data` already has rownames")
  if (!(var %in% colnames(.data))) stop("Column `", var, "` could not be found in `.data`.")
  rownames(.data) <- .data[[var]]
  .data[[var]] <- NULL
  .data
}

#' @rdname rownames
#' @export
remove_rownames <- function(.data) {
  check_is_dataframe(.data)
  rownames(.data) <- NULL
  .data
}

#' @rdname rownames
#' @export
has_rownames <- function(.data) {
  .row_names_info(.data) > 0L
}

state_x77 <- as.data.frame(state.x77) %>% rownames_to_column(var = "State")

res <- state_x77 %>%
  pivot_longer(cols = -State, names_to = "Statistic", values_to = "Value")

expect_equal(dim(res), c(400, 3))
expect_equal(colnames(res), c("State", "Statistic", "Value"))

df <- data.frame(id = c(1, 2), v1.act = c(0.4, 0.8), v2.act = c(0.5, 0.7), v1.fix = c(1, 0), v2.fix = c(0, 1))
res <- pivot_longer(df, -id, names_to = c("var", "type"), names_pattern = "v(.).(.*)")
expect_equal(dim(res), c(8, 4))
expect_equal(colnames(res), c("id", "var", "type", "value"))

res <- pivot_longer(df, -id, names_to = c("var", "type"), names_sep = ".")
expect_equal(dim(res), c(8, 4))
expect_equal(colnames(res), c("id", "var", "type", "value"))

df <- data.frame(x = 1:2, y = 3:4)
pv <- pivot_longer(df, x:y)
expect_equal(names(pv), c("name", "value"))
expect_equal(pv$name, rep(names(df), 2))
expect_equal(pv$value, c(1, 3, 2, 4))


df <- data.frame(
  x = c(1, 2),
  y = c(10, 20),
  z = c(100, 200)
)
pv <- pivot_longer(df, 1:3)
expect_equal(pv$value, c(1, 10, 100, 2, 20, 200))


df <- data.frame(x = 1:2, y = 2, z = 1:2)
pv <- pivot_longer(df, y:z)
expect_equal(names(pv), c("x", "name", "value"))
expect_equal(pv$x, rep(df$x, each = 2))


df <- data.frame(x = c(1, NA), y = c(NA, 2))
pv <- pivot_longer(df, x:y, values_drop_na = TRUE)
expect_equal(pv$name, c("x", "y"))
expect_equal(pv$value, c(1, 2))


df <- data.frame(x = factor("a"), y = factor("b"))
pv <- pivot_longer(df, x:y)
expect_equal(pv$value, factor(c("a", "b")))


df <- data.frame(x = 1, y = 2)

expect_error(
  pivot_longer(df, y, names_to = "x"),
  pattern = "Some values of the columns specified in 'names_to' are already present"
)

expect_false(has_rownames(iris), info = "has_rownames() returns FALSE correctly")
expect_true(has_rownames(mtcars), info = "has_rownames() returns TRUE correctly")
expect_false(has_rownames(1:10), info = "has_rownames() works on other objects, returning FALSE")

expect_false(has_rownames(remove_rownames(mtcars)), info = "remove_rownames() acts accordingly")
expect_false(
  has_rownames(remove_rownames(iris)),
  info = "remove_rownames() acts accordingly for datasets without rownames"
)

res <- rowid_to_column(mtcars)
expect_false(has_rownames(res), info = "rownames are correctly removed")
expect_equal(
  mtcars %>% group_by(am) %>% rowid_to_column() %>% class(),
  c("grouped_df", "data.frame"),
  info = "rowid_to_column() retains classes"
)
expect_equal(res$rowid, seq_len(nrow(mtcars)), info = "rowid is correctly set")

res <- rownames_to_column(mtcars)
expect_false(has_rownames(res), info = "rownames have indeed been removed")
expect_equal(res$rowname, rownames(mtcars), info = "rownames have been correctly converted to a column")
expect_equal(
  mtcars %>% group_by(am) %>% rownames_to_column() %>% class(),
  c("grouped_df", "data.frame"),
  info = "rownames_to_column() retains class"
)

res <- column_to_rownames(res, "rowname")
expect_true(has_rownames(res), info = "column_to_rownames() does set rownames")
expect_equal(rownames(res), rownames(mtcars), info = "rownames are correctly set")
expect_equal(res, mtcars)
expect_false("rowname" %in% colnames(res), info = "The old column is removed")
expect_equal(
  mtcars %>% group_by(am) %>% rownames_to_column() %>% column_to_rownames("rowname") %>% class(),
  c("grouped_df", "data.frame"),
  info = "column_to_rownames() retains class"
)

expect_error(rownames_to_column(mtcars, "mpg"), info = "rownames_to_column: column already exists")
expect_error(rowid_to_column(mtcars, "mpg"), info = "rowid_to_column: columns already exists")
expect_error(column_to_rownames(mtcars, "mpg"), info = "column_to_rownames: data already has rownames")
expect_error(column_to_rownames(iris, "fake_column"), info = "column_to_rownames: column cannot be found")

#' Compute lagged or leading values
#'
#' Find the "previous" (`lag()`) or "next" (`lead()`) values in a vector. Useful for comparing values behind of or ahead
#' of the current values.
#'
#' @param x A `vector` of values
#' @param n A positive `integer(1)`, giving the number of positions to lead or lag by.
#' @param default The value used for non-existent rows (default: `NA`).
#'
#' @examples
#' lag(1:5)
#' lead(1:5)
#'
#' x <- 1:5
#' data.frame(behind = lag(x), x, ahead = lead(x))
#'
#' # If you want to look more rows behind or ahead, use `n`
#' lag(1:5, n = 1)
#' lag(1:5, n = 2)
#'
#' lead(1:5, n = 1)
#' lead(1:5, n = 2)
#'
#' # If you want to define a value for non-existing rows, use `default`
#' lag(1:5)
#' lag(1:5, default = 0)
#'
#' lead(1:5)
#' lead(1:5, default = 6)
#'
#' @export
lag <- function(x, n = 1L, default = NA) {
  if (inherits(x, "ts")) stop("`x` must be a vector, not a `ts` object, do you want `stats::lag()`?")
  if (length(n) != 1L || !is.numeric(n) || n < 0L) stop("`n` must be a nonnegative integer scalar")
  if (n == 0L) return(x)
  tryCatch(
    storage.mode(default) <- typeof(x),
    warning = function(w) {
      stop("Cannot convert `default` <", typeof(default), "> to `x` <", typeof(x), ">")
    }
  )
  xlen <- length(x)
  n <- pmin(n, xlen)
  res <- c(rep(default, n), x[seq_len(xlen - n)])
  attributes(res) <- attributes(x)
  res
}

#' @rdname lag
#' @export
lead <- function(x, n = 1L, default = NA) {
  if (length(n) != 1L || !is.numeric(n) || n < 0L) stop("n must be a nonnegative integer scalar")
  if (n == 0L) return(x)
  tryCatch(
    storage.mode(default) <- typeof(x),
    warning = function(w) {
      stop("Cannot convert `default` <", typeof(default), "> to `x` <", typeof(x), ">")
    }
  )
  xlen <- length(x)
  n <- pmin(n, xlen)
  res <- c(x[-seq_len(n)], rep(default, n))
  attributes(res) <- attributes(x)
  res
}

expect_error(lag(1:5, n = -1), info = "Lag needs a non-negative integer")
expect_error(lead(1:5, n = -1), info = "Lead needs a non-negative integer")

expect_equal(
  lag(1:5, n = 2),
  c(NA, NA, 1, 2, 3),
  info = "Lag works for larger values of `n`"
)
expect_equal(
  lead(1:5, n = 2),
  c(3, 4, 5, NA, NA),
  info = "Lead works for larger values of `n`"
)

expect_equal(
  lag(1:5, default = 0),
  0:4,
  info = "Lag accepts non-NA defaults"
)
expect_equal(
  lead(1:5, default = 6),
  2:6,
  info = "Lead accepts non-NA defaults"
)

expect_error(
  lag(1:5, "NA"),
  info = "Lag performs _some_ type checking"
)
expect_error(
  lead(1:5, "NA"),
  info = "Lead performs _some_ type checking"
)

x <- factor(c("a", "b", "c"))
expect_equal(
  levels(lag(x)),
  c("a", "b", "c"),
  info = "Lag preserves factor levels"
)
expect_equal(
  levels(lead(x)),
  c("a", "b", "c"),
  info = "Lead preserves factor levels"
)
rm(x)

x <- as.Date("2013-01-01") + 1:3
y <- as.POSIXct(x)
expect_equal(class(lag(x)), "Date", info = "Lag preserves dates")
expect_equal(class(lead(x)), "Date", info = "Lead preserves dates")
expect_equal(class(lag(y)), c("POSIXct", "POSIXt"), info = "Lag preserves dates and times")
expect_equal(class(lead(y)), c("POSIXct", "POSIXt"), info = "Lead preserves dates and times")
rm(x, y)

data <- data.frame(
  name = c("Rob", "Pete", "Rob", "John", "Rob", "Pete", "John", "Pete", "John", "Pete", "Rob", "Rob"),
  time = c(3, 2, 5, 3, 2, 3, 2, 4, 1, 1, 4, 1),
  stringsAsFactors = FALSE
)
res <- data %>% group_by(name) %>% mutate(lag_time = lag(time))
expect_equal(
  res$lag_time[res$name == "Rob"],
  c(NA, head(data$time[data$name == "Rob"], -1)),
  info = "Lag works in grouped operations"
)
expect_equal(
  res$lag_time[res$name == "Pete"],
  c(NA, head(data$time[data$name == "Pete"], -1)),
  info = "Lag works in grouped operations"
)
expect_equal(
  res$lag_time[res$name == "John"],
  c(NA, head(data$time[data$name == "John"], -1)),
  info = "Lag works in grouped operations"
)
rm(data, res)

data <- data.frame(
  name = rep(c("Al", "Jen"), 3),
  score = rep(c(100, 80, 60), 2),
  stringsAsFactors = FALSE
)
res <- data %>% group_by(name) %>% mutate(next.score = lead(score))
expect_equal(
  res$next.score[res$name == "Al"],
  c(tail(data$score[data$name == "Al"], -1), NA),
  info = "Lead works in grouped operations"
)
expect_equal(
  res$next.score[res$name == "Jen"],
  c(tail(data$score[data$name == "Jen"], -1), NA),
  info = "Lead works in grouped operations"
)
rm(data, res)

#' Get a glimpse of your data
#'
#' `glimpse()` is like a transposed version of print(): columns run down the page, and data runs across. This makes it
#' possible to see every column in a `data.frame`. It is no more than a wrapper around [utils::str()] only it returns
#' the input (invisibly) meaning it can be used within a data pipeline.
#'
#' @param x An object to glimpse at.
#' @param width `integer(1)`. Width of the output.
#' @param ... Additional parameters to pass to [utils::str()].
#'
#' @return
#' `x`, invisibly.
#'
#' @examples
#' glimpse(mtcars)
#'
#' @export
glimpse <- function(x, width = getOption("width"), ...) {
  UseMethod("glimpse")
}

#' @export
glimpse.default <- function(x, width = getOption("width"), max.level = 3, ...) {
  utils::str(x, width = width, max.level = max.level, ...)
  invisible(x)
}

#' @export
glimpse.data.frame <- function(x, width = getOption("width"), ...) {
  utils::str(x, width = width, ...)
  invisible(x)
}

#' Subset columns using their names and types
#'
#' Select (and optionally rename) variables in a `data.frame`, using a concise mini-language that makes it easy to refer
#' to variables based on their name (e.g. `a:f` selects all columns from `a` on the left to `f` on the right). You can
#' also use predicate functions like [is.numeric()] to select variables based on their properties.
#'
#' ## Overview of selection features
#'
#' {poorman} selections implement a dialect of R where operators make it easy to select variables:
#'
#' - `:` for selecting a range of consecutive variables.
#' - `!` for taking the complement of a set of variables.
#' - `&` and `|` for selecting the intersection or the union of two sets of variables.
#' - `c()` for combining selections.
#'
#' In addition, you can use __selection helpers__. Some helpers select specific columns:
#' * [`everything()`]: Matches all variables.
#' * [`last_col()`]: Select last variable, possibly with an offset.
#'
#' These helpers select variables by matching patterns in their names:
#' * [`starts_with()`]: Starts with a prefix.
#' * [`ends_with()`]: Ends with a suffix.
#' * [`contains()`]: Contains a literal string.
#' * [`matches()`]: Matches a regular expression.
#' * [`num_range()`]: Matches a numerical range like `x01`, `x02`, `x03`.
#'
#' These helpers select variables from a character vector:
#' * [`all_of()`]: Matches variable names in a character vector. All names must be present, otherwise an out-of-bounds
#' error is thrown.
#' * [`any_of()`]: Same as `all_of()`, except that no error is thrown for names that don't exist.
#'
#' This helper selects variables with a function:
#' * [`where()`]: Applies a function to all variables and selects those for which the function returns `TRUE`.
#'
#' @param .data A `data.frame`.
#' @param ... <[`poor-select`][select_helpers]> One or more unquoted expressions separated by commas. Variable names can
#' be used as if they were positions in the data frame, so expressions like `x:y` can be used to select a range of
#' variables.
#'
#' @return
#' An object of the same type as `.data`. The output has the following properties:
#'
#' * Rows are not affected.
#' * Output columns are a subset of input columns, potentially with a different order. Columns will be renamed if
#' `new_name = old_name` form is used.
#' * Data frame attributes are preserved.
#' * Groups are maintained; you can't select off grouping variables.
#'
#' @examples
#' # Here we show the usage for the basic selection operators. See the
#' # specific help pages to learn about helpers like [starts_with()].
#'
#' # Select variables by name:
#' mtcars %>% select(mpg)
#'
#' # Select multiple variables by separating them with commas. Note
#' # how the order of columns is determined by the order of inputs:
#' mtcars %>% select(disp, gear, am)
#'
#' # Rename variables:
#' mtcars %>% select(MilesPerGallon = mpg, everything())
#'
#' # The `:` operator selects a range of consecutive variables:
#' select(mtcars, mpg:cyl)
#'
#' # The `!` operator negates a selection:
#' mtcars %>% select(!(mpg:qsec))
#' mtcars %>% select(!ends_with("p"))
#'
#' # `&` and `|` take the intersection or the union of two selections:
#' iris %>% select(starts_with("Petal") & ends_with("Width"))
#' iris %>% select(starts_with("Petal") | ends_with("Width"))
#'
#' # To take the difference between two selections, combine the `&` and
#' # `!` operators:
#' iris %>% select(starts_with("Petal") & !ends_with("Width"))
#'
#' @export
select <- function(.data, ...) {
  col_pos <- select_positions(.data, ..., .group_pos = TRUE)
  res <- .data[, col_pos, drop = FALSE]
  if (length(names(res)) != 0) colnames(res) <- names(col_pos)
  if (has_groups(.data)) res <- groups_set(res, group_vars(.data))
  res
}

#' Efficiently bind multiple `data.frame`s by row and column
#'
#' @param ... `data.frame`s to combine.
#'
#' Each argument can either be a `data.frame`, a `list` that could be a `data.frame`, or a `list` of `data.frame`s.
#'
#' When row-binding, columns are matched by name, and any missing columns will be filled with `NA`.
#'
#' When column-binding, rows are matched by position, so all `data.frame`s must have the same number of rows. To match
#' by value, not position, see [mutate_joins].
#' @param .id `character(1)`. `data.frame` identifier.
#'
#' When `.id` is supplied, a new column of identifiers is created to link each row to its original `data.frame`. The
#' labels are taken from the named arguments to `bind_rows()`. When a `list` of `data.frame`s is supplied, the labels
#' are taken from the names of the `list`. If no names are found a numeric sequence is used instead.
#'
#' @examples
#' one <- mtcars[1:4, ]
#' two <- mtcars[9:12, ]
#'
#' # You can supply data frames as arguments:
#' bind_rows(one, two)
#'
#' # The contents of lists are spliced automatically:
#' bind_rows(list(one, two))
#' bind_rows(split(mtcars, mtcars$cyl))
#' bind_rows(list(one, two), list(two, one))
#'
#' # In addition to data frames, you can supply vectors. In the rows
#' # direction, the vectors represent rows and should have inner
#' # names:
#' bind_rows(
#'   c(a = 1, b = 2),
#'   c(a = 3, b = 4)
#' )
#'
#' # You can mix vectors and data frames:
#' bind_rows(
#'   c(a = 1, b = 2),
#'   data.frame(a = 3:4, b = 5:6),
#'   c(a = 7, b = 8)
#' )
#'
#' # When you supply a column name with the `.id` argument, a new
#' # column is created to link each row to its original data frame
#' bind_rows(list(one, two), .id = "id")
#' bind_rows(list(a = one, b = two), .id = "id")
#' bind_rows("group 1" = one, "group 2" = two, .id = "groups")
#'
#' \dontrun{
#' # Rows need to match when column-binding
#' bind_cols(data.frame(x = 1:3), data.frame(y = 1:2))
#'
#' # even with 0 columns
#' bind_cols(data.frame(x = 1:3), data.frame())
#' }
#'
#' bind_cols(one, two)
#' bind_cols(list(one, two))
#'
#' @name bind
NULL

#' @rdname bind
#' @export
bind_cols <- function(...) {
  lsts <- list(...)
  lsts <- squash(lsts)
  lsts <- Filter(Negate(is.null), lsts)
  if (length(lsts) == 0L) return(data.frame())
  lapply(lsts, function(x) is_df_or_vector(x))
  lsts <- do.call(cbind, lsts)
  if (!is.data.frame(lsts)) lsts <- as.data.frame(lsts)
  lsts
}

#' @rdname bind
#' @export
bind_rows <- function(..., .id = NULL) {
  lsts <- list(...)
  lsts <- flatten(lsts)
  lsts <- Filter(Negate(is.null), lsts)
  lapply(lsts, function(x) is_df_or_vector(x))
  lapply(lsts, function(x) if (is.atomic(x) && !is_named(x)) stop("Vectors must be named."))

  if (!missing(.id)) {
    lsts <- lapply(seq_along(lsts), function(i) {
      nms <- names(lsts)
      id_df <- data.frame(id = if (is.null(nms)) as.character(i) else nms[i], stringsAsFactors = FALSE)
      colnames(id_df) <- .id
      cbind(id_df, lsts[[i]])
    })
  }

  nms <- unique(unlist(lapply(lsts, names)))
  lsts <- lapply(
    lsts,
    function(x) {
      if (!is.data.frame(x)) x <- data.frame(as.list(x), stringsAsFactors = FALSE)
      for (i in nms[!nms %in% names(x)]) x[[i]] <- NA
      x
    }
  )
  names(lsts) <- NULL
  do.call(rbind, lsts)
}

one <- mtcars[1:4, ]
two <- mtcars[9:12, ]

expect_equal(
  bind_rows(one, two),
  rbind(one, two),
  info = "data.frames can be supplied as arguments"
)

expect_equal(
  bind_rows(list(one, two)),
  rbind(one, two),
  info = "data.frames can be supplied in a list"
)
expect_equal(
  bind_rows(split(mtcars, mtcars$cyl)),
  mtcars[order(mtcars$cyl), ],
  info = "contents of lists are spliced automatically"
)
expect_equal(
  bind_rows(list(one, two), list(two, one)),
  rbind(one, two, two, one),
  info = "contents of lists are spliced automatically"
)

expect_equal(
  bind_rows(c(a = 1, b = 2), c(a = 3, b = 4)),
  data.frame(a = c(1, 3), b = c(2, 4)),
  info = "Vectors can be supplied"
)

expect_error(
  bind_rows(c(1, 2), c(a = 3, b = 4)),
  info = "Vectors must be named"
)

expect_equal(
  bind_rows(c(a = 1, b = 2), data.frame(a = 3:4, b = 5:6), c(a = 7, b = 8)),
  data.frame(a = c(1, 3, 4, 7), b = c(2, 5, 6, 8)),
  info = "A mixture of vectors and data.frames can be supplied"
)

expect_equal(
  bind_rows(list(one, two), .id = "id"),
  {
    res <- rbind(one, two)
    res[, "id"] <- c(rep("1", 4), rep("2", 4))
    res[, c("id", "mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb")]
  },
  info = "If no names are found a numeric sequence is used."
)
expect_equal(
  bind_rows(list(a = one, b = two), .id = "id"),
  {
    res <- rbind(one, two)
    res[, "id"] <- c(rep("a", 4), rep("b", 4))
    res[, c("id", "mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb")]
  },
  info = "When a list of data.frames is supplied, the labels are taken from the names of the list."
)
expect_equal(
  bind_rows("group 1" = one, "group 2" = two, .id = "groups"),
  {
    res <- rbind(one, two)
    res[, "groups"] <- c(rep("group 1", 4), rep("group 2", 4))
    res[, c("groups", "mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb")]
  },
  info = "Named dots are used for groups when supplied."
)

df <- data.frame(x = 1)
expect_equal(bind_rows(df, NULL), df, info = "bind_rows() handles NULL")
expect_equal(bind_rows(list(df, NULL)), df, info = "bind_rows() handles NULL")
rm(df)

expect_equal(
  bind_rows(c(z = 2), data.frame(x = 1:3), data.frame(y = 1:4)),
  data.frame(
    z = c(2, rep(NA_real_, 7)),
    x = c(NA_integer_, 1L:3L, rep(NA_integer_, 4)),
    y = c(rep(NA_integer_, 4), 1L:4L)
  ),
  info = "Columns missing from subsequent data.frames are filled with NAs"
)

expect_error(
  bind_rows(c(z = 2), data.frame(x = 1:3), y ~ x),
  info = "You must pass vector(s) and/or data.frame(s)."
)

# -- bind_cols() ---------------------------------------------------------------

expect_equal(
  bind_cols(one, two),
  cbind(one, two),
  info = "data.frames can be supplied as arguments"
)

expect_equal(
  bind_cols(c(1, 2), c(3, 4)),
  data.frame(V1 = c(1, 2), V2 = c(3, 4)),
  info = "vectors can be supplied as arguments"
)

expect_equal(
  bind_cols(one, test = c(1, 2, 3, 4)),
  {
    res <- mtcars[1:4, ]
    res[, "test"] <- 1:4
    res
  },
  info = c("A mixture of data.frames and lists can be passed")
)

expect_equal(
  bind_cols(list(one, two)),
  cbind(one, two),
  info = "A list of data.frames can be supplied as arguments"
)

expect_identical(
  bind_cols(NULL),
  data.frame(),
  info = "bind_cols() handles NULL"
)
expect_identical(
  bind_cols(list(a = NULL, b = NULL)),
  data.frame(),
  info = "bind_cols handles all-NULL values"
)

expect_equal(
  bind_cols(list(y = data.frame(x = 1:2), z = data.frame(y = 1:2))),
  data.frame(x = 1:2, y = 1:2),
  info = "bind_cols unpacks data.frames"
)

expect_error(
  bind_cols(data.frame(x = 1:3), data.frame(y = 1:2)),
  info = "Rows need to match when column-binding"
)

expect_error(
  bind_cols(data.frame(x = 1:3), data.frame()),
  info = "Rows need to match when column-binding even with 0 columns"
)

expect_error(
  bind_cols(list(y = data.frame(x = 1:2), y ~ x)),
  info = "You must pass vector(s) and/or data.frame(s)"
)

rm(one, two)

expect_equal(
  union_all(1:3, 4:6),
  1:6,
  info = "union_all() on vectors concatenates"
)

df1 <- data.frame(x = 1:2)
df2 <- data.frame(y = 1:2)
expect_equal(
  union_all(df1, df2),
  bind_rows(df1, df2),
  info = "union_all() on data frames calls bind rows"
)

#' Create or transform variables
#'
#' `mutate()` adds new variables and preserves existing ones; `transmute()` adds new variables and drops existing ones.
#' Both functions preserve the number of rows of the input. New variables overwrite existing variables of the same name.
#' Variables can be removed by setting their value to `NULL`.
#'
#' @section Useful mutate functions:
#'
#' * [`+`], [`-`], [log()], etc., for their usual mathematical meanings
#'
#' * [lead()], [lag()]
#'
#' * [dense_rank()], [min_rank()], [percent_rank()], [row_number()], [cume_dist()], [ntile()]
#'
#' * [cumsum()], [cummin()], [cummax()]
#'
#' * [na_if()], [coalesce()]
#'
#' * [if_else()], [recode()], [case_when()]
#'
#' @param .data A `data.frame`.
#' @param ... Name-value pairs of expressions, each with length `1L`. The name of each argument will be the name of a
#' new column and the value will be its corresponding value. Use a `NULL` value in `mutate` to drop a variable. New
#' variables overwrite existing variables of the same name.
#'
#' @examples
#' mutate(mtcars, mpg2 = mpg * 2)
#' mtcars %>% mutate(mpg2 = mpg * 2)
#' mtcars %>% mutate(mpg2 = mpg * 2, cyl2 = cyl * 2)
#'
#' # Newly created variables are available immediately
#' mtcars %>% mutate(mpg2 = mpg * 2, mpg4 = mpg2 * 2)
#'
#' # You can also use mutate() to remove variables and modify existing variables
#' mtcars %>% mutate(
#'   mpg = NULL,
#'   disp = disp * 0.0163871 # convert to litres
#' )
#'
#' # By default, new columns are placed on the far right.
#' # You can override this with `.before` or `.after`.
#' df <- data.frame(x = 1, y = 2)
#' df %>% mutate(z = x + y)
#' df %>% mutate(z = x + y, .before = 1)
#' df %>% mutate(z = x + y, .after = x)
#'
#' # By default, mutate() keeps all columns from the input data.
#' # You can override with `.keep`
#' df <- data.frame(
#'   x = 1, y = 2, a = "a", b = "b",
#'   stringsAsFactors = FALSE
#' )
#' df %>% mutate(z = x + y, .keep = "all") # the default
#' df %>% mutate(z = x + y, .keep = "used")
#' df %>% mutate(z = x + y, .keep = "unused")
#' df %>% mutate(z = x + y, .keep = "none") # same as transmute()
#'
#' # mutate() vs transmute --------------------------
#' # mutate() keeps all existing variables
#' mtcars %>%
#'   mutate(displ_l = disp / 61.0237)
#'
#' # transmute keeps only the variables you create
#' mtcars %>%
#'   transmute(displ_l = disp / 61.0237)
#'
#' @name mutate
#' @export
mutate <- function(.data, ...) {
  UseMethod("mutate")
}

#' @rdname mutate
#' @param .keep  This argument allows you to control which columns from `.data` are retained in the output:
#'
#' * `"all"`, the default, retains all variables.
#' * `"used"` keeps any variables used to make new variables; it's useful for checking your work as it displays inputs
#'   and outputs side-by-side.
#' * `"unused"` keeps only existing variables **not** used to make new variables.
#' * `"none"`, only keeps grouping keys (like [transmute()]).
#'
#' Grouping variables are always kept, unconditional to `.keep`.
#' @param .before,.after <[`poor-select`][select_helpers]> Optionally, control where new columns should appear (the
#' default is to add to the right hand side). See [relocate()] for more details.
#' @export
mutate.data.frame <- function(
  .data,
  ...,
  .keep = c("all", "used", "unused", "none"),
  .before = NULL,
  .after = NULL
) {
  keep <- match.arg(arg = .keep, choices = c("all", "used", "unused", "none"), several.ok = FALSE)

  res <- mutate_df(.data = .data, ...)
  data <- res$data
  new_cols <- res$new_cols

  .before <- substitute(.before)
  .after <- substitute(.after)
  if (!is.null(.before) || !is.null(.after)) {
    new <- setdiff(new_cols, names(.data))
    data <- do.call(relocate, c(list(.data = data), new, .before = .before, .after = .after))
  }

  if (keep == "all") {
    data
  } else if (keep == "unused") {
    unused <- setdiff(colnames(.data), res$used_cols)
    keep <- intersect(colnames(data), c(group_vars(.data), unused, new_cols))
    select(.data = data, keep)
  } else if (keep == "used") {
    keep <- intersect(colnames(data), c(group_vars(.data), res$used_cols, new_cols))
    select(.data = data, keep)
  } else if (keep == "none") {
    keep <- c(setdiff(group_vars(.data), new_cols), intersect(new_cols, colnames(data)))
    select(.data = data, keep)
  }
}

#' @export
mutate.grouped_df <- function(.data, ...) {
  context$group_env <- parent.frame(n = 1)
  on.exit(rm(list = c("group_env"), envir = context), add = TRUE)
  rows <- rownames(.data)
  res <- apply_grouped_function("mutate", .data, drop = TRUE, ...)
  res[rows, , drop = FALSE]
}

## -- Helpers ----------------------------------------------------------------------------------------------------------

mutate_df <- function(.data, ...) {
  conditions <- dotdotdot(..., .impute_names = TRUE)
  cond_nms <- names(dotdotdot(..., .impute_names = FALSE))
  if (length(conditions) == 0L) {
    return(list(
      data = .data,
      used_cols = NULL,
      new_cols = NULL
    ))
  }
  used <- unname(do.call(c, lapply(conditions, find_used)))
  used <- used[used %in% colnames(.data)]
  context$setup(.data)
  on.exit(context$clean(), add = TRUE)
  for (i in seq_along(conditions)) {
    not_named <- (is.null(cond_nms) || cond_nms[i] == "")
    res <- eval(
      conditions[[i]],
      envir = context$as_env(),
      enclos = if (!is.null(context$group_env)) context$group_env else parent.frame(n = 2)
    )
    res_nms <- names(res)
    if (is.data.frame(res)) {
      if (not_named) {
        context$.data[, res_nms] <- res
      } else {
        context$.data[[cond_nms[i]]] <- res
      }
    } else if (is.atomic(res)) {
      cond_nms[i] <- names(conditions)[[i]]
      context$.data[[cond_nms[i]]] <- res
    } else {
      context$.data[[names(conditions)[[i]]]] <- res
    }
  }
  list(
    data = context$.data,
    used_cols = used,
    new_cols = cond_nms
  )
}

#' Recursively find used variables in an expression
#'
#' @param expr An expression.
#'
#' @return `character(n)`.
#'
#' @examples
#' \dontrun{
#' expr <- quote(x * var + ifelse(x < y, 1, 0) + n())
#' find_used(expr)
#' }
#'
#' @author Mark T Fairbanks, \email{mark.t.fairbanks@@gmail.com}
#'
#' @noRd
find_used <- function(expr) {
  if (is.symbol(expr)) {
    as.character(expr)
  } else {
    unique(unlist(lapply(expr[-1], find_used)))
  }
}

df <- data.frame(x = 1)
gf <- group_by(df, x)
expect_equal(mutate(df), df, info = "empty mutate() returns input")
expect_equal(mutate(gf), gf, info = "empty mutate() returns input for grouped data")

expect_equal(
  mtcars %>% mutate(mpg2 = mpg * 2),
  {
    res <- mtcars
    res[, "mpg2"] <- res$mpg * 2
    res
  },
  info = "Test mutate() can add a new column"
)

df <- data.frame(x = 1:3, y = 1:3)
gf <- group_by(df, x)
expect_equal(df %>% mutate(y = NULL), df[1], info = "mutate() can remove variables with NULL")
expect_equal(
  df %>% mutate(z = NULL),
  df,
  info = "mutate() can remove variables with NULL even if it doesn't exist"
)
expect_equal(
  df %>% mutate(z = 1, z = NULL),
  df,
  info = "mutate() can remove variables with NULL even if it was just created"
)

expect_equal(
  mtcars %>% mutate(mpg = mpg * 2),
  {
    res <- mtcars
    res[, "mpg"] <- res$mpg * 2
    res
  },
  info = "Test mutate() can transform existing columns"
)

df <- data.frame(x = 1)
expect_equal(
  df %>% mutate(y = x + 1, z = y + 1),
  data.frame(x = 1, y = 2, z = 3),
  info = "mutations applied progressively: 1"
)
expect_equal(
  df %>% mutate(x = x + 1, x = x + 1),
  data.frame(x = 3),
  info = "mutations applied progressively: 2"
)
expect_equal(
  df %>% mutate(x = 2, y = x),
  data.frame(x = 2, y = 2),
  info = "mutations applied progressively: 3"
)
df <- data.frame(x = 1, y = 2)
expect_equal(
  df %>% mutate(x2 = x, x3 = x2 + 1),
  df %>% mutate(x2 = x + 0, x3 = x2 + 1),
  info = "mutations applied progressively: 4"
)

expect_equal(
  mtcars %>% mutate(1L),
  {
    res <- mtcars
    res[, "1L"] <- 1L
    res
  },
  info = "Test that unnamed conditions are evaluated"
)

# Grouped Operations
expect_equal(
  mtcars %>% group_by(am, cyl) %>% mutate(mpg2 = mpg * 2) %>% ungroup(),
  {
    res <- mtcars
    res <- do.call(rbind, unname(lapply(
      split(res, list(res$am, res$cyl)),
      function(x) {
        x[, "mpg2"] <- x$mpg * 2
        x
      }
    )))
    res[rownames(mtcars), ]
  },
  info = "Test grouped mutations"
)

expect_equal(
  mtcars %>% group_by(am, cyl, gear) %>% mutate(col = 1) %>% ungroup(),
  {
    res <- mtcars
    res <- do.call(rbind, unname(lapply(
      split(res, list(res$am, res$cyl)),
      function(x) {
        x[, "col"] <- 1
        x
      }
    )))
    res[rownames(mtcars), ]
  },
  info = "mutate() works when there are missing groups"
)

df <- data.frame(x = 1, y = 2)
out <- df %>% mutate(z = {x <- 10; x})
expect_equal(out, data.frame(x = 1, y = 2, z = 10), info = "assignments don't overwrite variables")

df <- structure(list(), class = "data.frame", row.names = c(NA, -2L), .Names = character(0))
colnames(df) <- NULL
expect_equal(
  mutate(df, x = 1),
  data.frame(x = c(1, 1)),
  info = "can mutate a data frame with zero columns and `NULL` column names"
)

df <- data.frame(x = c(1, NA, NaN))
out <- mutate(df, y = x * 1)
expect_equal(out$y, df$x, info = "mutate() disambiguates NA and NaN")

expect_identical(
  data.frame(a = 1) %>% mutate(data.frame(b = 2)),
  data.frame(a = 1, b = 2),
  info = "unnamed data frames are automatically unspliced: 1"
)
expect_identical(
  data.frame(a = 1) %>% mutate(data.frame(b = 2), data.frame(b = 3)),
  data.frame(a = 1, b = 3),
  info = "unnamed data frames are automatically unspliced: 2"
)
expect_identical(
  data.frame(a = 1) %>% mutate(data.frame(b = 2), c = b),
  data.frame(a = 1, b = 2, c = 2),
  info = "unnamed data frames are automatically unspliced: 3"
)

df <- data.frame(x = 1)
out <- df %>% mutate(y = data.frame(a = x))
expect_equal(
  out,
  {
    res <- data.frame(x = 1)
    res[["y"]] <- data.frame(a = 1)
    res
  },
  info = "named data frames are packed"
)

gf <- group_by(data.frame(x = 1:2, y = 2), x)
out <- mutate(gf, x = 1)
expect_equal(group_vars(out), "x", info = "mutate() preserves grouping: 1")
expect_equal(nrow(group_data(out)), 1, info = "mutate() preserves grouping: 2")
out <- mutate(gf, z = 1)
expect_equal(group_data(out), group_data(gf), info = "mutate() preserves grouping: 3")

df <- data.frame()
res <- df %>% mutate()
expect_equal(nrow(res), 0L, info = "mutate() works on empty data frames: 1")
expect_equal(length(res), 0L, info = "mutate() works on empty data frames: 2")
res <- df %>% mutate(x = numeric())
expect_equal(names(res), "x", info = "mutate() works on empty data frames: 3")
expect_equal(nrow(res), 0L, info = "mutate() works on empty data frames: 4")
expect_equal(length(res), 1L, info = "mutate() works on empty data frames: 5")

d <- data.frame(subject = c("Jack", "Jill"), id = c(2, 1)) %>% group_by(subject)
a1 <- names(attributes(d))
a2 <- names(attributes(d %>% mutate(foo = 1)))
expect_equal(setdiff(a1, a2), character(0), info = "grouped mutate does not drop grouping attributes")

df <- data.frame(x = c(1, 2), y = c(1, NA))
res <- df %>% group_by(x) %>% mutate(z = ifelse(y > 1, 1, 2))
expect_true(is.na(res$z[2]), info = "mutate() coerces results from one group with all NA values: 1")
expect_true(inherits(res$z, "numeric"), info = "mutate() coerces results from one group with all NA values: 2")

# Evaluation envs
pass_through <- function(x) x
test_function <- function(mtcars_df) {
  pass_through_internal <- function(x) x
  mutate(.data = mtcars_df, carb = pass_through(carb), gear = pass_through_internal(gear))
}
expect_equal(
  test_function(mtcars),
  mtcars,
  info = "Ensure mutations occur in the correct enclosing environment"
)

test_function_grouped <- function(mtcars_df) {
  pass_through_internal_group <- function(x) x

  mtcars_df %>%
    group_by(cyl) %>%
    mutate(carb = pass_through(carb), gear = pass_through_internal_group(gear)) %>%
    ungroup()
}
expect_equal(
  test_function_grouped(mtcars),
  mtcars,
  info = "Ensure mutations occur in the correct enclosing environment for grouped data (#68)"
)

# List columns

df <- structure(list(), class = "data.frame", row.names = c(NA, -3L), .Names = character(0))
df[["x"]] <- list(1, 2:3, 4:6)
df[["y"]] <- 1:3
expect_equal(
  df %>% mutate(l = length(x)) %>% .[["l"]],
  c(3, 3, 3),
  info = "List columns can be mutated: 1"
)
expect_equal(
  df %>% mutate(l = lengths(x)) %>% .[["l"]],
  c(1, 2, 3),
  info = "List columns can be mutated: 2"
)
models <- mtcars %>% nest_by(cyl) %>% mutate(model = list(lm(mpg ~ wt, data = data)))
expect_equal(
  lapply(models$model, class),
  list("lm", "lm", "lm"),
  info = "List columns can be mutated: 3"
)

# .keep

df <- data.frame(x = 1, y = 2)
out <- mutate(df, x1 = x + 1, y = y, .keep = "unused")
expect_equal(colnames(out), c("y", "x1"), info = "`.keep` = 'unused' keeps variables explicitly mentioned")

df <- data.frame(x = 1, y = 2, z = 3, a = "a", b = "b", c = "c", stringsAsFactors = FALSE)
out <- mutate(df, across(where(is.numeric), identity), .keep = "unused")
expect_equal(colnames(out), colnames(df), info = ".keep = 'used' not affected by across()")

df <- data.frame(a = 1, b = 2, c = 3, x = 1, y = 2)
out <- mutate(df, xy = x + y, .keep = "used")
expect_equal(colnames(out), c("x", "y", "xy"), info = ".keep = 'used' keeps variables used in expressions")

test_df <- data.frame(x = 1:3, y = 1:3, z = 1:3)
var <- 1
test_expr <- quote(x * var + ifelse(x < y, 1, 0) + n())
used <- poorman:::find_used(test_expr)
expect_equal(
  used[used %in% colnames(test_df)],
  c("x", "y"),
  info = "find_used() works for variables used inside of functions"
)
rm(test_df, var, test_expr, used)

df <- data.frame(x = 1, y = 2)
gf <- group_by(df, x)
expect_equal(
  colnames(mutate(df, z = 1, .keep = "none")),
  "z",
  info = ".keep = 'none' only keeps mutated variables"
)
expect_equal(
  colnames(mutate(gf, z = 1, .keep = "none")),
  c("x", "z"),
  info = ".keep = 'none' only keeps grouping and mutated variables"
)

df <- data.frame(x = 1, y = 2)
expect_equal(
  colnames(mutate(df, y = 1, x = 2, .keep = "none")),
  c("y", "x"),
  info = ".keep = 'none' prefers new order"
)

gf <- group_by(df, x)
expect_equal(
  colnames(mutate(gf, y = 1, x = 2, .keep = "none")),
  c("y", "x"),
  info = ".keep = 'none' prefers new order even when grouped"
)

df <- data.frame(x = 1, y = 2, z = 3) %>% group_by(z)
expect_equal(
  df %>% mutate(a = x + 1, .keep = "none"),
  data.frame(z = 3, a = 2) %>% group_by(z),
  info = ".keep = 'none' always retains grouping variables"
)
expect_equal(
  df %>% mutate(a = x + 1, .keep = "all"),
  data.frame(x = 1, y = 2, z = 3, a = 2) %>% group_by(z),
  info = ".keep = 'all' always retains grouping variables"
)
expect_equal(
  df %>% mutate(a = x + 1, .keep = "used"),
  data.frame(x = 1, z = 3, a = 2) %>% group_by(z),
  info = ".keep = 'used' always retains grouping variables"
)
expect_equal(
  df %>% mutate(a = x + 1, .keep = "unused"),
  data.frame(y = 2, z = 3, a = 2) %>% group_by(z),
  info = ".keep = 'unused' always retains grouping variables"
)

# .before, .after

df <- data.frame(x = 1, y = 2)
expect_equal(
  colnames(mutate(df, z = 1)),
  c("x", "y", "z"),
  info = "lack of .before and .after doesn't affect column position"
)
expect_equal(
  colnames(mutate(df, z = 1, .before = 1)),
  c("z", "x", "y"),
  info = "can use .before to control column position"
)
expect_equal(
  colnames(mutate(df, z = 1, .after = 1)),
  c("x", "z", "y"),
  info = "can use .after to control column position"
)
expect_equal(
  colnames(mutate(df, x = 1, .after = y)),
  c("x", "y"),
  info = ".after/.before doesn't affect order of existing columns"
)

# Errors

df <- data.frame(x = 1:2, y = 1:2)
expect_error(df %>% mutate(y = NULL, a = sum(y)), info = "setting column to NULL makes it unavailable")
expect_error(
  df %>% group_by(x) %>% mutate(y = NULL, a = sum(y)),
  info = "setting column to NULL makes it unavailable for grouped df"
)

reconstruct_attrs <- function(data, template) {
  data <- remove_attributes(data)
  return(reconstruct_attrs_dispatch(data, template))
  UseMethod(generic = "reconstruct_attrs", object = template)
}

reconstruct_attrs_dispatch <- function(data, template) {
  UseMethod("reconstruct_attrs", template)
}

reconstruct_attrs.data.frame <- function(data, template) {
  attrs <- attributes(template)
  attrs$names <- names(data)
  attrs$row.names <- .row_names_info(data, type = 0L)
  attributes(data) <- attrs
  data
}

reconstruct_attrs.grouped_df <- function(data, template) {
  group_vars <- intersect(group_vars(template), colnames(data))
  structure(
    groups_set(data, group_vars, drop = group_by_drop_default(template)),
    class = c("grouped_df", "data.frame")
  )
}

remove_attributes <- function(data) {
  attrs <- attributes(data)
  foreign <- which(!names(attrs) %in% c("names", "row.names", "class"))
  if (length(foreign) > 0L) {
    for (i in foreign) {
      attr(data, names(attrs)[i]) <- NULL
    }
  }
  as.data.frame(data)
}

reconstruct_attrs <- poorman:::reconstruct_attrs
reconstruct_attrs.data.frame <- poorman:::reconstruct_attrs.data.frame

expect_identical(
  reconstruct_attrs(data.frame(), data.frame()),
  data.frame(),
  info = "classes are restored for data.frames"
)
expect_identical(
  reconstruct_attrs(data.frame(), structure(data.frame(), class = c("foo", "data.frame"))),
  structure(data.frame(), class = c("foo", "data.frame")),
  info = "classes are restored for additional classes"
)

expect_identical(
  reconstruct_attrs(
    structure(list(), .Names = character(0), row.names = c(NA, -1L), class = "data.frame"),
    structure(list(), class = "data.frame", row.names = integer(0), .Names = character(0), foo = 1)
  ),
  structure(list(), class = "data.frame", row.names = c(NA, -1L), .Names = character(0), foo = 1),
  info = "attributes of `template` are kept"
)

data <- data.frame(a = c(1, 2))
template <- data.frame()
x <- reconstruct_attrs(data, template)
expect <- data.frame(a = c(1, 2))
expect_identical(x, expect, info = "compact row names are retained")
expect_identical(
  .row_names_info(x, type = 0L),
  .row_names_info(expect, type = 0L),
  info = "Explicitly ensure internal row name structure is identical"
)

df <- structure(list(x = 1), class = c("tmp", "data.frame"), row.names = 1L, foo = "bar")
out <- poorman:::remove_attributes(df)
expect_identical(
  out,
  data.frame(x = 1),
  info = "remove_attributes() strips attributes before dispatch"
)

df <- structure(list(x = 1), class = c("tmp", "data.frame"), row.names = "a", foo = "bar")
out <- poorman:::remove_attributes(df)
expect_identical(
  out,
  data.frame(x = 1, row.names = "a"),
  info = "remove_attributes() strips attributes before dispatch"
)

replace_with <- function(x, i, val, arg_name) {
  if (is.null(val)) return(x)
  check_length(val, x, arg_name)
  check_type(val, x, arg_name)
  check_class(val, x, arg_name)
  i[is.na(i)] <- FALSE
  if (length(val) == 1L) {
    x[i] <- val
  } else {
    x[i] <- val[i]
  }
  x
}

check_length <- function(x, y, arg_name) {
  length_x <- length(x)
  length_y <- length(y)
  if (all(length_x %in% c(1L, length_y))) return()
  if (length_y == 1) {
    stop(arg_name, " must be length 1, not ", paste(length_x, sep = ", "))
  } else {
    stop(arg_name, " must be length ", length_y, " or 1, not ", length_x)
  }
}

check_type <- function(x, y, arg_name) {
  x_type <- typeof(x)
  y_type <- typeof(y)
  if (identical(x_type, y_type)) return()
  stop(arg_name, " must be `", y_type, "`, not `", x_type, "`")
}

check_class <- function(x, y, arg_name) {
  if (!is.object(x)) return()
  exp_classes <- class(y)
  out_classes <- class(x)
  if (identical(out_classes, exp_classes)) return()
  stop(arg_name, " must have class `", exp_classes, "`, not class `", out_classes, "`")
}

#' Pivot data from long to wide
#'
#' `pivot_wider()` "widens" data, increasing the number of columns and decreasing the number of rows. The inverse
#' transformation is [pivot_longer()].
#'
#' @param data `data.frame`. The data to pivot.
#' @param id_cols `character(1)`. The name of the column that identifies the rows. If `NULL`, it will use all the unique
#' rows.
#' @param names_from `character(n)`. The name of the column(s) that contains the levels to be used as future column
#' names.
#' @param names_prefix `character(1)`. String added to the start of every variable name. This is particularly useful if
#' `names_from` is a numeric vector and you want to create syntactic variable names.
#' @param names_sep `character(1)`. If `names_from` or `values_from` contains multiple variables, this will be used to
#' join their values together into a single string to use as a column name.
#' @param names_glue `character(1)`. Instead of `names_sep` and `names_prefix`, you can supply a
#' [glue specification](https://glue.tidyverse.org/index.html) that uses the `names_from` columns to create custom
#' column names. Note that the only delimiters supported by `names_glue` are curly brackets, `{` and `}`.
#' @param values_from `character(n)`. The name of the column that contains the values to be used as future variable
#' values.
#' @param values_fill `numeric(n)`. Optionally, a (scalar) value that will be used to replace missing values in the new
#' columns created.
#' @param ... Not used for now.
#'
#' @return If a tibble was provided as input, `pivot_wider()` also returns a
#' tibble. Otherwise, it returns a data frame.
#'
#' @examples
#' data_long <- read.table(header = TRUE, text = "
#'  subject sex condition measurement
#'        1   M   control         7.9
#'        1   M     cond1        12.3
#'        1   M     cond2        10.7
#'        2   F   control         6.3
#'        2   F     cond1        10.6
#'        2   F     cond2        11.1
#'        3   F   control         9.5
#'        3   F     cond1        13.1
#'        3   F     cond2        13.8
#'        4   M   control        11.5
#'        4   M     cond1        13.4
#'        4   M     cond2        12.9")
#'
#'
#' pivot_wider(
#'   data_long,
#'   id_cols = "subject",
#'   names_from = "condition",
#'   values_from = "measurement"
#' )
#'
#' pivot_wider(
#'   data_long,
#'   id_cols = "subject",
#'   names_from = "condition",
#'   values_from = "measurement",
#'   names_prefix = "Var.",
#'   names_sep = "."
#' )
#'
#' production <- expand.grid(
#'   product = c("A", "B"),
#'   country = c("AI", "EI"),
#'   year = 2000:2014
#' ) %>%
#'   filter((product == "A" & country == "AI") | product == "B") %>%
#'   mutate(production = rnorm(nrow(.)))
#'
#' pivot_wider(
#'   production,
#'   names_from = c("product", "country"),
#'   values_from = "production",
#'   names_glue = "prod_{product}_{country}"
#' )
#'
#' @export
pivot_wider <- function(
  data,
  id_cols = NULL,
  values_from = "Value",
  names_from = "Name",
  names_sep = "_",
  names_prefix = "",
  names_glue = NULL,
  values_fill = NULL,
  ...
) {

  old_names <- names(data)

  names_from <- names(eval_select_pos(data, substitute(names_from)))
  values_from <- names(eval_select_pos(data, substitute(values_from)))

  # Preserve attributes
  variable_attr <- lapply(data, attributes)

  # Create an id for stats::reshape
  if (is.null(id_cols)) {
    row_index <- do.call(
      paste, 
      c(data[, !names(data) %in% c(values_from, names_from), drop = FALSE], sep = "_")
    )
    if (length(row_index) == 0) row_index <- rep("", nrow(data))
    data[["_Rows"]] <- row_index
    id_cols <- "_Rows"
  }

  # create pattern of column names - stats::reshape renames columns that concatenates "v.names" + values - we only want
  # values
  current_colnames <- colnames(data)
  current_colnames <- current_colnames[current_colnames != "_Rows"]
  if (is.null(names_glue)) {
    future_colnames <- unique(do.call(paste, c(data[, names_from, drop = FALSE], sep = names_sep)))
  } else {
    vars <- regmatches(names_glue, gregexpr("\\{\\K[^{}]+(?=\\})", names_glue, perl = TRUE))[[1]]
    tmp_data <- unique(data[, vars])
    future_colnames <- unique(apply(tmp_data, 1, function(x) {
      tmp_vars <- list()
      for (i in seq_along(vars)) {
        tmp_vars[[i]] <- x[vars[i]]
      }

      tmp_colname <- gsub("\\{\\K[^{}]+(?=\\})", "", names_glue, perl = TRUE)
      tmp_colname <- gsub("\\{\\}", "%s", tmp_colname)
      do.call(sprintf, c(fmt = tmp_colname, tmp_vars))
    }))
  }

  # stop if some column names would be duplicated (follow tidyr workflow)
  if (any(future_colnames %in% current_colnames)) {
    stop(
      paste0(
        "Some values of the columns specified in 'names_from' are already present
        as column names. Either use `name_prefix` or rename the following columns: ",
        paste(current_colnames[which(current_colnames %in% future_colnames)], sep = ", ")
      ),
      call. = FALSE
    )
  }

  # stats::reshape works strangely when several variables are in idvar/timevar so we unite all ids in a single temporary
  # column that will be used by stats::reshape
  data$new_time <- do.call(paste, c(data[, names_from, drop = FALSE], sep = "_"))
  data[, names_from] <- NULL

  wide <- stats::reshape(
    data,
    v.names = values_from,
    idvar = id_cols,
    timevar = "new_time",
    sep = names_sep,
    direction = "wide"
  )

  # Clean
  if ("_Rows" %in% names(wide)) wide[["_Rows"]] <- NULL
  rownames(wide) <- NULL

  if (length(values_from) == 1) {
    to_rename <- which(startsWith(names(wide), paste0(values_from, names_sep)))
    names(wide)[to_rename] <- future_colnames
  }

  # Order columns as in tidyr
  if (length(values_from) > 1) {
    for (i in values_from) {
      tmp1 <- wide[, which(!startsWith(names(wide), i))]
      tmp2 <- wide[, which(startsWith(names(wide), i))]
      wide <- cbind(tmp1, tmp2)
      # TODO: Fix relocate() when using numeric values for .before and .after
      # wide <- relocate(wide, starts_with(i), .after = -1)
    }
  }

  new_cols <- setdiff(names(wide), old_names)
  names(wide)[which(names(wide) %in% new_cols)] <- paste0(names_prefix, new_cols)

  # Fill missing values
  if (!is.null(values_fill)) {
    if (length(values_fill) == 1) {
      if (is.numeric(wide[[new_cols[1]]])) {
        if (!is.numeric(values_fill)) {
          stop(paste0("`values_fill` must be of type numeric."), call. = FALSE)
        } else {
          for (i in new_cols) {
            wide[[i]] <- replace_na(wide[[i]], replace = values_fill)
          }
        }
      } else if (is.character(wide[[new_cols[1]]])) {
        if (!is.character(values_fill)) {
          stop(paste0("`values_fill` must be of type character."), call. = FALSE)
        } else {
          for (i in new_cols) {
            wide[[i]] <- replace_na(wide[[i]], replace = values_fill)
          }
        }
      } else if (is.factor(wide[[new_cols[1]]])) {
        if (!is.factor(values_fill)) {
          stop(paste0("`values_fill` must be of type factor."), call. = FALSE)
        } else {
          for (i in new_cols) {
            wide[[i]] <- replace_na(wide[[i]], replace = values_fill)
          }
        }
      }
    } else {
      stop("`values_fill` must be of length 1.", call. = FALSE)
    }
  }

  # Remove reshape attributes
  attributes(wide)$reshapeWide <- NULL

  # add back attributes where possible
  for (i in colnames(wide)) {
    attributes(wide[[i]]) <- variable_attr[[i]]
  }

  wide
}

df <- data.frame(key = c("x", "y", "z"), val = 1:3)
pv <- pivot_wider(df, names_from = "key", values_from = "val")

expect_equal(names(pv), c("x", "y", "z"))
expect_equal(nrow(pv), 1)

df <- data.frame(a = 1, key = c("x", "y"), val = 1:2)
pv <- pivot_wider(df, names_from = "key", values_from = "val")

expect_equal(names(pv), c("a", "x", "y"))
expect_equal(nrow(pv), 1)

df <- data.frame(a = 1:2, key = c("x", "y"), val = 1:2)
pv <- pivot_wider(df, names_from = "key", values_from = "val")

expect_equal(pv$a, c(1, 2))
expect_equal(pv$x, c(1, NA))
expect_equal(pv$y, c(NA, 2))

df <- data.frame(
  a = c(1, 1),
  key = c("a", "b"),
  val = c(1, 2)
)

expect_error(
  pivot_wider(df, names_from = "key", values_from = "val"),
  pattern = "Some values of the columns specified"
)

df <- data.frame(
  variable = c("a", "b", "c"),
  date = c("2015-01-01", "2015-01-01", "2015-01-02"),
  value = c(1, 2, 3),
  stringsAsFactors = FALSE
)

res <- df %>% pivot_wider(
  names_from = "date",
  values_from = "value",
  values_fill = 0
)

expect_equal(
  res,
  structure(
    list(variable = c("a", "b", "c"), `2015-01-01` = c(1, 2, 0), `2015-01-02` = c(0, 0, 3)),
    row.names = c(NA, 3L),
    class = "data.frame"
  )
)


production <- expand.grid(
  product = c("A", "B"),
  country = c("AI", "EI"),
  year = 2000:2014
) %>%
  filter((product == "A" & country == "AI") | product == "B")

production$production <- rnorm(nrow(production))

res <- production %>%
  pivot_wider(
    names_from = c("product", "country"),
    values_from = "production"
  )
expect_equal(dim(res), c(15, 4))
expect_equal(colnames(res), c("year", "A_AI", "B_AI", "B_EI"))

res <- production %>%
  pivot_wider(
    names_from = c("product", "country"),
    values_from = "production",
    names_glue = "prod_{product}_{country}"
  )
expect_equal(colnames(res), c("year", "prod_A_AI", "prod_B_AI", "prod_B_EI"))

res <- production %>%
  pivot_wider(
    names_from = c("product", "country"),
    values_from = "production",
    names_sep = "."
  )

expect_equal(colnames(res), c("year", "A.AI", "B.AI", "B.EI"))

contacts <- data.frame(
  field = c("name", "company", "name", "company", "email", "name"),
  value = c("Jiena McLellan", "Toyota", "John Smith", "google", "john@google.com", "Huxley Ratcliffe"),
  stringsAsFactors = FALSE
)
contacts$person_id <- cumsum(contacts$field == "name")

res <- contacts %>%
  pivot_wider(names_from = "field", values_from = "value")

expect_equal(
  res,
  structure(
    list(
      person_id = 1:3,
      name = c("Jiena McLellan", "John Smith", "Huxley Ratcliffe"),
      company = c("Toyota", "google", NA),
      email = c(NA, "john@google.com", NA)
    ),
    row.names = c(NA, 3L),
    class = "data.frame"
  )
)

df <- data.frame(
  food = c("banana", "banana", "banana", "banana", "cheese", "cheese", "cheese", "cheese"),
  binary = c(rep(c("yes", "no"), 4)),
  car = c("toyota", "subaru", "mazda", "skoda", "toyota", "subaru", "mazda", "skoda"),
  fun = c(2, 4, 3, 6, 2, 4, 2, 3),
  stringsAsFactors = FALSE
)

res <- df %>%
  pivot_wider(
    id_cols = "food",
    names_from = c("car", "binary"),
    names_glue = "{binary}_{car}",
    values_from = "fun"
  )

expect_equal(
  res,
  structure(
    list(
      food = c("banana", "cheese"), yes_toyota = c(2, 2), no_subaru = c(4, 4), yes_mazda = c(3, 2), no_skoda = c(6, 3)
    ),
    row.names = 1:2,
    class = "data.frame"
  )
)

#' Select Helpers
#'
#' These functions allow you to select variables based on their names.
#' * `starts_with()`: Starts with a prefix.
#' * `ends_with()`: Ends with a prefix.
#' * `contains()`: Contains a literal string.
#' * `matches()`: Matches a regular expression.
#' * `all_of()`: Matches variable names in a character vector. All names must be present, otherwise an error is thrown.
#' * `any_of()`: The same as `all_of()` except it doesn't throw an error.
#' * `everything()`: Matches all variables.
#' * `last_col()`: Select the last variable, possibly with an offset.
#'
#' @param match `character(n)`. If length > 1, the union of the matches is taken.
#' @param ignore.case `logical(1)`. If `TRUE`, the default, ignores case when matching names.
#' @param vars `character(n)`. A character vector of variable names. When called from inside selecting functions such as
#' `select()`, these are automatically set to the names of the table.
#'
#' @seealso
#' [select()], [relocate()], [where()], [group_cols()]
#'
#' @return
#' An integer vector giving the position of the matched variables.
#'
#' @examples
#' mtcars %>% select(starts_with("c"))
#' mtcars %>% select(starts_with(c("c", "h")))
#' mtcars %>% select(ends_with("b"))
#' mtcars %>% relocate(contains("a"), .before = mpg)
#' iris %>% select(matches(".t."))
#' mtcars %>% select(last_col())
#'
#' # `all_of()` selects the variables in a character vector:
#' iris %>% select(all_of(c("Petal.Length", "Petal.Width")))
#' # `all_of()` is strict and will throw an error if the column name isn't found
#' try({iris %>% select(all_of(c("Species", "Genres")))})
#' # However `any_of()` allows missing variables
#' iris %>% select(any_of(c("Species", "Genres")))
#'
#' @name select_helpers
#'
#' @export
starts_with <- function(match, ignore.case = TRUE, vars = peek_vars()) {
  grep(pattern = paste0("^", paste0(match, collapse = "|^")), x = vars, ignore.case = ignore.case)
}

#' @name select_helpers
#' @export
ends_with <- function(match, ignore.case = TRUE, vars = peek_vars()) {
  grep(pattern = paste0(paste0(match, collapse = "$|"), "$"), x = vars, ignore.case = ignore.case)
}

#' @name select_helpers
#' @export
contains <- function(match, ignore.case = TRUE, vars = peek_vars()) {
  matches <- lapply(
    match,
    function(x) {
      if (isTRUE(ignore.case)) {
        match_u <- toupper(x)
        match_l <- tolower(x)
        pos_u <- grep(pattern = match_u, x = toupper(vars), fixed = TRUE)
        pos_l <- grep(pattern = match_l, x = tolower(vars), fixed = TRUE)
        unique(c(pos_l, pos_u))
      } else {
        grep(pattern = x, x = vars, fixed = TRUE)
      }
    }
  )
  unique(unlist(matches))
}

#' @param perl `logical(1)`. Should Perl-compatible regexps be used?
#'
#' @name select_helpers
#' @export
matches <- function(match, ignore.case = TRUE, perl = FALSE, vars = peek_vars()) {
  grep(pattern = match, x = vars, ignore.case = ignore.case, perl = perl)
}

#' @param prefix A prefix which starts the numeric range.
#' @param range `integer(n)`. A sequence of integers, e.g. `1:5`.
#' @param width `numeric(1)`. Optionally, the "width" of the numeric range. For example, a range of 2 gives "01", a
#' range of three "001", etc.
#'
#' @name select_helpers
#' @export
num_range <- function(prefix, range, width = NULL, vars = peek_vars()) {
  if (!is.null(width)) {
    range <- sprintf(paste0("%0", width, "d"), range)
  }
  find <- paste0(prefix, range)
  if (any(duplicated(vars))) {
    stop("Column names must be unique")
  } else {
    x <- match(find, vars)
    x[!is.na(x)]
  }
}

#' @param x `character(n)`. A vector of column names.
#' @name select_helpers
#' @export
all_of <- function(x, vars = peek_vars()) {
  x_ <- !x %in% vars
  if (any(x_)) {
    which_x_ <- which(x_)
    if (length(which_x_) == 1L) {
      stop("The column ", x[which_x_], " does not exist.")
    } else {
      stop("The columns ", paste(x[which_x_], collapse = ", "), " do not exist.")
    }
  } else {
    which(vars %in% x)
  }
}

#' @name select_helpers
#' @export
any_of <- function(x, vars = peek_vars()) {
  which(vars %in% x)
}

#' @name select_helpers
#' @export
everything <- function(vars = peek_vars()) {
  seq_along(vars)
}

#' @param offset `integer(1)`. Select the `n`th variable from the end of the `data.frame`.
#'
#' @name select_helpers
#' @export
last_col <- function(offset = 0L, vars = peek_vars()) {
  if (!is_wholenumber(offset)) stop("`offset` must be an integer")
  n <- length(vars)
  if (offset && n <= offset) {
    stop("`offset` must be smaller than the number of `vars`")
  } else if (n == 0) {
    stop("Can't select last column when `vars` is empty")
  } else {
    n - offset
  }
}

#' Peek at variables in the selection context
#'
#' Return the vector of column names of the data currently available for selection.
#'
#' @return
#' A vector of column names.
#'
#' @export
peek_vars <- function() {
  select_env$get_colnames()
}

expect_equal(
  poorman:::select_positions(mtcars, 1L),
  c(mpg = 1),
  info = "Integer selections work"
)

expect_equal(
  poorman:::select_positions(mtcars, 0L),
  integer(0),
  info = "Selecting the 0th integer column returns an empty data.frame"
)

expect_equal(
  poorman:::select_positions(mtcars, 1),
  c(mpg = 1),
  info = "Numeric selections work"
)
expect_equal(
  poorman:::select_positions(mtcars, 0),
  integer(0),
  info = "Selecting the 0th numeric column returns an empty data.frame"
)

expect_equal(
  poorman:::select_positions(mtcars, "mpg"),
  c(mpg = 1),
  info = "Selecting by character string works"
)

expect_equal(
  poorman:::select_positions(mtcars, mpg),
  c(mpg = 1),
  info = "Selecting by symbol works"
)

expect_equal(
  poorman:::select_positions(mtcars, starts_with("m")),
  c(mpg = 1),
  info = "Selecting columns using an expression"
)

expect_equal(
  poorman:::select_positions(mtcars, 1:3),
  c(mpg = 1, cyl = 2, disp = 3),
  info = "Selecting with a numeric sequence"
)
expect_equal(
  poorman:::select_positions(mtcars, "mpg":"cyl"),
  c(mpg = 1, cyl = 2),
  info = "Selecting with a character sequence"
)
expect_equal(
  poorman:::select_positions(mtcars, mpg:cyl),
  c(mpg = 1, cyl = 2),
  info = "Selecting with a symbol sequence"
)
expect_equal(
  poorman:::select_positions(mtcars, -1:-3),
  c(hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a numeric sequence"
)
expect_equal(
  poorman:::select_positions(mtcars, -mpg:-cyl),
  c(disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a symbol sequence"
)
expect_equal(
  poorman:::select_positions(mtcars, -"mpg":-"cyl"),
  c(disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a charatcer sequence"
)

expect_equal(
  poorman:::select_positions(mtcars, !mpg),
  c(cyl = 2, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a symbol"
)
expect_equal(
  poorman:::select_positions(mtcars, !1),
  c(cyl = 2, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a numeric"
)
expect_equal(
  poorman:::select_positions(mtcars, !"mpg"),
  c(cyl = 2, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a character vector"
)
expect_equal(
  poorman:::select_positions(mtcars, !mpg:!cyl),
  c(disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with symbol sequences"
)
expect_equal(
  poorman:::select_positions(mtcars, !"mpg":!"cyl"),
  c(disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a character sequence"
)
expect_equal(
  poorman:::select_positions(mtcars, !c(mpg, cyl)),
  c(disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns using a negated c()"
)

expect_equal(
  poorman:::select_positions(mtcars, -1),
  c(cyl = 2, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a numeric"
)
expect_equal(
  poorman:::select_positions(mtcars, -1L),
  c(cyl = 2, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with an integer"
)
expect_equal(
  poorman:::select_positions(mtcars, -mpg),
  c(cyl = 2, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a symbol"
)
expect_equal(
  poorman:::select_positions(mtcars, -"mpg"),
  c(cyl = 2, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns with a character vector"
)
expect_equal(
  poorman:::select_positions(mtcars, -starts_with("m")),
  c(cyl = 2, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns using an expression"
)

expect_equal(
  poorman:::select_positions(mtcars, c(1, 2)),
  c(mpg = 1, cyl = 2),
  info = "Selecting columns with c() and numerics"
)
expect_equal(
  poorman:::select_positions(mtcars, c("mpg", "cyl")),
  c(mpg = 1, cyl = 2),
  info = "Selecting columns with c() and character strings"
)
expect_equal(
  poorman:::select_positions(mtcars, c(mpg, cyl)),
  c(mpg = 1, cyl = 2),
  info = "Selecting columns with c() and symbols"
)
expect_equal(
  poorman:::select_positions(mtcars, c(starts_with("m"), starts_with("c"))),
  c(mpg = 1, cyl = 2, carb = 11),
  info = "Selecting columns with c() and expressions"
)
expect_equal(
  poorman:::select_positions(mtcars, c(-mpg, -cyl)),
  c(disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns using c() and symbols"
)
expect_equal(
  poorman:::select_positions(mtcars, !c(mpg, cyl)),
  c(disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns using a negated c()"
)

expect_equal(
  poorman:::select_positions(mtcars, -(1:2)),
  c(disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns using a negative, bracketed sequence"
)
expect_equal(
  poorman:::select_positions(mtcars, !(1:2)),
  c(disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Dropping columns using a negated, bracketed sequence"
)

expect_equal(
  poorman:::select_positions(mtcars, 1L, 2, "disp", "hp", starts_with("dr"), wt:qsec),
  c(mpg = 1, cyl = 2, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7),
  info = "Test selecting with a mixture of selection options"
)

expect_equal(
  poorman:::select_positions(mtcars, -2, 3),
  c(mpg = 1, disp = 3, hp = 4, drat = 5, wt = 6, qsec = 7, vs = 8, am = 9, gear = 10, carb = 11),
  info = "Mixture of negative and positives select only negatives"
)

expect_equal(
  poorman:::select_positions(mtcars, 3, -2),
  c(disp = 3),
  info = "Mixture of positives and negatives select only negatives"
)

expect_equal(
  poorman:::select_positions(mtcars, NULL),
  integer(0),
  info = "NULL returns zero column positions"
)

expect_equal(
  poorman:::select_positions(mtcars, am, NULL, cyl),
  c("am" = 9, "cyl" = 2),
  info = "combinations of NULL and other parameter names ignore the NULLs"
)

expect_error(
  poorman:::select_positions(mtcars, 100),
  info = "Out of range columns error"
)

expect_error(
  poorman:::select_positions(mtcars, TRUE),
  info = "Logical selections do not work"
)

expect_equal(
  mtcars %>% select(starts_with("d")),
  mtcars[, c("disp", "drat")],
  info = "Test select() using starts_with() with a single value"
)

expect_equal(
  mtcars %>% select(starts_with(c("d", "c"))),
  mtcars[, c("cyl", "disp", "drat", "carb")],
  info = "Test select() using starts_with() for multiple values"
)

expect_equal(
  mtcars %>% relocate(starts_with("c"), .before = mpg),
  mtcars[, c("cyl", "carb", "mpg", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear")],
  info = "Test relocate() with starts_with()"
)

expect_equal(
  mtcars %>% select(ends_with("g")),
  mtcars[, "mpg", drop = FALSE],
  info = "Test select() using ends_with() with a single value"
)

expect_equal(
  mtcars %>% select(ends_with(c("g", "b"))),
  mtcars[, c("mpg", "carb")],
  info = "Test select() using ends_with() with a multiple values"
)

expect_equal(
  mtcars %>% relocate(starts_with("c"), .before = mpg),
  mtcars[, c("cyl", "carb", "mpg", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear")],
  info = "Test relocate() with ends_with()"
)

expect_equal(
  mtcars %>% select(starts_with("m"), 2, disp, "gear", ends_with("b")),
  mtcars[, c("mpg", "cyl", "disp", "gear", "carb")],
  info = "Test multiple *_with() functions in combination with other selection methods"
)

expect_equal(
  mtcars %>% select(contains("a")),
  mtcars[, c("drat", "am", "gear", "carb")],
  info = "Test select() with contains() and a single variable"
)

expect_equal(
  mtcars %>% select(contains(c("a", "m"))),
  mtcars[, c("drat", "am", "gear", "carb", "mpg")],
  info = "Test select() with contains() and multiple variables"
)

expect_equal(
  mtcars %>% select(contains("M", ignore.case = FALSE)),
  mtcars[, 0],
  info = "Test select() with contains() and no match"
)

expect_equal(
  iris %>% select(matches(".t.")),
  iris[, 1:4],
  info = "Test matches() selects columns based on a regular expression"
)

expect_equal(
  mtcars %>% select(all_of(c("disp", "qsec"))),
  mtcars[, c("disp", "qsec")],
  info = "Test `select()` with `all_of()` and valid columns"
)

expect_error(
  mtcars %>% relocate(all_of(c("disp", "gear", "tmp"))),
  info = "Test errors are returned when `all_of()` has a single incorrect column"
)

expect_error(
  mtcars %>% relocate(all_of(c("disp", "gear", "tmp", "blah"))),
  info = "Test errors are returned when `all_of()` has a multiple incorrect columns"
)

expect_equal(
  mtcars %>% relocate(any_of(c("disp", "gear", "tmp", "cyl")), .before = mpg),
  {
    cols <- c("cyl", "disp", "gear")
    mtcars[, c(cols, colnames(mtcars)[which(!colnames(mtcars) %in% cols)])]
  },
  info = "Test that `any_of()` selects columns regardless of their validity"
)

expect_equal(
  data.frame("V1" = 1, "V2" = 2, "V3" = 3, "V4" = 4) %>% select(num_range("V", 2:3)),
  data.frame("V2" = 2, "V3" = 3),
  info = "Test num_range() within a selection"
)

expect_equal(
  data.frame("V001" = 1, "V002" = 2, "V003" = 3, "V004" = 4) %>% select(num_range("V", 2:3, width = 3)),
  data.frame("V002" = 2, "V003" = 3),
  info = "Test that num_range() works for a given width"
)

expect_error(
  data.frame("a1" = 1, "a1" = 2, "a2" = 3, check.names = FALSE) %>% select(num_range("a", 1)),
  info = "Test that num_range() only works when column names are unique"
)

expect_equal(
  mtcars %>% select(last_col()),
  mtcars[, "carb", drop = FALSE],
  info = "Test select() with last_col()"
)

expect_equal(
  mtcars %>% select(last_col(2)),
  mtcars[, "am", drop = FALSE],
  info = "Test select() with last_col() and an offset"
)

expect_error(
  mtcars %>% select(last_col(1.2)),
  info = "Test last_col() returns an error when not given a whole number"
)

expect_error(
  mtcars %>% select(last_col(100)),
  info = "Test an error occurs when the user attempts to select using `last_col()` with too large an offset"
)

expect_error(
  mtcars %>% select(last_col(vars = c())),
  info = "Test an error occurs when no variables are provided to match column names against"
)

expect_equal(
  mtcars %>% select(everything()),
  mtcars,
  info = "Test everything() returns all columns in the same order"
)

expect_equal(
  mtcars %>% select(starts_with("q"), ends_with("s"), everything()),
  mtcars[, c("qsec", "vs", colnames(mtcars)[!colnames(mtcars) %in% c("qsec", "vs")])],
  info = "Test everything() returns the rest of the columns after the initial selection"
)

expect_equal(
  mtcars %>% select(group_cols()),
  mtcars[, 0],
  info = "group_cols() with non-grouped data returns a data.frame with zero columns"
)

expect_equivalent(
  mtcars %>% group_by(am, cyl) %>% select(group_cols()),
  mtcars[, c("am", "cyl")],
  info = "selection of grouped data using group_cols() returns only the group columns"
)

expect_silent(
  mtcars %>% group_by(am, cyl) %>% select(group_cols()),
  info = "message to say adding missing grouping variables should not appear"
)

expect_equal(
  iris %>% select(ends_with("Width") & ends_with("Length")),
  iris[, c("Sepal.Width", "Petal.Width", "Sepal.Length", "Petal.Length")],
  info = "`&` works for all positive column positions"
)

expect_equal(
  iris %>% select(starts_with("Petal") & !ends_with("Width")),
  iris[, "Petal.Length", drop = FALSE],
  info = "`&` works for a mixture of positive and negative column positions"
)

expect_equal(
  iris %>% select(!ends_with("Width") & !ends_with("Length")),
  iris[, "Species", drop = FALSE],
  info = "`&` works for all negative column positions"
)

expect_equal(
  select(mtcars, poorman::everything()),
  mtcars,
  info = "pkg::fun() selecters do not fail"
)

expect_equal(
  mtcars %>% relocate(drat),
  mtcars[, c("drat", colnames(mtcars)[!colnames(mtcars) %in% "drat"])],
  info = "Test default relocate() moves column to the start of the data.frame"
)

expect_equal(
  mtcars %>% relocate(drat, .before = cyl),
  mtcars[, c("mpg", "drat", "cyl", "disp", "hp", "wt", "qsec", "vs", "am", "gear", "carb")],
  info = "Test .before functionality for single variable"
)

expect_equal(
  mtcars %>% relocate(drat, .after = cyl),
  mtcars[, c("mpg", "cyl", "drat", "disp", "hp", "wt", "qsec", "vs", "am", "gear", "carb")],
  info = "Test .after functionality for single variable"
)

expect_equal(
  mtcars %>% relocate(drat, gear, .after = cyl),
  mtcars[, c("mpg", "cyl", "drat", "gear", "disp", "hp", "wt", "qsec", "vs", "am", "carb")],
  info = "Test .after functionality for multiple variables"
)

expect_equal(
  iris %>% relocate(contains("Petal"), .after = Species),
  iris[, c("Sepal.Length", "Sepal.Width", "Species", "Petal.Length", "Petal.Width")],
  info = "Test .after works when .after is the last column"
)

expect_equal(
  mtcars %>% relocate(gear, .before = 1),
  mtcars[, c("gear", "mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "carb")],
  info = ".before works with a numeric value"
)

expect_equal(
  mtcars %>% relocate(gear, .after = 1),
  mtcars[, c("mpg", "gear", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "carb")],
  info = ".after works with a numeric value"
)

expect_error(
  mtcars %>% relocate(gear, .after = mpg, .before = cyl),
  info = "relocate() fails when .after and .before are both given"
)

as_function <- function(x, env = parent.frame()) {
  if (is.function(x)) return(x)
  if (is_formula(x)) {
    if (length(x) > 2) stop("Can't convert a two-sided formula to a function")
    env <- attr(x, ".Environment", exact = TRUE)
    rhs <- as.list(x)[[2]]
    return(as.function(list(... = substitute(), .x = quote(..1), .y = quote(..2), . = quote(..1), rhs), envir = env))
  }
  if (is_string(x)) return(get(x, envir = env, mode = "function"))
  stop("Can't convert an object of class ", class(x), " to a function.")
}

is_formula <- function(x) {
  inherits(x, "formula")
}

is_string <- function(x) {
  is.character(x) && length(x) == 1L
}

is_wholenumber <- function(x) {
  x %% 1L == 0L
}

names_are_invalid <- function(x) {
  x == "" | is.na(x)
}

is_named <- function(x) {
  nms <- names(x)
  if (is.null(nms)) return(FALSE)
  if (any(names_are_invalid(nms))) return(FALSE)
  TRUE
}

have_name <- function(x) {
  nms <- names(x)
  if (is.null(nms)) rep(FALSE, length(x)) else !names_are_invalid(nms)
}

is_empty_list <- function(x) {
  inherits(x, "list") && length(x) == 0L
}

#' Turn an am atomic vector into a list of symbols
#'
#' @return
#' A `list` of symbols.
#'
#' @example
#' as_symbols(c("am", "cyl"))
#'
#' @noRd
as_symbols <- function(x) {
  lapply(x, as.symbol)
}

#' Check whether the input is an atomic vector or a data.frame
#' @noRd
is_df_or_vector <- function(x) {
  res <- is.data.frame(x) || is.atomic(x)
  if (!isTRUE(res)) stop("You must pass vector(s) and/or data.frame(s).")
  TRUE
}

expect_equal(
  poorman:::as_function("mean"),
  mean,
  "character vectors can be converted to functions"
)

expect_equal(
  poorman:::as_function(~mean)(x),
  mean,
  info = "formulas can be converted to functions"
)

expect_equal(
  poorman:::as_function(mean),
  mean,
  info = "functions are returned as is"
)

## Errors

expect_error(
  poorman:::as_function(x ~ mean),
  info = "cannot convert a two-sided formula"
)

expect_error(
  poorman:::as_function(2),
  info = "cannot convert objects which aren't functions, formulas or strings"
)

#' Change column order
#'
#' Use `relocate()` to change column positions, using the same syntax as [select()] to make it easy to move blocks of
#' columns at once.
#'
#' @inheritParams select
#' @param ... <[`poor-select`][select_helpers]> Columns to move.
#' @param .before,.after <[`poor-select`][select_helpers]> Destination of columns selected by `...`. Supplying neither
#' will move columns to the left-hand side; specifying both will result in an error.
#'
#' @return
#' An object of the same type as `.data`. The output has the following properties:
#'
#' * Rows are not affected.
#' * The same columns appear in the output, but (usually) in a different place.
#' * Data frame attributes are preserved.
#' * Groups are not affected.
#'
#' @examples
#' df <- data.frame(
#'   a = 1, b = 1, c = 1, d = "a", e = "a", f = "a",
#'   stringsAsFactors = FALSE
#' )
#' df %>% relocate(f)
#' df %>% relocate(a, .after = c)
#' df %>% relocate(f, .before = b)
#' df %>% relocate(a, .after = last_col())
#'
#' # Can also select variables based on their type
#' df %>% relocate(where(is.character))
#' df %>% relocate(where(is.numeric), .after = last_col())
#' # Or with any other select helper
#' df %>% relocate(any_of(c("a", "e", "i", "o", "u")))
#'
#' # When .before or .after refers to multiple variables they will be
#' # moved to be immediately before/after the selected variables.
#' df2 <- data.frame(
#'   a = 1, b = "a", c = 1, d = "a",
#'   stringsAsFactors = FALSE
#' )
#' df2 %>% relocate(where(is.numeric), .after = where(is.character))
#' df2 %>% relocate(where(is.numeric), .before = where(is.character))
#'
#' @export
relocate <- function(.data, ..., .before = NULL, .after = NULL) {
  UseMethod("relocate")
}

#' @export
relocate.data.frame <- function(.data, ..., .before = NULL, .after = NULL) {
  data_names <- colnames(.data)
  col_pos <- select_positions(.data, ...)

  if (!missing(.before)) .before <- colnames(.data)[eval_select_pos(.data, substitute(.before))]
  if (!missing(.after)) .after <- colnames(.data)[eval_select_pos(.data, substitute(.after))]

  has_before <- !is.null(.before)
  has_after <- !is.null(.after)

  if (has_before && has_after) {
    stop("You must supply only one of `.before` and `.after`")
  } else if (has_before) {
    where <- min(match(.before, data_names))
    col_pos <- c(setdiff(col_pos, where), where)
  } else if (has_after) {
    where <- max(match(.after, data_names))
    col_pos <- c(where, setdiff(col_pos, where))
  } else {
    where <- 1L
    col_pos <- union(col_pos, where)
  }
  lhs <- setdiff(seq(1L, where - 1L), col_pos)
  rhs <- setdiff(seq(where + 1L, ncol(.data)), col_pos)
  col_pos <- unique(c(lhs, col_pos, rhs))
  col_pos <- col_pos[col_pos <= length(data_names)]

  res <- .data[col_pos]
  if (has_groups(.data)) res <- groups_set(res, group_vars(.data))
  res
}

#' Replace missing values
#'
#' Replace missing values in a `data.frame` or `vector`.
#'
#' @param data A `data.frame` or `vector`.
#' @param replace If `data` is a `data.frame`, a named `list` giving the value to replace `NA` with for each column. If
#' `data` is a `vector`, a single value used for replacement.
#' @param ... Additional arguments passed onto methods; not currently used.
#'
#' @seealso
#' [na_if()] to replace specified values with a `NA`.
#'
#' [coalesce()] to replace missing values within subsequent `vector`(s) of value(s).
#'
#' @return
#' If `data` is a `data.frame`, `replace_na()` returns a `data.frame`. If `data` is a `vector`, `replace_na()` returns a
#' `vector` of class determined by the union of `data` and `replace`.
#'
#' @examples
#' df <- data.frame(x = c(1, 2, NA), y = c("a", NA, "b"), stringsAsFactors = FALSE)
#' df %>% replace_na(list(x = 0, y = "unknown"))
#' df %>% mutate(x = replace_na(x, 0))
#'
#' df$x %>% replace_na(0)
#' df$y %>% replace_na("unknown")
#'
#' @export
replace_na <- function(data, replace, ...) {
  UseMethod("replace_na")
}

#' @export
replace_na.default <- function(data, replace = NA, ...) {
  check_replacement(replace, deparse(substitute(data)))
  data[is.na(data)] <- replace
  data
}

#' @export
replace_na.data.frame <- function(data, replace = list(), ...) {
  stopifnot(is.list(replace))
  replace_vars <- intersect(names(replace), names(data))
  for (var in replace_vars) {
    check_replacement(replace[[var]], var)
    data[[var]][is.na(data[[var]])] <- replace[[var]]
  }
  data
}

check_replacement <- function(x, var) {
  n <- length(x)
  if (n == 1L) return()
  stop("Replacement for `", var, "` is length ", n, ", not length 1")
}

x <- c(1, NA)
expect_equal(
  replace_na(x),
  x,
  info = "Empty replacement for `vector` does nothing"
)

expect_equal(
  replace_na(x, 0),
  c(1, 0),
  info = "Missing values are replaced in a `vector`"
)

expect_error(
  replace_na(1, 1:10),
  info = "Replacement length should be 1 for `vector`s"
)

df <- data.frame(x = c(1, NA))
expect_equal(
  replace_na(df),
  df,
  info = "Empty replacement for `data.frame` does nothing"
)

expect_equal(
  replace_na(df, list(x = 0))$x,
  c(1, 0),
  info = "Missing values are replaced in a `data.frame`"
)

expect_equal(
  replace_na(df, list(x = 100, b = 0))$x,
  c(1, 100),
  info = "Don't complain about variables that don't exist"
)

rm(df, x)

#' Pull out a single variable
#'
#' This is a direct replacement for `[[.data.frame`.
#'
#' @param .data A `data.frame`.
#' @param var A variable specified as:
#' * a literal variable name
#' * a positive integer, giving the position counting from the left
#' * a negative integer, giving the position counting from the right
#'
#' The default returns the last column (on the assumption that's the column you've created most recently).
#'
#' @examples
#' mtcars %>% pull(-1)
#' mtcars %>% pull(1)
#' mtcars %>% pull(cyl)
#' mtcars %>% pull("cyl")
#'
#' @export
pull <- function(.data, var = -1) {
  var_list <- as.list(seq_along(.data))
  names(var_list) <- names(.data)
  .var <- eval(substitute(var), var_list)
  if (.var < 0L) .var <- length(var_list) + .var + 1L
  .data[[.var]]
}

expect_error(
  mtcars %>% pull(mpg, cyl),
  info = "Expect error when pulling multiple columns"
)

expect_equal(
  mtcars %>% pull(mpg),
  mtcars[, "mpg"],
  info = "Test pulling a single nse column"
)

expect_equal(
  mtcars %>% pull("mpg"),
  mtcars[, "mpg"],
  info = "Test pulling a single quoted column"
)

expect_equal(
  mtcars %>% pull(1L),
  mtcars[, "mpg"],
  info = "Test pulling a single column with an integer"
)

expect_equal(
  mtcars %>% pull(1),
  mtcars[, "mpg"],
  info = "Test pulling a single column with a numeric"
)

expect_equal(
  mtcars %>% pull(-1L),
  mtcars[, "carb"],
  info = "Test pulling a single column with a negative integer"
)

expect_equal(
  mtcars %>% pull(-1),
  mtcars[, "carb"],
  info = "Test pulling a single column with a negative numeric"
)

expect_equal(
  {
    var <- "mpg"
    mtcars %>% pull(var)
  },
  mtcars[, "mpg"],
  info = "Can pull using a variable"
)

#' Nest By
#'
#' `nest_by()` is similar to [group_by()] however instead of storing the group structure in the metadata, it is made
#' explicit in the data. Each group key is given a single row within the `data.frame` and the group's data is stored
#' within a list-column of the `data.frame`.
#'
#' Currently there is no pretty-printing provided for the results of `nest_by()` and they are not useable with other
#' functions such as [mutate()].
#'
#' @inheritParams group_split
#' @param .key `character(1)`. The name of the column in which to nest the data (default: "data").
#'
#' @examples
#' mtcars %>% nest_by(am, cyl)
#' # Or equivalently
#' mtcars %>% group_by(am, cyl) %>% nest_by()
#'
#' @export
nest_by <- function(.data, ..., .key = "data", .keep = FALSE) {
  if (length(.key) != 1L || !is.character(.key)) stop("`.key` should be `character(1)`")
  if (length(.keep) != 1L || !is.logical(.keep)) stop("`.keep` should be `logical(1)`")
  UseMethod("nest_by")
}

#' @export
nest_by.data.frame <- function(.data, ..., .key = "data", .keep = FALSE) {
  .data <- group_by(.data, ...)
  nest_by(.data, .key = .key, .keep = .keep)
}

#' @export
nest_by.grouped_df <- function(.data, ..., .key = "data", .keep = FALSE) {
  if (!missing(...)) {
    stop("Can't re-group while nesting. Either `ungroup()` first or don't supply arguments to `nest_by()`")
  }
  nests <- group_split(.data, ..., .keep = TRUE)
  groups <- group_vars(.data)
  res <- lapply(
    nests,
    function(x) {
      df <- x[1, groups, drop = FALSE]
      df[[.key]] <- list(x[, !colnames(x) %in% groups, drop = FALSE])
      df
    }
  )
  res <- do.call(rbind, res)
  rownames(res) <- NULL
  res <- do.call(arrange, list(res, as.symbol(groups)))
  do.call(group_by, list(res, as.symbol(groups)))
}

#' Get integer column positions
#'
#' Within the context of a selection function (`select`, `relocate`, `rename`), return the integer position of the
#' columns selected.
#'
#' @details
#' This function accepts
#'
#' * `integer()`s
#' * `numeric()`s
#' * `character()`s
#' * `symbol()`s
#' * `call()`s
#'
#' Each type is handled separately.
#'
#' @inheritParams select
#' @param group_pos `logical(1)`. Should grouping variable positions be returned (default: `FALSE`)?
#'
#' @return
#' A vector of named positive `integer`s.
#'
#' @examples
#' select_positions(mtcars, mpg)
#' select_positions(mtcars, "mpg")
#' select_positions(mtcars, starts_with("m"))
#' select_positions(mtcars, -mpg)
#' select_positions(mtcars, mpg:drat)
#'
#' @noRd
select_positions <- function(.data, ..., .group_pos = FALSE) {
  cols <- dotdotdot(...)
  cols <- cols[!vapply(cols, is.null, FALSE)]
  if (length(cols) == 0L) return(integer(0))
  select_env$setup(.data = .data, calling_frame = parent.frame(2L))
  on.exit(select_env$clean(), add = TRUE)
  data_names <- select_env$get_colnames()
  pos <- unlist(lapply(cols, eval_expr))
  if (length(pos) > 0) pos <- if (pos[1] >= 0) pos[pos >= 0] else pos[pos < 0]
  col_len <- select_env$get_ncol()
  if (any(pos > col_len)) {
    oor <- pos[which(pos > col_len)]
    oor_len <- length(oor)
    stop(
      "Location", if (oor_len > 1) "s " else " ", collapse_to_sentence(oor),
      if (oor_len > 1) " don't " else " doesn't ", "exist. There are only ", col_len, " columns."
    )
  }
  if (isTRUE(.group_pos)) {
    groups <- group_vars(.data)
    missing_groups <- !(groups %in% cols)
    if (any(missing_groups)) {
      sel_missing <- groups[missing_groups]
      readd <- match(sel_missing, data_names)
      readd <- readd[!(readd %in% pos)]
      if (length(readd) > 0L) {
        message("Adding missing grouping variables: `", paste(sel_missing, collapse = "`, `"), "`")
        if (length(names(cols)) > 0L) names(readd) <- data_names[readd]
        pos <- c(readd, pos)
      }
    }
  }
  if (length(data_names[pos]) != 0L) {
    nm_pos <- names(pos)
    if (any(nm_pos == "")) {
      names(pos)[which(nm_pos == "")] <- data_names[pos[which(nm_pos == "")]]
    }
    if (is.null(nm_pos)) {
      names(pos) <- data_names[abs(pos)]
    }
  }
  uniques <- pos[!duplicated(pos)]
  res_nms <- data_names[uniques]
  res <- match(res_nms, data_names)
  if (length(res) != 0L) {
    res <- if (length(setdiff(names(uniques), data_names)) > 0L) {
      if (all(uniques > 0L)) structure(res, names = names(uniques)) else structure(res, names = res_nms)
    } else {
      structure(res, names = res_nms)
    }
  }
  res
}

eval_expr <- function(x) {
  type <- typeof(x)
  switch(
    type,
    "integer" = x,
    "double" = as.integer(x),
    "character" = select_char(x),
    "symbol" = select_symbol(x),
    "language" = eval_call(x),
    stop("Expressions of type <", typeof(x), "> cannot be evaluated for use when subsetting.")
  )
}

select_char <- function(expr) {
  pos <- match(expr, select_env$get_colnames())
  if (any(is.na(pos))) stop("The following columns do not exist:\n    ", paste(expr, collapse = "\n    "))
  pos
}

select_symbol <- function(expr) {
  expr_name <- as.character(expr)
  if (grepl("^is\\.", expr_name) && is.function(expr)) {
    stop(
      "Predicate functions must be wrapped in `where()`.\n\n",
      sprintf("  data %%>%% select(where(%s))", expr_name)
    )
  }
  res <- try(select_char(as.character(expr)), silent = TRUE)
  if (inherits(res, "try-error")) {
    res <- tryCatch(
      unlist(lapply(eval(expr, envir = select_env$calling_frame), eval_expr)),
      error = function(e) stop("Column ", expr, " does not exist.")
    )
  }
  res
}

eval_call <- function(x) {
  type <- as.character(x[[1]])
  if (length(type) > 1L) {
    # This helps when pkg::fn is used in a select helper
    type <- "context"
  }
  switch(
    type,
    `:` = select_seq(x),
    `!` = select_negate(x),
    `-` = select_minus(x),
    `c` = select_c(x),
    `(` = select_bracket(x),
    `&` = select_and(x),
    select_context(x)
  )
}

select_and <- function(expr) {
  exprs <- as.list(expr)[-1]
  res <- do.call(c, lapply(exprs, eval_expr))
  if (all(res > 0) || all(res < 0)) return(unique(res))
  res <- res[!(duplicated(abs(res)) | duplicated(abs(res), fromLast = TRUE))]
  res[res > 0]
}

select_seq <- function(expr) {
  x <- eval_expr(expr[[2]])
  y <- eval_expr(expr[[3]])
  x:y
}

select_negate <- function(expr) {
  x <- if (is_negated_colon(expr)) {
    expr <- call(":", expr[[2]][[2]], expr[[2]][[3]][[2]])
    eval_expr(expr)
  } else {
    eval_expr(expr[[2]])
  }
  x * -1L
}

is_negated_colon <- function(expr) {
  expr[[1]] == "!" && length(expr[[2]]) > 1L && expr[[2]][[1]] == ":" && expr[[2]][[3]][[1]] == "!"
}

select_minus <- function(expr) {
  x <- eval_expr(expr[[2]])
  x * -1L
}

select_c <- function(expr) {
  lst_expr <- as.list(expr)
  lst_expr[[1]] <- NULL
  unlist(lapply(lst_expr, eval_expr))
}

select_bracket <- function(expr) {
  eval_expr(expr[[2]])
}

select_context <- function(expr) {
  eval(expr, envir = select_env$.data)
}

# -- Environment ---------------------------------------------------------------

select_env <- new.env()
select_env$setup <- function(.data, calling_frame) {
  select_env$.data <- .data
  select_env$calling_frame <- calling_frame
}
select_env$clean <- function() {
  rm(list = c(".data", "calling_frame"), envir = select_env)
}
select_env$get_colnames <- function() colnames(select_env$.data)
select_env$get_nrow <- function() nrow(select_env$.data)
select_env$get_ncol <- function() ncol(select_env$.data)

# -- Helpers -------------------------------------------------------------------

#' A cleaner interface to evaluating select_positions when column names are not passed via ...
#' @noRd
eval_select_pos <- function(.data, .cols, .group_pos = FALSE) {
  do.call(select_positions, list(.data = .data, .cols, .group_pos = .group_pos))
}

#' Rename columns
#'
#' `rename()` changes the names of individual variables using `new_name = old_name` syntax.
#' `rename_with()` renames columns using a function.
#'
#' @param .data A `data.frame`
#' @param ...
#'   For `rename()`: comma separated key-value pairs in the form of `new_name = old_name` to rename selected variables.
#'
#'   For `rename_with()`: additional arguments passed onto `.fn`.
#'
#' @return
#' A `data.frame` with the following properties:
#'
#' * Rows are not affected.
#' * Column names are changed; column order is preserved.
#' * `data.frame` attributes are preserved.
#' * Groups are updated to reflect new names.
#'
#' @examples
#' rename(mtcars, MilesPerGallon = mpg)
#' rename(mtcars, Cylinders = cyl, Gears = gear)
#' mtcars %>% rename(MilesPerGallon = mpg)
#'
#' @export
rename <- function(.data, ...) {
  UseMethod("rename")
}

#' @export
rename.data.frame <- function(.data, ...) {
  new_names <- names(dotdotdot(...))
  if (length(new_names) == 0L) {
    warning("You didn't give any new names")
    return(.data)
  }
  col_pos <- select_positions(.data, ...)
  old_names <- colnames(.data)[col_pos]
  new_names_zero <- nchar(new_names) == 0L
  if (any(new_names_zero)) {
    warning("You didn't provide new names for: ", paste0("`", old_names[new_names_zero], collapse = ", "), "`")
    new_names[new_names_zero] <- old_names[new_names_zero]
  }
  colnames(.data)[col_pos] <- new_names
  .data
}

#' @param .fn A `function()` used to transform the selected `.cols`. Should return a character vector the same length as
#' the input.
#' @param .cols Columns to rename; defaults to all columns.
#'
#' @examples
#' rename_with(mtcars, toupper)
#' rename_with(mtcars, toupper, starts_with("c"))
#'
#' @rdname rename
#' @export
rename_with <- function(.data, .fn, .cols = everything(), ...) {
  UseMethod("rename_with")
}

#' @export
rename_with.data.frame <- function(.data, .fn, .cols = everything(), ...) {
  if (!is.function(.fn)) stop("`", .fn, "` is not a valid function")
  grouped <- is.grouped_df(.data)
  if (grouped) grp_pos <- which(colnames(.data) %in% group_vars(.data))
  col_pos <- eval_select_pos(.data = .data, .group_pos = TRUE, .cols = substitute(.cols))
  cols <- colnames(.data)[col_pos]
  new_cols <- .fn(cols, ...)
  if (any(duplicated(new_cols))) {
    stop("New names must be unique however `", deparse(substitute(.fn)), "` returns duplicate column names")
  }
  colnames(.data)[col_pos] <- new_cols
  if (grouped) .data <- groups_set(.data, colnames(.data)[grp_pos])
  .data
}

expect_equal(
  mtcars %>% rename(MilesPerGallon = mpg),
  {
    res <- mtcars
    colnames(res)[1] <- "MilesPerGallon"
    res
  },
  info = "Test renaming a single column"
)

expect_warning(
  mtcars %>% rename(),
  info = "Test that a warning is given when no new names are passed to rename()"
)

expect_warning(
  mtcars %>% rename(Cyls = cyl, mpg),
  info = "Test that a warning is given when a column is not provided a new name in rename()"
)

# Grouped Operations
expect_equal(
  mtcars %>% rename(Gears = gear, Auto = am),
  {
    res <- mtcars
    colnames(res)[which(colnames(res) %in% c("am", "gear"))] <- c("Auto", "Gears")
    res
  },
  info = "Test renaming multiple columns"
)

#' Arrange rows by variables
#'
#' Order rows of a `data.frame` by an expression involving its variables.
#'
#' @param .data A `data.frame`.
#' @param ... A comma separated vector of unquoted name(s) to order the data by.
#'
#' @examples
#' arrange(mtcars, mpg)
#' mtcars %>% arrange(mpg)
#' mtcars %>% arrange(cyl, mpg)
#'
#' @return
#' A `data.frame`.
#'
#' @export
arrange <- function(.data, ...) {
  UseMethod("arrange")
}

#' @export
arrange.data.frame <- function(.data, ..., .by_group = FALSE) {
  dots <- dotdotdot(...)
  is_grouped <- has_groups(.data)
  if (isTRUE(.by_group)) dots <- c(groups(.data), dots)
  rows <- arrange_rows(.data = .data, dots)
  row_number <- attr(.data, "row.names") # row.names returns a character vector
  out <- .data[rows, , drop = FALSE]
  if (is.numeric(row_number)) {
    row.names(out) <- row_number
  }
  if (is_grouped) {
    attr(out, "groups") <- calculate_groups(out, group_vars(out))
  }
  out
}

## -- Helpers ------------------------------------------------------------------

arrange_rows <- function(.data, dots) {

  if (length(dots) == 0L) return(seq_len(nrow(.data)))

  for (i in seq_along(dots)) {
    tmp <- deparse(dots[[i]])
    if (startsWith(tmp, "desc(")) {
      tmp <- gsub("^desc\\(", "-", tmp)
      tmp <- gsub("\\)$", "", tmp)
    }
    dots[[i]] <- parse(text = tmp, keep.source = FALSE)[[1]]
  }

  used <- unname(do.call(c, lapply(dots, find_used)))
  used <- used[used %in% colnames(.data)]
  for (i in seq_along(dots)) {
    if (is.character(.data[[used[[i]]]])) {
      .data[[used[[i]]]] <- factor(.data[[used[[i]]]])
    }
    if (is.factor(.data[[used[[i]]]]) &&
        (startsWith(deparse(dots[[i]]), "desc(") ||
         startsWith(deparse(dots[[i]]), "-"))) {
      dots[[i]] <- bquote(-xtfrm(.(as.name(used[[i]]))))
    }
  }

  data <- do.call(transmute, c(list(.data = ungroup(.data)), dots))
  do.call(order, c(data, list(decreasing = FALSE, na.last = TRUE)))

}

expect_equal(
  desc(1:10),
  -1:-10,
  info = "desc() returns negative vectors"
)

expect_equal(
  mtcars %>% arrange(desc(mpg)),
  mtcars[order(-mtcars$mpg), ],
  info = "desc() arranges a data.frame in descending order of the chosen column"
)

#' Return rows with matching conditions
#'
#' Use `filter()` to choose rows/cases where conditions are `TRUE`.
#'
#' @param .data A `data.frame`.
#' @param ... Logical predicated defined in terms of the variables in `.data`. Multiple conditions are combined with
#' `&`. Arguments within `...` are automatically quoted and evaluated within the context of the `data.frame`.
#' @param .preserve `logical(1)`. Relevant when the .data input is grouped. If `.preserve = FALSE` (the default), the
#' grouping structure is recalculated based on the resulting data, otherwise the grouping is kept as is.
#'
#' @section Useful filter functions:
#'
#' * `==`, `>`, `>=`, etc.
#' * `&`, `|`, `!`, `xor()`
#' * `is.na()`
#'
#' @examples
#' filter(mtcars, am == 1)
#' mtcars %>% filter(cyl == 4)
#' mtcars %>% filter(cyl <= 5 & am > 0)
#' mtcars %>% filter(cyl == 4 | cyl == 8)
#' mtcars %>% filter(!(cyl %in% c(4, 6)), am != 0)
#'
#' @return
#' A `data.frame`.
#'
#' @export
filter <- function(.data, ..., .preserve = FALSE) {
  UseMethod("filter")
}

#' @export
filter.data.frame <- function(.data, ..., .preserve = FALSE) {
  conditions <- dotdotdot(...)
  if (length(conditions) == 0L) return(.data)
  check_filter(conditions)
  cond_class <- vapply(conditions, typeof, NA_character_)
  cond_class <- cond_class[!cond_class %in% c("language", "logical")]
  if (length(cond_class) > 0L) stop("Conditions must be logical vectors")
  context$setup(.data)
  on.exit(context$clean(), add = TRUE)
  eval_env$env <- parent.frame()
  on.exit(rm(list = "env", envir = eval_env), add = TRUE)
  rows <- lapply(
    conditions,
    function(cond, frame) eval(cond, context$.data, frame),
    frame = eval_env$env
  )
  rows <- Reduce("&", rows)
  .data[rows & !is.na(rows), ]
}

#' @export
filter.grouped_df <- function(.data, ..., .preserve = FALSE) {
  rows <- rownames(.data)
  res <- apply_grouped_function("filter", .data, drop = TRUE, ...)
  res <- res[rows[rows %in% rownames(res)], ]

  groups <- group_vars(.data)
  pre_filtered_groups <- group_data(.data)
  post_filtered_groups <- calculate_groups(res, groups)

  if (!(!.preserve && isTRUE(attr(pre_filtered_groups, ".drop")))) {
    filtered_groups <- anti_join(pre_filtered_groups, post_filtered_groups, by = groups)
    filtered_groups <- filtered_groups[, groups, drop = FALSE]
    filtered_groups[[".rows"]] <- rep(list(integer()), length.out = nrow(filtered_groups))
    post_filtered_groups <- bind_rows(post_filtered_groups, filtered_groups)
    ordered <- do.call(arrange_rows, list(post_filtered_groups, as_symbols(groups)))
    post_filtered_groups <- post_filtered_groups[ordered, ]
  }

  attr(res, "groups") <- post_filtered_groups
  res
}

# -- Helpers -------------------------------------------------------------------

check_filter <- function(conditions) {
  named <- have_name(conditions)
  for (i in which(named)) {
    if (!is.logical(conditions[[i]])) {
      stop(
        sprintf("Problem with `filter()` input `..%s`.\n", i),
        sprintf("Input `..%s` is named.\n", i),
        "This usually means that you've used `=` instead of `==`.\n",
        sprintf("Did you mean `%s == %s`?", names(conditions)[[i]], conditions[[i]])
      )
    }
  }
}

#' Pivot data from wide to long
#'
#' `pivot_longer()` "lengthens" data, increasing the number of rows and decreasing the number of columns. The inverse
#' transformation is [pivot_wider()].
#'
#' @param data `data.frame`. The data to pivot.
#' @param cols <[`poor-select`][select_helpers]>. Columns to pivot into longer format.
#' @param names_to `character(n)`. The name of the new column(s) that will contain the column names.
#' @param names_prefix `character(1)`. A regular expression used to remove matching text from the start of each variable
#' name.
#' @param names_sep,names_pattern `character(1)`. If `names_to` contains multiple values, this argument controls how the
#' column name is broken up. `names_pattern` takes a regular expression containing matching groups (`()`).
#' @param values_to `character(n)`. The name of the new column(s) that will contain the values of the pivoted variables.
#' @param values_drop_na `logical(1)`. If `TRUE`, will drop rows that contain only `NA` in the `values_to` column. This
#' effectively converts explicit missing values to implicit missing values, and should generally be used only when
#' missing values in data were created by its structure.
#' @param ... Additional arguments passed on to methods.
#'
#' @return A `data.frame`.
#'
#' @examples
#' wide_data <- data.frame(replicate(5, rnorm(10)))
#' # Customizing the names
#' pivot_longer(
#'   data = wide_data,
#'   cols = c(1, 2),
#'   names_to = "Column",
#'   values_to = "Numbers"
#' )
#'
#' @export
pivot_longer <- function(
  data,
  cols,
  names_to = "name",
  names_prefix = NULL,
  names_sep = NULL,
  names_pattern = NULL,
  values_to = "value",
  values_drop_na = FALSE,
  ...
) {

  if (missing(cols)) {
    stop("`cols` must select at least one column.")
  }

  cols <- names(eval_select_pos(data, substitute(cols)))

  if (any(names_to %in% setdiff(names(data), cols))) {
    stop(
      paste0(
        "Some values of the columns specified in 'names_to' are already present
        as column names. Either use another value in `names_to` or rename the
        following columns: ",
        paste(names_to[which(names_to %in% setdiff(names(data), cols))], sep = ", ")
      ),
      call. = FALSE)
  }

  # Sanity checks ----------------

  # nothing to select?
  if (length(cols) == 0L) {
    stop("No columns found for reshaping data.", call. = FALSE)
  }

  # Reshaping ---------------------
  # Create Index column as needed by reshape
  data[["_Row"]] <- as.numeric(rownames(data))

  # Create a new index for cases with length(names_to) > 1
  names_to_2 <- paste(names_to, collapse = "_")

  # Reshape
  long <- stats::reshape(
    data,
    varying = cols,
    idvar = "_Row",
    v.names = values_to,
    timevar = names_to_2,
    direction = "long"
  )

  # Cleaning --------------------------
  # Sort the dataframe (to match pivot_longer's output)
  long <- long[do.call(order, long[, c("_Row", names_to_2)]), ]

  long[["_Row"]] <- NULL

  # Re-insert col names as levels
  long[[names_to_2]] <- cols[long[[names_to_2]]]

  # if several variable in names_to, split the names either with names_sep or with names_pattern
  if (length(names_to) > 1) {
    if (is.null(names_pattern)) {
      for (i in seq_along(names_to)) {
        new_vals <- unlist(lapply(
          strsplit(unique(long[[names_to_2]]), names_sep, fixed = TRUE),
          function(x) x[i]
        ))
        long[[names_to[i]]] <- new_vals
      }
    } else {
      tmp <- regmatches(
        unique(long[[names_to_2]]),
        regexec(names_pattern, unique(long[[names_to_2]]))
      )
      tmp <- as.data.frame(do.call(rbind, tmp), stringsAsFactors = FALSE)
      names(tmp) <- c(names_to_2, names_to)
      # faster than merge
      long <- cbind(long, tmp[match(long[[names_to_2]], tmp[[names_to_2]]), -1])
    }
    long[[names_to_2]] <- NULL
  }

  # reorder
  long <- relocate(.data = long, values_to, .after = -1)

  # remove names prefix if specified
  if (!is.null(names_prefix)) {
    if (length(names_to) > 1) {
      stop("`names_prefix` only works when `names_to` is of length 1.", call. = FALSE)
    }
    long[[names_to]] <- gsub(paste0("^", names_prefix), "", long[[names_to]])
  }

  if (values_drop_na) {
    long <- long[!is.na(long[, values_to]), ]
  }

  # Reset row names
  rownames(long) <- NULL

  # Remove reshape attributes
  attributes(long)$reshapeLong <- NULL

  long
}

#' Union All
#'
#' Union all elements of R objects together.
#'
#' @param x,y objects to union all elements of (ignoring order)
#' @param ... other arguments passed on to methods
#'
#' @examples
#' first <- mtcars[1:20, ]
#' second <- mtcars[10:32, ]
#' union_all(first, second)
#'
#' # union_all does not remove duplicates
#' a <- data.frame(column = c(1:10, 10))
#' b <- data.frame(column = c(1:5, 5))
#' union_all(a, b)
#'
#' @export
union_all <- function(x, y, ...) {
  UseMethod("union_all")
}

#' @export
union_all.default <- function(x, y, ...) {
  c(x, y)
}

#' @export
union_all.data.frame <- function(x, y, ...) {
  out <- bind_rows(x, y)
  reconstruct_attrs(out, x)
}

#' @title Grouping metadata
#'
#' @param .data,x A `data.frame`.
#'
#' @examples
#' df <- data.frame(x = c(1,1,2,2))
#' group_vars(df)
#' group_rows(df)
#' group_data(df)
#'
#' gf <- group_by(df, x)
#' group_vars(gf)
#' group_rows(gf)
#' group_data(gf)
#'
#' @seealso See [context] for equivalent functions that return values for the current group.
#'
#' @name group_metadata
NULL

#' @description
#' * `group_data()` returns a data frame that defines the grouping structure. The columns give the values of the
#'   grouping variables. The last column, always called `.rows`, is a list of integer vectors that gives the location of
#'   the rows in each group.
#' @rdname group_metadata
#' @export
group_data <- function(.data) {
  UseMethod("group_data")
}

#' @export
group_data.data.frame <- function(.data) {
  structure(list(.rows = list(seq_len(nrow(.data)))), class = "data.frame", row.names = c(NA, -1L))
}

#' @export
group_data.grouped_df <- function(.data) {
  attr(.data, "groups")
}

#' @description
#' * `group_rows()` returns the rows which each group contains.
#' @rdname group_metadata
#' @export
group_rows <- function(.data) {
  group_data(.data)[[".rows"]]
}

#' @description
#' * `group_indices()` returns an integer vector the same length as `.data` that gives the group that each row belongs
#'   to.
#' @rdname group_metadata
#' @export
group_indices <- function(.data) {
  if (!has_groups(.data)) return(rep(1L, nrow(.data)))
  groups <- group_vars(.data)
  res <- unique(.data[, groups, drop = FALSE])
  res <- res[do.call(order, lapply(groups, function(x) res[, x])), , drop = FALSE]
  class(res) <- "data.frame"
  nrow_data <- nrow(.data)
  rows <- rep(NA, nrow_data)
  for (i in seq_len(nrow_data)) {
    rows[i] <- which(interaction(res[, groups]) %in% interaction(.data[i, groups]))
  }
  rows
}

#' @description
#' * `group_vars()` gives names of grouping variables as character vector.
#' @rdname group_metadata
#' @export
group_vars <- function(x) {
  groups <- attr(x, "groups", exact = TRUE)
  if (is.null(groups)) character(0) else colnames(groups)[!colnames(groups) %in% c(".group_id", ".rows")]
}

#' @description
#' * `groups()` gives the names as a list of symbols.
#' @rdname group_metadata
#' @export
groups <- function(x) {
  as_symbols(group_vars(x))
}

#' @description
#' * `group_size()` gives the size of each group.
#' @rdname group_metadata
#' @export
group_size <- function(x) {
  lengths(group_rows(x))
}

#' @description
#' * `n_groups()` gives the total number of groups.
#' @rdname group_metadata
#' @export
n_groups <- function(x) {
  nrow(group_data(x))
}

# group_data() -----------------------------------------------------------------

res <- data.frame(x = 1:3) %>% group_data()
expect_identical(class(res), "data.frame", info = "group_data() returns a data frame")
expect_equal(
  res,
  structure(list(.rows = list(1:3)), class = "data.frame", row.names = c(NA, -1L)),
  info = "group_data() returns all rows from .data"
)

res <- data.frame(x = c(1, 1, 2)) %>% group_by(x) %>% group_data()
expect_identical(class(res), "data.frame")
expect_equivalent(
  res,
  structure(list(x = c(1, 2), .rows = list(1:2, 3L)), row.names = c(NA, -2L), class = "data.frame"),
  info = "group_data() returns the unique rows for each group"
)

# group_rows() -----------------------------------------------------------------

df <- data.frame(x = 1:2, y = 1:2) %>% group_by(x, y) %>% group_data()
gf <- group_by(df, x, y)
gd <- group_data(gf)
expect_equal(group_rows(gf), gd[[".rows"]], info = "group_rows() partitions group_data()")

# group_indices() ---------------------------------------------------------

df <- data.frame(x = c("b", "a", "b"), stringsAsFactors = FALSE)
gf <- group_by(df, x)
expect_equal(group_indices(df), c(1, 1, 1), info = "group_indices() returns expected values for ungrouped data")
expect_equal(group_indices(gf), c(2, 1, 2), info = "group_indices() returns expected values for grouped data")

# group_size --------------------------------------------------------------

df <- data.frame(x = rep(1:3, each = 10), y = rep(1:6, each = 5))
expect_equal(n_groups(df), 1L, info = "Ungrouped data has 1 group")
expect_equal(group_size(df), 30, info = "Ungrouped data has group_size() equal to nrow(data)")

df <- data.frame(x = rep(1:3, each = 10), y = rep(1:6, each = 5)) %>% group_by(x)
expect_equal(n_groups(df), 3L, info = "n_groups() is correct for grouped data")
expect_equal(group_size(df), rep(10, 3), info = "group_size() is correct for grouped data")

# groups -----------------------------------------------------------------------
expect_equal(
  df %>% groups(),
  list(as.symbol("x")),
  info = "groups() returns list of group variables as symbols"
)

rm(res, df, gd, gf)

expect_equal(
  mtcars %>% summarise(mean(mpg)),
  data.frame("mean(mpg)" = mean(mtcars$mpg), check.names = FALSE),
  info = "Test single summaries"
)

expect_equal(
  mtcars %>% summarise(meanMpg = mean(mpg)),
  data.frame("meanMpg" = mean(mtcars$mpg)),
  info = "Test single summaries with column name"
)

expect_equal(
  mtcars %>% summarise(mean(mpg), sum(disp)),
  data.frame("mean(mpg)" = mean(mtcars$mpg), "sum(disp)" = sum(mtcars$disp), check.names = FALSE),
  info = "Test multiple summarise"
)

expect_equal(
  mtcars %>% summarise(meanMpg = mean(mpg), sumDisp = sum(disp)),
  data.frame("meanMpg" = mean(mtcars$mpg), "sumDisp" = sum(mtcars$disp), check.names = FALSE),
  info = "Test multiple summarise with column names"
)

expect_equal(
  summarise(data.frame(x = 1:10), y = mean(x), z = y + 1),
  data.frame(y = 5.5, z = 6.5),
  info = "Can use freshly create variables"
)

expect_equal(
  mtcars %>% summarise(n(), range(mpg)),
  structure(list(`n()` = c(32L, 32L), `range(mpg)` = c(10.4, 33.9)), class = "data.frame", row.names = c(NA, -2L)),
  info = "Functions returning multiple values get split over rows"
)

expect_equal(
  mtcars %>% summarise(n(), list(range(mpg))),
  structure(list(`n()` = 32L, `list(range(mpg))` = list(c(10.4, 33.9))), class = "data.frame", row.names = c(NA, -1L)),
  info = "Functions returning multiple values that are wrapped in list are returned as nested columns"
)

# Grouped Operations
res <- mtcars %>% group_by(am, cyl, gear) %>% summarise(meanMpg = mean(mpg), sumDisp = sum(disp))
gd <- group_data(res)
expect_equal(
  gd,
  structure(
    list(
      am = c(0, 0, 0, 1, 1, 1),
      cyl = c(4, 6, 8, 4, 6, 8),
      .rows = list(1:2, 3:4, 5L, 6:7, 8:9, 10L)
    ), row.names = c(NA, 6L), class = "data.frame", .drop = TRUE
  ),
  info = "Ensure the summarised data still contains group attributes"
)

attr(res, "groups") <- NULL
expect_equal(
  res,
  structure(
    list(
      am = c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1),
      cyl = c(4, 4, 6, 6, 8, 4, 4, 6, 6, 8),
      gear = c(3, 4, 3, 4, 3, 4, 5, 4, 5, 5),
      meanMpg = c(21.5, 23.6, 19.75, 18.5, 15.05, 28.0333333333333, 28.2, 21, 19.7, 15.4),
      sumDisp = c(120.1, 287.5, 483, 335.2, 4291.4, 533.5, 215.4, 320, 145, 652)
    ),
    row.names = c(NA, 10L),
    class = c("grouped_df", "data.frame")
  ),
  info = "Test multiple groups and multiple summary functions"
)

expect_equal(
  summarise(mtcars),
  data.frame(),
  info = "empty summarise returns empty data.frame"
)

res <- mtcars %>% group_by(am) %>% summarise()
expect_equal(
  class(res),
  c("grouped_df", "data.frame"),
  info = "empty grouped summarise() returns groups #1"
)
expect_equal(
  {
    attr(res, "groups") <- NULL
    class(res) <- "data.frame"
    res
  },
  data.frame(am = c(0, 1)),
  info = "empty grouped summarise() returns groups #2"
)

# .groups
df <- data.frame(x = 1, y = 2) %>% group_by(x, y)
expect_equal(
  df %>% summarise() %>% group_vars(),
  "x",
  info = ".groups = NULL drops the last grouping variable when the results return a single value"
)
expect_equal(
  df %>% summarise(.groups = "drop_last") %>% group_vars(),
  "x",
  info = ".groups = 'drop_last' drops the last grouping variable"
)
expect_equal(
  df %>% summarise(.groups = "drop") %>% group_vars(),
  character(),
  info = ".groups = 'drop' drops all grouping variables"
)
expect_equal(
  df %>% summarise(.groups = "keep") %>% group_vars(),
  c("x", "y"),
  info = ".groups = 'keep' keeps all grouping variables"
)
expect_equal(
  mtcars %>%
    group_by(cyl, vs) %>%
    summarise(cyl_n = n()) %>%
    group_vars(),
  "cyl",
  info = ".groups = NULL drops the last grouping variable when the results return a single value for multiple groups"
)

#' @rdname mutate
#' @export
transmute <- function(.data, ...) {
  UseMethod("transmute")
}

#' @export
transmute.data.frame <- function(.data, ...) {
  mutate(.data, ..., .keep = "none")
}

#' @export
transmute.grouped_df <- function(.data, ...) {
  rows <- rownames(.data)
  res <- apply_grouped_function("transmute", .data, drop = TRUE, ...)
  res[rows, ]
}

#' Subset rows by position
#'
#' Subset rows by their original position in the `data.frame`. Grouped `data.frame`s use the position within each group.
#'
#' @param .data A `data.frame`.
#' @param ... For `slice()`: integer row values.
#'
#' Provide either positive values to keep, or negative values to drop. The values provided must be either all positive
#' or negative. Indices beyond the number of rows in the input are silently ignored.
#' @param n,prop Provide either `n`, the number of rows, or `prop`, the proportion of rows to select. If neither are
#' supplied, `n = 1` will be used.
#'
#' If `n` is greater than the number of rows in the group (or `prop > 1`), the result will be silently truncated to the
#' group size. If the `prop`ortion of a group size is not an integer, it is rounded down.
#'
#' @return
#' An object of the same type as `.data`. The output has the following properties:
#'
#' * Each row may appear 0, 1, or many times in the output.
#' * Columns are not modified.
#' * Groups are not modified.
#' * Data frame attributes are preserved.
#'
#' @examples
#' slice(mtcars, c(1, 2, 3))
#' mtcars %>% slice(1:3)
#'
#' # Similar to head(mtcars, 1)
#' mtcars %>% slice(1L)
#'
#' # Similar to tail(mtcars, 1):
#' mtcars %>% slice(n())
#' mtcars %>% slice(5:n())
#' # Rows can be dropped with negative indices:
#' slice(mtcars, -(1:4))
#'
#' # First and last rows based on existing order
#' mtcars %>% slice_head(n = 5)
#' mtcars %>% slice_tail(n = 5)
#'
#' # Grouped operations:
#' mtcars %>% group_by(am, cyl, gear) %>% slice_head(n = 2)
#'
#' @importFrom utils head
#'
#' @name slice
NULL

#' @rdname slice
#' @export
slice <- function(.data, ...) {
  UseMethod("slice")
}

#' @export
slice.data.frame <- function(.data, ...) {
  if (nrow(.data) == 0L) return(.data)
  pos <- slice_positions(.data, ...)
  .data[pos, , drop = FALSE]
}

#' @export
slice.grouped_df <- function(.data, ...) {
  apply_grouped_function("slice", .data, drop = TRUE, ...)
}


#' @rdname slice
#' @export
slice_head <- function(.data, ..., n, prop) {
  UseMethod("slice_head")
}

#' @export
slice_head.data.frame <- function(.data, ..., n, prop) {
  size <- check_slice_size(n, prop)
  idx <- switch(
    size$type,
    n = function(n) seq2(1, min(size$n, n)),
    prop = function(n) seq2(1, min(size$prop * n, n))
  )
  slice(.data, idx(poorman::n()))
}

#' @export
slice_head.grouped_df <- function(.data, ..., n, prop) {
  apply_grouped_function("slice_head", .data, drop = TRUE, n = n, prop = prop, ...)
}

#' @rdname slice
#' @export
slice_tail <- function(.data, ..., n, prop) {
  UseMethod("slice_tail")
}

#' @export
slice_tail.data.frame <- function(.data, ..., n, prop) {
  size <- check_slice_size(n, prop)
  idx <- switch(
    size$type,
    n = function(n) seq2(max(n - size$n + 1, 1), n),
    prop = function(n) seq2(max(ceiling(n - size$prop * n) + 1, 1), n)
  )
  slice(.data, idx(poorman::n()))
}

#' @export
slice_tail.grouped_df <- function(.data, ..., n, prop) {
  apply_grouped_function("slice_tail", .data, drop = TRUE, n = n, prop = prop, ...)
}

#' @param order_by The variable to order by.
#' @param with_ties `logical(1)`. Should ties be kept together? The default, `TRUE`, may return more rows than you
#' request. Use `FALSE` to ignore ties, and return the first `n` rows.
#'
#' @rdname slice
#' @export
slice_min <- function(.data, order_by, ..., n, prop, with_ties = TRUE) {
  UseMethod("slice_min")
}

#' @export
slice_min.data.frame <- function(.data, order_by, ..., n, prop, with_ties = TRUE) {
  if (missing(order_by)) stop("argument `order_by` is missing, with no default.")

  size <- check_slice_size(n, prop)
  idx <- if (isTRUE(with_ties)) {
    switch(
      size$type,
      n = function(x, n) head(order(x), smaller_ranks(x, size$n)),
      prop = function(x, n) head(order(x), smaller_ranks(x, size$prop * n))
    )
  } else {
    switch(
      size$type,
      n = function(x, n) head(order(x), size$n),
      prop = function(x, n) head(order(x), size$prop * n)
    )
  }
  order_by <- .data[, deparse_var(order_by)]
  slice(.data, idx(order_by, poorman::n()))
}

#' @export
slice_min.grouped_df <- function(.data, order_by, ..., n, prop, with_ties = TRUE) {
  eval_env$env <- environment()
  on.exit(rm(list = "env", envir = eval_env), add = TRUE)
  apply_grouped_function(
    "slice_min", .data, drop = TRUE, order_by = order_by, n = n, prop = prop, with_ties = with_ties, ...
  )
}

#' @rdname slice
#' @export
slice_max <- function(.data, order_by, ..., n, prop, with_ties = TRUE) {
  UseMethod("slice_max")
}

#' @export
slice_max.data.frame <- function(.data, order_by, ..., n, prop, with_ties = TRUE) {
  if (missing(order_by)) stop("argument `order_by` is missing, with no default.")

  size <- check_slice_size(n, prop)
  idx <- if (isTRUE(with_ties)) {
    switch(
      size$type,
      n = function(x, n) head(order(x, decreasing = TRUE), smaller_ranks(desc(x), size$n)),
      prop = function(x, n) head(order(x, decreasing = TRUE), smaller_ranks(desc(x), size$prop * n))
    )
  } else {
    switch(
      size$type,
      n = function(x, n) head(order(x, decreasing = TRUE), size$n),
      prop = function(x, n) head(order(x, decreasing = TRUE), size$prop * n)
    )
  }
  order_by <- .data[, deparse_var(order_by)]
  slice(.data, idx(order_by, poorman::n()))
}

#' @export
slice_max.grouped_df <- function(.data, order_by, ..., n, prop, with_ties = TRUE) {
  eval_env$env <- environment()
  on.exit(rm(list = "env", envir = eval_env), add = TRUE)
  apply_grouped_function(
    "slice_max", .data, drop = TRUE, order_by = order_by, n = n, prop = prop, with_ties = with_ties, ...
  )
}

#' @param replace `logical(1)`. Should sampling be performed with (`TRUE`) or without (`FALSE`, the default)
#' replacement.
#' @param weight_by Sampling weights. This must evaluate to a vector of non-negative numbers the same length as the
#' input. Weights are automatically standardised to sum to 1.
#'
#' @rdname slice
#' @export
slice_sample <- function(.data, ..., n, prop, weight_by = NULL, replace = FALSE) {
  UseMethod("slice_sample")
}

#' @export
slice_sample.data.frame <- function(.data, ..., n, prop, weight_by = NULL, replace = FALSE) {
  size <- check_slice_size(n, prop)
  idx <- switch(
    size$type,
    n = function(x, n) sample_int(n, size$n, replace = replace, wt = x),
    prop = function(x, n) sample_int(n, size$prop * n, replace = replace, wt = x),
  )
  weight_by <- deparse_var(weight_by)
  if (!is.null(weight_by)) weight_by <- .data[, weight_by]
  slice(.data, idx(weight_by, poorman::n()))
}

#' @export
slice_sample.grouped_df <- function(.data, ..., n, prop, weight_by = NULL, replace = FALSE) {
  eval_env$env <- environment()
  on.exit(rm(list = "env", envir = eval_env), add = TRUE)
  apply_grouped_function(
    "slice_sample", .data, drop = TRUE, n = n, prop = prop, weight_by = weight_by, replace = replace, ...
  )
}

# helpers ----------------------------------------------------------------------

slice_positions <- function(.data, ...) {
  conditions <- dotdotdot(...)
  context$setup(.data)
  on.exit(context$clean(), add = TRUE)
  if (length(conditions) == 0L) return(seq_len(n()))

  frame <- parent.frame(2L)
  rows <- lapply(
    conditions,
    function(cond, frame) {
      res <- eval(cond, context$.data, frame)
      if (is.logical(res) && all(is.na(res))) {
        res <- integer()
      } else if (is.numeric(res)) {
        res <- as.integer(res)
      } else if (!is.integer(res)) {
        stop("`slice()` expressions should return indices (positive or negative integers).")
      }
    },
    frame = frame
  )
  rows <- do.call(c, rows)
  if (length(rows) == 0L) {
    # do nothing
  } else if (all(rows >= 0, na.rm = TRUE)) {
    rows <- rows[!is.na(rows) & rows <= n() & rows > 0]
  } else if (all(rows <= 0, na.rm = TRUE)) {
    rows <- setdiff(seq_len(n()), -rows)
  } else {
    stop("`slice()` expressions should return either all positive or all negative.")
  }
  rows
}

check_slice_size <- function(n, prop) {
  if (missing(n) && missing(prop)) {
    list(type = "n", n = 1L)
  } else if (!missing(n) && missing(prop)) {
    if (!is.numeric(n) || length(n) != 1) {
      stop("`n` must be a single number.")
    }
    if (is.na(n) || n < 0) {
      stop("`n` must be a non-missing positive number.")
    }

    list(type = "n", n = n)
  } else if (!missing(prop) && missing(n)) {
    if (!is.numeric(prop) || length(prop) != 1) {
      stop("`prop` must be a single number.")
    }
    if (is.na(prop) || prop < 0) {
      stop("`prop` must be a non-missing positive number.")
    }
    list(type = "prop", prop = prop)
  } else {
    stop("Must supply exactly one of `n` and `prop` arguments.")
  }
}

sample_int <- function(n, size, replace = FALSE, wt = NULL) {
  if (isTRUE(replace)) {
    sample.int(n, size, prob = wt, replace = TRUE)
  } else {
    sample.int(n, min(size, n), prob = wt)
  }
}

smaller_ranks <- function(x, y) {
  sum(min_rank(x) <= y, na.rm = TRUE)
}

expect_equal(
  mtcars %>% slice(),
  mtcars,
  info = "Empty slice() returns the input data"
)

expect_equal(
  mtcars %>% slice(1, 5, 9),
  mtcars[c(1, 5, 9), ],
  info = "Test that slice() handles numeric input"
)

expect_equal(
  mtcars %>% group_by(am, cyl) %>% slice(1, 2) %>% ungroup(),
  do.call(rbind, unname(lapply(split(mtcars, list(mtcars$am, mtcars$cyl)), function(x) x[1:2, ]))),
  info = "Test that grouped slice() operations handle numeric input"
)

expect_equal(
  mtcars %>% slice(-(1:2)),
  mtcars[-(1:2), ],
  info = "Test that slice() works with negative indices"
)

expect_error(
  mtcars %>% slice(-1, 1),
  info = "`slice()` expressions should return either all positive or all negative."
)

expect_error(
  mtcars %>% slice(TRUE),
  info = "`slice()` expressions should return indices (positive or negative integers)."
)

expect_equal(
  mtcars %>% slice(1:3),
  mtcars[1:3, ],
  info = "Test slicing rows with sequence"
)

expect_equal(
  mtcars %>% slice(1:100),
  mtcars,
  info = "Test that extra rows aren't returned if the user slices more rows than are available"
)

expect_equal(
  mtcars %>% slice(2, 2, 2),
  mtcars[c(2, 2, 2), ],
  info = "Test that slicing the same row number n times duplicates that row n times"
)

expect_equal(
  mtcars %>% slice(NA),
  mtcars[0, ],
  info = "Test that slice() handles NA values"
)

expect_equal(
  mtcars %>% slice(1, NA),
  mtcars[1, ],
  info = "Test that slice() handles a mixture of NA and numeric values"
)

expect_equal(
  data.frame(x = numeric()) %>% slice(1:3),
  structure(list(x = numeric(0)), row.names = integer(0), class = "data.frame"),
  info = "Test that slice() handles empty data.frames"
)

expect_equal(
  data.frame(a = 1:3) %>% select(a) %>% slice(1) %>% nrow(),
  1L,
  info = "Test that slice() handles 0-column data.frames"
)

# Slice variants

df <- data.frame(x = 1:5)
expect_equal(df %>% slice_head(n = 6) %>% nrow(), 5, info = "slice_head() silently truncates results")
expect_equal(df %>% slice_tail(n = 6) %>% nrow(), 5, info = "slice_tail() silently truncates results")
expect_equal(df %>% slice_sample(n = 6) %>% nrow(), 5, info = "slice_sample() silently truncates results")
expect_equal(df %>% slice_min(x, n = 6) %>% nrow(), 5, info = "slice_min() silently truncates results")
expect_equal(df %>% slice_max(x, n = 6) %>% nrow(), 5, info = "slice_max() silently truncates results")

df <- data.frame(x = 1:10)
expect_equal(df %>% slice_head(prop = 0.11) %>% nrow(), 1, info = "slice_head() correctly computes the proportion")
expect_equal(df %>% slice_tail(prop = 0.11) %>% nrow(), 1, info = "slice_tail() correctly computes the proportion")
expect_equal(df %>% slice_sample(prop = 0.11) %>% nrow(), 1, info = "slice_sample() correctly computes the proportion")
expect_equal(df %>% slice_min(x, prop = 0.11) %>% nrow(), 1, info = "slice_min() correctly computes the proportion")
expect_equal(df %>% slice_max(x, prop = 0.11) %>% nrow(), 1, info = "slice_max() correctly computes the proportion")
expect_equal(
  df %>% slice_min(x, prop = 0.11, with_ties = FALSE) %>% nrow(),
  1,
  info = "slice_min(with_ties = FALSE) correctly computes the proportion"
)
expect_equal(
  df %>% slice_max(x, prop = 0.11, with_ties = FALSE) %>% nrow(),
  1,
  info = "slice_max(with_ties = FALSE) correctly computes the proportion"
)

df <- data.frame(x = c(1, 1, 1, 2, 2))
expect_equal(df %>% slice_min(x) %>% nrow(), 3, info = "slice_min() returns ties by default")
expect_equal(df %>% slice_max(x) %>% nrow(), 2, info = "slice_max() returns ties by default")
expect_equal(
  df %>% slice_min(x, with_ties = FALSE) %>% nrow(),
  1,
  info = "slice_min(with_ties = FALSE) does not return ties"
)
expect_equal(
  df %>% slice_max(x, with_ties = FALSE) %>% nrow(),
  1,
  info = "slice_max(with_ties = FALSE) does not return ties"
)

df <- data.frame(id = 1:4, x = c(2, 3, 1, 2))
expect_equal(df %>% slice_min(x, n = 2) %>% pull(id), c(3, 1, 4), info = "slice_min() reorders results")
expect_equal(
  df %>% slice_min(x, n = 2, with_ties = FALSE) %>% pull(id),
  c(3, 1),
  info = "slice_min(with_ties = FALSE) reorders results"
)
expect_equal(df %>% slice_max(x, n = 2) %>% pull(id), c(2, 1, 4), info = "slice_max() reorders results")
expect_equal(
  df %>% slice_max(x, n = 2, with_ties = FALSE) %>% pull(id),
  c(2, 1),
  info = "slice_max(with_ties = FALSE) reorders results"
)

df <- data.frame(id = 1:4, x = c(2, NA, 1, 2), y = c(NA, NA, NA, NA))
expect_equal(df %>% slice_min(x, n = 2) %>% pull(id), c(3, 1, 4), info = "slice_min() ignores NAs")
expect_equal(df %>% slice_min(y, n = 2) %>% nrow(), 0, info = "slice_min() ignores NAs")
expect_equal(df %>% slice_max(x, n = 2) %>% pull(id), c(1, 4), info = "slice_max() ignores NAs")
expect_equal(df %>% slice_max(y, n = 2) %>% nrow(), 0, info = "slice_max() ignores NAs")

df <- data.frame(x = 1:100, wt = c(1, rep(0, 99)))
expect_equal(
  df %>% slice_sample(n = 1, weight_by = wt) %>% pull(x),
  1,
  info = "Arguments to slice_sample() are passed along"
)
expect_equal(
  df %>% slice_sample(n = 2, weight_by = wt, replace = TRUE) %>% pull(x),
  c(1, 1),
  info = "Arguments to slice_sample() are passed along"
)

expect_error(
  poorman:::check_slice_size(n = c(1, 2)),
  info = "`n` must be a single number."
)

expect_error(
  poorman:::check_slice_size(n = NA),
  info = "`n` must be a non-missing positive number."
)

expect_error(
  poorman:::check_slice_size(n = -1),
  info = "`n` must be a non-missing positive number."
)

expect_error(
  poorman:::check_slice_size(prop = c(1, 2)),
  info = "`prop` must be a single number."
)

expect_error(
  poorman:::check_slice_size(prop = NA),
  info = "`prop` must be a non-missing positive number."
)

expect_error(
  poorman:::check_slice_size(prop = -1),
  info = "`prop` must be a non-missing positive number."
)

expect_error(
  poorman:::check_slice_size(n = 1, prop = 0.1),
  info = "Must supply exactly one of `n` and `prop` arguments."
)

rm(df)

#' Vectorised if
#'
#' This is a wrapper around `ifelse()` which checks that `true` and `false` are of the same type, making the output more
#' predictable.
#'
#' @param condition A `logical(n)` vector.
#' @param true,false Values to use for `TRUE` and `FALSE` in `condition`. They must either be the same length as
#' `condition` or be length 1. They must also be the same type.
#' @param missing If not `NULL` (the default), this will replace any missing values.
#'
#' @return A vector the same length as `condition` with values for `TRUE` and `FALSE` replaced by those specified in
#' `true` and `false`, respectively.
#'
#' @examples
#' x <- c(-5:5, NA)
#' if_else(x < 0, NA_integer_, x)
#' if_else(x < 0, "negative", "positive", "missing")
#'
#' # Unlike ifelse, if_else preserves types
#' x <- factor(sample(letters[1:5], 10, replace = TRUE))
#' ifelse(x %in% c("a", "b", "c"), x, factor(NA))
#' # Attributes are taken from the `true` vector
#' if_else(x %in% c("a", "b", "c"), x, factor(NA))
#'
#' @export
if_else <- function(condition, true, false, missing = NULL) {
  if (!is.logical(condition)) stop("`condition` must be a logical vector.")
  cls_true <- class(true)
  cls_false <- class(false)
  cls_missing <- class(missing)
  if (!identical(cls_true, cls_false)) {
    stop("The class of `true` <", class(true), "> is not the same as the class of `false` <", class(false), ">")
  }
  if (!is.null(missing) && !identical(cls_true, cls_missing)) {
    stop("`missing` must be a ", cls_true, " vector, not a ", cls_missing, " vector.")
  }
  res <- ifelse(condition, true, false)
  if (!is.null(missing)) res[is.na(res)] <- missing
  attributes(res) <- attributes(true)
  res
}

expect_equal(
  if_else(c(TRUE, TRUE, FALSE, FALSE), 1, 2),
  c(1, 1, 2, 2),
  info = "Scalar true and false are vectorised"
)

expect_error(
  if_else(c(TRUE, TRUE, FALSE, FALSE), 1, "2"),
  info = "true and false should be the same type"
)

expect_equal(
  {
    x <- c(-1, 0, 1)
    if_else(x < 0, x, 0)
  },
  c(-1, 0, 0),
  info = "Vectorised true works"
)

expect_equal(
  {
    x <- c(-1, 0, 1)
    if_else(x > 0, x, 0)
  },
  c(0, 0, 1),
  info = "Vectorised false works"
)

expect_equal(
  if_else(c(TRUE, NA, FALSE), -1, 1),
  c(-1, NA, 1),
  info = "Missing values are missing by default"
)

expect_equal(
  if_else(c(TRUE, NA, FALSE), -1, 1, 0),
  c(-1, 0, 1),
  info = "Missing values are replaced"
)

expect_error(
  if_else(c(TRUE, NA, FALSE), -1, 1, "should fail"),
  info = "missing should be the same type as true and false"
)

expect_equal(
  {
    x <- factor(letters[1:5])
    if_else(x %in% c("a", "b", "c"), x, factor(NA))
  },
  factor(c("a", "b", "c", NA, NA), levels = letters[1:5]),
  info = "if_else works with factors"
)

expect_equal(
  {
    x <- list(1, 2, 3)
    if_else(c(TRUE, TRUE, FALSE), x, list(NULL))
  },
  list(1, 2, NULL),
  info = "if_else works with lists"
)

#' Extract the first, last or nth value from a vector
#'
#' These are straightforward wrappers around \code{\link{[[}}. The main advantage is that you can provide an optional
#' secondary vector that defines the ordering, and provide a default value to use when the input is shorter than
#' expected.
#'
#' @param x A vector
#' @param n For `nth()`, a single integer specifying the position. Negative integers index from the end (i.e. `-1L`
#' will return the last value in the vector).
#'
#' If a double is supplied, it will be silently truncated.
#' @param order_by An optional vector used to determine the order
#' @param default A default value to use if the position does not exist in the input. This is guessed by default for
#' base vectors, where a missing value of the appropriate type is returned, and for lists, where a `NULL` is return.
#'
#' For more complicated objects, you'll need to supply this value. Make sure it is the same type as `x`.
#'
#' @return A single value. `[[` is used to do the subsetting.
#'
#' @examples
#' x <- 1:10
#' y <- 10:1
#'
#' first(x)
#' last(y)
#'
#' nth(x, 1)
#' nth(x, 5)
#' nth(x, -2)
#' nth(x, 11)
#'
#' last(x)
#' # Second argument provides optional ordering
#' last(x, y)
#'
#' # These functions always return a single value
#' first(integer())
#'
#' @export
nth <- function(x, n, order_by = NULL, default = default_missing(x)) {
  if (length(n) != 1 || !is.numeric(n)) stop("`n` must be a single integer.")
  n <- trunc(n)
  if (n == 0 || n > length(x) || n < -length(x)) return(default)
  if (n < 0) n <- length(x) + n + 1
  if (is.null(order_by)) x[[n]] else x[[order(order_by)[[n]]]]
}

#' @export
#' @rdname nth
first <- function(x, order_by = NULL, default = default_missing(x)) {
  nth(x, 1L, order_by = order_by, default = default)
}

#' @export
#' @rdname nth
last <- function(x, order_by = NULL, default = default_missing(x)) {
  nth(x, -1L, order_by = order_by, default = default)
}

default_missing <- function(x) {
  UseMethod("default_missing")
}

#' @export
default_missing.default <- function(x) {
  if (!is.object(x) && is.list(x)) NULL else x[NA_real_]
}

#' @export
default_missing.data.frame <- function(x) {
  rep(NA, nrow(x))
}

x <- list(1, 2, 3)
expect_equal(nth(x, 1), 1, info = "nth works with lists #1")
expect_equal(nth(x, 4), NULL, info = "nth works with lists #2")
expect_equal(nth(x, 4, default = 1), 1, info = "nth works with lists #3")

x <- 1:5
expect_equal(nth(x, -1), 5, info = "negative values index from end #1")
expect_equal(nth(x, -3), 3, info = "negative values index from end #2")

expect_equal(nth(1:4, 5), NA_integer_, info = "indexing past ends returns default value #1")
expect_equal(nth(1:4, -5), NA_integer_, info = "indexing past ends returns default value #2")
expect_equal(nth(1:4, -10), NA_integer_, info = "indexing past ends returns default value #3")

expect_equal(first(logical()), NA, info = "first() uses default value for 0 length logical vectors")
expect_equal(first(integer()), NA_integer_, info = "first() uses default value for 0 length integer vectors")
expect_equal(first(numeric()), NA_real_, info = "first() uses default value for 0 length numeric vectors")
expect_equal(first(character()), NA_character_, info = "first() uses default value for 0 length character vectors")
expect_equal(first(list()), NULL, info = "first() uses default value for 0 length lists")

fc <- factor("a")[0]
expect_equal(first(fc[0]), fc[NA], info = "first() uses default value for 0 length augmented vectors #1")
dt <- Sys.Date()
expect_equal(first(dt[0]), dt[NA], info = "first() uses default value for 0 length augmented vectors #2")
tm <- Sys.time()
expect_equal(first(tm[0]), tm[NA], info = "first() uses default value for 0 length augmented vectors #3")

expect_error(nth(1:10, "x"), info = "nth() gives meaningful error message")

#' Split data.frame by groups
#'
#' @description
#' `group_split()` works like [base::split()] but
#'
#' * it uses the grouping structure from [group_by()] and is therefore subject to the data mask
#' * it does not name the elements of the list based on the grouping as this typically loses information and is
#'   confusing
#'
#' @details
#' **Grouped `data.frame`s:**
#'
#' The primary use case for `group_split()` is with already grouped `data.frame`s, typically a result of [group_by()].
#' In this case, `group_split()` only uses the first argument, the grouped `data.frame`, and warns when `...` is used.
#'
#' Because some of these groups may be empty, it is best paired with `group_keys()` which identifies the representatives
#' of each grouping variable for the group.
#'
#' **Ungrouped `data.frame`s:**
#'
#' When used on ungrouped `data.frame`s, `group_split()` forwards the `...` to `group_by()` before the split, therefore
#' the `...` are subject to the data mask.
#'
#' @param .data A `data.frame`.
#' @param ... Grouping specification, forwarded to [group_by()].
#' @param .keep `logical(1)`. Should the grouping columns be kept (default: TRUE)?
#'
#' @return
#' * `group_split()` returns a list of `data.frame`s. Each `data.frame` contains the rows of `.data` with the associated
#'   group and all the columns, including the grouping variables.
#' * `group_keys()` returns a `data.frame` with one row per group, and one column per grouping variable
#'
#' @seealso [group_by()]
#'
#' @examples
#' # Grouped data.frames:
#' mtcars %>% group_by(cyl, am) %>% group_split()
#' mtcars %>% group_by(cyl, am) %>% group_split(.keep = FALSE)
#' mtcars %>% group_by(cyl, am) %>% group_keys()
#'
#' # Ungrouped data.frames:
#' mtcars %>% group_split(am, cyl)
#'
#' @export
group_split <- function(.data, ..., .keep = TRUE) {
  dots_len <- length(dotdotdot(...)) > 0L
  if (has_groups(.data) && isTRUE(dots_len)) {
    warning("... is ignored in group_split(<grouped_df>), please use group_by(..., .add = TRUE) %>% group_split()")
  }
  if (!has_groups(.data) && isTRUE(dots_len)) {
    .data <- group_by(.data, ...)
  }
  if (!has_groups(.data) && !isTRUE(dots_len)) {
    return(list(.data))
  }
  context$setup(.data)
  on.exit(context$clean(), add = TRUE)
  groups <- group_vars(.data)
  attr(context$.data, "groups") <- NULL
  res <- split_into_groups(context$.data, groups)
  names(res) <- NULL
  if (!isTRUE(.keep)) {
    res <- lapply(res, function(x) x[, !colnames(x) %in% groups, drop = FALSE])
  }
  any_empty <- unlist(lapply(res, function(x) !(nrow(x) == 0L)))
  res[any_empty]
}

#' @rdname group_split
#' @export
group_keys <- function(.data) {
  groups <- group_vars(.data)
  context$setup(.data)
  res <- context$get_columns(context$get_colnames() %in% groups)
  res <- res[!duplicated(res), , drop = FALSE]
  if (nrow(res) == 0L) return(res)
  class(res) <- "data.frame"
  res <- res[do.call(order, lapply(groups, function(x) res[, x])), , drop = FALSE]
  rownames(res) <- NULL
  res
}

# -- Helpers -------------------------------------------------------------------

#' Split a `data.frame` into groups.
#'
#' @param .data A `data.frame`.
#' @param groups `character(n)`. A `vector` of grouping variables to split by.
#' @param drop `logical(1)`. Drop levels that do not occur?
#' @param ... Additional parameters to be passed to [split()].
#'
#' @return A `list` with a `data.frame` in each level.
#' @seealso [split()]
#' @noRd
split_into_groups <- function(.data, groups, drop = FALSE, ...) {
  class(.data) <- "data.frame"
  group_factors <- lapply(groups, function(x, .data) as.factor(.data[, x]), .data)
  split(x = .data, f = group_factors, drop = drop, ...)
}

res <- split(mtcars, mtcars$cyl)
names(res) <- NULL
expect_equal(
  mtcars %>% group_by(cyl) %>% group_split(),
  res,
  info = "group_split() works"
)

expect_equal(
  mtcars %>% group_split(cyl),
  res,
  info = "Non-grouped group_split() works"
)
rm(res)

expect_equal(
  mtcars %>% group_by(cyl, am) %>% group_split(.keep = FALSE),
  {
    res <- split(mtcars, list(mtcars$cyl, mtcars$am))
    res <- lapply(res, function(x) x[, !colnames(x) %in% c("cyl", "am")])
    names(res) <- NULL
    res
  },
  info = ".keep = FALSE drops the grouping columns"
)

expect_equal(
  mtcars %>% group_by(cyl, am, gear) %>% group_split(),
  {
    res <- split(mtcars, list(mtcars$cyl, mtcars$am, mtcars$gear))
    names(res) <- NULL
    any_empty <- unlist(lapply(res, function(x) !(nrow(x) == 0L)))
    res[any_empty]
  },
  info = "Empty group combinations are dropped"
)

expect_equal(
  mtcars %>% group_split(),
  list(mtcars),
  info = "Non-grouped data with no groups given returns the whole dataset as a list"
)

expect_equal(
  mtcars %>% group_by(cyl, am) %>% group_keys(),
  structure(
    list(cyl = c(4, 4, 6, 6, 8, 8), am = c(0, 1, 0, 1, 0, 1)),
    row.names = c(NA, -6L),
    class = "data.frame"
  ),
  info = "group_keys() works"
)

expect_warning(
  mtcars %>% group_by(am) %>% group_split(cyl),
  info = "group_split() warns if data are already grouped"
)

#' Count observations by group
#'
#' `count()` lets you quickly count the unique values of one or more variables:
#' `df %>% count(a, b)` is roughly equivalent to
#' `df %>% group_by(a, b) %>% summarise(n = n())`.
#' `count()` is paired with `tally()`, a lower-level helper that is equivalent to `df %>% summarise(n = n())`. Supply
#' `wt` to perform weighted counts, switching the summary from from `n = n()` to `n = sum(wt)`.
#' `add_count()` and `add_tally()` are equivalent to `count()` and `tally()` but use `mutate()` instead of `summarise()`
#' so that they add a new column with group-wise counts.
#'
#' @param x A `data.frame`.
#' @param ... Variables to group by.
#' @param wt If omitted, will count the number of rows. If specified, will perform a "weighted" count by summing the
#' (non-missing) values of variable `wt`. If omitted, and column `n` exists, it will automatically be used as a
#' weighting variable, although you will have to specify `name` to provide a new name for the output.
#' @param sort `logical(1)`. If `TRUE`, will show the largest groups at the top.
#' @param name `character(1)`. The name of the new column in the output. If omitted, it will default to `n`. If there's
#' already a column called `n`, it will error, and require you to specify the name.
#'
#' @examples
#' # count() is a convenient way to get a sense of the distribution of
#' # values in a dataset
#' mtcars %>% count(cyl)
#' mtcars %>% count(cyl, sort = TRUE)
#' mtcars %>% count(cyl, am, sort = TRUE)
#' # Note that if the data are already grouped, count() adds an additional grouping variable
#' # which is removed afterwards
#' mtcars %>% group_by(gear) %>% count(cyl)
#'
#' # tally() is a lower-level function that assumes you've done the grouping
#' mtcars %>% tally()
#' mtcars %>% group_by(cyl) %>% tally()
#'
#' # both count() and tally() have add_ variants that work like mutate() instead of summarise
#' mtcars %>% add_count(cyl, wt = am)
#' mtcars %>% add_tally(wt = am)
#'
#' @return A `data.frame`. `count()` and `add_count()` have the same groups as the input.
#'
#' @importFrom stats setNames
#'
#' @export
count <- function(x, ..., wt = NULL, sort = FALSE, name = NULL) {
  groups <- group_vars(x)
  if (!missing(...)) x <- group_by(x, ..., .add = TRUE)
  wt <- deparse_var(wt)
  res <- do.call(tally, list(x, wt, sort, name))
  if (length(groups) > 0L) res <- do.call(group_by, list(res, as.name(groups)))
  res
}

#' @rdname count
#' @export
tally <- function(x, wt = NULL, sort = FALSE, name = NULL) {
  name <- check_name(x, name)
  wt <- deparse_var(wt)
  res <- do.call(summarise, setNames(list(x, tally_n(x, wt)), c(".data", name)))
  res <- ungroup(res)
  if (isTRUE(sort)) res <- do.call(arrange, list(res, call("desc", as.name(name))))
  rownames(res) <- NULL
  res
}

#' @export
#' @rdname count
add_count <- function(x, ..., wt = NULL, sort = FALSE, name = NULL) {
  name <- check_name(x, name)
  row_names <- rownames(x)
  wt <- deparse_var(wt)
  if (!missing(...)) x <- group_by(x, ..., .add = TRUE)
  res <- do.call(add_tally, list(x, wt, sort, name))
  res[row_names, ]
}

#' @export
#' @rdname count
add_tally <- function(x, wt = NULL, sort = FALSE, name = NULL) {
  wt <- deparse_var(wt)
  n <- tally_n(x, wt)
  name <- check_name(x, name)
  res <- do.call(mutate, setNames(list(x, n), c(".data", name)))

  if (isTRUE(sort)) {
    do.call(arrange, list(res, call("desc", as.name(name))))
  } else {
    res
  }
}

tally_n <- function(x, wt) {
  if (is.null(wt) && "n" %in% colnames(x)) {
    message("Using `n` as weighting variable")
    wt <- "n"
  }
  context$setup(.data = x)
  on.exit(context$clean(), add = TRUE)
  if (is.null(wt)) {
    call("n")
  } else {
    call("sum", as.name(wt), na.rm = TRUE)
  }
}

check_name <- function(df, name) {
  if (is.null(name)) {
    if ("n" %in% colnames(df)) {
      stop(
        "Column 'n' is already present in output\n",
        "* Use `name = \"new_name\"` to pick a new name"
      )
    }
    return("n")
  }

  if (!is.character(name) || length(name) != 1) {
    stop("`name` must be a single string")
  }

  name
}

expect_equal(
  mtcars %>% count(cyl),
  data.frame(cyl = c(4, 6, 8), n = c(11L, 7L, 14L)),
  info = "count() works on a single grouping variable"
)

res <- mtcars %>% group_by(gear) %>% count(cyl)
expect_equal(
  attr(res, "groups"),
  structure(
    list(gear = c(3, 4, 5), .rows = list(1:3, 4:5, 6:8)),
    class = "data.frame", row.names = c(NA, -3L), .drop = TRUE
  ),
  info = "count() works on a grouped data.frame"
)
expect_equal(
  res %>% ungroup(),
  structure(
    list(gear = c(3, 3, 3, 4, 4, 5, 5, 5), cyl = c(4, 6, 8, 4, 6, 4, 6, 8), n = c(1L, 2L, 12L, 8L, 4L, 2L, 1L, 2L)),
    class = "data.frame", row.names = c(NA, -8L)
  ),
  info = "count() works on a grouped data.frame"
)

expect_equal(
  mtcars %>% count(cyl, wt = am),
  data.frame(cyl = c(4, 6, 8), n = c(8, 3, 2)),
  info = "count() can perform weighted counts"
)

expect_equal(
  mtcars %>% count(cyl, sort = TRUE, name = "freq"),
  data.frame(cyl = c(8, 4, 6), freq = c(14, 11, 7)),
  info = "count() is able to sort and rename the frequency column"
)

expect_equal(
  mtcars %>% tally(),
  data.frame(n = 32),
  info = "tally() works on non-grouped data.frames"
)

expect_equal(
  mtcars %>% group_by(cyl) %>% tally(),
  data.frame(cyl = c(4, 6, 8), n = c(11L, 7L, 14L)),
  info = "tally() works on grouped data.frames"
)

expect_equal(
  mtcars %>% tally(wt = am),
  data.frame(n = 13),
  info = "tally() can perform weighted counts"
)

expect_equal(
  mtcars %>% group_by(cyl) %>% tally(wt = am),
  data.frame(cyl = c(4, 6, 8), n = c(8, 3, 2)),
  info = "tally() can perform weighted counts on grouped data.frames"
)

expect_equal(
  mtcars %>% group_by(cyl) %>% tally(sort = TRUE, name = "freq"),
  data.frame(cyl = c(8, 4, 6), freq = c(14L, 11L, 7L)),
  info = "tally() can sort and rename the frequency variable"
)

exp <- mtcars %>% add_count(cyl, wt = am, sort = TRUE, name = "freq")
expect_equal(
  exp %>% ungroup(),
  {
    res <- mtcars
    res[, "freq"] <- c(3, 3, 8, 3, 2, 3, 2, 8, 8, 3, 3, 2, 2, 2, 2, 2, 2, 8, 8, 8, 8, 2, 2, 2, 2, 8, 8, 8, 2, 3, 2, 8)
    res
  },
  info = "add_count() returns the expected data.frame"
)
expect_equal(
  group_vars(exp),
  "cyl",
  info = "add_count() adds the expected groups"
)

expect_equal(
  mtcars %>% group_by(cyl) %>% add_tally(wt = am, sort = TRUE, name = "freq") %>% ungroup(),
  {
    res <- mtcars
    res <- res[order(mtcars$cyl), ]
    res[, "freq"] <- c(8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2)
    res
  },
  info = "tally() returns the expected data.frame"
)

expect_equal(
  mtcars %>% group_by(cyl) %>% add_tally() %>% ungroup(),
  {
    res <- mtcars
    res[, "n"] <- c(
      7L, 7L, 11L, 7L, 14L, 7L, 14L, 11L, 11L, 7L, 7L, 14L, 14L, 14L, 14L, 14L, 14L, 11L, 11L, 11L, 11L, 14L, 14L, 14L,
      14L, 11L, 11L, 11L, 14L, 7L, 14L, 11L
    )
    res
  }
)

expect_error(
  mtcars %>% count(cyl, name = c("name_one", "name_two"))
)

expect_error(
  {
    res <- mtcars
    colnames(res)[1] <- "n"
    res %>% tally()
  }
)

expect_message(
  {
    res <- mtcars
    colnames(res)[1] <- "n"
    res %>% add_tally(name = "freq")
  }
)

#' Descending order
#'
#' Transform a vector into a format that will be sorted in descending order. This is useful within [arrange()].
#'
#' @param x A vector to transform.
#'
#' @examples
#' desc(1:10)
#' desc(factor(letters))
#'
#' first_day <- seq(as.Date("1910/1/1"), as.Date("1920/1/1"), "years")
#' desc(first_day)
#'
#' mtcars %>% arrange(desc(mpg))
#'
#' @return
#' A vector of the same length as `x`.
#'
#' @export
desc <- function(x) -xtfrm(x)

expect_equal(
  iris %>% head,
  head(iris),
  info = "Function calls without brackets works."
)

expect_equal(
  iris %>% head(10),
  head(iris, 10),
  info = "Use with lhs as first argument."
)

expect_equal(
  "Ceci n'est pas une pipe" %>% gsub("une", "un", .),
  "Ceci n'est pas un pipe",
  info = "Using the dot place-holder."
)

expect_equal(
  1:10 %>% paste0(LETTERS[.]),
  paste0(1:10, LETTERS[1:10]),
  info = "When dot is nested, lhs is still placed first."
)

expect_equal(
  iris %>%
    {
      rbind(head(., n = 1), tail(., n = 1))
    },
  rbind(head(iris, n = 1), tail(iris, n = 1)),
  info = "Lambda expressions"
)

expect_equal(
  iris %>%
    {
      my_data <- .
      size <- 1
      rbind(head(my_data, size), tail(my_data, size))
    },
  rbind(head(iris, n = 1), tail(iris, n = 1)),
  info = "Renaming in lambda expressions"
)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarise.R
\name{summarise}
\alias{summarise}
\alias{summarize}
\title{Reduce multiple values down to a single value}
\usage{
summarise(.data, ..., .groups = NULL)

summarize(.data, ..., .groups = NULL)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{Name-value pairs of summary functions. The name will be the name of the variable in the result.}

\item{.groups}{\code{character(1)}. Grouping structure of the result.
\itemize{
\item \code{"drop_last"}: drops the last level of grouping.
\item \code{"drop"}: all levels of grouping are dropped.
\item \code{"keep"}: keeps the same grouping structure as \code{.data}.
}

When \code{.groups} is not specified, it is chosen based on the number of rows of the results:
\itemize{
\item If all the results have 1 row, you get \code{"drop_last"}.
\item If the number of rows varies, you get \code{"keep"}.
}

In addition, a message informs you of that choice, unless the result is ungrouped, the option
\code{"poorman.summarise.inform"} is set to \code{FALSE}.

The value can be:
\itemize{
\item A vector of length \code{1}, e.g. \code{min(x)}, \code{n()}, or \code{sum(is.na(y))}.
\item A vector of length \code{n}, e.g. \code{quantile()}.
}}
}
\description{
Create one or more scalar variables summarising the variables of an existing \code{data.frame}. Grouped \code{data.frame}s will
result in one row in the output for each group.
}
\details{
\code{summarise()} and \code{summarize()} are synonyms.
}
\examples{
# A summary applied to ungrouped tbl returns a single row
mtcars \%>\%
  summarise(mean = mean(disp), n = n())

# Usually, you'll want to group first
mtcars \%>\%
  group_by(cyl) \%>\%
  summarise(mean = mean(disp), n = n())

# You can summarise to more than one value:
mtcars \%>\%
   group_by(cyl) \%>\%
   summarise(qs = quantile(disp, c(0.25, 0.75)), prob = c(0.25, 0.75))

# You use a data frame to create multiple columns so you can wrap
# this up into a function:
my_quantile <- function(x, probs) {
  data.frame(x = quantile(x, probs), probs = probs)
}
mtcars \%>\%
  group_by(cyl) \%>\%
  summarise(my_quantile(disp, c(0.25, 0.75)))

# Each summary call removes one grouping level (since that group
# is now just a single row)
mtcars \%>\%
  group_by(cyl, vs) \%>\%
  summarise(cyl_n = n()) \%>\%
  group_vars()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/na_if.R
\name{na_if}
\alias{na_if}
\title{Convert values to NA}
\usage{
na_if(x, y)
}
\arguments{
\item{x}{The vector to modify.}

\item{y}{The value to replace with \code{NA}.}
}
\value{
A modified version of \code{x} that replaces any values that are equal to \code{y} with \code{NA}.
}
\description{
This is a translation of the SQL command \code{NULLIF}. It is useful if you want to convert an annoying value to \code{NA}.
}
\examples{
na_if(1:5, 5:1)

x <- c(1, -1, 0, 10)
100 / x
100 / na_if(x, 0)

y <- c("abc", "def", "", "ghi")
na_if(y, "")

# na_if() is particularly useful inside mutate(),
# and is meant for use with vectors rather than entire data.frames
mtcars \%>\%
  mutate(cyl = na_if(cyl, 6))

}
\seealso{
\code{\link[=coalesce]{coalesce()}} to replace missing values within subsequent \code{vector}(s) of value(s). \code{\link[=replace_na]{replace_na()}} to replace \code{NA} with
a value.

\code{\link[=replace_na]{replace_na()}} to replace \code{NA} with a value.

\code{\link[=recode]{recode()}} to more generally replace values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_by.R, R/ungroup.R
\name{group_by}
\alias{group_by}
\alias{ungroup}
\title{Group by one or more variables}
\usage{
group_by(.data, ..., .add = FALSE, .drop = group_by_drop_default(.data))

ungroup(x, ...)
}
\arguments{
\item{.data}{\code{data.frame}. The data to group.}

\item{...}{One or more unquoted column names to group/ungroup the data by.}

\item{.add}{\code{logical(1)}. When \code{FALSE} (the default) \code{group_by()} will override existing groups. To add to existing
groups, use \code{.add = TRUE}.}

\item{.drop}{\code{logical(1)}. Drop groups formed by factor levels that don't appear in the data? The default is \code{TRUE}
except when \code{.data} has been previously grouped with \code{.drop = FALSE}. See \code{\link[=group_by_drop_default]{group_by_drop_default()}} for details.}

\item{x}{A \code{data.frame}.}
}
\value{
When using \code{\link[=group_by]{group_by()}}, a \code{data.frame}, grouped by the grouping variables.

When using \code{\link[=ungroup]{ungroup()}}, a \code{data.frame}.
}
\description{
Determine the groups within a \code{data.frame} to perform operations on. \code{\link[=ungroup]{ungroup()}} removes the grouping levels.
}
\examples{
group_by(mtcars, am, cyl)
ungroup(mutate(group_by(mtcars, am, cyl), sumMpg = sum(mpg)))
mtcars \%>\%
  group_by(am, cyl) \%>\%
  mutate(sumMpg = sum(mpg)) \%>\%
  ungroup()
mtcars \%>\%
  group_by(carb) \%>\%
  filter(any(gear == 5))

# You can group by expressions: this is just short-hand for
# a mutate() followed by a group_by()
mtcars \%>\% group_by(vsam = vs + am)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pivot_wider.R
\name{pivot_wider}
\alias{pivot_wider}
\title{Pivot data from long to wide}
\usage{
pivot_wider(
  data,
  id_cols = NULL,
  values_from = "Value",
  names_from = "Name",
  names_sep = "_",
  names_prefix = "",
  names_glue = NULL,
  values_fill = NULL,
  ...
)
}
\arguments{
\item{data}{\code{data.frame}. The data to pivot.}

\item{id_cols}{\code{character(1)}. The name of the column that identifies the rows. If \code{NULL}, it will use all the unique
rows.}

\item{values_from}{\code{character(n)}. The name of the column that contains the values to be used as future variable
values.}

\item{names_from}{\code{character(n)}. The name of the column(s) that contains the levels to be used as future column
names.}

\item{names_sep}{\code{character(1)}. If \code{names_from} or \code{values_from} contains multiple variables, this will be used to
join their values together into a single string to use as a column name.}

\item{names_prefix}{\code{character(1)}. String added to the start of every variable name. This is particularly useful if
\code{names_from} is a numeric vector and you want to create syntactic variable names.}

\item{names_glue}{\code{character(1)}. Instead of \code{names_sep} and \code{names_prefix}, you can supply a
\href{https://glue.tidyverse.org/index.html}{glue specification} that uses the \code{names_from} columns to create custom
column names. Note that the only delimiters supported by \code{names_glue} are curly brackets, \verb{\{} and \verb{\}}.}

\item{values_fill}{\code{numeric(n)}. Optionally, a (scalar) value that will be used to replace missing values in the new
columns created.}

\item{...}{Not used for now.}
}
\value{
If a tibble was provided as input, \code{pivot_wider()} also returns a
tibble. Otherwise, it returns a data frame.
}
\description{
\code{pivot_wider()} "widens" data, increasing the number of columns and decreasing the number of rows. The inverse
transformation is \code{\link[=pivot_longer]{pivot_longer()}}.
}
\examples{
data_long <- read.table(header = TRUE, text = "
 subject sex condition measurement
       1   M   control         7.9
       1   M     cond1        12.3
       1   M     cond2        10.7
       2   F   control         6.3
       2   F     cond1        10.6
       2   F     cond2        11.1
       3   F   control         9.5
       3   F     cond1        13.1
       3   F     cond2        13.8
       4   M   control        11.5
       4   M     cond1        13.4
       4   M     cond2        12.9")


pivot_wider(
  data_long,
  id_cols = "subject",
  names_from = "condition",
  values_from = "measurement"
)

pivot_wider(
  data_long,
  id_cols = "subject",
  names_from = "condition",
  values_from = "measurement",
  names_prefix = "Var.",
  names_sep = "."
)

production <- expand.grid(
  product = c("A", "B"),
  country = c("AI", "EI"),
  year = 2000:2014
) \%>\%
  filter((product == "A" & country == "AI") | product == "B") \%>\%
  mutate(production = rnorm(nrow(.)))

pivot_wider(
  production,
  names_from = c("product", "country"),
  values_from = "production",
  names_glue = "prod_{product}_{country}"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipe.R
\name{pipe}
\alias{pipe}
\alias{\%>\%}
\title{Forward-pipe operator}
\usage{
lhs \%>\% rhs
}
\arguments{
\item{lhs}{The result you are piping.}

\item{rhs}{Where you are piping the result to.}
}
\description{
Pipe an object forward into a function or call expression.
}
\examples{
# Basic use:
iris \%>\% head

# Use with lhs as first argument
iris \%>\% head(10)

# Using the dot place-holder
"Ceci n'est pas une pipe" \%>\% gsub("une", "un", .)

# When dot is nested, lhs is still placed first:
sample(1:10) \%>\% paste0(LETTERS[.])

# This can be avoided:
rnorm(100) \%>\% {c(min(.), mean(.), max(.))} \%>\% floor

# Lambda expressions:
iris \%>\%
  {
    size <- sample(1:10, size = 1)
    rbind(head(., size), tail(., size))
  }

# renaming in lambdas:
iris \%>\%
  {
    my_data <- .
    size <- sample(1:10, size = 1)
    rbind(head(my_data, size), tail(my_data, size))
  }

}
\author{
Nathan Eastwood and Antoine Fabri \email{antoine.fabri@gmail.com}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_helpers.R
\name{select_helpers}
\alias{select_helpers}
\alias{starts_with}
\alias{ends_with}
\alias{contains}
\alias{matches}
\alias{num_range}
\alias{all_of}
\alias{any_of}
\alias{everything}
\alias{last_col}
\title{Select Helpers}
\usage{
starts_with(match, ignore.case = TRUE, vars = peek_vars())

ends_with(match, ignore.case = TRUE, vars = peek_vars())

contains(match, ignore.case = TRUE, vars = peek_vars())

matches(match, ignore.case = TRUE, perl = FALSE, vars = peek_vars())

num_range(prefix, range, width = NULL, vars = peek_vars())

all_of(x, vars = peek_vars())

any_of(x, vars = peek_vars())

everything(vars = peek_vars())

last_col(offset = 0L, vars = peek_vars())
}
\arguments{
\item{match}{\code{character(n)}. If length > 1, the union of the matches is taken.}

\item{ignore.case}{\code{logical(1)}. If \code{TRUE}, the default, ignores case when matching names.}

\item{vars}{\code{character(n)}. A character vector of variable names. When called from inside selecting functions such as
\code{select()}, these are automatically set to the names of the table.}

\item{perl}{\code{logical(1)}. Should Perl-compatible regexps be used?}

\item{prefix}{A prefix which starts the numeric range.}

\item{range}{\code{integer(n)}. A sequence of integers, e.g. \code{1:5}.}

\item{width}{\code{numeric(1)}. Optionally, the "width" of the numeric range. For example, a range of 2 gives "01", a
range of three "001", etc.}

\item{x}{\code{character(n)}. A vector of column names.}

\item{offset}{\code{integer(1)}. Select the \code{n}th variable from the end of the \code{data.frame}.}
}
\value{
An integer vector giving the position of the matched variables.
}
\description{
These functions allow you to select variables based on their names.
\itemize{
\item \code{starts_with()}: Starts with a prefix.
\item \code{ends_with()}: Ends with a prefix.
\item \code{contains()}: Contains a literal string.
\item \code{matches()}: Matches a regular expression.
\item \code{all_of()}: Matches variable names in a character vector. All names must be present, otherwise an error is thrown.
\item \code{any_of()}: The same as \code{all_of()} except it doesn't throw an error.
\item \code{everything()}: Matches all variables.
\item \code{last_col()}: Select the last variable, possibly with an offset.
}
}
\examples{
mtcars \%>\% select(starts_with("c"))
mtcars \%>\% select(starts_with(c("c", "h")))
mtcars \%>\% select(ends_with("b"))
mtcars \%>\% relocate(contains("a"), .before = mpg)
iris \%>\% select(matches(".t."))
mtcars \%>\% select(last_col())

# `all_of()` selects the variables in a character vector:
iris \%>\% select(all_of(c("Petal.Length", "Petal.Width")))
# `all_of()` is strict and will throw an error if the column name isn't found
try({iris \%>\% select(all_of(c("Species", "Genres")))})
# However `any_of()` allows missing variables
iris \%>\% select(any_of(c("Species", "Genres")))

}
\seealso{
\code{\link[=select]{select()}}, \code{\link[=relocate]{relocate()}}, \code{\link[=where]{where()}}, \code{\link[=group_cols]{group_cols()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutate.R, R/transmute.R
\name{mutate}
\alias{mutate}
\alias{mutate.data.frame}
\alias{transmute}
\title{Create or transform variables}
\usage{
mutate(.data, ...)

\method{mutate}{data.frame}(
  .data,
  ...,
  .keep = c("all", "used", "unused", "none"),
  .before = NULL,
  .after = NULL
)

transmute(.data, ...)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{Name-value pairs of expressions, each with length \code{1L}. The name of each argument will be the name of a
new column and the value will be its corresponding value. Use a \code{NULL} value in \code{mutate} to drop a variable. New
variables overwrite existing variables of the same name.}

\item{.keep}{This argument allows you to control which columns from \code{.data} are retained in the output:
\itemize{
\item \code{"all"}, the default, retains all variables.
\item \code{"used"} keeps any variables used to make new variables; it's useful for checking your work as it displays inputs
and outputs side-by-side.
\item \code{"unused"} keeps only existing variables \strong{not} used to make new variables.
\item \code{"none"}, only keeps grouping keys (like \code{\link[=transmute]{transmute()}}).
}

Grouping variables are always kept, unconditional to \code{.keep}.}

\item{.before, .after}{<\code{\link[=select_helpers]{poor-select}}> Optionally, control where new columns should appear (the
default is to add to the right hand side). See \code{\link[=relocate]{relocate()}} for more details.}
}
\description{
\code{mutate()} adds new variables and preserves existing ones; \code{transmute()} adds new variables and drops existing ones.
Both functions preserve the number of rows of the input. New variables overwrite existing variables of the same name.
Variables can be removed by setting their value to \code{NULL}.
}
\section{Useful mutate functions}{

\itemize{
\item \code{\link{+}}, \code{\link{-}}, \code{\link[=log]{log()}}, etc., for their usual mathematical meanings
\item \code{\link[=lead]{lead()}}, \code{\link[=lag]{lag()}}
\item \code{\link[=dense_rank]{dense_rank()}}, \code{\link[=min_rank]{min_rank()}}, \code{\link[=percent_rank]{percent_rank()}}, \code{\link[=row_number]{row_number()}}, \code{\link[=cume_dist]{cume_dist()}}, \code{\link[=ntile]{ntile()}}
\item \code{\link[=cumsum]{cumsum()}}, \code{\link[=cummin]{cummin()}}, \code{\link[=cummax]{cummax()}}
\item \code{\link[=na_if]{na_if()}}, \code{\link[=coalesce]{coalesce()}}
\item \code{\link[=if_else]{if_else()}}, \code{\link[=recode]{recode()}}, \code{\link[=case_when]{case_when()}}
}
}

\examples{
mutate(mtcars, mpg2 = mpg * 2)
mtcars \%>\% mutate(mpg2 = mpg * 2)
mtcars \%>\% mutate(mpg2 = mpg * 2, cyl2 = cyl * 2)

# Newly created variables are available immediately
mtcars \%>\% mutate(mpg2 = mpg * 2, mpg4 = mpg2 * 2)

# You can also use mutate() to remove variables and modify existing variables
mtcars \%>\% mutate(
  mpg = NULL,
  disp = disp * 0.0163871 # convert to litres
)

# By default, new columns are placed on the far right.
# You can override this with `.before` or `.after`.
df <- data.frame(x = 1, y = 2)
df \%>\% mutate(z = x + y)
df \%>\% mutate(z = x + y, .before = 1)
df \%>\% mutate(z = x + y, .after = x)

# By default, mutate() keeps all columns from the input data.
# You can override with `.keep`
df <- data.frame(
  x = 1, y = 2, a = "a", b = "b",
  stringsAsFactors = FALSE
)
df \%>\% mutate(z = x + y, .keep = "all") # the default
df \%>\% mutate(z = x + y, .keep = "used")
df \%>\% mutate(z = x + y, .keep = "unused")
df \%>\% mutate(z = x + y, .keep = "none") # same as transmute()

# mutate() vs transmute --------------------------
# mutate() keeps all existing variables
mtcars \%>\%
  mutate(displ_l = disp / 61.0237)

# transmute keeps only the variables you create
mtcars \%>\%
  transmute(displ_l = disp / 61.0237)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind.R
\name{bind}
\alias{bind}
\alias{bind_cols}
\alias{bind_rows}
\title{Efficiently bind multiple \code{data.frame}s by row and column}
\usage{
bind_cols(...)

bind_rows(..., .id = NULL)
}
\arguments{
\item{...}{\code{data.frame}s to combine.

Each argument can either be a \code{data.frame}, a \code{list} that could be a \code{data.frame}, or a \code{list} of \code{data.frame}s.

When row-binding, columns are matched by name, and any missing columns will be filled with \code{NA}.

When column-binding, rows are matched by position, so all \code{data.frame}s must have the same number of rows. To match
by value, not position, see \link{mutate_joins}.}

\item{.id}{\code{character(1)}. \code{data.frame} identifier.

When \code{.id} is supplied, a new column of identifiers is created to link each row to its original \code{data.frame}. The
labels are taken from the named arguments to \code{bind_rows()}. When a \code{list} of \code{data.frame}s is supplied, the labels
are taken from the names of the \code{list}. If no names are found a numeric sequence is used instead.}
}
\description{
Efficiently bind multiple \code{data.frame}s by row and column
}
\examples{
one <- mtcars[1:4, ]
two <- mtcars[9:12, ]

# You can supply data frames as arguments:
bind_rows(one, two)

# The contents of lists are spliced automatically:
bind_rows(list(one, two))
bind_rows(split(mtcars, mtcars$cyl))
bind_rows(list(one, two), list(two, one))

# In addition to data frames, you can supply vectors. In the rows
# direction, the vectors represent rows and should have inner
# names:
bind_rows(
  c(a = 1, b = 2),
  c(a = 3, b = 4)
)

# You can mix vectors and data frames:
bind_rows(
  c(a = 1, b = 2),
  data.frame(a = 3:4, b = 5:6),
  c(a = 7, b = 8)
)

# When you supply a column name with the `.id` argument, a new
# column is created to link each row to its original data frame
bind_rows(list(one, two), .id = "id")
bind_rows(list(a = one, b = two), .id = "id")
bind_rows("group 1" = one, "group 2" = two, .id = "groups")

\dontrun{
# Rows need to match when column-binding
bind_cols(data.frame(x = 1:3), data.frame(y = 1:2))

# even with 0 columns
bind_cols(data.frame(x = 1:3), data.frame())
}

bind_cols(one, two)
bind_cols(list(one, two))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/across.R
\name{across}
\alias{across}
\alias{if_any}
\alias{if_all}
\title{Apply a function (or functions) across multiple columns}
\usage{
across(.cols = everything(), .fns = NULL, ..., .names = NULL)

if_any(.cols, .fns = NULL, ..., .names = NULL)

if_all(.cols, .fns = NULL, ..., .names = NULL)
}
\arguments{
\item{.fns}{Functions to apply to each of the selected columns.
Possible values are:
\itemize{
\item \code{NULL}, to returns the columns untransformed.
\item A function, e.g. \code{mean}.
\item A lambda, e.g. \code{~ mean(.x, na.rm = TRUE)}
\item A list of functions/lambdas, e.g. \verb{list(mean = mean, n_miss = ~ sum(is.na(.x))}
}

Within these functions you can use \code{\link[=cur_column]{cur_column()}} and \code{\link[=cur_group]{cur_group()}} to access the current column and grouping keys
respectively.}

\item{...}{Additional arguments for the function calls in \code{.fns}.}

\item{.names}{A glue specification that describes how to name the output
columns. This can use \code{{.col}} to stand for the selected column name, and \code{{.fn}}
to stand for the name of the function being applied. The default (\code{NULL}) is
equivalent to "\code{{.col}}" for the single function case and "\verb{\{.col\}_\{.fn\}}" for
the case where a list is used for \code{.fns}.}

\item{cols, .cols}{<\code{\link[=select_helpers]{poor-select}}> Columns to transform. Because \code{across()} is used within functions
like \code{summarise()} and \code{mutate()}, you can't select or compute upon grouping variables.}
}
\value{
\code{across()} returns a \code{data.frame} with one column for each column in \code{.cols} and each function in \code{.fns}.

\code{if_any()} and \code{if_all()} return a logical vector.
}
\description{
\code{across()} makes it easy to apply the same transformation to multiple columns, allowing you to use \code{\link[=select]{select()}}
semantics inside in "data-masking" functions like \code{\link[=summarise]{summarise()}} and \code{\link[=mutate]{mutate()}}.

\code{if_any()} and \code{if_all()} are used to apply the same predicate function to a selection of columns and combine the
results into a single logical vector.

\code{across()} supersedes the family of \code{dplyr} "scoped variants" like \code{summarise_at()}, \code{summarise_if()}, and
\code{summarise_all()} and therefore these functions will not be implemented in \code{poorman}.
}
\examples{
# across() -----------------------------------------------------------------
iris \%>\%
  group_by(Species) \%>\%
  summarise(across(starts_with("Sepal"), mean))
iris \%>\%
  mutate(across(where(is.factor), as.character))

# Additional parameters can be passed to functions
iris \%>\%
  group_by(Species) \%>\%
  summarise(across(starts_with("Sepal"), mean, na.rm = TRUE))

# A named list of functions
iris \%>\%
  group_by(Species) \%>\%
  summarise(across(starts_with("Sepal"), list(mean = mean, sd = sd)))

# Use the .names argument to control the output names
iris \%>\%
  group_by(Species) \%>\%
  summarise(
    across(starts_with("Sepal"),
    mean,
    .names = c("mean_sepal_length", "mean_sepal_width"))
  )

# if_any() and if_all() ----------------------------------------------------
iris \%>\%
  filter(if_any(ends_with("Width"), ~ . > 4))
iris \%>\%
  filter(if_all(ends_with("Width"), ~ . > 2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pivot_longer.R
\name{pivot_longer}
\alias{pivot_longer}
\title{Pivot data from wide to long}
\usage{
pivot_longer(
  data,
  cols,
  names_to = "name",
  names_prefix = NULL,
  names_sep = NULL,
  names_pattern = NULL,
  values_to = "value",
  values_drop_na = FALSE,
  ...
)
}
\arguments{
\item{data}{\code{data.frame}. The data to pivot.}

\item{cols}{<\code{\link[=select_helpers]{poor-select}}>. Columns to pivot into longer format.}

\item{names_to}{\code{character(n)}. The name of the new column(s) that will contain the column names.}

\item{names_prefix}{\code{character(1)}. A regular expression used to remove matching text from the start of each variable
name.}

\item{names_sep, names_pattern}{\code{character(1)}. If \code{names_to} contains multiple values, this argument controls how the
column name is broken up. \code{names_pattern} takes a regular expression containing matching groups (\verb{()}).}

\item{values_to}{\code{character(n)}. The name of the new column(s) that will contain the values of the pivoted variables.}

\item{values_drop_na}{\code{logical(1)}. If \code{TRUE}, will drop rows that contain only \code{NA} in the \code{values_to} column. This
effectively converts explicit missing values to implicit missing values, and should generally be used only when
missing values in data were created by its structure.}

\item{...}{Additional arguments passed on to methods.}
}
\value{
A \code{data.frame}.
}
\description{
\code{pivot_longer()} "lengthens" data, increasing the number of rows and decreasing the number of columns. The inverse
transformation is \code{\link[=pivot_wider]{pivot_wider()}}.
}
\examples{
wide_data <- data.frame(replicate(5, rnorm(10)))
# Customizing the names
pivot_longer(
  data = wide_data,
  cols = c(1, 2),
  names_to = "Column",
  values_to = "Numbers"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_split.R
\name{group_split}
\alias{group_split}
\alias{group_keys}
\title{Split data.frame by groups}
\usage{
group_split(.data, ..., .keep = TRUE)

group_keys(.data)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{Grouping specification, forwarded to \code{\link[=group_by]{group_by()}}.}

\item{.keep}{\code{logical(1)}. Should the grouping columns be kept (default: TRUE)?}
}
\value{
\itemize{
\item \code{group_split()} returns a list of \code{data.frame}s. Each \code{data.frame} contains the rows of \code{.data} with the associated
group and all the columns, including the grouping variables.
\item \code{group_keys()} returns a \code{data.frame} with one row per group, and one column per grouping variable
}
}
\description{
\code{group_split()} works like \code{\link[base:split]{base::split()}} but
\itemize{
\item it uses the grouping structure from \code{\link[=group_by]{group_by()}} and is therefore subject to the data mask
\item it does not name the elements of the list based on the grouping as this typically loses information and is
confusing
}
}
\details{
\strong{Grouped \code{data.frame}s:}

The primary use case for \code{group_split()} is with already grouped \code{data.frame}s, typically a result of \code{\link[=group_by]{group_by()}}.
In this case, \code{group_split()} only uses the first argument, the grouped \code{data.frame}, and warns when \code{...} is used.

Because some of these groups may be empty, it is best paired with \code{group_keys()} which identifies the representatives
of each grouping variable for the group.

\strong{Ungrouped \code{data.frame}s:}

When used on ungrouped \code{data.frame}s, \code{group_split()} forwards the \code{...} to \code{group_by()} before the split, therefore
the \code{...} are subject to the data mask.
}
\examples{
# Grouped data.frames:
mtcars \%>\% group_by(cyl, am) \%>\% group_split()
mtcars \%>\% group_by(cyl, am) \%>\% group_split(.keep = FALSE)
mtcars \%>\% group_by(cyl, am) \%>\% group_keys()

# Ungrouped data.frames:
mtcars \%>\% group_split(am, cyl)

}
\seealso{
\code{\link[=group_by]{group_by()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slice.R
\name{slice}
\alias{slice}
\alias{slice_head}
\alias{slice_tail}
\alias{slice_min}
\alias{slice_max}
\alias{slice_sample}
\title{Subset rows by position}
\usage{
slice(.data, ...)

slice_head(.data, ..., n, prop)

slice_tail(.data, ..., n, prop)

slice_min(.data, order_by, ..., n, prop, with_ties = TRUE)

slice_max(.data, order_by, ..., n, prop, with_ties = TRUE)

slice_sample(.data, ..., n, prop, weight_by = NULL, replace = FALSE)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{For \code{slice()}: integer row values.

Provide either positive values to keep, or negative values to drop. The values provided must be either all positive
or negative. Indices beyond the number of rows in the input are silently ignored.}

\item{n, prop}{Provide either \code{n}, the number of rows, or \code{prop}, the proportion of rows to select. If neither are
supplied, \code{n = 1} will be used.

If \code{n} is greater than the number of rows in the group (or \code{prop > 1}), the result will be silently truncated to the
group size. If the \code{prop}ortion of a group size is not an integer, it is rounded down.}

\item{order_by}{The variable to order by.}

\item{with_ties}{\code{logical(1)}. Should ties be kept together? The default, \code{TRUE}, may return more rows than you
request. Use \code{FALSE} to ignore ties, and return the first \code{n} rows.}

\item{weight_by}{Sampling weights. This must evaluate to a vector of non-negative numbers the same length as the
input. Weights are automatically standardised to sum to 1.}

\item{replace}{\code{logical(1)}. Should sampling be performed with (\code{TRUE}) or without (\code{FALSE}, the default)
replacement.}
}
\value{
An object of the same type as \code{.data}. The output has the following properties:
\itemize{
\item Each row may appear 0, 1, or many times in the output.
\item Columns are not modified.
\item Groups are not modified.
\item Data frame attributes are preserved.
}
}
\description{
Subset rows by their original position in the \code{data.frame}. Grouped \code{data.frame}s use the position within each group.
}
\examples{
slice(mtcars, c(1, 2, 3))
mtcars \%>\% slice(1:3)

# Similar to head(mtcars, 1)
mtcars \%>\% slice(1L)

# Similar to tail(mtcars, 1):
mtcars \%>\% slice(n())
mtcars \%>\% slice(5:n())
# Rows can be dropped with negative indices:
slice(mtcars, -(1:4))

# First and last rows based on existing order
mtcars \%>\% slice_head(n = 5)
mtcars \%>\% slice_tail(n = 5)

# Grouped operations:
mtcars \%>\% group_by(am, cyl, gear) \%>\% slice_head(n = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/desc.R
\name{desc}
\alias{desc}
\title{Descending order}
\usage{
desc(x)
}
\arguments{
\item{x}{A vector to transform.}
}
\value{
A vector of the same length as \code{x}.
}
\description{
Transform a vector into a format that will be sorted in descending order. This is useful within \code{\link[=arrange]{arrange()}}.
}
\examples{
desc(1:10)
desc(factor(letters))

first_day <- seq(as.Date("1910/1/1"), as.Date("1920/1/1"), "years")
desc(first_day)

mtcars \%>\% arrange(desc(mpg))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cumulative.R
\name{cummean}
\alias{cummean}
\alias{cumany}
\alias{cumall}
\title{Cumulative versions of any, all, and mean}
\usage{
cummean(x)

cumany(x)

cumall(x)
}
\arguments{
\item{x}{For \code{cumall()} and \code{cumany()}, a logical vector; for \code{cummean()} an integer or numeric vector.}
}
\value{
A vector the same length as \code{x}.
}
\description{
{poorman} provides \code{cumall()}, \code{cumany()}, and \code{cummean()} to complete R's set of cumulative functions.
}
\section{Cumulative logical functions}{


These are particularly useful in conjunction with \code{filter()}:
\itemize{
\item \code{cumall(x)}: all cases until the first \code{FALSE}.
\item \code{cumall(!x)}: all cases until the first \code{TRUE}.
\item \code{cumany(x)}: all cases after the first \code{TRUE}.
\item \code{cumany(!x)}: all cases after the first \code{FALSE}.
}
}

\examples{
# `cummean()` returns a numeric/integer vector of the same length
# as the input vector.
x <- c(1, 3, 5, 2, 2)
cummean(x)
cumsum(x) / seq_along(x)

# `cumall()` and `cumany()` return logicals
cumall(x < 5)
cumany(x == 3)

# `cumall()` vs. `cumany()`
df <- data.frame(
  date = as.Date("2020-01-01") + 0:6,
  balance = c(100, 50, 25, -25, -50, 30, 120)
)
# all rows after first overdraft
df \%>\% filter(cumany(balance < 0))
# all rows until first overdraft
df \%>\% filter(cumall(!(balance < 0)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replace_na.R
\name{replace_na}
\alias{replace_na}
\title{Replace missing values}
\usage{
replace_na(data, replace, ...)
}
\arguments{
\item{data}{A \code{data.frame} or \code{vector}.}

\item{replace}{If \code{data} is a \code{data.frame}, a named \code{list} giving the value to replace \code{NA} with for each column. If
\code{data} is a \code{vector}, a single value used for replacement.}

\item{...}{Additional arguments passed onto methods; not currently used.}
}
\value{
If \code{data} is a \code{data.frame}, \code{replace_na()} returns a \code{data.frame}. If \code{data} is a \code{vector}, \code{replace_na()} returns a
\code{vector} of class determined by the union of \code{data} and \code{replace}.
}
\description{
Replace missing values in a \code{data.frame} or \code{vector}.
}
\examples{
df <- data.frame(x = c(1, 2, NA), y = c("a", NA, "b"), stringsAsFactors = FALSE)
df \%>\% replace_na(list(x = 0, y = "unknown"))
df \%>\% mutate(x = replace_na(x, 0))

df$x \%>\% replace_na(0)
df$y \%>\% replace_na("unknown")

}
\seealso{
\code{\link[=na_if]{na_if()}} to replace specified values with a \code{NA}.

\code{\link[=coalesce]{coalesce()}} to replace missing values within subsequent \code{vector}(s) of value(s).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/context.R
\name{context}
\alias{context}
\alias{n}
\alias{cur_data}
\alias{cur_data_all}
\alias{cur_group}
\alias{cur_group_id}
\alias{cur_group_rows}
\alias{cur_column}
\title{Context dependent expressions}
\usage{
n()

cur_data()

cur_data_all()

cur_group()

cur_group_id()

cur_group_rows()

cur_column()
}
\description{
These functions return information about the "current" group or "current" variable, so only work inside specific
contexts like \code{\link[=summarise]{summarise()}} and \code{\link[=mutate]{mutate()}}.

\itemize{
\item \code{n()} gives the number of observations in the current group.
}

\itemize{
\item \code{cur_data()} gives the current data for the current group (excluding grouping variables).
}

\itemize{
\item \code{cur_data_all()} gives the current data for the current group (including grouping variables).
}

\itemize{
\item \code{cur_group()} gives the group keys, a single row \code{data.frame} containing a column for each grouping variable and
its value.
}

\itemize{
\item \code{cur_group_id()} gives a unique numeric identifier for the current group.
}

\itemize{
\item \code{cur_group_rows()} gives the rows the groups appear in the data.
}

\itemize{
\item \code{cur_column()} gives the name of the current column (in \code{\link[=across]{across()}} only).
}
}
\section{\code{data.table}}{

If you're familiar with \code{data.table}:
\itemize{
\item \code{cur_data()} <-> \code{.SD}
\item \code{cur_group_id()} <-> \code{.GRP}
\item \code{cur_group()} <-> \code{.BY}
\item \code{cur_group_rows()} <-> \code{.I}
}
}

\examples{
df <- data.frame(
  g = sample(rep(letters[1:3], 1:3)),
  x = runif(6),
  y = runif(6),
  stringsAsFactors = FALSE
)
gf <- df \%>\% group_by(g)

gf \%>\% summarise(n = n())

gf \%>\% mutate(id = cur_group_id())
gf \%>\% summarise(row = cur_group_rows())
gf \%>\% summarise(data = list(cur_group()))
gf \%>\% summarise(data = list(cur_data()))
gf \%>\% summarise(data = list(cur_data_all()))

gf \%>\% mutate(across(everything(), ~ paste(cur_column(), round(.x, 2))))

}
\seealso{
See \code{\link[=group_data]{group_data()}} for equivalent functions that return values for all groups.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nest_by.R
\name{nest_by}
\alias{nest_by}
\title{Nest By}
\usage{
nest_by(.data, ..., .key = "data", .keep = FALSE)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{Grouping specification, forwarded to \code{\link[=group_by]{group_by()}}.}

\item{.key}{\code{character(1)}. The name of the column in which to nest the data (default: "data").}

\item{.keep}{\code{logical(1)}. Should the grouping columns be kept (default: TRUE)?}
}
\description{
\code{nest_by()} is similar to \code{\link[=group_by]{group_by()}} however instead of storing the group structure in the metadata, it is made
explicit in the data. Each group key is given a single row within the \code{data.frame} and the group's data is stored
within a list-column of the \code{data.frame}.
}
\details{
Currently there is no pretty-printing provided for the results of \code{nest_by()} and they are not useable with other
functions such as \code{\link[=mutate]{mutate()}}.
}
\examples{
mtcars \%>\% nest_by(am, cyl)
# Or equivalently
mtcars \%>\% group_by(am, cyl) \%>\% nest_by()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lst.R
\name{lst}
\alias{lst}
\title{Build a list}
\usage{
lst(...)
}
\arguments{
\item{...}{Named or unnamed elements of a list. If the element is unnamed, its
expression will be used as its name.}
}
\value{
A named list.
}
\description{
\code{lst()} constructs a list, similar to \code{\link[base:list]{base::list()}}, but where components
are built sequentially. When defining a component, you can refer to components
created earlier in the call. \code{lst()} also generates missing names
automatically.
}
\examples{
# the value of n can be used immediately in the definition of x
lst(n = 5, x = runif(n))

# missing names are constructed from user's input
lst(1:3, z = letters[4:6], runif(3))

a <- 1:3
b <- letters[4:6]
lst(a, b)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count-tally.R
\name{count}
\alias{count}
\alias{tally}
\alias{add_count}
\alias{add_tally}
\title{Count observations by group}
\usage{
count(x, ..., wt = NULL, sort = FALSE, name = NULL)

tally(x, wt = NULL, sort = FALSE, name = NULL)

add_count(x, ..., wt = NULL, sort = FALSE, name = NULL)

add_tally(x, wt = NULL, sort = FALSE, name = NULL)
}
\arguments{
\item{x}{A \code{data.frame}.}

\item{...}{Variables to group by.}

\item{wt}{If omitted, will count the number of rows. If specified, will perform a "weighted" count by summing the
(non-missing) values of variable \code{wt}. If omitted, and column \code{n} exists, it will automatically be used as a
weighting variable, although you will have to specify \code{name} to provide a new name for the output.}

\item{sort}{\code{logical(1)}. If \code{TRUE}, will show the largest groups at the top.}

\item{name}{\code{character(1)}. The name of the new column in the output. If omitted, it will default to \code{n}. If there's
already a column called \code{n}, it will error, and require you to specify the name.}
}
\value{
A \code{data.frame}. \code{count()} and \code{add_count()} have the same groups as the input.
}
\description{
\code{count()} lets you quickly count the unique values of one or more variables:
\code{df \%>\% count(a, b)} is roughly equivalent to
\code{df \%>\% group_by(a, b) \%>\% summarise(n = n())}.
\code{count()} is paired with \code{tally()}, a lower-level helper that is equivalent to \code{df \%>\% summarise(n = n())}. Supply
\code{wt} to perform weighted counts, switching the summary from from \code{n = n()} to \code{n = sum(wt)}.
\code{add_count()} and \code{add_tally()} are equivalent to \code{count()} and \code{tally()} but use \code{mutate()} instead of \code{summarise()}
so that they add a new column with group-wise counts.
}
\examples{
# count() is a convenient way to get a sense of the distribution of
# values in a dataset
mtcars \%>\% count(cyl)
mtcars \%>\% count(cyl, sort = TRUE)
mtcars \%>\% count(cyl, am, sort = TRUE)
# Note that if the data are already grouped, count() adds an additional grouping variable
# which is removed afterwards
mtcars \%>\% group_by(gear) \%>\% count(cyl)

# tally() is a lower-level function that assumes you've done the grouping
mtcars \%>\% tally()
mtcars \%>\% group_by(cyl) \%>\% tally()

# both count() and tally() have add_ variants that work like mutate() instead of summarise
mtcars \%>\% add_count(cyl, wt = am)
mtcars \%>\% add_tally(wt = am)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_helpers.R
\name{peek_vars}
\alias{peek_vars}
\title{Peek at variables in the selection context}
\usage{
peek_vars()
}
\value{
A vector of column names.
}
\description{
Return the vector of column names of the data currently available for selection.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pull.R
\name{pull}
\alias{pull}
\title{Pull out a single variable}
\usage{
pull(.data, var = -1)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{var}{A variable specified as:
\itemize{
\item a literal variable name
\item a positive integer, giving the position counting from the left
\item a negative integer, giving the position counting from the right
}

The default returns the last column (on the assumption that's the column you've created most recently).}
}
\description{
This is a direct replacement for \verb{[[.data.frame}.
}
\examples{
mtcars \%>\% pull(-1)
mtcars \%>\% pull(1)
mtcars \%>\% pull(cyl)
mtcars \%>\% pull("cyl")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/relocate.R
\name{relocate}
\alias{relocate}
\title{Change column order}
\usage{
relocate(.data, ..., .before = NULL, .after = NULL)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{<\code{\link[=select_helpers]{poor-select}}> Columns to move.}

\item{.before, .after}{<\code{\link[=select_helpers]{poor-select}}> Destination of columns selected by \code{...}. Supplying neither
will move columns to the left-hand side; specifying both will result in an error.}
}
\value{
An object of the same type as \code{.data}. The output has the following properties:
\itemize{
\item Rows are not affected.
\item The same columns appear in the output, but (usually) in a different place.
\item Data frame attributes are preserved.
\item Groups are not affected.
}
}
\description{
Use \code{relocate()} to change column positions, using the same syntax as \code{\link[=select]{select()}} to make it easy to move blocks of
columns at once.
}
\examples{
df <- data.frame(
  a = 1, b = 1, c = 1, d = "a", e = "a", f = "a",
  stringsAsFactors = FALSE
)
df \%>\% relocate(f)
df \%>\% relocate(a, .after = c)
df \%>\% relocate(f, .before = b)
df \%>\% relocate(a, .after = last_col())

# Can also select variables based on their type
df \%>\% relocate(where(is.character))
df \%>\% relocate(where(is.numeric), .after = last_col())
# Or with any other select helper
df \%>\% relocate(any_of(c("a", "e", "i", "o", "u")))

# When .before or .after refers to multiple variables they will be
# moved to be immediately before/after the selected variables.
df2 <- data.frame(
  a = 1, b = "a", c = 1, d = "a",
  stringsAsFactors = FALSE
)
df2 \%>\% relocate(where(is.numeric), .after = where(is.character))
df2 \%>\% relocate(where(is.numeric), .before = where(is.character))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/where.R
\name{where}
\alias{where}
\title{Select variables with a function}
\usage{
where(fn)
}
\arguments{
\item{fn}{A function that returns \code{TRUE} or \code{FALSE}.}
}
\value{
A vector of \code{integer} column positions which are the result of the \code{fn} evaluation.
}
\description{
This selection helper selects the variables for which a function returns \code{TRUE}.
}
\examples{
iris \%>\% select(where(is.numeric))
iris \%>\% select(where(function(x) is.numeric(x)))
iris \%>\% select(where(function(x) is.numeric(x) && mean(x) > 3.5))

}
\seealso{
\link{select_helpers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nth.R
\name{nth}
\alias{nth}
\alias{first}
\alias{last}
\title{Extract the first, last or nth value from a vector}
\usage{
nth(x, n, order_by = NULL, default = default_missing(x))

first(x, order_by = NULL, default = default_missing(x))

last(x, order_by = NULL, default = default_missing(x))
}
\arguments{
\item{x}{A vector}

\item{n}{For \code{nth()}, a single integer specifying the position. Negative integers index from the end (i.e. \code{-1L}
will return the last value in the vector).

If a double is supplied, it will be silently truncated.}

\item{order_by}{An optional vector used to determine the order}

\item{default}{A default value to use if the position does not exist in the input. This is guessed by default for
base vectors, where a missing value of the appropriate type is returned, and for lists, where a \code{NULL} is return.

For more complicated objects, you'll need to supply this value. Make sure it is the same type as \code{x}.}
}
\value{
A single value. \code{[[} is used to do the subsetting.
}
\description{
These are straightforward wrappers around \code{\link{[[}}. The main advantage is that you can provide an optional
secondary vector that defines the ordering, and provide a default value to use when the input is shorter than
expected.
}
\examples{
x <- 1:10
y <- 10:1

first(x)
last(y)

nth(x, 1)
nth(x, 5)
nth(x, -2)
nth(x, 11)

last(x)
# Second argument provides optional ordering
last(x, y)

# These functions always return a single value
first(integer())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/with_groups.R
\name{with_groups}
\alias{with_groups}
\title{Perform an operation with temporary groups}
\usage{
with_groups(.data, .groups, .f, ...)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{.groups}{<\code{\link[=select_helpers]{poor-select}}> One or more variables to group by. Unlike \code{\link[=group_by]{group_by()}}, you can
only group by existing variables, and you can use \code{poor-select} syntax like \code{c(x, y, z)} to select multiple
variables.

Use \code{NULL} to temporarily \strong{un}group.}

\item{.f}{A \code{function} to apply to regrouped data. Supports lambda-style \code{~} syntax.}

\item{...}{Additional arguments passed on to \code{.f}.}
}
\description{
This function allows you to modify the grouping variables for a single operation.
}
\examples{
df <- data.frame(g = c(1, 1, 2, 2, 3), x = runif(5))
df \%>\% with_groups(g, mutate, x_mean = mean(x))
df \%>\% with_groups(g, ~ mutate(.x, x_mean = mean(x)))

df \%>\%
  group_by(g) \%>\%
  with_groups(NULL, mutate, x_mean = mean(x))

# NB: grouping can't be restored if you remove the grouping variables
df \%>\%
  group_by(g) \%>\%
  with_groups(NULL, mutate, g = NULL)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/if_else.R
\name{if_else}
\alias{if_else}
\title{Vectorised if}
\usage{
if_else(condition, true, false, missing = NULL)
}
\arguments{
\item{condition}{A \code{logical(n)} vector.}

\item{true, false}{Values to use for \code{TRUE} and \code{FALSE} in \code{condition}. They must either be the same length as
\code{condition} or be length 1. They must also be the same type.}

\item{missing}{If not \code{NULL} (the default), this will replace any missing values.}
}
\value{
A vector the same length as \code{condition} with values for \code{TRUE} and \code{FALSE} replaced by those specified in
\code{true} and \code{false}, respectively.
}
\description{
This is a wrapper around \code{ifelse()} which checks that \code{true} and \code{false} are of the same type, making the output more
predictable.
}
\examples{
x <- c(-5:5, NA)
if_else(x < 0, NA_integer_, x)
if_else(x < 0, "negative", "positive", "missing")

# Unlike ifelse, if_else preserves types
x <- factor(sample(letters[1:5], 10, replace = TRUE))
ifelse(x \%in\% c("a", "b", "c"), x, factor(NA))
# Attributes are taken from the `true` vector
if_else(x \%in\% c("a", "b", "c"), x, factor(NA))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_cols.R
\name{group_cols}
\alias{group_cols}
\title{Select Grouping Variables}
\usage{
group_cols()
}
\description{
This selection helper matches grouping variables. It can be used within \code{\link[=select]{select()}} and \code{\link[=relocate]{relocate()}} selections.
}
\examples{
mtcars \%>\% group_by(am, cyl) \%>\% select(group_cols())

}
\seealso{
\code{\link[=groups]{groups()}} and \code{\link[=group_vars]{group_vars()}} for retrieving the grouping variables outside selection contexts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select.R
\name{select}
\alias{select}
\title{Subset columns using their names and types}
\usage{
select(.data, ...)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{<\code{\link[=select_helpers]{poor-select}}> One or more unquoted expressions separated by commas. Variable names can
be used as if they were positions in the data frame, so expressions like \code{x:y} can be used to select a range of
variables.}
}
\value{
An object of the same type as \code{.data}. The output has the following properties:
\itemize{
\item Rows are not affected.
\item Output columns are a subset of input columns, potentially with a different order. Columns will be renamed if
\code{new_name = old_name} form is used.
\item Data frame attributes are preserved.
\item Groups are maintained; you can't select off grouping variables.
}
}
\description{
Select (and optionally rename) variables in a \code{data.frame}, using a concise mini-language that makes it easy to refer
to variables based on their name (e.g. \code{a:f} selects all columns from \code{a} on the left to \code{f} on the right). You can
also use predicate functions like \code{\link[=is.numeric]{is.numeric()}} to select variables based on their properties.
}
\details{
\subsection{Overview of selection features}{

{poorman} selections implement a dialect of R where operators make it easy to select variables:
\itemize{
\item \code{:} for selecting a range of consecutive variables.
\item \code{!} for taking the complement of a set of variables.
\item \code{&} and \code{|} for selecting the intersection or the union of two sets of variables.
\item \code{c()} for combining selections.
}

In addition, you can use \strong{selection helpers}. Some helpers select specific columns:
\itemize{
\item \code{\link[=everything]{everything()}}: Matches all variables.
\item \code{\link[=last_col]{last_col()}}: Select last variable, possibly with an offset.
}

These helpers select variables by matching patterns in their names:
\itemize{
\item \code{\link[=starts_with]{starts_with()}}: Starts with a prefix.
\item \code{\link[=ends_with]{ends_with()}}: Ends with a suffix.
\item \code{\link[=contains]{contains()}}: Contains a literal string.
\item \code{\link[=matches]{matches()}}: Matches a regular expression.
\item \code{\link[=num_range]{num_range()}}: Matches a numerical range like \code{x01}, \code{x02}, \code{x03}.
}

These helpers select variables from a character vector:
\itemize{
\item \code{\link[=all_of]{all_of()}}: Matches variable names in a character vector. All names must be present, otherwise an out-of-bounds
error is thrown.
\item \code{\link[=any_of]{any_of()}}: Same as \code{all_of()}, except that no error is thrown for names that don't exist.
}

This helper selects variables with a function:
\itemize{
\item \code{\link[=where]{where()}}: Applies a function to all variables and selects those for which the function returns \code{TRUE}.
}
}
}
\examples{
# Here we show the usage for the basic selection operators. See the
# specific help pages to learn about helpers like [starts_with()].

# Select variables by name:
mtcars \%>\% select(mpg)

# Select multiple variables by separating them with commas. Note
# how the order of columns is determined by the order of inputs:
mtcars \%>\% select(disp, gear, am)

# Rename variables:
mtcars \%>\% select(MilesPerGallon = mpg, everything())

# The `:` operator selects a range of consecutive variables:
select(mtcars, mpg:cyl)

# The `!` operator negates a selection:
mtcars \%>\% select(!(mpg:qsec))
mtcars \%>\% select(!ends_with("p"))

# `&` and `|` take the intersection or the union of two selections:
iris \%>\% select(starts_with("Petal") & ends_with("Width"))
iris \%>\% select(starts_with("Petal") | ends_with("Width"))

# To take the difference between two selections, combine the `&` and
# `!` operators:
iris \%>\% select(starts_with("Petal") & !ends_with("Width"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/union_all.R
\name{union_all}
\alias{union_all}
\title{Union All}
\usage{
union_all(x, y, ...)
}
\arguments{
\item{x, y}{objects to union all elements of (ignoring order)}

\item{...}{other arguments passed on to methods}
}
\description{
Union all elements of R objects together.
}
\examples{
first <- mtcars[1:20, ]
second <- mtcars[10:32, ]
union_all(first, second)

# union_all does not remove duplicates
a <- data.frame(column = c(1:10, 10))
b <- data.frame(column = c(1:5, 5))
union_all(a, b)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coalesce.R
\name{coalesce}
\alias{coalesce}
\title{Find first non-missing element}
\usage{
coalesce(...)
}
\arguments{
\item{...}{Vectors. Inputs should be recyclable (either be length \code{1L} or \code{n}) and coercible to a common type.}
}
\description{
Given a set of vectors, \code{coalesce()} finds the first non-missing value at each position. This is inspired by the SQL
\code{COALESCE} function which does the same thing for \code{NULL}s.
}
\details{
Currently, \code{coalesce()} type checking does not take place.
}
\examples{
# Use a single value to replace all missing vectors
x <- sample(c(1:5, NA, NA, NA))
coalesce(x, 0L)

# Or match together a complete vector from missing pieces
y <- c(1, 2, NA, NA, 5)
z <- c(NA, NA, 3, 4, 5)
coalesce(y, z)

}
\seealso{
\code{\link[=na_if]{na_if()}} to replace specified values to a \code{NA}.

\code{\link[=replace_na]{replace_na()}} to replace a \code{NA} with a value.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/joins.R
\name{mutate_joins}
\alias{mutate_joins}
\alias{inner_join}
\alias{left_join}
\alias{right_join}
\alias{full_join}
\title{Mutating Joins}
\usage{
inner_join(
  x,
  y,
  by = NULL,
  suffix = c(".x", ".y"),
  ...,
  na_matches = c("na", "never")
)

left_join(
  x,
  y,
  by = NULL,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE,
  na_matches = c("na", "never")
)

right_join(
  x,
  y,
  by = NULL,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE,
  na_matches = c("na", "never")
)

full_join(
  x,
  y,
  by = NULL,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE,
  na_matches = c("na", "never")
)
}
\arguments{
\item{x, y}{The \code{data.frame}s to join.}

\item{by}{A character vector of variables to join by. If \code{NULL}, the default, \verb{*_join()} will do a natural join,
using all variables with common names across the two tables. A message lists the variables so that you can check
they're right (to suppress the message, simply explicitly list the variables that you want to join).

To join by different variables on x and y use a named vector. For example, \code{by = c("a" = "b")} will match \code{x.a} to
\code{y.b}.

To join by multiple variables, use a vector with length > 1. For example, \code{by = c("a", "b")} will match \code{x$a} to
\code{y$a} and \code{x$b} to \code{y$b}. Use a named vector to match different variables in \code{x} and \code{y}. For example,
\code{by = c("a" = "b", "c" = "d")} will match \code{x$a} to \code{y$b} and \code{x$c} to \code{y$d}.

To perform a cross-join, generating all combinations of \code{x} and \code{y}, use \code{by = character()}.}

\item{suffix}{\code{character(2)}. If there are non-joined duplicate variables in \code{x} and \code{y}, these suffixes will be
added to the output to disambiguate them.}

\item{...}{Additional arguments to pass to \code{\link[=merge]{merge()}}}

\item{na_matches}{Should \code{NA} and \code{NaN} values match one another?

The default, \code{"na"}, treats two \code{NA} or \code{NaN} values as equal, like \code{\%in\%}, \code{\link[=match]{match()}}, \code{\link[=merge]{merge()}}.

Use \code{"never"} to always treat two \code{NA} or \code{NaN} values as different, like joins for database sources, similarly to
\code{merge(incomparables = FALSE)}.}

\item{keep}{\code{logical(1)}. Should the join keys from both \code{x} and \code{y} be preserved in the output? Only applies to
\code{left_join()}, \code{right_join()}, and \code{full_join()}.}
}
\value{
A \code{data.frame}. The order of the rows and columns of \code{x} is preserved as much as possible. The output has the
following properties:
\itemize{
\item For \code{inner_join()}, a subset of \code{x} rows.
For \code{left_join()}, all \code{x} rows.
For \code{right_join()}, a subset of \code{x} rows, followed by unmatched \code{y} rows.
For \code{full_join()}, all \code{x} rows, followed by unmatched \code{y} rows.
\item For all joins, rows will be duplicated if one or more rows in \code{x} matches multiple rows in \code{y}.
\item Output columns include all \code{x} columns and all \code{y} columns. If columns in \code{x} and \code{y} have the same name (and
aren't included in \code{by}), \code{suffix}es are added to disambiguate.
\item Output columns included in \code{by} are coerced to common type across \code{x} and \code{y}.
\item Groups are taken from \code{x}.
}
}
\description{
The mutating joins add columns from \code{y} to \code{x}, matching rows based on the keys:
\itemize{
\item \code{inner_join()}: includes all rows in \code{x} and \code{y}.
\item \code{left_join()}: includes all rows in \code{x}.
\item \code{right_join()}: includes all rows in \code{y}.
\item \code{full_join()}: includes all rows in \code{x} or \code{y}.
}

If a row in \code{x} matches multiple rows in \code{y}, all the rows in \code{y} will be returned once for each matching row in \code{x}.
}
\examples{
# If a row in `x` matches multiple rows in `y`, all the rows in `y` will be
# returned once for each matching row in `x`
df1 <- data.frame(x = 1:3)
df2 <- data.frame(x = c(1, 1, 2), y = c("first", "second", "third"))
df1 \%>\% left_join(df2)

# By default, NAs match other NAs so that there are two
# rows in the output of this join:
df1 <- data.frame(x = c(1, NA), y = 2)
df2 <- data.frame(x = c(1, NA), z = 3)
left_join(df1, df2)

# You can optionally request that NAs don't match, giving a
# a result that more closely resembles SQL joins
left_join(df1, df2, na_matches = "never")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/n_distinct.R
\name{n_distinct}
\alias{n_distinct}
\title{Count the number of unique values in a set of vectors}
\usage{
n_distinct(..., na.rm = FALSE)
}
\arguments{
\item{...}{Vectors of values.}

\item{na.rm}{\code{logical(1)}. If \code{TRUE} missing values don't count.}
}
\description{
This is the equivalent of \code{length(unique(x))} for multiple vectors.
}
\examples{
x <- sample(1:10, 1e5, rep = TRUE)
length(unique(x))
n_distinct(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distinct.R
\name{distinct}
\alias{distinct}
\title{Subset distinct/unique rows}
\usage{
distinct(.data, ..., .keep_all = FALSE)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{Optional variables to use when determining uniqueness. If there are multiple rows for a given combination
of inputs, only the first row will be preserved. If omitted, will use all variables.}

\item{.keep_all}{\code{logical(1)}. If \code{TRUE}, keep all variables in \code{.data}. If a combination of \code{...} is not distinct,
this keeps the first row of values.}
}
\value{
A \code{data.frame} with the following properties:
\itemize{
\item Rows are a subset of the input but appear in the same order.
\item Columns are not modified if \code{...} is empty or \code{.keep_all} is \code{TRUE}. Otherwise, \code{distinct()} first calls \code{mutate()}
to create new columns.
\item Groups are not modified.
\item \code{data.frame} attributes are preserved.
}
}
\description{
Select only distinct/unique rows from a \code{data.frame}.
}
\examples{
df <- data.frame(
  x = sample(10, 100, rep = TRUE),
  y = sample(10, 100, rep = TRUE)
)
nrow(df)
nrow(distinct(df))
nrow(distinct(df, x, y))

distinct(df, x)
distinct(df, y)

# You can choose to keep all other variables as well
distinct(df, x, .keep_all = TRUE)
distinct(df, y, .keep_all = TRUE)

# You can also use distinct on computed variables
distinct(df, diff = abs(x - y))

# The same behaviour applies for grouped data frames,
# except that the grouping variables are always included
df <- data.frame(
  g = c(1, 1, 2, 2),
  x = c(1, 1, 2, 1)
) \%>\% group_by(g)
df \%>\% distinct(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/between.R
\name{between}
\alias{between}
\title{Do values in a numeric vector fall in specified range?}
\usage{
between(x, left, right)
}
\arguments{
\item{x}{A \code{numeric} vector of values.}

\item{left, right}{Boundary values.}
}
\value{
A \code{logical} vector the same length as \code{x}.
}
\description{
This is a shortcut for \code{x >= left & x <= right}.
}
\examples{
between(1:12, 7, 9)

x <- rnorm(1e2)
x[between(x, -1, 1)]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recode.R
\name{recode}
\alias{recode}
\alias{recode_factor}
\title{Recode values}
\usage{
recode(.x, ..., .default = NULL, .missing = NULL)

recode_factor(.x, ..., .default = NULL, .missing = NULL, .ordered = FALSE)
}
\arguments{
\item{.x}{A vector to modify}

\item{...}{Replacements. For \code{character} and \code{factor} \code{.x}, these should be named and replacement is based only on
their name. For \code{numeric} \code{.x}, these can be named or not. If not named, the replacement is done based on position
i.e. \code{.x} represents positions to look for in replacements. See examples.

When named, the argument names should be the current values to be replaced, and the argument values should be the new
(replacement) values.

All replacements must be the same type, and must have either length one or the same length as \code{.x}.}

\item{.default}{If supplied, all values not otherwise matched will be given this value. If not supplied and if the
replacements are the same type as the original values in \code{.x}, unmatched values are not changed. If not supplied and
if the replacements are not compatible, unmatched values are replaced with \code{NA}.

\code{.default} must be either length 1 or the same length as \code{.x}.}

\item{.missing}{If supplied, any missing values in \code{.x} will be replaced by this value. Must be either length 1 or
the same length as \code{.x}.}

\item{.ordered}{\code{logical(1)}. If \code{TRUE}, \code{recode_factor()} creates an ordered \code{factor}.}
}
\value{
A vector the same length as \code{.x}, and the same type as
the first of \code{...}, \code{.default}, or \code{.missing}.
\code{recode_factor()} returns a factor whose levels are in the same order as
in \code{...}. The levels in \code{.default} and \code{.missing} come last.
}
\description{
This is a vectorised version of \code{\link[=switch]{switch()}}: you can replace \code{numeric} values based on their position or their name,
and \code{character} or \code{factor} values only by their name. This is an S3 generic: \code{{poorman}} provides methods for
\code{numeric},  \code{character}, and \code{factor}s. For \code{logical} vectors, use \code{\link[=if_else]{if_else()}}. For more complicated criteria, use
\code{\link[=case_when]{case_when()}}.

You can use \code{recode()} directly with \code{factor}s; it will preserve the existing order of levels while changing the
values. Alternatively, you can use \code{recode_factor()}, which will change the order of levels to match the order of
replacements.

This is a direct port of the \code{dplyr::recode()} function.
}
\examples{
# For character values, recode values with named arguments only. Unmatched
# values are unchanged.
char_vec <- sample(c("a", "b", "c"), 10, replace = TRUE)
recode(char_vec, a = "Apple")
recode(char_vec, a = "Apple", b = "Banana")

# Use .default as replacement for unmatched values. Note that NA and
# replacement values need to be of the same type.
recode(char_vec, a = "Apple", b = "Banana", .default = NA_character_)

# Throws an error as NA is logical, not character.
\dontrun{
recode(char_vec, a = "Apple", b = "Banana", .default = NA)
}

# For numeric values, named arguments can also be used
num_vec <- c(1:4, NA)
recode(num_vec, `2` = 20L, `4` = 40L)

# Or if you don't name the arguments, recode() matches by position.
# (Only works for numeric vector)
recode(num_vec, "a", "b", "c", "d")
# .x (position given) looks in (...), then grabs (... value at position)
# so if nothing at position (here 5), it uses .default or NA.
recode(c(1, 5, 3), "a", "b", "c", "d", .default = "nothing")

# Note that if the replacements are not compatible with .x,
# unmatched values are replaced by NA and a warning is issued.
recode(num_vec, `2` = "b", `4` = "d")
# use .default to change the replacement value
recode(num_vec, "a", "b", "c", .default = "other")
# use .missing to replace missing values in .x
recode(num_vec, "a", "b", "c", .default = "other", .missing = "missing")

# For factor values, use only named replacements
# and supply default with levels()
factor_vec <- factor(c("a", "b", "c"))
recode(factor_vec, a = "Apple", .default = levels(factor_vec))

# Use recode_factor() to create factors with levels ordered as they
# appear in the recode call. The levels in .default and .missing
# come last.
recode_factor(num_vec, `1` = "z", `2` = "y", `3` = "x")
recode_factor(num_vec, `1` = "z", `2` = "y", `3` = "x", .default = "D")
recode_factor(num_vec, `1` = "z", `2` = "y", `3` = "x", .default = "D", .missing = "M")

# When the input vector is a compatible vector (character vector or
# factor), it is reused as default.
recode_factor(letters[1:3], b = "z", c = "y")
recode_factor(factor(letters[1:3]), b = "z", c = "y")

}
\seealso{
\code{\link[=na_if]{na_if()}} to replace specified values with a \code{NA}.

\code{\link[=coalesce]{coalesce()}} to replace missing values with a specified value.

\code{\link[=replace_na]{replace_na()}} to replace \code{NA} with a value.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arrange.R
\name{arrange}
\alias{arrange}
\title{Arrange rows by variables}
\usage{
arrange(.data, ...)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{A comma separated vector of unquoted name(s) to order the data by.}
}
\value{
A \code{data.frame}.
}
\description{
Order rows of a \code{data.frame} by an expression involving its variables.
}
\examples{
arrange(mtcars, mpg)
mtcars \%>\% arrange(mpg)
mtcars \%>\% arrange(cyl, mpg)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glimpse.R
\name{glimpse}
\alias{glimpse}
\title{Get a glimpse of your data}
\usage{
glimpse(x, width = getOption("width"), ...)
}
\arguments{
\item{x}{An object to glimpse at.}

\item{width}{\code{integer(1)}. Width of the output.}

\item{...}{Additional parameters to pass to \code{\link[utils:str]{utils::str()}}.}
}
\value{
\code{x}, invisibly.
}
\description{
\code{glimpse()} is like a transposed version of print(): columns run down the page, and data runs across. This makes it
possible to see every column in a \code{data.frame}. It is no more than a wrapper around \code{\link[utils:str]{utils::str()}} only it returns
the input (invisibly) meaning it can be used within a data pipeline.
}
\examples{
glimpse(mtcars)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename.R
\name{rename}
\alias{rename}
\alias{rename_with}
\title{Rename columns}
\usage{
rename(.data, ...)

rename_with(.data, .fn, .cols = everything(), ...)
}
\arguments{
\item{.data}{A \code{data.frame}}

\item{...}{For \code{rename()}: comma separated key-value pairs in the form of \code{new_name = old_name} to rename selected variables.

For \code{rename_with()}: additional arguments passed onto \code{.fn}.}

\item{.fn}{A \verb{function()} used to transform the selected \code{.cols}. Should return a character vector the same length as
the input.}

\item{.cols}{Columns to rename; defaults to all columns.}
}
\value{
A \code{data.frame} with the following properties:
\itemize{
\item Rows are not affected.
\item Column names are changed; column order is preserved.
\item \code{data.frame} attributes are preserved.
\item Groups are updated to reflect new names.
}
}
\description{
\code{rename()} changes the names of individual variables using \code{new_name = old_name} syntax.
\code{rename_with()} renames columns using a function.
}
\examples{
rename(mtcars, MilesPerGallon = mpg)
rename(mtcars, Cylinders = cyl, Gears = gear)
mtcars \%>\% rename(MilesPerGallon = mpg)

rename_with(mtcars, toupper)
rename_with(mtcars, toupper, starts_with("c"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/joins_filtering.R
\name{filter_joins}
\alias{filter_joins}
\alias{anti_join}
\alias{semi_join}
\title{Filtering joins filter rows from \code{x} based on the presence or absence of
matches in \code{y}:}
\usage{
anti_join(x, y, by = NULL)

semi_join(x, y, by = NULL)
}
\arguments{
\item{x, y}{The \code{data.frame}s to join.}

\item{by}{A character vector of variables to join by. If \code{NULL}, the default,
\verb{*_join()} will do a natural join, using all variables with common names
across the two tables. A message lists the variables so that you can check
they're right (to suppress the message, simply explicitly list the variables
that you want to join).}
}
\description{
\itemize{
\item \code{semi_join()} return all rows from \code{x} with a match in \code{y}.
\item \code{anti_join()} return all rows from \code{x} with\emph{out} a match in \code{y}.
}
}
\examples{
table1 <- data.frame(
  pupil = rep(1:3, each = 2),
  test = rep(c("A", "B"), 3),
  score = c(60, 70, 65, 80, 85, 70),
  stringsAsFactors = FALSE
)
table2 <- table1[c(1, 3, 4), ]

table1 \%>\% anti_join(table2, by = c("pupil", "test"))
table1 \%>\% semi_join(table2, by = c("pupil", "test"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fill.R
\name{fill}
\alias{fill}
\title{Fill in missing values with previous or next value}
\usage{
fill(data, ..., .direction = c("down", "up", "downup", "updown"))
}
\arguments{
\item{data}{A \code{data.frame}.}

\item{...}{Columns to fill.}

\item{.direction}{Direction in which to fill missing values. Currently either \code{"down"} (the default), \code{"up"},
\code{"downup"} (i.e. first down and then up) or \code{"updown"} (first up and then down).}
}
\description{
Fills missing values in selected columns using the next or previous entry. This is useful in the common output format
where values are not repeated, and are only recorded when they change.
}
\details{
Missing values are replaced in atomic vectors; \code{NULL}s are replaced in lists.
}
\examples{
# Value (year) is recorded only when it changes
sales <- data.frame(
  quarter = c(
    "Q1", "Q2", "Q3", "Q4", "Q1", "Q2", "Q3", "Q4", "Q1", "Q2",
    "Q3", "Q4", "Q1", "Q2", "Q3", "Q4"
  ),
  year = c(2000, NA, NA, NA, 2001, NA, NA, NA, 2002, NA, NA, NA, 2004, NA, NA, NA),
  sales = c(
    66013, 69182, 53175, 21001, 46036, 58842, 44568, 50197, 39113, 41668, 30144,
    52897, 32129, 67686, 31768, 49094
  )
)

# `fill()` defaults to replacing missing data from top to bottom
sales \%>\% fill(year)

# Value (pet_type) is missing above
tidy_pets <- data.frame(
  rank = c(1L, 2L, 3L, 4L, 5L, 6L, 1L, 2L, 3L, 4L, 5L, 6L),
  pet_type = c(NA, NA, NA, NA, NA, "Dog", NA, NA, NA, NA, NA, "Cat"),
  breed = c(
    "Boston Terrier", "Retrievers (Labrador)", "Retrievers (Golden)",
    "French Bulldogs", "Bulldogs", "Beagles", "Persian", "Maine Coon",
    "Ragdoll", "Exotic", "Siamese", "American Short"
  )
)

# For values that are missing above you can use `.direction = "up"`
tidy_pets \%>\%
  fill(pet_type, .direction = "up")

# Value (n_squirrels) is missing above and below within a group
squirrels <- data.frame(
  group = c(1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3),
  name = c(
    "Sam", "Mara", "Jesse", "Tom", "Mike", "Rachael", "Sydekea",
    "Gabriela", "Derrick", "Kara", "Emily", "Danielle"
  ),
  role = c(
    "Observer", "Scorekeeper", "Observer", "Observer", "Observer",
    "Observer", "Scorekeeper", "Observer", "Observer", "Scorekeeper",
    "Observer", "Observer"
  ),
  n_squirrels = c(NA, 8, NA, NA, NA, NA, 14, NA, NA, 9, NA, NA)
)

# The values are inconsistently missing by position within the group
# Use .direction = "downup" to fill missing values in both directions
squirrels \%>\%
  group_by(group) \%>\%
  fill(n_squirrels, .direction = "downup") \%>\%
  ungroup()

# Using `.direction = "updown"` accomplishes the same goal in this example

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rownames.R
\name{rownames}
\alias{rownames}
\alias{rownames_to_column}
\alias{rowid_to_column}
\alias{column_to_rownames}
\alias{remove_rownames}
\alias{has_rownames}
\title{Tools for working with row names}
\usage{
rownames_to_column(.data, var = "rowname")

rowid_to_column(.data, var = "rowid")

column_to_rownames(.data, var = "rowname")

remove_rownames(.data)

has_rownames(.data)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{var}{\code{character(1)}. The name of the column to use for row names.}
}
\value{
\itemize{
\item \code{column_to_rownames()} always returns a \code{data.frame}.
\item \code{has_rownames()} returns a \code{logical(1)}.
\item All other functions return an object of the same class as the input.
}
}
\description{
In some quarters, it is considered best to avoid row names, because they are effectively a character column with
different semantics than every other column.
These functions allow to you detect if a \code{data.frame} has row names (\code{has_rownames()}), remove them
(\code{remove_rownames()}), or convert them back-and-forth between an explicit column (\code{rownames_to_column()} and
\code{column_to_rownames()}). Also included is \code{rowid_to_column()}, which adds a column at the start of the dataframe of
ascending sequential row ids starting at 1. Note that this will remove any existing row names.
}
\examples{
# Detect row names
has_rownames(mtcars)
has_rownames(iris)

# Remove row names
remove_rownames(mtcars) \%>\% has_rownames()

# Convert between row names and column
mtcars <- rownames_to_column(mtcars, var = "car")
column_to_rownames(mtcars, var = "car") \%>\% head()

# Adding rowid as a column
rowid_to_column(iris) \%>\% head()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/near.R
\name{near}
\alias{near}
\title{Compare two numeric vectors}
\usage{
near(x, y, tol = .Machine$double.eps^0.5)
}
\arguments{
\item{x, y}{Numeric vectors to compare}

\item{tol}{Tolerance of comparison.}
}
\description{
This is a safe way of comparing if two vectors of floating point numbers are (pairwise) equal.  This is safer than
using \code{==}, because it has a built in tolerance.
}
\examples{
sqrt(2) ^ 2 == 2
near(sqrt(2) ^ 2, 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter.R
\name{filter}
\alias{filter}
\title{Return rows with matching conditions}
\usage{
filter(.data, ..., .preserve = FALSE)
}
\arguments{
\item{.data}{A \code{data.frame}.}

\item{...}{Logical predicated defined in terms of the variables in \code{.data}. Multiple conditions are combined with
\code{&}. Arguments within \code{...} are automatically quoted and evaluated within the context of the \code{data.frame}.}

\item{.preserve}{\code{logical(1)}. Relevant when the .data input is grouped. If \code{.preserve = FALSE} (the default), the
grouping structure is recalculated based on the resulting data, otherwise the grouping is kept as is.}
}
\value{
A \code{data.frame}.
}
\description{
Use \code{filter()} to choose rows/cases where conditions are \code{TRUE}.
}
\section{Useful filter functions}{

\itemize{
\item \code{==}, \code{>}, \code{>=}, etc.
\item \code{&}, \code{|}, \code{!}, \code{xor()}
\item \code{is.na()}
}
}

\examples{
filter(mtcars, am == 1)
mtcars \%>\% filter(cyl == 4)
mtcars \%>\% filter(cyl <= 5 & am > 0)
mtcars \%>\% filter(cyl == 4 | cyl == 8)
mtcars \%>\% filter(!(cyl \%in\% c(4, 6)), am != 0)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/case_when.R
\name{case_when}
\alias{case_when}
\title{A General Vetorised \verb{if()}}
\usage{
case_when(...)
}
\arguments{
\item{...}{A sequence of two-sided formulas. The left hand side (LHS) determines which values match this case. The
right hand side (RHS) provides the replacement value.

The LHS must evaluate to a logical vector. The RHS does not need to be logical, but all RHSs must evaluate to the
same type of vector.

Both LHS and RHS may have the same length of either 1 or n. The value of n must be consistent across all cases. The
case of \code{n == 0} is treated as a variant of \code{n != 1}.

\code{NULL} inputs are ignored.}
}
\value{
A vector of length 1 or n, matching the length of the logical input or output vectors, with the type (and attributes)
of the first RHS. Inconsistent lengths or types will generate an error.
}
\description{
This function allows you to vectorise multiple \code{\link[=if_else]{if_else()}} statements. It is an R equivalent of the SQL \verb{CASE WHEN}
statement. If no cases match, \code{NA} is returned.
}
\examples{
x <- 1:50
case_when(
  x \%\% 35 == 0 ~ "fizz buzz",
  x \%\% 5 == 0 ~ "fizz",
  x \%\% 7 == 0 ~ "buzz",
  TRUE ~ as.character(x)
)

# Like an if statement, the arguments are evaluated in order, so you must
# proceed from the most specific to the most general. This won't work:
case_when(
  TRUE ~ as.character(x),
  x \%\%  5 == 0 ~ "fizz",
  x \%\%  7 == 0 ~ "buzz",
  x \%\% 35 == 0 ~ "fizz buzz"
)

# If none of the cases match, NA is used:
case_when(
  x \%\%  5 == 0 ~ "fizz",
  x \%\%  7 == 0 ~ "buzz",
  x \%\% 35 == 0 ~ "fizz buzz"
)

# Note that NA values in the vector x do not get special treatment. If you want
# to explicitly handle NA values you can use the `is.na` function:
x[2:4] <- NA_real_
case_when(
  x \%\% 35 == 0 ~ "fizz buzz",
  x \%\% 5 == 0 ~ "fizz",
  x \%\% 7 == 0 ~ "buzz",
  is.na(x) ~ "nope",
  TRUE ~ as.character(x)
)

# All RHS values need to be of the same type. Inconsistent types will throw an error.
# This applies also to NA values used in RHS: NA is logical, use
# typed values like NA_real_, NA_complex, NA_character_, NA_integer_ as appropriate.
case_when(
  x \%\% 35 == 0 ~ NA_character_,
  x \%\% 5 == 0 ~ "fizz",
  x \%\% 7 == 0 ~ "buzz",
  TRUE ~ as.character(x)
)
case_when(
  x \%\% 35 == 0 ~ 35,
  x \%\% 5 == 0 ~ 5,
  x \%\% 7 == 0 ~ 7,
  TRUE ~ NA_real_
)

# case_when() evaluates all RHS expressions, and then constructs its
# result by extracting the selected (via the LHS expressions) parts.
# In particular NaN are produced in this case:
y <- seq(-2, 2, by = .5)
case_when(
  y >= 0 ~ sqrt(y),
  TRUE   ~ y
)

\dontrun{
case_when(
  x \%\% 35 == 0 ~ 35,
  x \%\% 5 == 0 ~ 5,
  x \%\% 7 == 0 ~ 7,
  TRUE ~ NA
)
}

# case_when is particularly useful inside mutate when you want to
# create a new variable that relies on a complex combination of existing
# variables
mtcars \%>\%
  mutate(
    efficient = case_when(
      mpg > 25 ~ TRUE,
      TRUE ~ FALSE
    )
  )

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unite.R
\name{unite}
\alias{unite}
\title{Unite Multiple Columns Into One}
\usage{
unite(data, col, ..., sep = "_", remove = TRUE, na.rm = FALSE)
}
\arguments{
\item{data}{A \code{data.frame}.}

\item{col}{\code{character(1)} or \code{symbol(1)}. The name of the new column.}

\item{...}{The columns to unite.}

\item{sep}{\code{character(1)}. Separator to use between the values.}

\item{remove}{\code{logical(1)}. If \code{TRUE}, remove the input columns from the output \code{data.frame}.}

\item{na.rm}{\code{logical(1)}. If \code{TRUE}, missing values will be remove prior to uniting each value.}
}
\value{
A \code{data.frame} with the columns passed via \code{...} pasted together in a new column.
}
\description{
Convenience function to paste together multiple columns.
}
\examples{
df <- data.frame(x = c("a", "a", NA, NA), y = c("b", NA, "b", NA))
df

df \%>\% unite("z", x:y, remove = FALSE)
# To remove missing values:
df \%>\% unite("z", x:y, na.rm = TRUE, remove = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lead-lag.R
\name{lag}
\alias{lag}
\alias{lead}
\title{Compute lagged or leading values}
\usage{
lag(x, n = 1L, default = NA)

lead(x, n = 1L, default = NA)
}
\arguments{
\item{x}{A \code{vector} of values}

\item{n}{A positive \code{integer(1)}, giving the number of positions to lead or lag by.}

\item{default}{The value used for non-existent rows (default: \code{NA}).}
}
\description{
Find the "previous" (\code{lag()}) or "next" (\code{lead()}) values in a vector. Useful for comparing values behind of or ahead
of the current values.
}
\examples{
lag(1:5)
lead(1:5)

x <- 1:5
data.frame(behind = lag(x), x, ahead = lead(x))

# If you want to look more rows behind or ahead, use `n`
lag(1:5, n = 1)
lag(1:5, n = 2)

lead(1:5, n = 1)
lead(1:5, n = 2)

# If you want to define a value for non-existing rows, use `default`
lag(1:5)
lag(1:5, default = 0)

lead(1:5)
lead(1:5, default = 6)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poorman-package.R
\docType{package}
\name{poorman}
\alias{poorman}
\alias{poorman-package}
\title{poorman: A Poor Man's Dependency Free Recreation of 'dplyr'}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A replication of key functionality from 'dplyr' and the wider 'tidyverse' using only 'base'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://nathaneastwood.github.io/poorman/}
  \item \url{https://github.com/nathaneastwood/poorman}
  \item Report bugs at \url{https://github.com/nathaneastwood/poorman/issues}
}

}
\author{
\strong{Maintainer}: Nathan Eastwood \email{nathan.eastwood@icloud.com}

Other contributors:
\itemize{
  \item Etienne Bacher \email{etienne.bacher@protonmail.com} (\href{https://orcid.org/0000-0002-9271-5075}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/window_rank.R
\name{window_rank}
\alias{window_rank}
\alias{cume_dist}
\alias{dense_rank}
\alias{min_rank}
\alias{ntile}
\alias{percent_rank}
\alias{row_number}
\title{Windowed Rank Functions}
\usage{
cume_dist(x)

dense_rank(x)

min_rank(x)

ntile(x = row_number(), n)

percent_rank(x)

row_number(x)
}
\arguments{
\item{x}{A vector of values to rank. Missing values are left as is. If you want to treat them as the smallest or
largest values, replace with \code{Inf} or \code{-Inf} before ranking.}

\item{n}{\code{integer(1)}. The number of groups to split up into.}
}
\description{
Six variations on ranking functions, mimicking the ranking functions described in SQL2003. They are currently
implemented using the built in \code{\link[=rank]{rank()}} function. All ranking functions map smallest inputs to smallest outputs. Use
\code{desc()} to reverse the direction.
}
\details{
\itemize{
\item \code{cume_dist()}: a cumulative distribution function. Proportion of all values less than or equal to the current rank.
\item \code{dense_rank()}: like \code{min_rank()}, but with no gaps between ranks
\item \code{min_rank()}: equivalent to \code{rank(ties.method = "min")}
\item \code{ntile()}: a rough rank, which breaks the input vector into \code{n} buckets. The size of the buckets may differ by up
to one, larger buckets have lower rank.
\item \code{percent_rank()}: a number between \code{0} and \code{1} computed by rescaling \code{min_rank} to \verb{[0, 1]}
\item \code{row_number()}: equivalent to \code{rank(ties.method = "first")}
}
}
\examples{
x <- c(5, 1, 3, 2, 2, NA)
row_number(x)
min_rank(x)
dense_rank(x)
percent_rank(x)
cume_dist(x)

ntile(x, 2)
ntile(1:8, 3)

# row_number can be used with single table verbs without specifying x
# (for data frames and databases that support windowing)
mutate(mtcars, row_number() == 1L)
mtcars \%>\% filter(between(row_number(), 1, 10))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_metadata.R
\name{group_metadata}
\alias{group_metadata}
\alias{group_data}
\alias{group_rows}
\alias{group_indices}
\alias{group_vars}
\alias{groups}
\alias{group_size}
\alias{n_groups}
\title{Grouping metadata}
\usage{
group_data(.data)

group_rows(.data)

group_indices(.data)

group_vars(x)

groups(x)

group_size(x)

n_groups(x)
}
\arguments{
\item{.data, x}{A \code{data.frame}.}
}
\description{
\itemize{
\item \code{group_data()} returns a data frame that defines the grouping structure. The columns give the values of the
grouping variables. The last column, always called \code{.rows}, is a list of integer vectors that gives the location of
the rows in each group.
}

\itemize{
\item \code{group_rows()} returns the rows which each group contains.
}

\itemize{
\item \code{group_indices()} returns an integer vector the same length as \code{.data} that gives the group that each row belongs
to.
}

\itemize{
\item \code{group_vars()} gives names of grouping variables as character vector.
}

\itemize{
\item \code{groups()} gives the names as a list of symbols.
}

\itemize{
\item \code{group_size()} gives the size of each group.
}

\itemize{
\item \code{n_groups()} gives the total number of groups.
}
}
\examples{
df <- data.frame(x = c(1,1,2,2))
group_vars(df)
group_rows(df)
group_data(df)

gf <- group_by(df, x)
group_vars(gf)
group_rows(gf)
group_data(gf)

}
\seealso{
See \link{context} for equivalent functions that return values for the current group.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_by.R
\name{group_by_drop_default}
\alias{group_by_drop_default}
\title{Default value for .drop argument of group_by}
\usage{
group_by_drop_default(.tbl)
}
\arguments{
\item{.tbl}{A \code{data.frame}.}
}
\value{
\code{TRUE} unless \code{.tbl} is a grouped \code{data.frame} that was previously obtained by \code{group_by(.drop = FALSE)}
}
\description{
Default value for .drop argument of group_by
}
\examples{
group_by_drop_default(iris)

iris \%>\%
  group_by(Species) \%>\%
  group_by_drop_default()

iris \%>\%
  group_by(Species, .drop = FALSE) \%>\%
  group_by_drop_default()

}
